//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2007-2010 by PGMed@FIT, BUT    \n
 *
 * Authors: Ondrej Siler, siler@fit.vutbr.cz    \n
 * Date:    yyyy/mm/dd                          \n
 *
 * $Id: TriContainerOSGGeometry.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Class for storing triangles created by cross section of tetrahedral mesh.
 */

#ifndef TRI_CONTAINER_OSG_GEOMETRY_H
#define TRI_CONTAINER_OSG_GEOMETRY_H

#include <MDSTk/CrossSection/mdsTriContainer.h>

//stl
#include <vector>

//OSG
#include <osg/Array>
#include <osg/Vec3d>
#include <osg/Geometry>
#include <osg/PrimitiveSet>
#include <osg/Material>

#include <MDSTk/CrossSection/mdsColorTable.h>

namespace mds
{
namespace cs
{

//! triangle container for OpenSceneGraph derived from osg::Geometry
//! data are stored in osg arrays
class TriContainerOSGGeometry : public mds::cs::CTriContainer, public osg::Geometry
{
public:
    //! constructor
    //! sets the default color, creates arrays and initalizes geometry for displaying triangles
	TriContainerOSGGeometry(osg::Vec4d color = osg::Vec4d(1.0, 1.0, 1.0, 1.0)) 
	: _color(color), _colorTable(NULL)
	{
        // geometry will be displayed as triangles
        _pPrimitiveSet = new osg::DrawArrays(osg::PrimitiveSet::TRIANGLES, 0, 0); 
        // creates array for vertices
		_pVertices = new osg::Vec3Array();
        // creates array for colors
        _pColors = new osg::Vec4Array();

		setVertexArray(_pVertices.get());
        setColorArray(_pColors.get());
        addPrimitiveSet(_pPrimitiveSet.get());
    	
        // disable display lists
        setUseDisplayList(false);

        // color binding
        setColorBinding(BIND_PER_PRIMITIVE);

        // enable blending
        osg::ref_ptr<osg::StateSet> ss = new osg::StateSet();
        ss->setMode(GL_BLEND, osg::StateAttribute::ON);
        setStateSet(ss.get());
	};
	
	//! set the pointer to a color table
	void SetColorTable(mds::cs::CColorTable<osg::Vec4d> * colorTable)	{_colorTable = colorTable;}

    //! stores coordinates and color of the trinagle
	virtual void pushTri(vctl::MCVertex * v1, vctl::MCVertex * v2, vctl::MCVertex * v3, vctl::MCTetra * tetra)
	{
		_pVertices->push_back(osg::Vec3d(v1->GetX(), v1->GetY(), v1->GetZ()));
		_pVertices->push_back(osg::Vec3d(v2->GetX(), v2->GetY(), v2->GetZ()));
		_pVertices->push_back(osg::Vec3d(v3->GetX(), v3->GetY(), v3->GetZ()));
		_pColors->push_back((_colorTable) ? (*_colorTable)[tetra->GetObject()] : _color);
		
		_pPrimitiveSet->setCount(_pPrimitiveSet->getCount() + 3);
	};
	
    //! clear vertex and color arrays
	virtual void clearTris() {_pVertices->clear();_pPrimitiveSet->setCount(0);_pColors->clear();};

protected:
    //! destructor
    ~TriContainerOSGGeometry() {};

    //! primitive set
	osg::ref_ptr<osg::DrawArrays> 	_pPrimitiveSet;

    //! vertices
	osg::ref_ptr<osg::Vec3Array> 	_pVertices;

    //! colors
    osg::ref_ptr<osg::Vec4Array> 	_pColors;

    //! default color
    osg::Vec4d _color;

    //! pointer to color table
    mds::cs::CColorTable<osg::Vec4d> * _colorTable;		
};

}
}

#endif
