//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2007-2010 by PGMed@FIT, BUT    \n
 *
 * Authors: Ondrej Siler, siler@fit.vutbr.cz    \n
 * Date:    yyyy/mm/dd                          \n
 *
 * $Id: mdsCrossSection.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Cross section of a plane and tetrahedral mesh.
 */

#ifndef MDS_CROSSSECTION_H
#define MDS_CROSSSECTION_H

//VCTL
#include <VectorEntity/mcentity.h>
#include <VectorEntity/mcvertices.h>
#include <VectorEntity/mctris.h>
#include <VectorEntity/mctetras.h>
#include <VectorEntity/mcplane.h>

//STL
#include <queue>

// tri container
#include <MDSTk/CrossSection/mdsTriContainer.h>

// namespace
namespace mds
{
namespace cs
{
/*! \brief Cross section of a plane and a tetrahedral mesh.
*/
class CCrossSection
{
public:    
	/*! \brief A default constructor*/
    CCrossSection()
        : _tetras(NULL), _firstFound(NULL), _radius(0.0), _flag(0), _tris(NULL) {};
		
	/*! \brief A constructor with a tetrahedral mesh and a starting tetrahedron.
        
        Searching for a first intersected tetrahedron starts from the tetrahedron passed to the constructor.
        If NULL or no tetraheron is passed searching starts from the first tetrahedron in the mesh.
    */
    CCrossSection(vctl::MCTetraS * tetras, vctl::MCTetra * startingTetra = NULL) 
        : _tetras(tetras), _firstFound(startingTetra), _flag(0), _tris(NULL) 
		{SetCenter(); 
		 ResetFlags(_flag);		 
		 };
	/*! \brief A constructor with a tetrahedral mesh, a plane and a tetrahedron.
        
        Searching for a first intersected tetrahedron starts from the tetrahedron passed to the constructor.
        If NULL or no tetraheron is passed searching starts from the first tetrahedron in the mesh.
    */
    CCrossSection(vctl::MCTetraS * tetras, vctl::MCPlane plane, vctl::MCTetra * startingTetra = NULL) 
        : _plane(plane), _tetras(tetras), _firstFound(startingTetra), _flag(0), _tris(NULL) 
		{SetCenter(); 
		 ResetFlags(_flag);		 
		};	
		

	/*! \brief A destructor.*/
    ~CCrossSection() {};
	
	/*! \brief Returns the plane*/
    vctl::MCPlane * GetPlane() {return & _plane;};
	/*! \brief Sets the plane*/
    void SetPlane(const vctl::MCPlane & plane) {_plane = plane;};
    /*! \brief Sets the tetraheral mesh*/
    void SetTetraS(vctl::MCTetraS * tetras) {_tetras = tetras; _firstFound = NULL; SetCenter(); _flag = 0; ResetFlags(_flag);};
    /*! \brief Sets a pointer to the container of triangles representing the cross section.	*/
    void SetTriContainer(CTriContainer * container) {_tris = container;};
	/*! \brief Returns the pointer to the container.*/
    CTriContainer * GetTris() {return _tris;};
    /*! \brief Computes a cross section if a tetrahedral mesh, a plane and a triangle container are set. Stores the triangular mesh to the container.*/
	void make();

protected:		
	/*! \brief A plane that intersects the tetrahedral mesh*/
    vctl::MCPlane _plane;    
    
    /*! \brief A pointer to a tetrahedral mesh */
    vctl::MCTetraS * _tetras;	
    /*! \brief A queue of tetrahedrons*/
    std::queue<vctl::MCTetra *> _tetraQueue;
	/*! \brief First tetrahedron intersected (or touched) with the plane*/
    vctl::MCTetra * _firstFound;
    /*! \brief Center of the mesh*/
    vctl::MCVertex _center;
	/*! \brief Radius of a bounding sphere of the mesh*/
    double _radius;
    /*! \brief A flag to mark tetrahedrons in the mesh to process*/    
    unsigned int _flag;
    /*! \brief A pointer to the container of triangles to store the result of the cross section*/
    CTriContainer * _tris;
    
    
    
    /*! \brief Makes cross section of the plane and one tetrahedron.*/
    bool MakeCrossSection(vctl::MCTetra * tetra);
   /*! \brief Returns a neighbour over a face that is nearest to the plane*/
    vctl::MCTetra * getClosestTetra(vctl::MCTetra * tetra);
	/*! \brief Sets the center and radius of a bounding sphere*/
    void SetCenter();
    /*! \brief Resets flag for each tetrahedron of the mesh*/
    void ResetFlags(unsigned int flag);
   
	void VertexIndicesToFaceIndices(int v1, int v2, unsigned char & f)
    {	     if(v1 == 0 && v2 == 1) f = 0x03;
	 	else if(v1 == 0 && v2 == 2) f = 0x09;
	 	else if(v1 == 0 && v2 == 3) f = 0x0a;
	 	else if(v1 == 1 && v2 == 2) f = 0x05;
		else if(v1 == 1 && v2 == 3) f = 0x06;
		else if(v1 == 2 && v2 == 3) f = 0x0c;}
		
private:
	vctl::MCVertex  _tmpVertex[6];
	vctl::MCVector3D _vec1;
	vctl::MCVector3D _vec2;
	vctl::MCVector3D _vec3; 
	
};

} //namespace
} //namespace

#endif
