//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2007-2010 by PGMed@FIT, BUT    \n
 *
 * Authors: Ondrej Siler, siler@fit.vutbr.cz    \n
 * Date:    yyyy/mm/dd                          \n
 *
 * $Id: mdsTriContainerGL.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Class for storing triangles created by cross section of tetrahedral mesh.
 */

#ifndef TRI_CONTAINER_GL_H
#define TRI_CONTAINER_GL_H

//vector entity
#include <MDSTk/CrossSection/mdsTriContainer.h>
#include <MDSTk/CrossSection/mdsColorTable.h>


#include <GL/gl.h>

#include <stdlib.h>

namespace mds
{
namespace cs
{

// RGBA color
struct colorGL4f
{
    colorGL4f() {};
    colorGL4f(GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha) : r(red), g(green), b(blue), a(alpha) {};
    
    GLfloat r, g, b, a;
};

//! triangle container for OpenGL. Colors and vertices coordinates are stored in arrays
class CTriContainerGL : public mds::cs::CTriContainer
{
public:
	//! constructor
	CTriContainerGL() 
    : m_defaultColor(1.0, 1.0, 1.0, 1.0), m_pcolorTable(NULL)
	{
		m_allocated_tris = 1024; 
		m_pvertices = (GLdouble *)malloc(sizeof(GLdouble)*m_allocated_tris*3*3); 
		m_pcolor = (GLfloat *)malloc(sizeof(GLfloat)*m_allocated_tris*4); 
		m_size = 0;
	}
	
	//! destructor
	virtual ~CTriContainerGL() {free(m_pvertices); free(m_pcolor);};
	
	//! does not delete data, resets the counter of stores triangles
	virtual void clearTris() {m_size = 0;};

	//! stores coordinates and color of the trinagle
	virtual void pushTri(vctl::MCVertex * v1, vctl::MCVertex * v2, vctl::MCVertex * v3, vctl::MCTetra * tetra)
	{
		//! changes the size of arrays
        if(m_size == m_allocated_tris)
		{
            m_allocated_tris *= 2;
			m_pvertices = (GLdouble *)realloc(m_pvertices, sizeof(GLdouble)*m_allocated_tris*3*3);
			m_pcolor = (GLfloat *)realloc(m_pcolor, sizeof(GLfloat)*m_allocated_tris*4); 
		}
		
		//! stores coordinates
		int i = m_size*3*3;
		v1->GetXYZ(m_pvertices[i   ], m_pvertices[i +1], m_pvertices[i +2]);
		v2->GetXYZ(m_pvertices[i +3], m_pvertices[i +4], m_pvertices[i +5]);
		v3->GetXYZ(m_pvertices[i +6], m_pvertices[i +7], m_pvertices[i +8]);
		
		//! get color of segment
        colorGL4f c = (m_pcolorTable) ? (*m_pcolorTable)[tetra->GetObject()] : m_defaultColor;

		//! stores the color of triangle
		i = m_size*4;
        m_pcolor[i   ] = c.r;
		m_pcolor[i +1] = c.g;
		m_pcolor[i +2] = c.b;
		m_pcolor[i +3] = c.a;
        		
        // increment counter of stored triangles
		m_size++;
	}

	//! returns number of triangles
    GLsizei getNumTriangles() {return m_size;};
    
    //! returns pointer to vertices coordinates
    GLdouble * getVertices() {return m_pvertices;};
    
    //! returns pointer to triangles colors
    GLfloat * getColors() {return m_pcolor;};
    
    //! sets the pointer to a color table
    void SetColorTable(mds::cs::CColorTable<colorGL4f> * colorTable)	{m_pcolorTable = colorTable;}
	
private:
	//! array of coordinates
	GLdouble * m_pvertices;
	
	//! array of colors
	GLfloat * m_pcolor;
	
	//! default color
	colorGL4f m_defaultColor;
	
	//! counter of triangles
	unsigned int m_size; 
	
	//! allocated memory
	unsigned int m_allocated_tris;
	
	//! pointer to color table
    mds::cs::CColorTable<colorGL4f> * m_pcolorTable;
};

} // namespace
} // namespace


#endif
