//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2006/04/03                          \n
 *
 * $Id: mdsDelaunayTetra.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - 3D Delaunay triangulation.
 */

#ifndef MDS_DELAUNAYTETRA_H
#define MDS_DELAUNAYTETRA_H

#include "mdsTetraData.h"

// MDSTk
#include <MDSTk/Base/mdsGlobalLog.h>
#include <MDSTk/Base/mdsSharedPtr.h>
#include <MDSTk/Math/mdsRandom.h>
#include <MDSTk/Module/mdsSerialization.h>

// VCTL
#include <VectorEntity/mctetras.h>

// STL
#include <vector>


namespace mds
{
namespace seg
{

//==============================================================================
/*
 * Global definitions.
 */

//! If this macro is defined, various informations are logged.
//#ifndef MDS_SEG_LOGGING_ENABLED
//#    define MDS_SEG_LOGGING_ENABLED
//#endif

//! Enables checking of tetrahedral mesh coherence.
//#ifndef MDS_CHECK_TETRAHEDRONS
//#    define MDS_CHECK_TETRAHEDRONS
//#endif

//! Enables checking of minimal edge length.
#ifndef MDS_CHECK_MIN_EDGE_LENGTH
#    define MDS_CHECK_MIN_EDGE_LENGTH
#endif


//==============================================================================
/*!
 * Class providing Delaunay triangulation in 3D space.
 */
class CDelaunayTetra : public mds::base::CObject, public mds::mod::CSerializable
{
public:
    //! Smart pointer type.
    //! - Declares type tSmartPtr.
    MDS_SHAREDPTR(CDelaunayTetra);

    //! Standard method getEntityName().
    MDS_ENTITY_NAME("DelaunayTetra");

    //! Standard method getEntityCompression().
    MDS_ENTITY_COMPRESSION(mds::mod::CC_RAW);

    //! Standard method getEntityBlockSize().
    MDS_ENTITY_BLOCK_SIZE(4096);

    //! Default methods parameters.
    static const mds::tSize DEFAULT_NUMBER_OF_NODES;

    //! Minimal length of a tetrahedron edge in the mesh.
    static const double DEFAULT_MIN_EDGE_LENGTH;

    //! Minimal dihedral angle used to check tetrahedra flatness.
    static const double DEFAULT_MIN_DIHEDRAL_ANGLE;

    //! Maximal allowed random permutation of vertex coordinates.
    static const double MAX_MOVEMENT;

    //! Flags used to mark boundary vertices.
    enum
    {
        XMIN_BOUNDARY   = 1 << 4,
        YMIN_BOUNDARY   = 1 << 5,
        ZMIN_BOUNDARY   = 1 << 6,
        XMAX_BOUNDARY   = 1 << 7,
        YMAX_BOUNDARY   = 1 << 8,
        ZMAX_BOUNDARY   = 1 << 9,
        BOUNDARY_FLAGS  = XMIN_BOUNDARY | YMIN_BOUNDARY | ZMIN_BOUNDARY
                        | XMAX_BOUNDARY | YMAX_BOUNDARY | ZMAX_BOUNDARY,
    };

    //! Tetrahedron processing function.
    typedef void (*tTetrahedronFunc)(vctl::MCTetra *pTetrahedron, CDelaunayTetra *pMesh);

    //! Function called to check minimal allowed edge length.
    typedef bool (*tMinEdgeLengthFunc)(vctl::MCPoint3D *p1, vctl::MCPoint3D *p2, CDelaunayTetra *pMesh);

    //! Function called to check constrained facets.
    typedef bool (*tConstraintFunc)(vctl::MCPoint3D *p1, vctl::MCPoint3D *p2, vctl::MCPoint3D *p3, CDelaunayTetra *pMesh);

public:
    //! Default constructor.
    CDelaunayTetra();

    //! Virtual destructor.
    virtual ~CDelaunayTetra() { clear(); }

    //! Returns current minimal allowed edge length.
    double getMinEdgeLength() { return m_dMinEdgeLength; }

    //! Sets minimal allowed edge length in the triangulation.
    void setMinEdgeLength(double dValue = DEFAULT_MIN_EDGE_LENGTH)
    {
        MDS_CHECK(dValue > 0.0, return);

        m_dMinEdgeLength = dValue;
    }

    //! Returns current minimal allowed dihedral angle [degrees].
    double getMinDihedralAngle() { return m_dMinDihedralAngle; }

    //! Sets minimal allowed dihedral angle in the triangulation [degrees].
    void setMinDihedralAngle(double dValue = DEFAULT_MIN_DIHEDRAL_ANGLE)
    {
        MDS_CHECK(dValue > 0.0, return);

        m_dMinDihedralAngle = dValue;
    }

    //! Sets function for checking minimal allowed edge length.
    void setMinEdgeLengthFunc(tMinEdgeLengthFunc MinEdgeLengthFunc)
    {
        MDS_CHECK(MinEdgeLengthFunc, return);

        m_MinEdgeLengthFunc = MinEdgeLengthFunc;
    }

    //! Sets default function for checking minimal allowed edge length.
    void setDefaultMinEdgeLengthFunc() { m_MinEdgeLengthFunc = checkMinEdgeLengthFunc; }


    //! Sets function called to check constrained facets.
    void setConstraintFunc(tConstraintFunc ConstraintFunc)
    {
        MDS_CHECK(ConstraintFunc, return);

        m_ConstraintFunc = ConstraintFunc;
    }

    //! Sets default function for constrained facets checking.
    void setDefaultConstraintFunc() { m_ConstraintFunc = checkConstraintFunc; }


    //! Modifies position of a given vertex.
    virtual bool attractVertex(vctl::MCPoint3D *pPoint);

    //! Returns pointer to a newly created vertex.
    vctl::MCVertex *newVertex(vctl::MCPoint3D *pPoint, int iFlags = 0)
    {
        vctl::MCVertex *pVertex = m_pVertices->New(pPoint->GetX(),
                                                   pPoint->GetY(),
                                                   pPoint->GetZ()
                                                   );
        pVertex->SetFlag(iFlags);
        return pVertex;
    }

    //! Returns pointer to a newly created vertex.
    vctl::MCVertex *newVertex(double x, double y, double z, int iFlags = 0)
    {
        vctl::MCVertex *pVertex = m_pVertices->New(x, y, z);
        pVertex->SetFlag(iFlags);
        return pVertex;
    }

    //! Generates randomly moved vertex.
    vctl::MCVertex *randomVertex(vctl::MCPoint3D *pPoint, double dMax = MAX_MOVEMENT)
    {
        return randomVertex(pPoint->GetX(), pPoint->GetY(), pPoint->GetZ(), dMax);
    }

    //! Generates randomly moved vertex.
    vctl::MCVertex *randomVertex(double x, double y, double z, double dMax = MAX_MOVEMENT);


    //! Returns pointer to the vertices container.
    vctl::MCVerticeS *getVertices() { return m_pVertices; }

    //! Returns the number of vertices.
    int getNumOfVertices() { return m_pVertices->GetNumber(); }

    //! Returns pointer to the first vertex.
    //! - Returns NULL if there is no vertex in the list.
    vctl::MCVertex *getFirstVertex() { return (vctl::MCVertex *)m_pVertices->GetFirst(); }

    //! Returns pointer to the container of tetrahedra.
    vctl::MCTetraS *getTetrahedra() { return &m_Tetrahedra; }

    //! Returns the number of all tetrahedra.
    int getNumOfTetrahedra() { return m_Tetrahedra.GetNumber(); }

    //! Returns pointer to the first tetrahedron.
    //! - Returns NULL if there is no tetrahedron in the list.
    vctl::MCTetra *getFirstTetrahedron() { return (vctl::MCTetra *)m_Tetrahedra.GetFirst(); }


    //! Calls function object for every vertex in the mesh.
    template <class Function>
    inline Function forEachVertex(Function Func)
    {
        vctl::MCVertex *pVertex = getFirstVertex();
        while( pVertex )
        {
            Func(*pVertex);
            pVertex = pVertex->GetNext();
        }
        return Func;
    }

    //! Calls function object for every tetrahedron in the mesh.
    template <class Function>
    inline Function forEachTetrahedron(Function Func)
    {
        vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
        while( pTetrahedron )
        {
            Func(*pTetrahedron);
            pTetrahedron = pTetrahedron->GetNext();
        }
        return Func;
    }


    //! Initializes the Delaunay triangulation.
    //! - Vertices pMin and pMax are points defining the minimal
    //!   and maximal vertex coordinates.
    //! - Removes all tetrahedra and vertices.
    bool init(vctl::MCPoint3D *pMin,
              vctl::MCPoint3D *pMax,
              mds::tSize NumOfNodes = DEFAULT_NUMBER_OF_NODES
              );

    //! Constructs the Delaunay triangulation from all vertices
    //! which are currently registered in the container of vertices.
    //! - Container of candidate vertices should be initiliazed first.
    //! - Incremental method is used.
    bool construct();

    //! Inserts a new vertex to the current triangulation satisfying
    //! the Delaunay criterion.
    //! - Functions called for each removed and newly inserted tetrahedron
    //!   can be specified.
    bool addVertex(vctl::MCVertex *pVertex,
                   int iBoundaryVertex = 0,
                   tTetrahedronFunc InsertFunc = NULL,
                   tTetrahedronFunc RemoveFunc = NULL
                   );

    //! Optimizes the triangulation.
    //! - Minimal required tetrahedron quality, the best value (= equilateral
    //!   tetrahedron) is equal to one.
    bool optimizeByVertexAdding(double dMinQuality, bool bRandomPosition = true);

    //! Removes all tetrahedra and vertices.
    void clear();


    //! Estimates minimal vertex coordinates in the mesh.
    bool getMinCoordinates(double& dX, double& dY, double& dZ);

    //! Estimates maximal vertex coordinates.
    bool getMaxCoordinates(double& dX, double& dY, double& dZ);


    //! Serializes the tetrahedron mesh.
    template <class S>
    void serialize(mds::mod::CChannelSerializer<S>& Writer)
    {
        // Remove all unused vertices
//        m_Tetrahedra.EraseVerticesNoTetras();

        // Serialize the triangular mesh
        m_Tetrahedra.serialize(Writer);
    }

    //! Deserializes all class members the tetrahedron mesh.
    template <class S>
    void deserialize(mds::mod::CChannelSerializer<S>& Reader)
    {
        // Clear the current mesh
        clear();

        // Deserialize the tetrahedral mesh
        m_Tetrahedra.deserialize(Reader);

        // Create info structs
        vctl::MCTetra *pTetra = getFirstTetrahedron();
        for( ; pTetra; pTetra = pTetra->GetNext() )
        {
            // Create the info struct
            STetrahedronInfo *pInfo = new STetrahedronInfo;
            pTetra->SetValuePtr(pInfo);

            // Get tetrahedron vertices
            vctl::MCVertex *pVertices[4];
            pTetra->GetVerticeS(pVertices);

            // Compute circumpshere
            try {
                pInfo->Circumsphere.MakeSphereTetra(pVertices[0], pVertices[1], pVertices[2], pVertices[3]);
            }
            catch( ... )
            {
                // Failed to compute circumsphere
                MDS_LOG_NOTE("CDelaunayTetra::deserialize(): Failed to compute tetrahedron circumsphere!");
                continue;
            }
        }

        // Minimal and maximal coordinates
        getMinCoordinates(m_dMinX, m_dMinY, m_dMinZ);
        getMaxCoordinates(m_dMaxX, m_dMaxY, m_dMaxZ);
    }

protected:
    //! Vector of pointers to tetrahedra.
    typedef std::vector<vctl::MCTetra *> tTetrahedra;

    //! Vector of pointers to vertices.
    typedef std::vector<vctl::MCVertex *> tVertices;

protected:
    //! Container of all tetrahedra in the mesh.
    vctl::MCTetraS m_Tetrahedra;

    //! Pointer to the container of vertices which is stored.
    //! in the container of vertices.
    vctl::MCVerticeS *m_pVertices;

    //! Random number generator.
    mds::math::CUniformPRNG m_Random;

    //! Minimal vertex coordinates in the mesh.
    double m_dMinX, m_dMinY, m_dMinZ;

    //! Maximal vertex coordinates in the mesh.
    double m_dMaxX, m_dMaxY, m_dMaxZ;

    //! Minimal allowed edge length.
    double m_dMinEdgeLength;

    //! Minimal allowed dihedral angle.
    double m_dMinDihedralAngle;

    //! Function for checking minimal length of an edge.
    tMinEdgeLengthFunc m_MinEdgeLengthFunc;

    //! Function for checking if a given facet is constrained.
    tConstraintFunc m_ConstraintFunc;

protected:
    //! Creates a new tetrahedron.
    //! - It is garanteed that normals of all triangles are oriented
    //!   to the tetrahedron center.
    //! - Returns pointer to the newly created tetrahedron.
    vctl::MCTetra *addTetrahedron(vctl::MCVertex *pV0,
                                  vctl::MCVertex *pV1,
                                  vctl::MCVertex *pV2,
                                  vctl::MCVertex *pV3
                                  );

    //! Removes a given existing tetrahedron.
    void removeTetrahedron(vctl::MCTetra *pTetrahedron, bool bEraseVertices = false);

    //! Checks if given vertices lie on a plane.
    //! - Returns true if the vertices are flat.
    bool checkFlatness(vctl::MCVertex *pV0,
                       vctl::MCVertex *pV1,
                       vctl::MCVertex *pV2,
                       vctl::MCVertex *pV3
                       );

    //! Checks if a tetrahedral mesh around a given tetrahedron is coherent.
    //! - Returns true if the mesh is OK.
    bool checkTetrahedron(vctl::MCTetra *pTetrahedron);

    //! Checks if a given vertex lies on the mesh boundary.
    //! - Normalizes vertex coordinates.
    //! - Returns appropriate boundary flag.
    int checkBoundary(vctl::MCPoint3D *pPoint);

    //! Uses a simple recursive tunneling algorithm to find out tetrahedron
    //! in which a given point lies.
    //! - Returns NULL on failure.
    vctl::MCTetra *findBaseTetrahedron(vctl::MCPoint3D *pPoint);

    //! Finds cavity of a given point.
    //! - Cavity means all tetrahedra whose circumsphere contains the point.
    //! - Note: It uses entity flag vctl::ENTITY_MARK_FLAG to stop recursion.
    //!   This flag is not cleared after the cavity is found.
    //! - Returns false if the point lies on circumsphere boundary
    //!   of some tetrahedron. In this case you should randomly move the point.
    bool findCavity(vctl::MCPoint3D *pPoint, tTetrahedra& Cavity);

    //! Finds triangles covering a given cavity.
    //! - Note: It uses entity flag vctl::ENTITY_MARK_FLAG to recognize
    //!   tetrahedra in the cavity. This flag is cleared after the envelope
    //!   is found.
    //! - Returns false on failure.
    bool findCavityEnvelope(const tTetrahedra& Cavity, tVertices& Envelope);

    //! Clears entity flags of all tetrahedra.
    void clearFlags();

    //! Clears entity flags of given tetrahedra.
    void clearFlags(const tTetrahedra& Tetrahedra);

    //! Computes tetrahedron quality.
    //! - Quotient of the longest tetrahedron edge and the inscribed
    //!   sphere radius.
    static double getTetrahedronQuality(vctl::MCTetra *pTetrahedron);

    //! Finds nearest vertex.
    //! - Returns NULL on failure.
    vctl::MCVertex *findNearestVertex(vctl::MCPoint3D *pPoint, bool bStartFromPrevious = false);

private:
    //! Same as above. The finding starts from a specified tetrahedron.
    vctl::MCTetra *findBaseTetrahedron(vctl::MCPoint3D *pPoint,
                                       vctl::MCTetra *pTetrahedron
                                       );

    //! Exhaustive search for a base tetrahedron in which a given point lies.
    //! - Returns NULL on failure.
    vctl::MCTetra *findBaseTetrahedronBruteForce(vctl::MCPoint3D *pPoint);

    //! Same as above. The finding starts from a specified tetrahedron.
    //! - Returns false if the point lies exactly on circumsphere boundary
    //!   of some tetrahedron. In this case you should move the point.
    bool findCavity(vctl::MCPoint3D *pPoint,
                    tTetrahedra& Cavity,
                    vctl::MCTetra *pTetrahedron
                    );

    //! Finds the base (one or more tetrahedra around a given vertex).
    //! The finding starts from a specified tetrahedron.
    bool findBase(vctl::MCPoint3D *pPoint,
                  tTetrahedra& Base,
                  vctl::MCTetra *pTetrahedron
                  );

    //! Finds the cavity starting from a given base.
    //! - Returns false if the point lies exactly on circumsphere
    //!   of some tetrahedron. In this case you should move the point
    //!   randomly.
    bool findCavityFromBase(vctl::MCPoint3D *pPoint,
                            tTetrahedra& Cavity,
                            const tTetrahedra& Base
                            );

private:
    //! Computes the tetrahedron quality and inserts it to the queue.
    static void normInsertFunc(vctl::MCTetra *pTetrahedron, CDelaunayTetra* pMesh);

    //! Unsorts the tetrahedron from the queue.
    static void normRemoveFunc(vctl::MCTetra *pTetrahedron, CDelaunayTetra* pMesh);

    //! Checks minimal edge length in the cavity.
    //! - Returns false if an edge is shorter then the minimal allowed edge length.
    static bool checkMinEdgeLengthFunc(vctl::MCPoint3D *p1,
                                       vctl::MCPoint3D *p2,
                                       CDelaunayTetra* pMesh
                                       );

    //! Checks if a given facet is constrained.
    static bool checkConstraintFunc(vctl::MCPoint3D *p1,
                                    vctl::MCPoint3D *p2,
                                    vctl::MCPoint3D *p3,
                                    CDelaunayTetra* pMesh
                                    );
};


//=============================================================================
/*
 * Basic template instances and type definitions.
 * - Using smart pointers.
 */

//! Tetrahedral mesh.
typedef CDelaunayTetra::tSmartPtr   CDelaunayTetraPtr;


} // namespace seg
} // namespace mds

#endif // MDS_DELAUNAYTETRA_H

