//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2005/07/26                          \n
 *
 * $Id: mdsDelaunayTri.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - 2D Delaunay triangulation.
 */

#ifndef MDS_DELAUNAYTRI_H
#define MDS_DELAUNAYTRI_H

#include "mdsTriData.h"

// MDSTk
#include <MDSTk/Base/mdsGlobalLog.h>
#include <MDSTk/Base/mdsSharedPtr.h>
#include <MDSTk/Math/mdsRandom.h>
#include <MDSTk/Module/mdsSerialization.h>

// VCTL
#include <VectorEntity/mctris.h>

// STL
#include <vector>


namespace mds
{
namespace seg
{

//==============================================================================
/*
 * Global definitions.
 */

//! If this macro is defined, various informations are logged.
//#ifndef MDS_SEG_LOGGING_ENABLED
//#    define MDS_SEG_LOGGING_ENABLED
//#endif


//==============================================================================
/*!
 * Class providing Delaunay triangulation.
 */
class CDelaunayTri : public mds::base::CObject, public mds::mod::CSerializable
{
public:
    //! Smart pointer type.
    //! - Declares type tSmartPtr.
    MDS_SHAREDPTR(CDelaunayTri);

    //! Standard method getEntityName().
    MDS_ENTITY_NAME("DelaunayTri");

    //! Standard method getEntityCompression().
    MDS_ENTITY_COMPRESSION(mds::mod::CC_RAW);

    //! Standard method getEntityBlockSize().
    MDS_ENTITY_BLOCK_SIZE(4096);

    //! Default methods parameters.
    static const tSize DEFAULT_NUMBER_OF_NODES;

    //! Minimal length of a triangle edge in the mesh.
    static const double DEFAULT_MIN_EDGE_LENGTH;

    //! Maximal allowed random permutation of vertex coordinates.
    static const double MAX_MOVEMENT;

    //! Flag used to mark boundary vertices.
    enum
    {
        XMIN_BOUNDARY   = 16,
        YMIN_BOUNDARY   = 32,
        XMAX_BOUNDARY   = 64,
        YMAX_BOUNDARY   = 128,
        BOUNDARY_FLAGS  = XMIN_BOUNDARY | YMIN_BOUNDARY | XMAX_BOUNDARY | YMAX_BOUNDARY,
    };

    //! Triangle processing function.
    typedef void (*tTriangleFunc)(vctl::MCTri *pTriangle, CDelaunayTri *pMesh);

public:
    //! Default constructor.
    CDelaunayTri();

    //! Virtual destructor.
    virtual ~CDelaunayTri() { clear(); }


    //! Returns pointer to a newly created vertex.
    vctl::MCVertex *newVertex(double x, double y)
    {
        return m_pVertices->New(x, y, 0.0);
    }
    vctl::MCVertex *newVertex(vctl::MCPoint3D *pPoint)
    {
        return m_pVertices->New(*pPoint);
    }

    //! Generates randomly moved vertex.
    vctl::MCVertex *randomVertex(vctl::MCPoint3D *pPoint, double dMax = MAX_MOVEMENT)
    {
        return randomVertex(pPoint->GetX(), pPoint->GetY(), dMax);
    }

    //! Generates randomly moved vertex.
    vctl::MCVertex *randomVertex(double x, double y, double dMax = MAX_MOVEMENT);

    //! Returns pointer to the vertices container.
    vctl::MCVerticeS *getVertices() { return m_pVertices; }

    //! Returns the number of vertices
    int getNumOfVertices() { return m_pVertices->GetNumber(); }

    //! Returns pointer to the first vertex
    //! - Returns NULL if there is no vertex in the list
    vctl::MCVertex *getFirstVertex() { return (vctl::MCVertex *)m_pVertices->GetFirst(); }

    //! Returns pointer to the vertices container
    vctl::MCTriS *getTriangles() { return &m_Triangles; }

    //! Returns the number of all triangles
    int getNumOfTriangles() { return m_Triangles.GetNumber(); }

    //! Returns pointer to the first triangle
    //! - Returns NULL if there is no triangle in the list
    vctl::MCTri *getFirstTriangle() { return (vctl::MCTri *)m_Triangles.GetFirst(); }


    //! Initializes the Delaunay triangulation
    //! - Vertices pMin and pMax are points defining the minimal
    //!   and maximal vertex coordinates.
    //! - Removes all triangles and vertices
    bool init(vctl::MCPoint3D *pMin,
              vctl::MCPoint3D *pMax,
              mds::tSize NumOfNodes = DEFAULT_NUMBER_OF_NODES,
              double dMinEdgeLength = DEFAULT_MIN_EDGE_LENGTH
              );

    //! Constructs Delaunay triangulation from all vertices
    //! which are currently registered in the container of vertices.
    //! - Container of candidate vertices should be initiliazed first. Value
    //!   of every vertex represents its significance later used for sorting
    //!   vertices in decreasing order.
    //! - Incremental method is used.
    bool construct();

    //! Inserts a new vertex to the current triangulation satisfying
    //! the Delaunay criterion.
    //! - Functions called for each removed and newly inserted triangle
    //!   can be specified.
    bool addVertex(vctl::MCVertex *pVertex,
                   int bBoundaryVertex = 0,
                   tTriangleFunc InsertFunc = NULL,
                   tTriangleFunc RemoveFunc = NULL
                   );

    //! Normalizes the triangular mesh
    //! - Minimal required triangle quality, the best value (= equilateral
    //!   triangle) is equal to one.
    //! - If the longest triangle edge is shorter then a given value,
    //!   the triangle is not optimized.
    bool normalize(double dMinQuality, double dMinEdgeLength);

    //! Removes all triangles and vertices.
    void clear();


    //! Estimates minimal vertex coordinates in the mesh
    bool getMinCoordinates(double& dX, double& dY);

    //! Estimates maximal vertex coordinates
    bool getMaxCoordinates(double& dX, double& dY);


    //! Serializes the triangular mesh.
    template <class S>
    void serialize(mds::mod::CChannelSerializer<S>& Writer)
    {
        // Remove all unused vertices
//        m_Triangles.EraseVerticesNoTris();

        // Serialize the triangular mesh
        m_Triangles.serialize(Writer);
    }

    //! Deserializes the triangular mesh.
    template <class S>
    void deserialize(mds::mod::CChannelSerializer<S>& Reader)
    {
        // Clear the current mesh
        clear();

        // Deserialize the triangular mesh
        m_Triangles.deserialize(Reader);

        // Create info structs
        vctl::MCTri *pTriangle = getFirstTriangle();
        for( ; pTriangle; pTriangle = pTriangle->GetNext() )
        {
            // Create the info struct
            STriangleInfo *pInfo = new STriangleInfo;
            pTriangle->SetValuePtr(pInfo);

            // Get triangle vertices
            vctl::MCVertex *pVertices[3];
            pTriangle->GetVerticeS(pVertices);

            // Compute triangle's circumpshere
            try {
                pInfo->Circumsphere.MakeSphereTri(pVertices[0], pVertices[1], pVertices[2]);
            }
            catch( ... )
            {
                // Failed to compute circumsphere
                MDS_LOG_NOTE("CDelaunayTri::deserialize(): Failed to compute triangle circumsphere!");
                return;
            }
        }

        // Minimal and maximal coordinates
        getMinCoordinates(m_dMinX, m_dMinY);
        getMaxCoordinates(m_dMaxX, m_dMaxY);
    }

protected:
    //! Vector of pointers to triangles
    typedef std::vector<vctl::MCTri *> tTriangles;

    //! Vector of pointers to vertices
    typedef std::vector<vctl::MCVertex *> tVertices;

protected:
    //! Container of all triangles in the mesh
    vctl::MCTriS m_Triangles;

    //! Pointer to the container of vertices which is stored
    //! in the triangular mesh.
    vctl::MCVerticeS *m_pVertices;

    //! Random number generator.
    mds::math::CUniformPRNG m_Random;

    //! Minimal vertex coordinates in the mesh.
    double m_dMinX, m_dMinY;

    //! Maximal vertex coordinates in the mesh.
    double m_dMaxX, m_dMaxY;

protected:
    //! Creates a new triangle
    //! - It is garanteed that normal of the triangle is coincident
    //!   with the z-axis (0,0,1).
    //! - Returns pointer to the newly created triangle
    vctl::MCTri *addTriangle(vctl::MCVertex *pV0,
                             vctl::MCVertex *pV1,
                             vctl::MCVertex *pV2
                             );

    //! Removes a given existing triangle
    void removeTriangle(vctl::MCTri *pTriangle, bool bEraseVertices = false);

    //! Checks if given vertices lie on a line.
    bool checkFlatness(vctl::MCVertex *pV0,
                       vctl::MCVertex *pV1,
                       vctl::MCVertex *pV2
                       );

    //! Checks if a given vertex lies on the mesh boundary.
    //! - Normalizes vertex coordinates.
    //! - Returns appropriate boundary flag.
    int checkBoundary(vctl::MCPoint3D *pPoint);

    //! Uses a simple recursive tunneling algorithm to find out triangle
    //! in which a given point lies
    //! - Returns NULL on failure
    vctl::MCTri *findBaseTriangle(vctl::MCPoint3D *pPoint);

    //! Finds cavity of a given point
    //! - Cavity means all triangles whose circumcircle contains the point
    //! - Note: It uses entity flag vctl::ENTITY_MARK_FLAG to stop recursion.
    //!   This flag is not cleared after the cavity is found.
    //! - Returns false if there is no such triangle
    bool findCavity(vctl::MCPoint3D *pPoint, tTriangles& Cavity);

    //! Finds edges covering a given cavity
    //! - Note: It uses entity flag vctl::ENTITY_MARK_FLAG to recognize
    //!   triangles in the cavity. This flag is cleared after the envelope
    //!   is found.
    //! - Returns false on failure
    bool findCavityEnvelope(const tTriangles& Cavity, tVertices& Envelope);

    //! Clears entity flags of all triangles
    void clearFlags();

    //! Clears entity flags of given triangles
    void clearFlags(const tTriangles& Triangles);

private:
    //! Same as above. The finding starts from a specified triangle.
    vctl::MCTri *findBaseTriangle(vctl::MCPoint3D *pPoint,
                                  vctl::MCTri *pTriangle
                                  );

    //! Same as above. The finding starts from a specified triangle.
    bool findCavity(vctl::MCPoint3D *pPoint,
                    tTriangles& Cavity,
                    vctl::MCTri *pTriangle
                    );

private:
    //! Computes triangle quality.
    //! - Quotient of the longest triangle edge and the inscribed
    //!   circle radius.
    static double getTriangleQuality(vctl::MCTri *pTriangle);

    //! Computes the triangle quality and inserts it to the queue.
    static void normInsertFunc(vctl::MCTri *pTriangle, CDelaunayTri* pMesh);

    //! Unsorts the triangle from the queue.
    static void normRemoveFunc(vctl::MCTri *pTriangle, CDelaunayTri* pMesh);
};


//=============================================================================
/*
 * Basic template instances and type definitions.
 * - Using smart pointers.
 */

//! Triangular mesh.
typedef CDelaunayTri::tSmartPtr CDelaunayTriPtr;


} // namespace seg
} // namespace mds

#endif // MDS_DELAUNAYTRI_H

