//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2006/09/24                          \n
 *
 * $Id: mdsTetraContainer.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Container encapsulating list of tetrahedra.
 */

#ifndef MDS_TETRACONTAINER_H
#define MDS_TETRACONTAINER_H

#include "mdsTetraFeatures.h"

// MDSTk
#include <MDSTk/Base/mdsIterator.h>

// VCTL
#include <VectorEntity/mcentity.h>


namespace mds
{

// Declare iterator traits first...
MDS_ITERATOR_TRAITS(seg, CTetraContainerIterator, seg::CTetraFeatures);

namespace seg
{

//==============================================================================
/*!
 * Iterator used to traverse feature vectors of all tetrahedra in a list.
 */
class CTetraContainerIterator
    : public mds::base::CIteratorBase<CTetraContainerIterator>
    , public mds::base::CPtrHolder<CTetraFeatures>
{
public:
    //! Iterator base.
    typedef mds::base::CPtrHolder<CTetraFeatures> base;

public:
    //! Constructor.
    CTetraContainerIterator(vctl::MCTetra *pFirstTetrahedron)
    {
        init(pFirstTetrahedron);
    }

    //! Copy constructor.
    CTetraContainerIterator(const CTetraContainerIterator& It)
    {
        base::m_pItem = It.m_pItem;
        m_pTetrahedron = It.m_pTetrahedron;
    }

    //! Destructor.
    ~CTetraContainerIterator() {}

    //! Assignment operator.
    CTetraContainerIterator& operator=(const CTetraContainerIterator& It)
    {
        base::m_pItem = It.m_pItem;
        m_pTetrahedron = It.m_pTetrahedron;
        return *this;
    }


    //! Returns pointer to the tetrahedron.
    vctl::MCTetra *getTetrahedron() { return m_pTetrahedron; }

    //! Returns the number of increments between the positions addressed.
    //! by two iterators.
    tSize getDistance(const CTetraContainerIterator& End) const
    {
        if( !m_pTetrahedron )
        {
            return 0;
        }
        
        // Count the distance
        tSize Count = 0;
        for( CTetraContainerIterator it = *this; it != End; ++it )
        {
            ++Count;
        }
        return Count;
    }

    //! Moves iterator to the next tetrahedron.
    void advance()
    {
        do {
            if( !base::m_pItem )
            {
                break;
            }
            else
            {
                next();
            }
        } while( base::m_pItem && base::m_pItem->isError() );
    }

protected:
    //! Pointer to the actual Tetrahedron.
    vctl::MCTetra *m_pTetrahedron;

protected:
    //! Initializes the tetrahedron iterator.
    void init(vctl::MCTetra *pTetrahedron)
    {
        m_pTetrahedron = pTetrahedron;
        if( !m_pTetrahedron )
        {
            m_pItem = NULL;
            return;
        }
        
        m_pItem = getTetrahedronFeatures(pTetrahedron);

        // Find the first evaluated tetrahedron.
        if( m_pItem->isError() )
        {
            advance();
        }
    }

    //! Moves iterator to the next tetrahedron.
    void next()
    {
        m_pTetrahedron = m_pTetrahedron->GetNext();
        m_pItem = (m_pTetrahedron) ? getTetrahedronFeatures(m_pTetrahedron) : NULL;
    }
};


//==============================================================================
/*!
 * Container encapsulating list of tetrahedra.
 */
class CTetraContainer
{
public:
    //! Iterator is declared!
    enum { ITERATOR_DECLARED };
    
    //! Iterator type.
    typedef CTetraContainerIterator tIterator;
    
    //! Const iterator type.
    typedef CTetraContainerIterator tConstIterator;

public:
    //! Constructor.
    CTetraContainer(vctl::MCTetra *pFirstTetrahedron) : m_pTetrahedron(pFirstTetrahedron)
    {
        MDS_ASSERT(m_pTetrahedron);
    }

    //! Destructor.
    ~CTetraContainer() {}


    //! Returns iterator that points at the first tetrahedron.
    tIterator getBegin()
    {
        return tIterator(m_pTetrahedron);
    }
    tConstIterator getBegin() const
    {
        return tConstIterator(m_pTetrahedron);
    }

    //! Returns iterator that points beyond the last tetrahedron in the list.
    tIterator getEnd()
    {
        return tIterator(NULL);
    }
    tConstIterator getEnd() const
    {
        return tConstIterator(NULL);
    }

protected:
    //! Pointer to the first Tetrahedron.
    vctl::MCTetra *m_pTetrahedron;
};


} // namespace seg
} // namespace mds

#endif // MDS_TETRACONTAINER_H

