//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2006/04/20                          \n
 *
 * $Id: mdsTetraEdgeIterator.h 1623 2010-02-11 14:23:49Z spanel $
 *
 * Description:
 * - Tetrahedron edge pixels iterator.
 */

#ifndef MDS_TETRAEDGEITERATOR_H
#define MDS_TETRAEDGEITERATOR_H

#include "mdsTetraBase.h"

// MDSTk
#include <MDSTk/Base/mdsMemory.h>
#include <MDSTk/Base/mdsIterator.h>
#include <MDSTk/Math/mdsBase.h>
#include <MDSTk/Image/mdsVector3.h>

// VCTL
#include <VectorEntity/mcentity.h>


namespace mds
{

// Declare iterator traits first...
MDS_ITERATOR_TRAITS(seg, CTetraEdgeIterator, seg::CTetraBase::tVoxel);
MDS_ITERATOR_TRAITS(seg, CTetraEdgeDDAIterator, seg::CTetraBase::tVoxel);

namespace seg
{

//==============================================================================
/*!
 * Iterator used to traverse points lying on a tetrahedron edge.
 * - TODO: forEach() concept.
 */
class CTetraEdgeIterator
    : public mds::base::CIteratorBase<CTetraEdgeIterator>
    , public mds::base::CContainerHolder<CTetraBase::tVolume, CTetraBase::tVoxel>
{
public:
    //! Iterator base.
    typedef mds::base::CContainerHolder<CTetraBase::tVolume, CTetraBase::tVoxel> base;

    //! Base types.
    typedef CTetraBase::tVolume tVolume;
    typedef CTetraBase::tVoxel tVoxel;
    typedef CTetraBase::tScale tScale;

public:
    //! Constructor.
    CTetraEdgeIterator(tVolume *pVolume, vctl::MCEdge *pEdge, tScale *pScale)
        : m_pScale(pScale)
    {
        MDS_ASSERT(pScale && pEdge);
        
        // Get vertices of the edge
        vctl::MCVertex *pVertices[2];
        pEdge->GetVerticeS(pVertices);

        init(pVolume, pVertices[0], pVertices[1]);
    }

    //! Constructor.
    CTetraEdgeIterator(tVolume *pVolume,
                       vctl::MCPoint3D *p1,
                       vctl::MCPoint3D *p2,
                       tScale *pScale
                       )
        : m_pScale(pScale)
    {
        MDS_ASSERT(pScale);
        
        init(pVolume, p1, p2);
    }

    //! Copy constructor
    CTetraEdgeIterator(const CTetraEdgeIterator& It)
        : base(It)
        , m_pScale(It.m_pScale)
    {
        mds::base::memCopy(&m_Impl, &It.m_Impl, 1);
    }

    //! Destructor
    ~CTetraEdgeIterator() {}

    //! Assignment operator
    CTetraEdgeIterator& operator=(const CTetraEdgeIterator& It)
    {
        base::m_pContainer = It.m_pContainer;
        base::m_pItem = It.m_pItem;
        mds::base::memCopy(&m_Impl, &It.m_Impl, 1);
        m_pScale = It.m_pScale;
        return *this;
    }


    //! Returns current iterator position.
    mds::tSize getX() const { return *m_Impl.m_pX; }
    mds::tSize getY() const { return *m_Impl.m_pY; }
    mds::tSize getZ() const { return *m_Impl.m_pZ; }

    //! Returns current real iterator position.
    double getRealX() const { return m_pScale->getRealX(*m_Impl.m_pX); }
    double getRealY() const { return m_pScale->getRealY(*m_Impl.m_pY); }
    double getRealZ() const { return m_pScale->getRealZ(*m_Impl.m_pZ); }

    //! Returns true if iterator points after the last edge pixel.
    bool isEnd() const { return (m_Impl.m_X > m_Impl.m_MaxX); }

protected:
    //! All data members.
    struct SDataMembers
    {
        //! Steps in y and z direction.
        mds::tSize m_StepY, m_StepZ;

        //! Parameters of the Bresenham's algorithm.
        mds::tSize m_K1Y, m_K2Y, m_PY;
        mds::tSize m_K1Z, m_K2Z, m_PZ;

        //! Maximal X coordinate of the edge.
        mds::tSize m_MaxX;

        //! Pixel position.
        mds::tSize m_X, m_Y, m_Z;

        //! Iterator position
        mds::tSize *m_pX, *m_pY, *m_pZ;
    };

    //! Data members.
    SDataMembers m_Impl;

    //! Pointer to the tetrahedron scaling object.
    tScale *m_pScale;

public:
    //! Moves iterator to the next edge pixel.
    void advance()
    {
        if( isEnd() )
        {
            return;
        }

        ++m_Impl.m_X;

        if( m_Impl.m_PY < 0 )
        {
            m_Impl.m_PY += m_Impl.m_K1Y;
        }
        else
        {
            m_Impl.m_PY += m_Impl.m_K2Y;
            m_Impl.m_Y += m_Impl.m_StepY;
        }

        if( m_Impl.m_PZ < 0 )
        {
            m_Impl.m_PZ += m_Impl.m_K1Z;
        }
        else
        {
            m_Impl.m_PZ += m_Impl.m_K2Z;
            m_Impl.m_Z += m_Impl.m_StepZ;
        }

        m_pItem = m_pContainer->getPtr(*m_Impl.m_pX, *m_Impl.m_pY, *m_Impl.m_pZ);
    }

protected:
    //! Initializes the edge iterator.
    void init(tVolume *pVolume, vctl::MCPoint3D *p1, vctl::MCPoint3D *p2)
    {
        MDS_ASSERT(pVolume && p1 && p2);

        // Round the line end-point positions
        int iX1, iY1, iZ1, iX2, iY2, iZ2;
        m_pScale->conv2Volume(p1, iX1, iY1, iZ1);
        m_pScale->conv2Volume(p2, iX2, iY2, iZ2);

        // Difference in x, y and z axes
        int iDX = mds::math::getAbs(iX2 - iX1);
        int iDY = mds::math::getAbs(iY2 - iY1);
        int iDZ = mds::math::getAbs(iZ2 - iZ1);

        // Exchange x and y axes
        if( iDY >= iDX && iDY >= iDZ )
        {
            m_Impl.m_pX = &m_Impl.m_Y;
            m_Impl.m_pY = &m_Impl.m_X;
            m_Impl.m_pZ = &m_Impl.m_Z;

            mds::math::swap(iX1, iY1);
            mds::math::swap(iX2, iY2);
            mds::math::swap(iDX, iDY);
        }
        else if ( iDZ >= iDX && iDZ >= iDY )
        {
            m_Impl.m_pX = &m_Impl.m_Z;
            m_Impl.m_pY = &m_Impl.m_Y;
            m_Impl.m_pZ = &m_Impl.m_X;

            mds::math::swap(iX1, iZ1);
            mds::math::swap(iX2, iZ2);
            mds::math::swap(iDX, iDZ);
        }
        else
        {
            m_Impl.m_pX = &m_Impl.m_X;
            m_Impl.m_pY = &m_Impl.m_Y;
            m_Impl.m_pZ = &m_Impl.m_Z;
        }

        // Exchange line end-points
        if( iX1 > iX2 )
        {
            mds::math::swap(iX1, iX2);
            mds::math::swap(iY1, iY2);
            mds::math::swap(iZ1, iZ2);
        }

        // Step in y and z axis
        if( iY1 > iY2 )
        {
            m_Impl.m_StepY = -1;
        }
        else
        {
            m_Impl.m_StepY = 1;
        }

        if( iZ1 > iZ2 )
        {
            m_Impl.m_StepZ = -1;
        }
        else
        {
            m_Impl.m_StepZ = 1;
        }

        // Bresenham's algorithm
        m_Impl.m_K1Y = 2 * iDY;
        m_Impl.m_K2Y = 2 * (iDY - iDX);
        m_Impl.m_PY  = 2 * iDY - iDX;

        m_Impl.m_K1Z = 2 * iDZ;
        m_Impl.m_K2Z = 2 * (iDZ - iDX);
        m_Impl.m_PZ  = 2 * iDZ - iDX;

        // Starting point
        m_Impl.m_X = iX1;
        m_Impl.m_Y = iY1;
        m_Impl.m_Z = iZ1;

        // End-point
        m_Impl.m_MaxX = iX2;

        // Save pointer to the volume
        m_pContainer = pVolume;

        // Initialize pointer to the first pixel
        m_pItem = pVolume->getPtr(*m_Impl.m_pX, *m_Impl.m_pY, *m_Impl.m_pZ);
    }
};


//==============================================================================
/*!
 * Iterator used to traverse points lying on a tetrahedron edge.
 * - TODO: forEach() concept.
 */
class CTetraEdgeDDAIterator
    : public mds::base::CIteratorBase<CTetraEdgeDDAIterator>
    , public mds::base::CContainerHolder<CTetraBase::tVolume, CTetraBase::tVoxel>
{
public:
    //! Iterator base.
    typedef mds::base::CContainerHolder<CTetraBase::tVolume, CTetraBase::tVoxel> base;

    //! Base types.
    typedef CTetraBase::tVolume tVolume;
    typedef CTetraBase::tVoxel tVoxel;
    typedef CTetraBase::tScale tScale;

public:
    //! Constructor.
    CTetraEdgeDDAIterator(tVolume *pVolume, vctl::MCEdge *pEdge, tScale *pScale)
        : m_pScale(pScale)
    {
        MDS_ASSERT(pScale && pEdge);
        
        // Get vertices of the edge
        vctl::MCVertex *pVertices[2];
        pEdge->GetVerticeS(pVertices);
        
        init(pVolume, pVertices[0], pVertices[1]);
    }

    //! Constructor.
    CTetraEdgeDDAIterator(tVolume *pVolume,
                          vctl::MCPoint3D *p1,
                          vctl::MCPoint3D *p2,
                          tScale *pScale
                          )
        : m_pScale(pScale)
    {
        MDS_ASSERT(pScale);
        
        init(pVolume, p1, p2);
    }

    //! Copy constructor
    CTetraEdgeDDAIterator(const CTetraEdgeDDAIterator& It)
        : base(It)
        , m_pScale(It.m_pScale)
    {
        mds::base::memCopy(&m_Impl, &It.m_Impl, 1);
    }

    //! Destructor
    ~CTetraEdgeDDAIterator() {}

    //! Assignment operator
    CTetraEdgeDDAIterator& operator=(const CTetraEdgeDDAIterator& It)
    {
        base::m_pContainer = It.m_pContainer;
        base::m_pItem = It.m_pItem;
        mds::base::memCopy(&m_Impl, &It.m_Impl, 1);
        m_pScale = It.m_pScale;
        return *this;
    }


    //! Returns current iterator position.
    mds::tSize getX() const { return mds::math::round2Int(m_Impl.m_Position.getX()); }
    mds::tSize getY() const { return mds::math::round2Int(m_Impl.m_Position.getY()); }
    mds::tSize getZ() const { return mds::math::round2Int(m_Impl.m_Position.getZ()); }

    //! Returns current real iterator position.
    double getRealX() const { return m_pScale->getRealX(m_Impl.m_Position.getX()); }
    double getRealY() const { return m_pScale->getRealY(m_Impl.m_Position.getY()); }
    double getRealZ() const { return m_pScale->getRealZ(m_Impl.m_Position.getZ()); }

    //! Returns true if iterator points after the last edge pixel.
//    bool isEnd() const { return (m_Impl.m_Position.getX() > m_Impl.m_dMaxX); }
    bool isEnd() const { return mds::math::getAbs(m_Impl.m_Position.getX() - m_Impl.m_dMaxX) < 1.0; }

protected:
    //! All data members.
    struct SDataMembers
    {
        //! Steps in x, y and z direction.
        mds::img::CVector3D m_Step;

        //! Pixel position.
        mds::img::CPoint3D m_Position;

        //! Maximal x coordinate of the edge.
        double m_dMaxX;

        //! Current interpolated pixel value.
        tVoxel m_Value;
    };

    //! Data members.
    SDataMembers m_Impl;

    //! Pointer to the tetrahedron scaling object.
    tScale *m_pScale;

public:
    //! Moves iterator to the next edge pixel.
    void advance()
    {
        if( isEnd() )
        {
            return;
        }

        // Update the position
        m_Impl.m_Position.x() += m_Impl.m_Step.x();
        m_Impl.m_Position.y() += m_Impl.m_Step.y();
        m_Impl.m_Position.z() += m_Impl.m_Step.z();

        // Initialize value of the first pixel
        m_Impl.m_Value = m_pContainer->interpolate(m_Impl.m_Position);
    }

protected:
    //! Initializes the edge iterator.
    void init(tVolume *pVolume, vctl::MCPoint3D *p1, vctl::MCPoint3D *p2)
    {
        MDS_ASSERT(pVolume && p1 && p2);

        // Round the line end-point positions
        int iX1, iY1, iZ1, iX2, iY2, iZ2;
        m_pScale->conv2Volume(p1, iX1, iY1, iZ1);
        m_pScale->conv2Volume(p2, iX2, iY2, iZ2);

        // Exchange line end-points
/*        if( iX1 > iX2 )
        {
            mds::math::swap(iX1, iX2);
            mds::math::swap(iY1, iY2);
            mds::math::swap(iZ1, iZ2);
        }*/

        // Steps in x, y and z axis
        m_Impl.m_Step.setX(iX2 - iX1);
        m_Impl.m_Step.setY(iY2 - iY1);
        m_Impl.m_Step.setZ(iZ2 - iZ1);

/*        m_Impl.m_Step.normalize();
        m_Impl.m_Step.x *= 0.5;
        m_Impl.m_Step.y *= 0.5;
        m_Impl.m_Step.z *= 0.5;
        m_Impl.m_Step.x += (m_Impl.m_Step.x > 0.0) ? 0.0000001 : -0.0000001;*/
        
        double dX = mds::math::getAbs(m_Impl.m_Step.getX());
        double dY = mds::math::getAbs(m_Impl.m_Step.getY());
        double dZ = mds::math::getAbs(m_Impl.m_Step.getZ());
        
        double dMax = mds::math::getMax(dX, dY, dZ);
        if( dMax < 1.0 )
        {
            m_Impl.m_Step.setX(1.0);
            m_Impl.m_Step.setY(1.0);
            m_Impl.m_Step.setZ(1.0);
        }
        else
        {
            double dInvMax = 1.0 / dMax;
            m_Impl.m_Step *= dInvMax;
        }
        
        // Starting point
        m_Impl.m_Position.setX(iX1);
        m_Impl.m_Position.setY(iY1);
        m_Impl.m_Position.setZ(iZ1);
        
        // End-point
        m_Impl.m_dMaxX = iX2;
        
        // Save pointer to the volume
        m_pContainer = pVolume;
        m_pItem = &(m_Impl.m_Value);
        
        // Initialize value of the first pixel
        m_Impl.m_Value = m_pContainer->interpolate(m_Impl.m_Position);
    }
};


} // namespace seg
} // namespace mds

#endif // MDS_TETRAEDGEITERATOR_H

