//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2006/04/04                          \n
 *
 * $Id: mdsTetraFeatures.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Tetrahedron feature vector characterizing image data.
 * - May be extracted for just one tetrahedron or for a group of tetrahedra.
 */

#ifndef MDS_TETRAFEATURES_H
#define MDS_TETRAFEATURES_H

#include "mdsTetraBase.h"
#include "mdsTetraData.h"

// MDSTk
#include <MDSTk/Base/mdsGlobalLog.h>
#include <MDSTk/Math/mdsStaticVector.h>

// VCTL
#include <VectorEntity/mcentity.h>

// STL
#include <string>
#include <ostream>


namespace mds
{
namespace seg
{

//! If the macro is defined, the variance is used for classification.
//#define USE_VARIANCE

// TODO: Predelat do flagu! Podedit z toolkitu!

//! Enables substitution of very small tetrahedra (mean value and variance
//! be evaluated) by voxels lying in the tetrahedron center.
#define SUBST_SMALL_TETRAS

//! Skips all edge voxels.
//#define SKIP_EDGE_VOXELS

//! Skips all tetrahedra adjacent to domain boundary.
//#define SKIP_BOUNDARY_TETRAS


//==============================================================================
/*
 * Global constants.
 */

namespace TetraFeatures
{

//! The number of basic tetrahedron features.
#ifdef USE_VARIANCE
    const int NUM_OF_BASIC_FEATURES     = 2;
#else
    const int NUM_OF_BASIC_FEATURES     = 1;
#endif // USE_VARIANCE

//! Total number of all tetrahedron features.
const int NUM_OF_FEATURES               = NUM_OF_BASIC_FEATURES;

//! Various constants.
//const int MIN_NUM_OF_VOXELS             = 16;
//const int MIN_NUM_OF_VOXELS             = 4;
const int MIN_NUM_OF_VOXELS             = 1;

}


//==============================================================================
/*!
 * Class encapsulating all triangle features.
 */
class CTetraFeatures : public CTetraBase, public mds::math::CStaticVector<double, TetraFeatures::NUM_OF_FEATURES>
{
public:
    //! Type of a value representing some feature.
    typedef double tFeature;

    //! Type of the used static vector.
    typedef mds::math::CStaticVector<double, TetraFeatures::NUM_OF_FEATURES> tVector;

    //! The number of features.
    enum { NUM_OF_FEATURES = TetraFeatures::NUM_OF_FEATURES };

    //! Indexes of the most important features.
    enum
    {
        MEAN_VALUE = 0,
        VARIANCE,
    };

    //! Error codes.
    enum
    {
        ERROR_NOT_EVALUATED     = 1,
        ERROR_CANNOT_EVALUATE   = 2,
    };

public:
    //! Default constructor.
    CTetraFeatures() : m_iError(ERROR_NOT_EVALUATED), m_iNumOfVoxels(0) {}

    //! Destructor.
    ~CTetraFeatures() {}


    //! Returns value of the i-th feature.
    tFeature& getFeature(mds::tSize i) { return get(i); }
    const tFeature& getFeature(tSize i) const { return get(i); }

    //! Sets value of the i-th feature.
    void setFeature(tSize i, tFeature Value) { set(i, Value); }


    //! Returns the number of voxels taken into account.
    int getNumOfVoxels() const { return m_iNumOfVoxels; }

    //! Sets the number of voxels.
    void setNumOfVoxels(int iValue) { m_iNumOfVoxels = iValue; }


    //! Returns true if the previous evaluation of features failed.
    bool isError() const { return (m_iError != 0); }

    //! Returns current error code.
    int getError() const { return m_iError; }

    //! Sets the error code.
    void setError(int iError) { m_iError = iError; }


    //! Returns mean value of all voxels in the tetrahedron.
    tFeature& getVoxelMean() { return get(MEAN_VALUE); }
    const tFeature& getVoxelMean() const { return get(MEAN_VALUE); }

    //! Sets the voxel mean.
    void setVoxelMean(tFeature Value) { set(MEAN_VALUE, Value); }

#ifdef USE_VARIANCE
    //! Returns variance of all voxels in the tetrahedron.
    tFeature& getVoxelVariance() { return get(VARIANCE); }
    const tFeature& getVoxelVariance() const { return get(VARIANCE); }

    //! Sets the voxel variance.
    void setVoxelVariance(tFeature Value) { set(VARIANCE, Value); }
#else
    //! Returns variance of all voxels in the tetrahedron.
    tFeature& getVoxelVariance() { return m_VoxelVariance; }
    const tFeature& getVoxelVariance() const { return m_VoxelVariance; }

    //! Sets the voxel variance.
    void setVoxelVariance(tFeature Value) { m_VoxelVariance = Value; }
#endif // USE_VARIANCE

protected:
    //! Errors while evaluating the features.
    int m_iError;

    //! Total number of voxels.
    int m_iNumOfVoxels;

#ifndef USE_VARIANCE
    //! Variance of the voxel values.
    tFeature m_VoxelVariance;
#endif // USE_VARIANCE
};


//==============================================================================
/*!
 * Class used to extract image features for a tetrahedron.
 */
class CTetraFeaturesExtractor : public CTetraBase
{
public:
    //! Type representing feature.
    typedef CTetraFeatures::tFeature tFeature;

    //! Feature vector.
    typedef CTetraFeatures tVector;

public:
    //! Default constructor.
    CTetraFeaturesExtractor() { clear(); }

    //! Destructor.
    ~CTetraFeaturesExtractor() {}

    //! Extracts and acumulates image charakteristics from given tetrahedron.
    void acumulate(tVolume *pVolume, vctl::MCTetra *pTetrahedron, tScale *pScale);

    //! Acumulates image charakteristics from a given feature extractor.
    void acumulate(const CTetraFeaturesExtractor& Extractor);

    //! Evaluates features according to the acumulated image characteristics
    //! and writes result to a given feature vector.
    //! - Returns false on failure.
    //! - Concrete error code is saved in the feature vector.
    bool evaluate(tVector *pVector, bool bCheckNumOfVoxels = true);

    //! Clears all internal structures and counters.
    void clear();

protected:
    //! Sum of all voxels.
    double m_dSum, m_dSumSqr;

    //! Number of voxels counted.
    int m_iNumOfVoxels;
};


//==============================================================================
/*!
 * Structure containing additional tetrahedron information.
 */
struct STetrahedronData
{
    //! Index of the tetrahedron segment.
//    int m_iIndex;

    //! Tetrahedron features.
    CTetraFeatures m_Features;

    //! Default constructor
//    STetrahedronData() : m_iIndex(-1) {}
    STetrahedronData() {}
};


//==============================================================================
/*
 * Global functions.
 */

//! Returns index of the tetrahedron region or -1 if additional tetrahedron
//! data doesn't exist.
inline int getTetrahedronRegion(vctl::MCTetra *pTetrahedron)
{
/*    STetrahedronData *pData = (STetrahedronData *)getTetrahedronDataPtr(pTetrahedron);

    return (pData) ? pData->m_iIndex : -1;*/

    return (pTetrahedron) ? pTetrahedron->GetObject() : -1;
}


//! Sets tetrahedron region index.
//! - Creates tetrahedron data if it doesn't exist.
inline void setTetrahedronRegion(vctl::MCTetra *pTetrahedron, int iIndex)
{
/*    STetrahedronData *pData = (STetrahedronData *)getTetrahedronDataPtr(pTetrahedron);

    if( !pData )
    {
        pData = new STetrahedronData;
        setTetrahedronDataPtr(pTetrahedron, pData);
    }

    pData->m_iIndex = iIndex;*/

    if( pTetrahedron )
    {
        pTetrahedron->SetObject(iIndex);
    }
}


//! Returns pointer to the tetrahedron feature vector.
inline CTetraFeatures *getTetrahedronFeatures(vctl::MCTetra *pTetrahedron)
{
    STetrahedronData *pData = (STetrahedronData *)getTetrahedronDataPtr(pTetrahedron);

    if( !pData )
    {
        pData = new STetrahedronData;
        setTetrahedronDataPtr(pTetrahedron, pData);
    }

    return &(pData->m_Features);
}


} // namespace seg
} // namespace mds

#endif // MDS_TETRAFEATURES_H

