//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2009 by Michal Spanel          \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2009/12/10                          \n
 *
 * $Id:$
 *
 * Description:
 * - Tetrahedron info extraction.
 */

#ifndef MDS_TetraInfo_H
#define MDS_TetraInfo_H

#include "mdsTetraFeatures.h"
#include "mdsTetraMesh.h"


namespace mds
{
namespace seg
{

//==============================================================================
/*!
 * Function object that writes information about tetrahedron to a given stream.
 */
class CWriteTetraInfo
{
public:
    //! Helper flags.
    enum EFlags
    {
        INFO                = 1,
        NEIGHBOURS          = 2,
        FEATURES            = 4,
        CHECK_EDGE_FACES    = 8,
        DEFAULT_FLAGS       = INFO | NEIGHBOURS | FEATURES,
    };

public:
    //! Constructor.
    CWriteTetraInfo(std::ostream& Stream, int Flags = DEFAULT_FLAGS)
        : m_Stream(Stream)
        , m_Flags(Flags)
    {}

    //! Sets internal pointer to the mesh.
    //! - Pointer to the mesh must be valid if you've specified
    //!   the CHECK_EDGE_FACES flag in the constructor!
    void setMesh(CTetraMesh *pMesh) { m_spMesh = pMesh; }

    //! Writes info about a specified tetrahedron.
    void operator ()(vctl::MCTetra& Tetrahedron)
    {
        m_Stream << Tetrahedron.GetObject() << "\t";

        if( checkFlag(INFO) )
        {
            vctl::MCPoint3D Center;
            Tetrahedron.GetCenter(Center);
            m_Stream
                << Center.GetX() << "\t" << Center.GetY() << "\t" << Center.GetZ() << "\t"
                << Tetrahedron.GetVolume() << "\t";
        }

        if( checkFlag(NEIGHBOURS) )
        {
            vctl::MCVertex *pVertices[3];
            for( int i = 0; i < 4; ++i )
            {
                if( !Tetrahedron.GetNeighbour(i) )
                {
                    m_Stream << -1 << "\t";
                    continue;
                }

                Tetrahedron.GetTri(i, pVertices);
                if( checkFlag(CHECK_EDGE_FACES) && m_spMesh.get() )
                {
//                    if( m_spMesh->checkBoundaryTriangle(pVertices[0], pVertices[1], pVertices[2], 0.1) )
                    if( m_spMesh->isBoundaryTriangle(pVertices[0], pVertices[1], pVertices[2]) )
                    {
                        m_Stream << -1 << "\t";
                        continue;
                    }
                }

                // Default output
                m_Stream << Tetrahedron.GetNeighbour(i)->GetObject() << "\t";
            }
        }

        if( checkFlag(FEATURES) )
        {
            CTetraFeatures *pFeatures = getTetrahedronFeatures(&Tetrahedron);
#ifdef USE_VARIANCE
            m_Stream
                << pFeatures->getVoxelMean() << "\t"
                << pFeatures->getVoxelVariance() << "\t"
                << pFeatures->getNumOfVoxels();
#else
            m_Stream
                << pFeatures->getVoxelMean() << "\t"
                << pFeatures->getNumOfVoxels();
#endif // USE_VARIANCE
        }
      
        m_Stream << std::endl;
    }

    //! Writes heading to the stream.
    void writeHeading()
    {
        std::string Info("Id\t");
        
        if( checkFlag(INFO) )
        {
            Info += "CX\tCY\tCZ\tVolume\t";
        }
      
        if( checkFlag(NEIGHBOURS) )
        {
            Info += "N0\tN1\tN2\tN3\t";
        }
      
        if( checkFlag(FEATURES) )
        {
#ifdef USE_VARIANCE
            Info += "Mean\tVariance\tCount";
#else
            Info += "Mean\tCount";
#endif // USE_VARIANCE
        }
      
        m_Stream << Info << std::endl;
    }

    //! Returns true if a given flag is present.
    bool checkFlag(int Flag) { return (m_Flags & Flag) == Flag; }

protected:
    //! Output stream.
    std::ostream& m_Stream;

    //! Internal flags.
    int m_Flags;

    //! Pointer to the mesh.
    CTetraMesh::tSmartPtr m_spMesh;
};


//==============================================================================
/*!
 * Function object assigns unique identifier to all tetrahedra.
 */
class CAssignTetraIndex
{
public:
    //! Constructor.
    CAssignTetraIndex(int StartValue = 0) : m_Counter(StartValue) {}

    //! Assigns identifier to a given tetrahedron.
    void operator ()(vctl::MCTetra& Tetrahedron) { Tetrahedron.SetObject(m_Counter++); }

protected:
    //! Internal counter;
    int m_Counter;
};


} // namespace seg
} // namespace mds

#endif // MDS_TetraInfo_H
