//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2006/04/04                          \n
 *
 * $Id: mdsTetraIterator.h 1629 2010-02-16 14:17:39Z spanel $
 *
 * Description:
 * - Tetrahedron voxels iterator.
 */

#ifndef MDS_TETRAITERATOR_H
#define MDS_TETRAITERATOR_H

#include "mdsTetraBase.h"

// MDSTk
#include <MDSTk/Base/mdsMemory.h>
#include <MDSTk/Base/mdsIterator.h>
#include <MDSTk/Math/mdsBase.h>

// VCTL
#include <VectorEntity/mcentity.h>


namespace mds
{

// Declare iterator traits first...
MDS_ITERATOR_TRAITS(seg, CTetraIterator, seg::CTetraBase::tVoxel);

namespace seg
{

//==============================================================================
/*!
 * Iterator used to traverse points in a tetrahedron.
 * - TODO: forEach() concept.
 */
class CTetraIterator
    : public mds::base::CIteratorBase<CTetraIterator>
    , public mds::base::CContainerHolder<CTetraBase::tVolume, CTetraBase::tVoxel>
{
public:
    //! Iterator base.
    typedef mds::base::CContainerHolder<CTetraBase::tVolume, CTetraBase::tVoxel> base;

    //! Base types.
    typedef CTetraBase::tVolume tVolume;
    typedef CTetraBase::tVoxel tVoxel;
    typedef CTetraBase::tScale tScale;

public:
    //! Constructor.
    CTetraIterator(tVolume *pVolume, vctl::MCTetra *pTetrahedron, tScale *pScale)
        : m_pScale(pScale)
    {
        MDS_ASSERT(pScale);
        
        init(pVolume, pTetrahedron);
    }

    //! Copy constructor.
    CTetraIterator(const CTetraIterator& It)
        : base(It)
        , m_pScale(It.m_pScale)
    {
        mds::base::memCopy(&m_Impl, &It.m_Impl, 1);
    }

    //! Destructor.
    ~CTetraIterator() {}

    //! Assignment operator.
    CTetraIterator& operator=(const CTetraIterator& It)
    {
        m_pContainer = It.m_pContainer;
        m_pItem = It.m_pItem;
        mds::base::memCopy(&m_Impl, &It.m_Impl, 1);
        m_pScale = It.m_pScale;
        return *this;
    }


    //! Returns current iterator position.
    mds::tSize getX() const { return m_Impl.m_iX; }
    mds::tSize getY() const { return m_Impl.m_iY; }
    mds::tSize getZ() const { return m_Impl.m_iZ; }

    //! Returns current real iterator position.
    double getRealX() const { return m_pScale->getRealX(m_Impl.m_iX); }
    double getRealY() const { return m_pScale->getRealY(m_Impl.m_iY); }
    double getRealZ() const { return m_pScale->getRealZ(m_Impl.m_iZ); }

    //! Returns true if iterator points after the last tetrahedron
    //! voxel.
    bool isEnd() const { return (m_Impl.m_iZ > m_Impl.m_iMaxZ); }

    //! Moves iterator to the next triangle pixel.
    void advance()
    {
        do {
            if( isEnd() )
            {
                break;
            }
            else
            {
                next();
            }
        } while( !isInner() );
    }

protected:
    //! All data members.
    struct SDataMembers
    {
        //! Bounding rectangle.
        int m_iMinX, m_iMaxX, m_iMinY, m_iMaxY, m_iMinZ, m_iMaxZ;

        //! Coefficients.
        float m_fA[4], m_fB[4], m_fC[4], m_fD[4];

        //! Flags.
        float m_fEX[4], m_fEY[4], m_fEZ[4];

        //! Iterator position.
        int m_iX, m_iY, m_iZ;
    };

    //! Data members.
    SDataMembers m_Impl;

    //! Pointer to the tetrahedron scaling object.
    tScale *m_pScale;

protected:
    //! Returns true if the current voxel is inside the tetrahedron.
    bool isInner()
    {
/*        return (m_Impl.m_fEX[0] >= 0.0f
                && m_Impl.m_fEX[1] >= 0.0f
                && m_Impl.m_fEX[2] >= 0.0f
                && m_Impl.m_fEX[3] >= 0.0f);*/
        return (m_Impl.m_fEX[0] > 0.0f
                && m_Impl.m_fEX[1] > 0.0f
                && m_Impl.m_fEX[2] > 0.0f
                && m_Impl.m_fEX[3] > 0.0f);
    }

    //! Initializes the triangle iterator.
    void init(tVolume *pVolume, vctl::MCTetra *pTetrahedron)
    {
        MDS_ASSERT(pVolume && pTetrahedron);
//        MDS_CHECK(pVolume && pTetrahedron && m_pScale, return);

        // Get all vertices
        vctl::MCVertex *pVertices[4];
        pTetrahedron->GetVerticeS(pVertices);

//        MDS_LOG_NOTE(pVertices[0] << " " << pVertices[1] << " " << pVertices[2] << " " << pVertices[3]);

        // Scale all vertices
        vctl::MCPoint3D Vertices[4];
        int i;
        for( i = 0; i < 4; ++i )
        {
            Vertices[i].SetX(pVertices[i]->GetX());
            Vertices[i].SetY(pVertices[i]->GetY());
            Vertices[i].SetZ(pVertices[i]->GetZ());
            m_pScale->conv2Volume(&Vertices[i]);
        }

        // Bounding rectangle
        m_Impl.m_iMaxX = m_Impl.m_iMinX = mds::math::round2Int(Vertices[0].GetX());
        m_Impl.m_iMaxY = m_Impl.m_iMinY = mds::math::round2Int(Vertices[0].GetY());
        m_Impl.m_iMaxZ = m_Impl.m_iMinZ = mds::math::round2Int(Vertices[0].GetZ());
        for( i = 1; i < 4; ++i )
        {
            m_Impl.m_iMinX = mds::math::getMin(m_Impl.m_iMinX, mds::math::round2Int(Vertices[i].GetX()));
            m_Impl.m_iMaxX = mds::math::getMax(m_Impl.m_iMaxX, mds::math::round2Int(Vertices[i].GetX()));
            m_Impl.m_iMinY = mds::math::getMin(m_Impl.m_iMinY, mds::math::round2Int(Vertices[i].GetY()));
            m_Impl.m_iMaxY = mds::math::getMax(m_Impl.m_iMaxY, mds::math::round2Int(Vertices[i].GetY()));
            m_Impl.m_iMinZ = mds::math::getMin(m_Impl.m_iMinZ, mds::math::round2Int(Vertices[i].GetZ()));
            m_Impl.m_iMaxZ = mds::math::getMax(m_Impl.m_iMaxZ, mds::math::round2Int(Vertices[i].GetZ()));
        }

        // Compute normal vectors of all tetrahedron sides
        vctl::MCVector3D Normal;
        for( i = 0; i < 4; ++i )
        {
            // Get normal of the tetrahedron side
            pTetrahedron->GetNormal2(i, Normal);
            
            // Inverted conversion to the volume coordinates :)
            m_pScale->conv2Real(&Normal);
            
            m_Impl.m_fA[i] = float(Normal.GetX());
            m_Impl.m_fB[i] = float(Normal.GetY());
            m_Impl.m_fC[i] = float(Normal.GetZ());
        }

        MDS_LOG_NOTE("after normals");

        // Constant
        for( i = 0; i < 4; ++i )
        {
            m_Impl.m_fD[i]  = -m_Impl.m_fA[i] * float(Vertices[i].GetX());
            m_Impl.m_fD[i] -=  m_Impl.m_fB[i] * float(Vertices[i].GetY());
            m_Impl.m_fD[i] -=  m_Impl.m_fC[i] * float(Vertices[i].GetZ());
        }

        MDS_LOG_NOTE("after plane");

        // Starting point
        m_Impl.m_iX = m_Impl.m_iMinX;
        m_Impl.m_iY = m_Impl.m_iMinY;
        m_Impl.m_iZ = m_Impl.m_iMinZ;

        // Initialization
        for( i = 0; i < 4; ++i )
        {
            m_Impl.m_fEZ[i] = m_Impl.m_fD[i] + m_Impl.m_fA[i] * m_Impl.m_iX + m_Impl.m_fB[i] * m_Impl.m_iY + m_Impl.m_fC[i] * m_Impl.m_iZ;
            m_Impl.m_fEY[i] = m_Impl.m_fEX[i] = m_Impl.m_fEZ[i];
        }

        MDS_LOG_NOTE("after init");

        // Initialize pointer to the starting voxel
        m_pContainer = pVolume;
        m_pItem = pVolume->getPtr(m_Impl.m_iX, m_Impl.m_iY, m_Impl.m_iZ);

        // Find the first voxel in the tetrahedron
        if( !isInner() )
        {
            advance();
        }
    }

    //! Moves iterator to the next pixel.
    void next()
    {
        ++m_pItem;

        if( ++m_Impl.m_iX > m_Impl.m_iMaxX )
        {
            m_Impl.m_iX = m_Impl.m_iMinX;

            if( ++m_Impl.m_iY > m_Impl.m_iMaxY )
            {
                m_Impl.m_iY = m_Impl.m_iMinY;
                ++m_Impl.m_iZ;

                m_Impl.m_fEZ[0] += m_Impl.m_fC[0];
                m_Impl.m_fEZ[1] += m_Impl.m_fC[1];
                m_Impl.m_fEZ[2] += m_Impl.m_fC[2];
                m_Impl.m_fEZ[3] += m_Impl.m_fC[3];

                m_Impl.m_fEY[0] = m_Impl.m_fEZ[0];
                m_Impl.m_fEY[1] = m_Impl.m_fEZ[1];
                m_Impl.m_fEY[2] = m_Impl.m_fEZ[2];
                m_Impl.m_fEY[3] = m_Impl.m_fEZ[3];
            }
            else
            {
                m_Impl.m_fEY[0] += m_Impl.m_fB[0];
                m_Impl.m_fEY[1] += m_Impl.m_fB[1];
                m_Impl.m_fEY[2] += m_Impl.m_fB[2];
                m_Impl.m_fEY[3] += m_Impl.m_fB[3];
            }

            m_Impl.m_fEX[0] = m_Impl.m_fEY[0];
            m_Impl.m_fEX[1] = m_Impl.m_fEY[1];
            m_Impl.m_fEX[2] = m_Impl.m_fEY[2];
            m_Impl.m_fEX[3] = m_Impl.m_fEY[3];

            m_pItem = m_pContainer->getPtr(m_Impl.m_iX, m_Impl.m_iY, m_Impl.m_iZ);
        }
        else
        {
            m_Impl.m_fEX[0] += m_Impl.m_fA[0];
            m_Impl.m_fEX[1] += m_Impl.m_fA[1];
            m_Impl.m_fEX[2] += m_Impl.m_fA[2];
            m_Impl.m_fEX[3] += m_Impl.m_fA[3];
        }
    }
};


} // namespace seg
} // namespace mds

#endif // MDS_TETRAITERATOR_H

