//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2006/12/21                          \n
 *
 * $Id: mdsTetraMembershipFCM2.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Membership function based on fuzzy c-Means clustering algorithm.
 * - Modified approach which classifies predefined sub-regions instead of
 *   tetrahedra.
 */

#ifndef MDS_TETRAMEMBERSHIPFCM2_H
#define MDS_TETRAMEMBERSHIPFCM2_H

#define FCM_LOGGING_ENABLED

#include "mdsTetraMembershipFunc.h"

// MDSTk
#include <MDSTk/Base/mdsSTLContainer.h>
#include <MDSTk/Math/Algorithm/mdsFuzzyCMeans.h>

// STL
#include <vector>


namespace mds
{
namespace seg
{

//==============================================================================
/*!
 * Region membership function.
 * - Fuzzy c-Means (FCM) clustering algorithm.
 * - Modified approach which classifies predefined sub-regions instead of
 *   tetrahedra.
 */
class CTetraMembershipFCM2 : public CTetraMembershipFunc
{
public:
    //! Vector of image features extracted for every sub-region.
    typedef std::vector<CTetraFeatures> tSubRegions;

public:
    //! Default constructor.
    CTetraMembershipFCM2(tSubRegions& SubRegions, double dWeight);

    //! Virtual destructor.
    virtual ~CTetraMembershipFCM2() {}

    //! Computes membership function of all tetrahedra in a given list.
    //! - Return false on failure.
    virtual bool operator()(vctl::MCTetra *pFirstTetrahedron);

    //! Computes membership function of all tetrahedra in a given list.
    //! - Predefined number of clusters.
    //! - Return false on failure.
    virtual bool operator()(vctl::MCTetra *pFirstTetrahedron,
                            mds::tSize NumOfClasses
                            );

protected:
    //! Wrapped STL container.
    typedef mds::base::CSTLContainer<tSubRegions> tContainer;

protected:
    //! Used clustering algorithm.
    mds::math::CFuzzyCMeans<tContainer, NUM_OF_FEATURES> m_FCM;

    //! Reference to the vector of image features.
    tSubRegions& m_SubRegions;
};


} // namespace seg
} // namespace mds

#endif // MDS_TETRAMEMBERSHIPFCM2_H

