//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)        \n
 * Copyright (c) 2003-2010 by Michal Spanel         \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz      \n
 * Date:    2006/04/04                              \n
 *
 * $Id: mdsTetraMesh.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Adaptive tetrahedron mesh based on 3D Delaunay triangulation.
 */

#ifndef MDS_TETRAMESH_H
#define MDS_TETRAMESH_H

#include "mdsDelaunayTetra.h"
#include "mdsTetraFeatures.h"
#include "mdsTetraScaling.h"
#include "mdsTetraSimilarityMeasure.h"
#include "mdsTetraHomogeneityMeasure.h"

// STL
#include <vector>


namespace mds
{
namespace seg
{

//==============================================================================
/*
 * Global definitions.
 */

//! If this macro is defined, various informations are logged.
//#ifndef MDS_SEG_LOGGING_ENABLED
//#    define MDS_SEG_LOGGING_ENABLED
//#endif


//==============================================================================
/*!
 * Class representing adaptive tetrahedral mesh.
 */
class CTetraMesh : public CDelaunayTetra, public CTetraScaling, public CTetraBase
{
public:
    //! Smart pointer type.
    //! - Declares type tSmartPtr.
    MDS_SHAREDPTR(CTetraMesh);

    //! Standard method getEntityName().
    MDS_ENTITY_NAME("TetraMesh");

    //! Standard method getEntityCompression().
    MDS_ENTITY_COMPRESSION(mds::mod::CC_RAW);

    //! Standard method getEntityBlockSize().
    MDS_ENTITY_BLOCK_SIZE(4096);

    //! Custom flags used to mark vertices in the mesh.
    enum
    {
        //! Flag used to mark vertices during splitting.
        MARKED              = 1 << 10,

        //! Flag used to mark boundary vertices during the variational meshing.
        DOMAIN_BOUNDARY     = 1 << 11,
    };

    //! Vector of image features.
    typedef std::vector<CTetraFeatures> tSegments;

    //! Default parameters.
    static const tSize DEFAULT_CELL;
    static const tSize DEFAULT_CELL_MARGIN;
    static const tSize DEFAULT_VERTEX_NEIGHBOURHOOD;
    static const double DEFAULT_INIT_THRESHOLD;
    static const double DEFAULT_SPLITTING_THRESHOLD;
    static const double DEFAULT_DISTANCE_THRESHOLD;
    static const double MIN_TETRAHEDRON_VOLUME;
    static const double DEFAULT_DELTA_THRESHOLD;
    static const double DEFAULT_CORNER_MIN_EDGE_LENGTH;
    static const double DEFAULT_AVG_TETRA_SIZE;
    static const double DEFAULT_K;

public:
    //! Default constructor.
    CTetraMesh();

    //! Virtual destructor.
    virtual ~CTetraMesh();

    //! Return size of a neighbourhood used to attract new vertex to an image edge.
    double getVertexNeighbourhood() { return m_VertexNeighbourhood; }

    //! Sets size of a neighbourhood used to attract new vertex to an image edge.
    void setVertexNeighbourhood(tSize Value = DEFAULT_VERTEX_NEIGHBOURHOOD)
    {
        MDS_CHECK(Value >= 0, return);

        m_VertexNeighbourhood = Value;
    }

    //! Returns the minimal edge length near corners.
    double getCornerMinEdgeLength() { return m_dCornerMinEdgeLength; }

    //! Sets minimal allowed edge length in the triangulation near corners.
    void setCornerMinEdgeLength(double dValue = DEFAULT_CORNER_MIN_EDGE_LENGTH)
    {
        MDS_CHECK(dValue > 0.0, return);

        m_dCornerMinEdgeLength = dValue;
    }


    //! Returns pointer to the underlying volume.
    mds::img::CDensityVolume *getVolumePtr() { return m_spVolume; }

    //! Selection of the volume for segmentation.
    void setVolume(mds::img::CDensityVolume *pVolume);

    //! Returns pointer to the underlying edge volume.
    mds::img::CDensityVolume *getEdgeVolumePtr() { return m_spEdgeVolume; }

    //! Selection of the volume for segmentation.
    void setEdgeVolume(mds::img::CDensityVolume *pEdgeVolume);


    //! Returns pointer to the underlying control space volume.
    mds::img::CDensityVolume *getControlSpacePtr() { return m_spControlSpace; }

    //! Selection of the volume for segmentation.
    void setControlSpace(mds::img::CDensityVolume *pControlSpace);

    //! Makes the control space necessary for anisotropic meshing.
    //! - Original volume and edge volume must be set!
    bool prepareControlSpace(double dK = DEFAULT_K,
                             double dAvgTetrahedronSize = DEFAULT_AVG_TETRA_SIZE
                             );


    //! Returns number of regions.
    int getNumOfSegments() const { return m_iNumOfSegments; }

    //! Sets the number of regions.
    void setNumOfSegments(int i) { m_iNumOfSegments = i; }

    //! Returns vector of image features.
    tSegments& getSegments() { return m_Segments; }


    //! Modifies position of a given vertex.
    //! - Attracts vertex to the closest image edge.
    //! - Edge volume must be set!
    virtual bool attractVertex(vctl::MCPoint3D *pPoint);

    //! Modifies position of a given vertex.
    //! - Attracts vertex to the closest image edge.
    //! - Data volume and edge volume must be set!
    bool subpixelAttractVertex(vctl::MCPoint3D *pPoint);

    //! Modifies position of a given vertex.
    //! - Attracts vertex to the closest local minima of the control space.
    //! - Simple gradient descent method is used to find the minima.
    //! - Control space and edge volume must be set!
    bool isotropicAttractVertex(vctl::MCPoint3D *pPoint);


    //! Initializes the Delaunay triangulation.
    //! - Uses vertex detector to initialize the list of vertices.
    //! - Removes all tetrahedra.
    //! - Underlying edge volume must be set!
    bool init(double dThreshold = DEFAULT_INIT_THRESHOLD,
              tSize CellSize = DEFAULT_CELL,
              tSize NumOfNodes = DEFAULT_NUMBER_OF_NODES
              );

    //! Evaluates features of all tetrahedra in the mesh.
    //! - Underlying volume must be set!
    bool evaluateFeatures(bool bSkipSmall = true);

    //! Evaluates features of all segments in the mesh.
    //! - Underlying image must be set!
    //! - Tetrahedra in the mesh must be classified to regions!
    bool evaluateSegmentFeatures(bool bSkipSmall = true);

    //! Adaptation of the triangular mesh to the underlying volume.
    //! - Edge splitting algorithm.
    //! - Both volumes must be set!
    bool edgeSplitting(double dOrientThreshold = DEFAULT_SPLITTING_THRESHOLD);

    //! Adaptation of the triangular mesh to the underlying volume.
    //! - Isotropic edge splitting algorithm based on the control space.
    //! - Control space and edge volume must be set!
    bool isotropicEdgeSplitting(double dThreshold, bool bBoundaryEdges = false);

    //! Adaptation of the mesh by means of a vertex moving.
    //! - Attracts vertices to the closest local minima of the control space.
    //! - Simple gradient descent method is used to find the minima.
    //! - Control space and edge volume must be set!
    bool isotropicVertexMoving();

    //! Adaptation of the mesh by means of a vertex moving.
    //! - Attracts vertices to the center of image edges.
    //! - Both volumes must be set!
    bool subpixelVertexMoving();

    //! Introduces new points lying on image edges to the mesh.
    //! - Edge volume must be set!
    bool approachEdges(bool bCheckIsotropicMinEdgeLength = false);

    //! Variational approach to tetrahedral isotropic meshing.
    //! - Optimizes vertex positions so that a global energy is minimal.
    //! - Control space must be set!
    bool variationalMeshing(bool bCheckIsotropicMinEdgeLength = false);

/*    //! Adaptation of the triangular mesh to the underlying volume.
    //! - Tetrahedron face splitting algorithm.
    //! - Both volumes must be set!
    bool faceSplitting(double dOrientThreshold = DEFAULT_SPLITTING_THRESHOLD);*/

    //! Adaptation of the triangular mesh to the underlying volume.
    //! - Tetrahedron splitting algorithm.
    //! - Both volumes must be set!
    bool tetrahedronSplitting(CTetraHomogeneityMeasure& Measure);

    //! Optimization of the mesh by means of a vertex moving.
    //! - Used to eliminate thin tetrahedra and slivers.
    //! - Finally, the Delaunay triangulation is reconstructed using
    //!   a set of unchanged and moved vertices.
    //! - Edge volume must be set!
    //! - Parameter specifies a minimal required tetrahedron quality, the best
    //!   value (= equilateral tetrahedron) is equal to one.
    bool optimizeByVertexMoving(double dMinQuality);

    //! Simple region growing segmentation. Every region grows while a given
    //! similarity predicate holds and boundary edge is not touched.
    //! - Image features must be evaluated!
    //! - Both images must be set!
    bool regionGrowing(CTetraSimilarityMeasure& Measure,
                       double dBThreshold = DEFAULT_DISTANCE_THRESHOLD
                       );

    //! Aglomerative merging of regions in the mesh.
    //! - Evaluates image features.
    //! - Both images must be set!
    bool regionMerging(CTetraSimilarityMeasure& Measure,
                       double dBThreshold = DEFAULT_DISTANCE_THRESHOLD,
                       double dDeltaThreshold = DEFAULT_DELTA_THRESHOLD,
                       tSize NumOfRegions = -1,
                       bool bMergeAdjacent = true
                       );

    //! Identifies individual isolated regions of tetrahedra in the
    //! pre-classified mesh.
    void findIsolatedRegions();

    //! Overrides classification of all tetrahedron in a specified region.
    void changeRegion(int iMinIndex, int iMaxIndex = -1, int iFinalRegion = -1);

    //! Eliminates unused indexes of regions shaking down all currently
    //! assigned indexes.
    void reassignRegions();

    //! Classify all non-classified tetrahedra with respect to the neighbours.
    //! - This method doesn't use image features for classification!
    //! - Both images must be set!
    bool classifyAdjacent(double dBThreshold = DEFAULT_DISTANCE_THRESHOLD);

    //! Removes noisy (wrong classified) tetrahedra in the mesh.
    //! - Principle of the noise removal algorithm is similar to the median
    //!   filtering approach which is well known in image processing.
    bool noiseReduction();

    //! Visualizes the triangular mesh.
    //! - Draws all tetrahedra as simple lines to a given volume.
    void visualize(mds::img::CDVolume *pVolume);


    //! Finds all vertices lying close to an image edge. These vertices
    //! are expected to form boundary/surface between regions.
    //! - Marks all found vertices using the DOMAIN_BOUDNARY flag.
    //! - Edge volume must be set!
    bool markBoundaryVertices();

    //! Basic test if a given triangle forms boundary between two regions.
    //! - Returns true if all triangle vertices are marked
    //!   with the DOMAIN_BOUNDARY flag.
    //! - The method markBoundaryVertices() must be called first!
    bool isBoundaryTriangle(vctl::MCVertex *p0,
                            vctl::MCVertex *p1,
                            vctl::MCVertex *p2
                            );


    //! Returns pointer to the helper list of triangles.
    vctl::MCTriS *getTriangles() { return m_Tetrahedra.GetTriS(); }

    //! Returns the number of triangles in the helper list.
    int getNumOfTriangles() { return m_Tetrahedra.GetTriS()->GetNumber(); }

    //! Returns pointer to the first triangle in the helper list.
    //! - Returns NULL if there is no triangle in the list.
    vctl::MCTri *getFirstTriangle() { return (vctl::MCTri *)m_Tetrahedra.GetTriS()->GetFirst(); }

    //! Creates the helper list of triangles.
    //! - If index of the tetrahedron segment is equal to -1, the list
    //!   of all triangles in the mesh is going to be created.
    void createTriangles(int iRegion = -1, bool bBoundary = false);

    //! Creates the helper list of triangles.
    //! - Faces of all tetrahedra on the positive side of a given
    //!   plane are generated.
    void createTriangles(double a, double b, double c, double d, int iRegion = -1);

    //! Creates the helper list of triangles.
    //! - Faces of all slivers.
    void createTriangles(double dQuality, int iRegion = -1);

    //! Clears the helper list of triangles.
    void clearTriangles();

    //! Writes the helper list of triangles to a given output channel in STL format.
    //! - Returns false on failure.
    bool saveSTL(mds::mod::CChannel& Channel);

    //! Writes the helper list of triangles to a given output channel in VRML format.
    //! - Returns false on failure.
    //! - Modified Premek Krsek's original code.
    bool saveVRML(mds::mod::CChannel& Channel);


    //! Serializes the tetrahedron mesh.
    template <class S>
    void serialize(mds::mod::CChannelSerializer<S>& Writer)
    {
        // Serialize the Delaunay trianglation
        CDelaunayTetra::serialize(Writer);
    }

    //! Deserializes all class members the tetrahedron mesh.
    template <class S>
    void deserialize(mds::mod::CChannelSerializer<S>& Reader)
    {
        // Deserialize the Delaunay trianglation
        CDelaunayTetra::deserialize(Reader);

        // Evaluate the number of regions
        int iMaxRegion = -1;
        vctl::MCTetra *pTetra = getFirstTetrahedron();
        for( ; pTetra; pTetra = pTetra->GetNext() )
        {
            // Get triangle region
            int iRegion = getTetrahedronRegion(pTetra);
            if( iRegion > iMaxRegion )
            {
                iMaxRegion = iRegion;
            }
        }

        // Set the number of regions
        m_iNumOfSegments = iMaxRegion + 1;
    }

protected:
    //! Vector of points.
    typedef std::vector<vctl::MCPoint3D> tPoints;

protected:
    //! Pointer to the underlying volume.
    mds::img::CDensityVolumePtr m_spVolume;

    //! Pointer to the underlying edge volume.
    mds::img::CDensityVolumePtr m_spEdgeVolume;

    //! Pointer to the control space.
    mds::img::CDensityVolumePtr m_spControlSpace;

    //! Number of segments in the mesh.
    int m_iNumOfSegments;

    //! Vector of image features extracted for every region.
    tSegments m_Segments;

    //! Size of a neighbourhood used to attract new vertex to an image edge.
    mds::tSize m_VertexNeighbourhood;

    //!! Minimal edge length near corners.
    double m_dCornerMinEdgeLength;

    //! Currently used homogeneity measure.
    CTetraHomogeneityMeasure *m_pHomogeneityMeasure;

//protected:
public:
    //! Checks a tetrahedron volume.
    bool checkTetrahedronVolume(vctl::MCTetra *pTetra);

    //! Checks if a given edge is a boundary edge.
    //! - Returns true if the edge and some image edge are parallel.
    bool checkBoundaryEdge(vctl::MCEdge *pEdge,
                           vctl::MCVector3D *pNormal,
                           double dThreshold = DEFAULT_DISTANCE_THRESHOLD
                           );

    //! Checks if a given triangle forms boundary between two regions.
    //! - Returns true if the triangle and some image edge are parallel.
    bool checkBoundaryTriangle(vctl::MCPoint3D *p0,
                               vctl::MCPoint3D *p1,
                               vctl::MCPoint3D *p2,
                               double dThreshold = DEFAULT_DISTANCE_THRESHOLD
                               );

    int evalEdgeDistancePoint(vctl::MCPoint3D *pPoint,
                              mds::img::CPoint3D& Normal,
                              int iMaxDistance
                              );

    //! Splits a given edge.
    //! - Returns true if the edge was divided.
    bool splitEdge(vctl::MCEdge *pEdge,
                   double dThreshold,
                   tPoints& Points
                   );

    //! Inserts tetrahedron edges to the list of edges.
    static void edgeSplittingInsertFunc(vctl::MCTetra *pTetrahedron, CDelaunayTetra *pMesh);

    //! Inserts tetrahedron edges to the list of edges.
    static void edgeSplittingRemoveFunc(vctl::MCTetra *pTetrahedron, CDelaunayTetra *pMesh);

    //! Anisotropic splitting of a given edge.
    //! - Returns true if the edge was divided.
    bool isotropicSplitEdge(vctl::MCVertex *pV0,
                            vctl::MCVertex *pV1,
                            double dThreshold,
                            tPoints& EdgePoints,
                            tPoints& Points,
                            bool bBoundary = false
                            );

    //! Anisotropic splitting of a given edge.
    //! - Returns true if the edge was divided.
    bool isotropicSplitSubEdge(vctl::MCPoint3D &p1,
                               vctl::MCPoint3D &p2,
                               double dThreshold,
                               double dRestrictedZone,
                               tPoints& Points
                               );

    //! Inserts tetrahedron edges to the list of edges.
    static void isotropicEdgeSplittingInsertFunc(vctl::MCTetra *pTetrahedron, CDelaunayTetra *pMesh);

    //! Inserts tetrahedron edges to the list of edges.
    static void isotropicEdgeSplittingRemoveFunc(vctl::MCTetra *pTetrahedron, CDelaunayTetra *pMesh);

    //! Returns length of the edge in the control space.
    double getControlSpaceLength(vctl::MCPoint3D& p1, vctl::MCPoint3D& p2);

    //! Returns subpixel control space value.
    double getControlSpaceValue(const mds::img::CPoint3D& Point);

    //! Splits a given tetrahedron.
    //! - Returns true if the tetrahedron was divided.
    bool splitTetrahedron(vctl::MCTetra *pTetrahedron);

    //! Evaluates tetrahedron homogeneity and inserts tetrahedron
    //! to the sorted queue.
    static void tetrahedronSplittingInsertFunc(vctl::MCTetra *pTetrahedron,
                                               CDelaunayTetra *pMesh
                                               );

    //! Unsorts tetrahedron from the queue.
    static void tetrahedronSplittingRemoveFunc(vctl::MCTetra *pTetrahedron,
                                               CDelaunayTetra *pMesh
                                               );

public:
    //! Checks minimal edge length in the cavity.
    static bool checkIsotropicMinEdgeLengthFunc(vctl::MCPoint3D *p1,
                                                vctl::MCPoint3D *p2,
                                                CDelaunayTetra* pMesh
                                                );
    static bool checkIsotropicMinEdgeLengthFunc2(vctl::MCPoint3D *p1,
                                                 vctl::MCPoint3D *p2,
                                                 CDelaunayTetra* pMesh
                                                 );

    //! Checks if edge intersects any image edge.
    static bool checkNoImageEdgeIntersectionFunc(vctl::MCPoint3D *p1,
                                                 vctl::MCPoint3D *p2,
                                                 CDelaunayTetra* pMesh
                                                 );

    //! Checks if a given facet forms boundary between two image regions.
    //! - Returns true if the facet is parallel with some image edge.
    static bool checkBoundaryFacetFunc(vctl::MCPoint3D *p1,
                                       vctl::MCPoint3D *p2,
                                       vctl::MCPoint3D *p3,
                                       CDelaunayTetra* pMesh
                                       );
};


//=============================================================================
/*
 * Basic template instances and type definitions.
 * - Using smart pointers.
 */

//! Tetrahedral mesh.
typedef CTetraMesh::tSmartPtr   CTetraMeshPtr;


} // namespace seg
} // namespace mds

#endif // MDS_TETRAMESH_H

