//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2006/06/20                          \n
 *
 * $Id: mdsTetraScaling.h 1629 2010-02-16 14:17:39Z spanel $
 *
 * Description:
 * - Base definitions.
 */

#ifndef MDS_TETRASCALING_H
#define MDS_TETRASCALING_H

// MDSTk
#include <MDSTk/Base/mdsTypes.h>
#include <MDSTk/Base/mdsAssert.h>
#include <MDSTk/Math/mdsBase.h>

// VCTL
#include <VectorEntity/mccoordinate3d.h>


namespace mds
{
namespace seg
{

//==============================================================================
/*!
 * Class for scaling tetrahedral mesh with respect to a real voxel size.
 * - Forward conversion changes volume coordinates to real coordinates.
 */
class CTetraScaling
{
public:
    //! Type of VCTL coordinates.
    typedef vctl::MCCoordinate3D<double> tCoordinates;

public:
    //! Default constructor.
    CTetraScaling()
        : m_dDX(1.0)
        , m_dDY(1.0)
        , m_dDZ(1.0)
        , m_dMin(1.0)
        , m_dInvMin(1.0)
        , m_dInvDX(1.0)
        , m_dInvDY(1.0)
        , m_dInvDZ(1.0)
    {
    }

    //! Constructor.
    CTetraScaling(double dDX, double dDY, double dDZ)
        : m_dDX(dDX)
        , m_dDY(dDY)
        , m_dDZ(dDZ)
        , m_dMin(mds::math::getMin(dDX, dDY, dDZ))
    {
        MDS_ASSERT(dDX > 0.0 && dDY > 0.0 && dDZ > 0.0);

        m_dInvDX = 1.0 / dDX;
        m_dInvDY = 1.0 / dDY;
        m_dInvDZ = 1.0 / dDZ;
        
        m_dInvMin = 1.0 / m_dMin;
    }

    //! Destructor.
    ~CTetraScaling() {}

    //! Sets the real voxel size.
    void setVoxelSize(double dDX, double dDY, double dDZ)
    {
        MDS_ASSERT(dDX > 0.0 && dDY > 0.0 && dDZ > 0.0);

        m_dDX = dDX;
        m_dDY = dDY;
        m_dDZ = dDZ;

        m_dMin = mds::math::getMin(dDX, dDY, dDZ);
        m_dInvMin = 1.0 / m_dMin;

        m_dInvDX = 1.0 / dDX;
        m_dInvDY = 1.0 / dDY;
        m_dInvDZ = 1.0 / dDZ;
    }


    //! Forward conversion of given coordinates.
    //! - Equal to backward conversion of a normal vector.
    double getRealX(double x) const { return x * m_dDX; }
    double getRealY(double y) const { return y * m_dDY; }
    double getRealZ(double z) const { return z * m_dDZ; }

    //! Forward conversion using minimal voxel size.
    double getMinRealSize(double Value) const { return Value * m_dMin; }

    //! Forward conversion of given coordinates.
    //! - Equal to backward conversion of a normal vector.
    void conv2Real(mds::tSize x, mds::tSize y, mds::tSize z,
                   double& dX,
                   double& dY,
                   double& dZ
                   ) const
    {
        dX = x * m_dDX;
        dY = y * m_dDY;
        dZ = z * m_dDZ;
    }

    //! Forward conversion of a given point.
    //! - Equal to backward conversion of a normal vector.
    void conv2Real(mds::tSize x, mds::tSize y, mds::tSize z, tCoordinates *pPoint) const
    {
        pPoint->SetX(x * m_dDX);
        pPoint->SetY(y * m_dDY);
        pPoint->SetZ(z * m_dDZ);
    }

    //! Forward conversion of a given point.
    //! - Equal to backward conversion of a normal vector.
    void conv2Real(double dX, double dY, double dZ, tCoordinates *pPoint) const
    {
        pPoint->SetX(dX * m_dDX);
        pPoint->SetY(dY * m_dDY);
        pPoint->SetZ(dZ * m_dDZ);
    }

    //! Forward conversion of a given point.
    //! - Equal to backward conversion of a normal vector.
    void conv2Real(tCoordinates *pPoint) const
    {
        pPoint->x() *= m_dDX;
        pPoint->y() *= m_dDY;
        pPoint->z() *= m_dDZ;
    }


    //! Backward conversion of given coordinates.
    //! - Equal to forward conversion of a normal vector.
    mds::tSize getVolumeX(double x) const { return mds::math::round2Int(x * m_dInvDX); }
    mds::tSize getVolumeY(double y) const { return mds::math::round2Int(y * m_dInvDY); }
    mds::tSize getVolumeZ(double z) const { return mds::math::round2Int(z * m_dInvDZ); }

    //! Backward conversion using minimal voxel size.
    double getMinVolumeSize(double Value) const { return Value * m_dInvMin; }

    //! Backward conversion of given coordinates.
    //! - Equal to forward conversion of a normal vector.
    void conv2Volume(double dX, double dY, double dZ,
                     mds::tSize& x,
                     mds::tSize& y,
                     mds::tSize& z
                     ) const
    {
        x = mds::math::round2Int(dX * m_dInvDX);
        y = mds::math::round2Int(dY * m_dInvDY);
        z = mds::math::round2Int(dZ * m_dInvDZ);
    }

    //! Backward conversion of a given point.
    //! - Equal to forward conversion of a normal vector.
    void conv2Volume(tCoordinates *pPoint, mds::tSize& x, mds::tSize& y, mds::tSize& z) const
    {
        x = mds::math::round2Int(pPoint->x() * m_dInvDX);
        y = mds::math::round2Int(pPoint->y() * m_dInvDY);
        z = mds::math::round2Int(pPoint->z() * m_dInvDZ);
    }

    //! Backward conversion of a given point.
    //! - Equal to forward conversion of a normal vector.
    void conv2Volume(tCoordinates *pPoint, double& x, double& y, double& z) const
    {
        x = pPoint->x() * m_dInvDX;
        y = pPoint->y() * m_dInvDY;
        z = pPoint->z() * m_dInvDZ;
    }

    //! Backward conversion of a given point.
    //! - Equal to forward conversion of a normal vector.
    void conv2Volume(tCoordinates *pPoint) const
    {
        pPoint->x() *= m_dInvDX;
        pPoint->y() *= m_dInvDY;
        pPoint->z() *= m_dInvDZ;
    }

protected:
    //! Real voxel size.
    double m_dDX, m_dDY, m_dDZ;
    
    //! Minimal voxel size.
    double m_dMin, m_dInvMin;
    
    //! Backward conversion of coordinates.
    double m_dInvDX, m_dInvDY, m_dInvDZ;    
};


} // namespace seg
} // namespace mds

#endif // MDS_TETRASCALING_H

