//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2006/12/21                          \n
 *
 * $Id: mdsTetraSimilarityMeasure.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Tetrahedra similarity measure.
 */

#ifndef MDS_TETRASIMILARITYMEASURE_H
#define MDS_TETRASIMILARITYMEASURE_H

#include "mdsTetraFeatures.h"

// MDSTk
#include <MDSTk/Math/mdsVectorFunctions.h>

// STL
#include <cmath>


namespace mds
{
namespace seg
{

//==============================================================================
/*!
 * Abstract class of a similarity measure.
 */
class CTetraSimilarityMeasure
{
public:
    //! Default constructor.
    CTetraSimilarityMeasure(double dThreshold = 0.5) : m_dThreshold(dThreshold) {}

    //! Virtual destructor.
    virtual ~CTetraSimilarityMeasure() {}

    //! Computes similarity of two given tetrahedron feature vectors.
    virtual double getSimilarity(const CTetraFeatures& v1, const CTetraFeatures& v2) = 0;

    //! Compares similarity of two given tetrahedron feature vectors against
    //! the internal threshold.
    //! - Returns true if the similarity is greater than the threshold.
    virtual bool checkSimilarity(const CTetraFeatures& v1, const CTetraFeatures& v2)
    {
        return (this->getSimilarity(v1, v2) > m_dThreshold);
    }

    //! Returns value of the internal threshold.
    double getThreshold() { return m_dThreshold; }

    //! Sets the internal threshold.
    void setThreshold(double dValue) { m_dThreshold = dValue; }

protected:
    //! Internal threshold.
    double m_dThreshold;
};


//==============================================================================
/*!
 * Basic similarity measure.
 * - Compares mean pixel values.
 */
class CTetraMeanValueSimilarity : public CTetraSimilarityMeasure
{
public:
    //! Default parameters.
    static const double DEFAULT_THRESHOLD;
    static const double DEFAULT_RHO;

public:
    //! Default constructor.
    CTetraMeanValueSimilarity(double dThreshold = DEFAULT_THRESHOLD,
                              double dRho = DEFAULT_RHO
                              );

    //! Virtual destructor.
    virtual ~CTetraMeanValueSimilarity() {}

    //! Computes similarity of two given feature vectors.
    virtual double getSimilarity(const CTetraFeatures& v1, const CTetraFeatures& v2);

protected:
    //! Parameter affecting sensitivity of the measure.
    double m_dInvRho;
};


//==============================================================================
/*!
 * Basic similarity measure.
 * - Statistical test of the similarity based on pixel value variance.
 */
class CTetraVarianceSimilarity : public CTetraSimilarityMeasure
{
public:
    //! Default parameters.
    static const double DEFAULT_THRESHOLD;

public:
    //! Default constructor.
    CTetraVarianceSimilarity(double dThreshold = DEFAULT_THRESHOLD);

    //! Virtual destructor.
    virtual ~CTetraVarianceSimilarity() {}

    //! Computes similarity of two given feature vectors.
    virtual double getSimilarity(const CTetraFeatures& v1, const CTetraFeatures& v2);
};


//==============================================================================
/*!
 * Complex similarity measure.
 */
class CTetraMixtureSimilarity : public CTetraSimilarityMeasure
{
public:
    //! Default constructor.
    CTetraMixtureSimilarity(CTetraSimilarityMeasure& M1,
                            CTetraSimilarityMeasure& M2,
                            double dW1,
                            double dW2
                            );

    //! Virtual destructor.
    virtual ~CTetraMixtureSimilarity() {}

    //! Computes similarity of two given feature vectors.
    virtual double getSimilarity(const CTetraFeatures& v1, const CTetraFeatures& v2);

    //! Compares similarity of two given tetrahedron feature vectors.
    virtual bool checkSimilarity(const CTetraFeatures& v1, const CTetraFeatures& v2);

protected:
    //! References to the similarity measures.
    CTetraSimilarityMeasure& m_Measure1;
    CTetraSimilarityMeasure& m_Measure2;
    
    //! Mixing weights.
    double m_dW1, m_dW2;
};


} // namespace seg
} // namespace mds

#endif // MDS_TETRASIMILARITYMEASURE_H

