//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2006/12/25                          \n
 *
 * $Id: mdsTetraTriIterator.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Tetrahedron face (triangle) voxels iterator.
 */

#ifndef MDS_TETRATRIITERATOR_H
#define MDS_TETRATRIITERATOR_H

#include "mdsTetraBase.h"

// MDSTk
#include <MDSTk/Base/mdsMemory.h>
#include <MDSTk/Base/mdsIterator.h>
#include <MDSTk/Math/mdsBase.h>

// VCTL
#include <VectorEntity/mcentity.h>


namespace mds
{

// Declare iterator traits first...
MDS_ITERATOR_TRAITS(seg, CTetraTriIterator, seg::CTetraBase::tVoxel);

namespace seg
{

//==============================================================================
/*!
 * Iterator used to traverse points on a tetrahedron face (triangle).
 * - TODO: forEach() concept.
 */
class CTetraTriIterator
    : public mds::base::CIteratorBase<CTetraTriIterator>
    , public mds::base::CContainerHolder<CTetraBase::tVolume, CTetraBase::tVoxel>
{
public:
    //! Iterator base.
    typedef mds::base::CContainerHolder<CTetraBase::tVolume, CTetraBase::tVoxel> base;

    //! Base types.
    typedef CTetraBase::tVolume tVolume;
    typedef CTetraBase::tVoxel tVoxel;
    typedef CTetraBase::tScale tScale;

public:
    //! Constructor.
    CTetraTriIterator(tVolume *pVolume, vctl::MCTri *pTriangle, tScale *pScale)
        : m_pScale(pScale)
    {
        MDS_ASSERT(pTriangle && pScale);

        // Get vertices of the triangle
        vctl::MCVertex *pVertices[3];
        pTriangle->GetVerticeS(pVertices);

        init(pVolume, pVertices[0], pVertices[1], pVertices[2]);
    }

    //! Constructor.
    CTetraTriIterator(tVolume *pVolume,
                      vctl::MCPoint3D *p0,
                      vctl::MCPoint3D *p1,
                      vctl::MCPoint3D *p2,
                      tScale *pScale
                      )
        : m_pScale(pScale)
    {
        MDS_ASSERT(pScale);
        
        init(pVolume, p0, p1, p2);
    }

    //! Copy constructor
    CTetraTriIterator(const CTetraTriIterator& It)
        : base(It)
        , m_pScale(It.m_pScale)
    {
        mds::base::memCopy(&m_Impl, &It.m_Impl, 1);
    }

    //! Destructor
    ~CTetraTriIterator() {}

    //! Assignment operator
    CTetraTriIterator& operator=(const CTetraTriIterator& It)
    {
        base::m_pContainer = It.m_pContainer;
        base::m_pItem = It.m_pItem;
        mds::base::memCopy(&m_Impl, &It.m_Impl, 1);
        m_pScale = It.m_pScale;
        return *this;
    }


    //! Returns current iterator position.
    mds::tSize getX() const { return *m_Impl.m_pX; }
    mds::tSize getY() const { return *m_Impl.m_pY; }
    mds::tSize getZ() const { return *m_Impl.m_pZ; }

    //! Returns current real iterator position.
    double getRealX() const { return m_pScale->getRealX(*m_Impl.m_pX); }
    double getRealY() const { return m_pScale->getRealY(*m_Impl.m_pY); }
    double getRealZ() const { return m_pScale->getRealZ(*m_Impl.m_pZ); }

    //! Returns true if iterator points after the last triangle
    //! pixel.
    bool isEnd() const { return (m_Impl.m_Y > m_Impl.m_iMaxY); }

protected:
    //! All data members.
    struct SDataMembers
    {
        //! Bounding rectangle.
        int m_iMinX, m_iMaxX, m_iMinY, m_iMaxY;
        
        //! Deltas.
        int m_iFDX1, m_iFDX2, m_iFDX3;
        int m_iFDY1, m_iFDY2, m_iFDY3;
        
        //! Flags.
        int m_iCY1, m_iCY2, m_iCY3;
        int m_iCX1, m_iCX2, m_iCX3;
        
        //! Current position.
        mds::tSize m_X, m_Y, m_Z;
        
        //! Real iterator position
        mds::tSize *m_pX, *m_pY, *m_pZ;
        
        //! Slope in the XZ and YZ plane.
        double m_dDX, m_dDY;
        
        //! Current z-position.
        double m_dY, m_dZ;
        
        //! Exchages normal orientation.
        int m_iInv1, m_iInv2, m_iInv3;
    };
    
    //! Data members.
    SDataMembers m_Impl;
    
    //! Pointer to the tetrahedron scaling object.
    tScale *m_pScale;

public:
    //! Moves iterator to the next triangle pixel.
    void advance()
    {
        do {
            if( isEnd() )
            {
                break;
            }
            else
            {
                next();
            }
        } while( !isInner() );
    }

protected:
    //! Returns true if the current pixel is inside the triangle.
    bool isInner()
    {
//        return (m_Impl.m_iCX1 <= 0 && m_Impl.m_iCX2 <= 0 && m_Impl.m_iCX3 <= 0);
        return (m_Impl.m_iCX1 * m_Impl.m_iInv1 <= 0
            && m_Impl.m_iCX2 * m_Impl.m_iInv2 <= 0
            && m_Impl.m_iCX3 * m_Impl.m_iInv3 <= 0);
    }
    
    //! Initializes the triangle iterator.
    void init(tVolume *pVolume,
              vctl::MCPoint3D *p0,
              vctl::MCPoint3D *p1,
              vctl::MCPoint3D *p2
              )
    {
        MDS_ASSERT(p0 && p1 && p2 && pVolume);
        
        // Vertices
        double dVertices[3][3];
        m_pScale->conv2Volume(p0, dVertices[0][0], dVertices[0][1], dVertices[0][2]);
        m_pScale->conv2Volume(p1, dVertices[1][0], dVertices[1][1], dVertices[1][2]);
        m_pScale->conv2Volume(p2, dVertices[2][0], dVertices[2][1], dVertices[2][2]);
        
        // Compute normal of the triangle
        vctl::MCVector3D v01(dVertices[1][0] - dVertices[0][0],
                             dVertices[1][1] - dVertices[0][1],
                             dVertices[1][2] - dVertices[0][2]
                             );
        vctl::MCVector3D v02(dVertices[2][0] - dVertices[0][0],
                             dVertices[2][1] - dVertices[0][1],
                             dVertices[2][2] - dVertices[0][2]
                             );
        vctl::MCVector3D Normal(v01 % v02);
        
        // Parameters of the plane
        double dPlane[4];
        dPlane[0] = Normal.GetX();
        dPlane[1] = Normal.GetY();
        dPlane[2] = Normal.GetZ();
        dPlane[3] = -dPlane[0] * dVertices[0][0];
        dPlane[3] -= dPlane[1] * dVertices[0][1];
        dPlane[3] -= dPlane[2] * dVertices[0][2];
        
        // Initialize mapping of coordinates
        int iMapping[3] = { 0, 1, 2 };
        m_Impl.m_pX = &m_Impl.m_X;
        m_Impl.m_pY = &m_Impl.m_Y;
        m_Impl.m_pZ = &m_Impl.m_Z;
        
        // Find plane which is parallel with the triangle 
        // = Find maximal coordinate of the normal vector
        // - Swap coordinates so that the most forward plane is represented by the XY plane
        if( mds::math::getAbs(dPlane[0]) > mds::math::getAbs(dPlane[1]) )
        {
            if( mds::math::getAbs(dPlane[0]) > mds::math::getAbs(dPlane[2]) )
            {
                mds::math::swap(iMapping[0], iMapping[2]);
                mds::math::swap(m_Impl.m_pX, m_Impl.m_pZ);
            }
        }
        else
        {
            if( mds::math::getAbs(dPlane[1]) > mds::math::getAbs(dPlane[2]) )
            {
                mds::math::swap(iMapping[1], iMapping[2]);
                mds::math::swap(m_Impl.m_pY, m_Impl.m_pZ);
            }
        }
/*        if( mds::math::getAbs(dPlane[0]) > mds::math::getAbs(dPlane[2]) )
        {
            mds::math::swap(iMapping[0], iMapping[2]);
            mds::math::swap(m_Impl.m_pX, m_Impl.m_pZ);
        }
        if( mds::math::getAbs(dPlane[1]) > mds::math::getAbs(dPlane[iMapping[2]]) )
        {
            mds::math::swap(iMapping[1], iMapping[2]);
            mds::math::swap(m_Impl.m_pY, m_Impl.m_pZ);
        }*/
        
        // Slope in the XZ and YZ plane
        double dInvC = -1.0 / dPlane[iMapping[2]];
        m_Impl.m_dDX = dPlane[iMapping[0]] * dInvC;
        m_Impl.m_dDY = dPlane[iMapping[1]] * dInvC;
        
        // Rasterization of the triangle in the XY plane (projection)
        // Fixed-point coordinates
        int iY1 = mds::math::round2Int(16.0 * dVertices[0][iMapping[1]]);
        int iY2 = mds::math::round2Int(16.0 * dVertices[1][iMapping[1]]);
        int iY3 = mds::math::round2Int(16.0 * dVertices[2][iMapping[1]]);
        
        int iX1 = mds::math::round2Int(16.0 * dVertices[0][iMapping[0]]);
        int iX2 = mds::math::round2Int(16.0 * dVertices[1][iMapping[0]]);
        int iX3 = mds::math::round2Int(16.0 * dVertices[2][iMapping[0]]);
        
        // Deltas
        int iDX1 = iX2 - iX1;
        int iDX2 = iX3 - iX2;
        int iDX3 = iX1 - iX3;
        
        int iDY1 = iY2 - iY1;
        int iDY2 = iY3 - iY2;
        int iDY3 = iY1 - iY3;
        
        // Fixed-point deltas
        m_Impl.m_iFDX1 = iDX1 << 4;
        m_Impl.m_iFDX2 = iDX2 << 4;
        m_Impl.m_iFDX3 = iDX3 << 4;
        
        m_Impl.m_iFDY1 = iDY1 << 4;
        m_Impl.m_iFDY2 = iDY2 << 4;
        m_Impl.m_iFDY3 = iDY3 << 4;
        
        // Bounding rectangle
        m_Impl.m_iMinX = mds::math::getMin(iX1, iX2, iX3) >> 4;
        m_Impl.m_iMaxX = mds::math::getMax(iX1, iX2, iX3) >> 4;
        m_Impl.m_iMinY = mds::math::getMin(iY1, iY2, iY3) >> 4;
        m_Impl.m_iMaxY = mds::math::getMax(iY1, iY2, iY3) >> 4;
        
        // Half-edge constants
        int iC1 = iDX1 * iY1 - iDY1 * iX1;
        int iC2 = iDX2 * iY2 - iDY2 * iX2;
        int iC3 = iDX3 * iY3 - iDY3 * iX3;

        // Exchange normal orientation
        m_Impl.m_iInv1 = 1;
        m_Impl.m_iInv2 = 1;
        m_Impl.m_iInv3 = 1;
        if( (iC1 - iDX1 * iY3 + iDY1 * iX3) > 0 )
        {
        	m_Impl.m_iInv1 = -1;
        }
        if( (iC2 - iDX2 * iY1 + iDY2 * iX1) > 0 )
        {
        	m_Impl.m_iInv2 = -1;
        }
        if( (iC3 - iDX3 * iY2 + iDY3 * iX2) > 0 )
        {
        	m_Impl.m_iInv3 = -1;
        }
        
        // Initialization
        m_Impl.m_iCY1 = iC1 - iDX1 * (m_Impl.m_iMinY << 4) + iDY1 * (m_Impl.m_iMinX << 4);
        m_Impl.m_iCY2 = iC2 - iDX2 * (m_Impl.m_iMinY << 4) + iDY2 * (m_Impl.m_iMinX << 4);
        m_Impl.m_iCY3 = iC3 - iDX3 * (m_Impl.m_iMinY << 4) + iDY3 * (m_Impl.m_iMinX << 4);
        
        m_Impl.m_iCX1 = m_Impl.m_iCY1;
        m_Impl.m_iCX2 = m_Impl.m_iCY2;
        m_Impl.m_iCX3 = m_Impl.m_iCY3;
        
        m_Impl.m_X = m_Impl.m_iMinX;
        m_Impl.m_Y = m_Impl.m_iMinY;
        
        // Initial z-position
        m_Impl.m_dZ = dPlane[iMapping[0]] * m_Impl.m_iMinX;
        m_Impl.m_dZ += dPlane[iMapping[1]] * m_Impl.m_iMinY;
        m_Impl.m_dZ += dPlane[3];
        m_Impl.m_dZ *= dInvC;
        m_Impl.m_dY = m_Impl.m_dZ;
        m_Impl.m_Z = mds::math::round2Int(m_Impl.m_dZ);
        
        // Initialize pointer to the voxel        
        m_pContainer = pVolume;
        m_pItem = pVolume->getPtr(*m_Impl.m_pX, *m_Impl.m_pY, *m_Impl.m_pZ);
        
        // Find the first pixel
        if( !isInner() )
        {
            advance();
        }
    }
    
    //! Moves iterator to the next pixel.
    void next()
    {
        if( ++m_Impl.m_X > m_Impl.m_iMaxX )
        {
            m_Impl.m_X = m_Impl.m_iMinX;
            ++m_Impl.m_Y;
            m_Impl.m_iCY1 -= m_Impl.m_iFDX1;
            m_Impl.m_iCY2 -= m_Impl.m_iFDX2;
            m_Impl.m_iCY3 -= m_Impl.m_iFDX3;
            m_Impl.m_iCX1 = m_Impl.m_iCY1;
            m_Impl.m_iCX2 = m_Impl.m_iCY2;
            m_Impl.m_iCX3 = m_Impl.m_iCY3;
            
            m_Impl.m_dY += m_Impl.m_dDY;
            m_Impl.m_dZ = m_Impl.m_dY;
        }
        else
        {
            m_Impl.m_iCX1 += m_Impl.m_iFDY1;
            m_Impl.m_iCX2 += m_Impl.m_iFDY2;
            m_Impl.m_iCX3 += m_Impl.m_iFDY3;
            
            m_Impl.m_dZ += m_Impl.m_dDX;
        }
        
        m_Impl.m_Z = mds::math::round2Int(m_Impl.m_dZ);
        m_pItem = m_pContainer->getPtr(*m_Impl.m_pX, *m_Impl.m_pY, *m_Impl.m_pZ);
    }
};


} // namespace seg
} // namespace mds

#endif // MDS_TETRATRIITERATOR_H

