//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2006/05/06                          \n
 *
 * $Id: mdsTetraVertexDetector.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Initial vertex detector.
 */

#ifndef MDS_TETRAVERTEXDETECTOR_H
#define MDS_TETRAVERTEXDETECTOR_H

#include "mdsTetraBase.h"

// MDSTk
#include <MDSTk/Base/mdsAssert.h>
#include <MDSTk/Base/mdsIdentifier.h>
#include <MDSTk/Math/mdsBase.h>
#include <MDSTk/Image/mdsVolumeFunctions.h>

// VCTL
#include <VectorEntity/mcvertices.h>

// STL
#include <sstream>


namespace mds
{
namespace seg
{

//==============================================================================
/*
 * Global definitions.
 */

//! Volume vertex detection algorithms.
MDS_DECLARE_ID(CELL_MAX);


//==============================================================================
/*!
 * Basic vertex detector template.
 * - Parameter 'Id' is type of the detection algorithm.
 */
template <typename Id>
class CTetraVertexDetector : public CTetraBase
{
public:
    //! Detects all vertices in a given image. Returned list is sorted
    //! in order of vertex strength.
    bool operator()(tVolume *pVolume,
                    vctl::MCVerticeS *pVertices,
                    tScale *pScale
                    );
};


//==============================================================================
/*!
 * Vertex detector template specialization.
 * - Just one candidate pixel (that which has maximal value) is taken
 *   from an image cell.
 */
template <>
class CTetraVertexDetector<CELL_MAX> : public CTetraBase
{
public:
    //! Default constructor
    //! - Input image is divided into cells, while just one candidate
    //!   vertex is taken from each cell. Parameter 'CellSize' represents
    //!   size of these subwindows in pixels.
    //! - Threshold affecting the overall number of pre-selected candidate
    //!   vertices.
    CTetraVertexDetector(double dThreshold,
                         mds::tSize CellSize,
                         mds::tSize CellMargin
                         )
        : m_dThreshold(dThreshold)
        , m_CellSize(CellSize)
        , m_CellMargin(CellMargin)
    {
        MDS_ASSERT(m_CellSize > 0 && m_CellMargin > 0);
    }

    //! Detects all vertices in a given edge image. Returned list is sorted
    //! in order of vertex significance.
    bool operator()(tVolume *pVolume,
                    vctl::MCVerticeS *pVertices,
                    tScale *pScale
                    );

protected:
    //! Threshold.
    double m_dThreshold;

    //! Cell size and margin.
    mds::tSize m_CellSize, m_CellMargin;
};


//=============================================================================
/*
 * Methods templates.
 */

// Include file containing methods templates
#include "mdsTetraVertexDetector.hxx"


} // namespace seg
} // namespace mds

#endif // MDS_TETRAVERTEXDETECTOR_H

