//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2006/09/24                          \n
 *
 * $Id: mdsTriContainer.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Container encapsulating list of triangles.
 */

#ifndef MDS_TRICONTAINER_H
#define MDS_TRICONTAINER_H

#include "mdsTriFeatures.h"

// MDSTk
#include <MDSTk/Base/mdsIterator.h>

// VCTL
#include <VectorEntity/mcentity.h>


namespace mds
{

// Declare iterator traits first...
MDS_ITERATOR_TRAITS(seg, CTriContainerIterator, seg::CTriFeatures);

namespace seg
{

//==============================================================================
/*!
 * Iterator used to traverse feature vectors of all triangles in a list.
 */
class CTriContainerIterator
    : public mds::base::CIteratorBase<CTriContainerIterator>
    , public mds::base::CPtrHolder<CTriFeatures>
{
public:
    //! Iterator base.
    typedef mds::base::CPtrHolder<CTriFeatures> base;

public:
    //! Constructor.
    CTriContainerIterator(vctl::MCTri *pFirstTriangle)
    {
        init(pFirstTriangle);
    }

    //! Copy constructor.
    CTriContainerIterator(const CTriContainerIterator& It)
    {
        base::m_pItem = It.m_pItem;
        m_pTriangle = It.m_pTriangle;
    }

    //! Destructor.
    ~CTriContainerIterator() {}

    //! Assignment operator.
    CTriContainerIterator& operator=(const CTriContainerIterator& It)
    {
        base::m_pItem = It.m_pItem;
        m_pTriangle = It.m_pTriangle;
        return *this;
    }


    //! Returns pointer to the triangle.
    vctl::MCTri *getTriangle() { return m_pTriangle; }

    //! Returns the number of increments between the positions addressed.
    //! by two iterators.
    tSize getDistance(const CTriContainerIterator& End) const
    {
        if( !m_pTriangle )
        {
            return 0;
        }
        
        // Count the distance
        tSize Count = 0;
        for( CTriContainerIterator it = *this; it != End; ++it )
        {
            ++Count;
        }
        return Count;
    }

    //! Moves iterator to the next triangle pixel.
    void advance()
    {
        do {
            if( !m_pItem )
            {
                break;
            }
            else
            {
                next();
            }
        } while( m_pItem && m_pItem->isError() );
    }

protected:
    //! Pointer to the actual triangle.
    vctl::MCTri *m_pTriangle;

protected:
    //! Initializes the triangle iterator.
    void init(vctl::MCTri *pTriangle)
    {
        m_pTriangle = pTriangle;
        if( !m_pTriangle )
        {
            m_pItem = NULL;
            return;
        }

        m_pItem = getTriangleFeatures(pTriangle);

        // Find the first evaluated triangle
        if( m_pItem->isError() )
        {
            advance();
        }
    }

    //! Moves iterator to the next pixel.
    void next()
    {
        m_pTriangle = m_pTriangle->GetNext();
        m_pItem = (m_pTriangle) ? getTriangleFeatures(m_pTriangle) : NULL;
    }
};


//==============================================================================
/*!
 * Container encapsulating list of triangles.
 */
class CTriContainer
{
public:
    //! Iterator is declared!
    enum { ITERATOR_DECLARED };
    
    //! Iterator type.
    typedef CTriContainerIterator tIterator;
    
    //! Const iterator type.
    typedef CTriContainerIterator tConstIterator;

public:
    //! Constructor.
    CTriContainer(vctl::MCTri *pFirstTriangle) : m_pTriangle(pFirstTriangle)
    {
        MDS_ASSERT(m_pTriangle);
    }

    //! Destructor.
    ~CTriContainer() {}


    //! Returns iterator that points at the first triangle.
    tIterator getBegin()
    {
        return tIterator(m_pTriangle);
    }
    tConstIterator getBegin() const
    {
        return tConstIterator(m_pTriangle);
    }

    //! Returns iterator that points beyond the last triangle in the list.
    tIterator getEnd()
    {
        return tIterator(NULL);
    }
    tConstIterator getEnd() const
    {
        return tConstIterator(NULL);
    }

protected:
    //! Pointer to the first triangle.
    vctl::MCTri *m_pTriangle;
};


} // namespace seg
} // namespace mds

#endif // MDS_TRICONTAINER_H

