//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2005/12/15                          \n
 *
 * $Id: mdsTriEdgeIterator.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Triangle edge pixels iterator.
 */

#ifndef MDS_TRIEDGEITERATOR_H
#define MDS_TRIEDGEITERATOR_H

#include "mdsTriBase.h"

// MDSTk
#include <MDSTk/Base/mdsMemory.h>
#include <MDSTk/Base/mdsIterator.h>
#include <MDSTk/Math/mdsBase.h>

// VCTL
#include <VectorEntity/mcentity.h>


namespace mds
{

// Declare iterator traits first...
MDS_ITERATOR_TRAITS(seg, CTriEdgeIterator, seg::CTriBase::tPixel);

namespace seg
{

//==============================================================================
/*!
 * Iterator used to traverse points lying on triangle edge.
 * - TODO: forEach() concept.
 */
class CTriEdgeIterator
    : public mds::base::CIteratorBase<CTriEdgeIterator>
    , public mds::base::CContainerHolder<CTriBase::tImage, CTriBase::tPixel>
{
public:
    //! Iterator base.
    typedef mds::base::CContainerHolder<CTriBase::tImage, CTriBase::tPixel> base;

    //! Base types.
    typedef CTriBase::tImage tImage;
    typedef CTriBase::tPixel tPixel;

public:
    //! Constructor.
    CTriEdgeIterator(tImage *pImage, vctl::MCEdge *pEdge)
    {
        init(pImage, pEdge);
    }

    //! Copy constructor
    CTriEdgeIterator(const CTriEdgeIterator& It)
        : base(It)
    {
        mds::base::memCopy(&m_Impl, &It.m_Impl, 1);
    }

    //! Destructor
    ~CTriEdgeIterator() {}

    //! Assignment operator
    CTriEdgeIterator& operator=(const CTriEdgeIterator& It)
    {
        base::m_pContainer = It.m_pContainer;
        base::m_pItem = It.m_pItem;
        mds::base::memCopy(&m_Impl, &It.m_Impl, 1);
        return *this;
    }


    //! Returns current iterator position.
    mds::tSize getX() const { return *m_Impl.m_pX; }
    mds::tSize getY() const { return *m_Impl.m_pY; }

    //! Returns true if iterator points after the last edge pixel.
    bool isEnd() const { return (m_Impl.m_X > m_Impl.m_MaxX); }

protected:
    //! All data members.
    struct SDataMembers
    {
        //! Step in Y direction.
        mds::tSize m_StepY;

        //! Parameters of the Bresenham's algorithm.
        mds::tSize m_K1, m_K2, m_P;

        //! Maximal X coordinate of the edge.
        mds::tSize m_MaxX;

        //! Pixel position.
        mds::tSize m_X, m_Y;

        //! Iterator position
        mds::tSize *m_pX, *m_pY;
    };

    //! Data members.
    SDataMembers m_Impl;

public:
    //! Moves iterator to the next edge pixel.
    void advance()
    {
        if( isEnd() )
        {
            return;
        }
        ++m_Impl.m_X;
        if( m_Impl.m_P < 0 )
        {
            m_Impl.m_P += m_Impl.m_K1;
        }
        else
        {
            m_Impl.m_P += m_Impl.m_K2;
            m_Impl.m_Y += m_Impl.m_StepY;
        }
        m_pItem = m_pContainer->getPtr(*m_Impl.m_pX, *m_Impl.m_pY);
    }

protected:
    //! Initializes the edge iterator.
    void init(tImage *pImage, vctl::MCEdge *pEdge)
    {
        MDS_ASSERT(pImage && pEdge);

        // Get vertices of the edge
        vctl::MCVertex *pVertices[2];
        pEdge->GetVerticeS(pVertices);

        // Round the line end-point positions
        int iX1 = mds::math::round2Int(pVertices[0]->GetX());
        int iY1 = mds::math::round2Int(pVertices[0]->GetY());
        int iX2 = mds::math::round2Int(pVertices[1]->GetX());
        int iY2 = mds::math::round2Int(pVertices[1]->GetY());

        // Difference in x and y axes
        int iDX = mds::math::getAbs(iX2 - iX1);
        int iDY = mds::math::getAbs(iY2 - iY1);

        // Exchange x and y axes
        if( iDY > iDX )
        {
            m_Impl.m_pX = &m_Impl.m_Y;
            m_Impl.m_pY = &m_Impl.m_X;

            mds::math::swap(iX1, iY1);
            mds::math::swap(iX2, iY2);
            mds::math::swap(iDX, iDY);
        }
        else
        {
            m_Impl.m_pX = &m_Impl.m_X;
            m_Impl.m_pY = &m_Impl.m_Y;
        }

        // Exchange line end-points
        if( iX1 > iX2 )
        {
            mds::math::swap(iX1, iX2);
            mds::math::swap(iY1, iY2);
        }

        // Step in y axis
        if( iY1 > iY2 )
        {
            m_Impl.m_StepY = -1;
        }
        else
        {
            m_Impl.m_StepY = 1;
        }

        // Bresenham's algorithm
        m_Impl.m_K1 = 2 * iDY;
        m_Impl.m_K2 = 2 * (iDY - iDX);
        m_Impl.m_P  = 2 * iDY - iDX;

        // Starting point
        m_Impl.m_X = iX1;
        m_Impl.m_Y = iY1;

        // End-point
        m_Impl.m_MaxX = iX2;

        // Save pointer to the image
        m_pContainer = pImage;

        // Initialize pointer to the first pixel
        m_pItem = pImage->getPtr(*m_Impl.m_pX, *m_Impl.m_pY);
    }
};


} // namespace seg
} // namespace mds

#endif // MDS_TRIEDGEITERATOR_H

