//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2005/10/17                          \n
 *
 * $Id: mdsTriFeatures.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Feature vector characterizing underlying image.
 * - May be extracted for just one triangle or for a group of triangles.
 */

#ifndef MDS_TRIFEATURES_H
#define MDS_TRIFEATURES_H

#include "mdsTriBase.h"
#include "mdsTriData.h"

// MDSTk
#include <MDSTk/Base/mdsGlobalLog.h>
#include <MDSTk/Math/mdsStaticVector.h>

// VCTL
#include <VectorEntity/mcentity.h>


namespace mds
{
namespace seg
{

//! If the macro is defined, the variance is used for classification.
//#define USE_VARIANCE

//! If the macro is defined, the local moments are used for classification.
//#define USE_LOCAL_MOMENTS

//! If the macro is defined, the Haralick's texture features are used for classification.
//#define USE_HARALICK_FEATURES

//! If the macro is defined, the LBPs are used for classification.
//#define USE_LBP


//==============================================================================
/*
 * Global constants.
 */

namespace TriFeatures
{

//! The number of basic triangle features.
#ifdef USE_VARIANCE
    const int NUM_OF_BASIC_FEATURES     = 2;
#else
    const int NUM_OF_BASIC_FEATURES     = 1;
#endif // USE_VARIANCE

//! The number of local moments (local texture features).
#ifdef USE_LOCAL_MOMENTS
    const int NUM_OF_MOMENTS            = 21;
#else
    const int NUM_OF_MOMENTS            = 0;
#endif // USE_LOCAL_MOMENTS

//! The number of local moments (local texture features).
#ifdef USE_HARALICK_FEATURES
    const int NUM_OF_HARALICK_FEATURES  = 16;
    const int NUM_OF_LEVELS             = 32;
#else
    const int NUM_OF_HARALICK_FEATURES  = 0;
    const int NUM_OF_LEVELS             = 0;
#endif // USE_HARALICK_FEATURES

//! Size of the LBP histogram.
#ifdef USE_LBP
    const int LBP_HISTOGRAM_SIZE        = 255;
#else
    const int LBP_HISTOGRAM_SIZE        = 0;
#endif // USE_LOCAL_MOMENTS

//! Total number of all triangle features.
const int NUM_OF_FEATURES               = NUM_OF_BASIC_FEATURES + NUM_OF_MOMENTS + NUM_OF_HARALICK_FEATURES + LBP_HISTOGRAM_SIZE;

//! Minimal number of pixels neccessary to evaluate texture features.
#if defined(USE_LOCAL_MOMENTS) || defined(USE_HARALICK_FEATURES) || defined(USE_LBP)
    const int MIN_NUM_OF_PIXELS         = 32;
#else
    const int MIN_NUM_OF_PIXELS         = 4;
#endif // USE_LOCAL_MOMENTS || USE_HARALICK_FEATURES || USE_LBP

}


//==============================================================================
/*!
 * Feature vector extracted from triangle(s).
 */
class CTriFeatures : public CTriBase, public mds::math::CStaticVector<double, TriFeatures::NUM_OF_FEATURES>
{
public:
    //! Type representing feature value.
    typedef double tFeature;

    //! Feature vector.
    typedef mds::math::CStaticVector<double, TriFeatures::NUM_OF_FEATURES> tVector;

    //! The number of features.
    enum { NUM_OF_FEATURES = TriFeatures::NUM_OF_FEATURES };

    //! Indexes of the most important features.
    enum
    {
        MEAN_VALUE              = 0,
        VARIANCE,
        FIRST_LOCAL_MOMENT      = MEAN_VALUE + TriFeatures::NUM_OF_BASIC_FEATURES,
        FIRST_HARALICK_FEATURE  = FIRST_LOCAL_MOMENT + TriFeatures::NUM_OF_MOMENTS
    };

    //! Error codes.
    enum
    {
        ERROR_NOT_EVALUATED     = 1,
        ERROR_CANNOT_EVALUATE   = 2
    };

public:
    //! Default constructor.
    CTriFeatures() : m_iError(ERROR_NOT_EVALUATED), m_iNumOfPixels(0) {}

    //! Destructor.
    ~CTriFeatures() {}


    //! Returns value of the i-th feature.
    tFeature& getFeature(tSize i) { return get(i); }
    const tFeature& getFeature(tSize i) const { return get(i); }

    //! Sets value of the i-th feature.
    void setFeature(tSize i, tFeature Value) { set(i, Value); }


    //! Returns the number of pixels taken into account.
    int getNumOfPixels() const { return m_iNumOfPixels; }

    //! Sets the number of pixels.
    void setNumOfPixels(int iValue) { m_iNumOfPixels = iValue; }


    //! Returns true if the previous evaluation of features failed.
    bool isError() const { return (m_iError != 0); }

    //! Returns current error code.
    int getError() const { return m_iError; }

    //! Sets the error code.
    void setError(int iError) { m_iError = iError; }


    //! Returns mean value of all pixels in the triangle.
    tFeature& getPixelMean() { return get(MEAN_VALUE); }
    const tFeature& getPixelMean() const { return get(MEAN_VALUE); }

    //! Sets the pixel mean.
    void setPixelMean(tFeature Value) { set(MEAN_VALUE, Value); }

#ifdef USE_VARIANCE
    //! Returns variance of all pixels in the triangle.
    tFeature& getPixelVariance() { return get(VARIANCE); }
    const tFeature& getPixelVariance() const { return get(VARIANCE); }

    //! Sets the pixel variance.
    void setPixelVariance(tFeature Value) { set(VARIANCE, Value); }
#else
    //! Returns variance of all pixels in the triangle.
    tFeature& getPixelVariance() { return m_PixelVariance; }
    const tFeature& getPixelVariance() const { return m_PixelVariance; }

    //! Sets the pixel variance.
    void setPixelVariance(tFeature Value) { m_PixelVariance = Value; }
#endif // USE_VARIANCE

#ifdef USE_LOCAL_MOMENTS
    //! Returns the i-th local moment.
    tFeature& getLocalMoment(tSize i) { return get(FIRST_LOCAL_MOMENT + i); }
    const tFeature& getLocalMoment(tSize i) const { return get(FIRST_LOCAL_MOMENT + i); }

    //! Sets the i-th local moment.
    void setLocalMoment(tSize i, tFeature Value) { return set(FIRST_LOCAL_MOMENT + i, Value); }
#endif // USE_LOCAL_MOMENTS

#ifdef USE_HARALICK_FEATURES
    //! Returns the i-th Haralick's feature.
    tFeature& getHaralickFeature(tSize i) { return get(FIRST_HARALICK_FEATURE + i); }
    const tFeature& getHaralickFeature(tSize i) const { return get(FIRST_HARALICK_FEATURE + i); }

    //! Sets the i-th Haralick's feature.
    void setHaralickFeature(tSize i, tFeature Value) { return set(FIRST_HARALICK_FEATURE + i, Value); }
#endif // USE_HARALICK_FEATURES

protected:
    //! Error while evaluating the features.
    int m_iError;

    //! Total number of pixels.
    int m_iNumOfPixels;

#ifndef USE_VARIANCE
    //! Variance of the pixel values.
    tFeature m_PixelVariance;
#endif // USE_VARIANCE
};


//==============================================================================
/*!
 * Class used to extract image features for a triangle.
 */
class CTriFeaturesExtractor : public CTriBase
{
public:
    //! Type representing feature.
    typedef CTriFeatures::tFeature tFeature;

    //! Feature vector.
    typedef CTriFeatures tVector;

public:
    //! Default constructor.
    CTriFeaturesExtractor() { clear(); }

    //! Destructor.
    ~CTriFeaturesExtractor() {}

    //! Extracts and acumulates image charakteristics from given triangle.
    void acumulate(tImage *pImage, vctl::MCTri *pTriangle);

    //! Acumulates image charakteristics from a given feature extractor.
    void acumulate(const CTriFeaturesExtractor& Extractor);

    //! Evaluates features according to the acumulated image characteristics
    //! and writes result to a given feature vector.
    //! - Returns false on failure.
    //! - Concrete error code is saved in the feature vector.
    bool evaluate(tVector *pVector, bool bCheckNumOfPixels = true);

    //! Clears all internal structures and counters.
    void clear();

protected:
    //! Sum of all pixels.
    double m_dSum, m_dSumSqr;

    //! Number of pixels counted.
    int m_iNumOfPixels;

#ifdef USE_LOCAL_MOMENTS
    //! Vector used to store minimal and maximal local moments.
    typedef mds::math::CStaticVector<double, TriFeatures::NUM_OF_MOMENTS> tLocalMoments;

    //! Minimal local moment, maximal moment and sum of all moments.
    tLocalMoments m_Moments;
#endif // USE_LOCAL_MOMENTS

#ifdef USE_HARALICK_FEATURES
    //! Co-occurence matrix.
    typedef mds::math::CStaticMatrix<int, TriFeatures::NUM_OF_LEVELS, TriFeatures::NUM_OF_LEVELS> tCooMatrix;

    //! Co-occurence matrices.
    tCooMatrix m_Matrix1, m_Matrix2, m_Matrix3, m_Matrix4;
#endif // USE_HARALICK_FEATURES
};


//==============================================================================
/*!
 * Structure containing additional triangle information.
 */
struct STriangleData
{
    //! Index of the triangle segment.
//    int m_iIndex;

    //! Triangle features.
    CTriFeatures m_Features;

    //! Default constructor
//    STriangleData() : m_iIndex(-1) {}
    STriangleData() {}
};


//==============================================================================
/*
 * Global functions.
 */

//! Returns index of the triangle region or -1 if additional triangle
//! data doesn't exist.
inline int getTriangleRegion(vctl::MCTri *pTriangle)
{
/*    STriangleData *pData = (STriangleData *)getTriangleDataPtr(pTriangle);

    return (pData) ? pData->m_iIndex : -1;*/
    
    return (pTriangle) ? pTriangle->GetObject() : -1;
}


//! Sets triangle region index.
//! - Creates triangle data if it doesn't exist.
inline void setTriangleRegion(vctl::MCTri *pTriangle, int iIndex)
{
/*    STriangleData *pData = (STriangleData *)getTriangleDataPtr(pTriangle);

    if( !pData )
    {
        pData = new STriangleData;
        setTriangleDataPtr(pTriangle, pData);
    }

    pData->m_iIndex = iIndex;*/
    
    if( pTriangle )
    {
        pTriangle->SetObject(iIndex);
    }
}


//! Returns pointer to the triangle feature vector.
inline CTriFeatures *getTriangleFeatures(vctl::MCTri *pTriangle)
{
    STriangleData *pData = (STriangleData *)getTriangleDataPtr(pTriangle);

    if( !pData )
    {
        pData = new STriangleData;
        setTriangleDataPtr(pTriangle, pData);
    }

    return &(pData->m_Features);
}


} // namespace seg
} // namespace mds

#endif // MDS_TRIFEATURES_H

