//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2006/12/14                          \n
 *
 * $Id: mdsTriHomogeneityMeasure.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Triangle homogeneity measure.
 */

#ifndef MDS_TRIHOMOGENEITYMEASURE_H
#define MDS_TRIHOMOGENEITYMEASURE_H

#include "mdsTriFeatures.h"

// MDSTk
#include <MDSTk/Math/mdsVectorFunctions.h>

// STL
#include <cmath>


namespace mds
{
namespace seg
{

//==============================================================================
/*!
 * Abstract class of a homogeneity measure.
 */
class CTriHomogeneityMeasure
{
public:
    //! Default constructor.
    CTriHomogeneityMeasure(double dThreshold = 0.5) : m_dThreshold(dThreshold) {}

    //! Virtual destructor.
    virtual ~CTriHomogeneityMeasure() {}

    //! Computes homogeneity of a given triangle feature vector.
    virtual double getHomogeneity(const CTriFeatures& v1) = 0;


    //! Compares homogeneity of a given triangle feature vector against
    //! the internal threshold.
    //! - Returns true if the homogeneity is greater than the threshold.
    bool checkHomogeneity(const CTriFeatures& v1)
    {
        return (this->getHomogeneity(v1) > m_dThreshold);
    }

    //! Returns value of the internal threshold.
    double getThreshold() { return m_dThreshold; }

    //! Sets the internal threshold.
    void setThreshold(double dValue) { m_dThreshold = dValue; }

protected:
    //! Internal threshold.
    double m_dThreshold;
};


//==============================================================================
/*!
 * Basic homogeneity measure.
 * - Checks variance of the triangle.
 */
class CTriVarianceHomogeneity : public CTriHomogeneityMeasure
{
public:
    //! Default parameters.
    static const double DEFAULT_THRESHOLD;
    static const double DEFAULT_RHO;

public:
    //! Default constructor.
    CTriVarianceHomogeneity(double dThreshold = DEFAULT_THRESHOLD,
                            double dRho = DEFAULT_RHO
                            );

    //! Virtual destructor.
    virtual ~CTriVarianceHomogeneity() {}

    //! Computes homogeneity of a given feature vector.
    virtual double getHomogeneity(const CTriFeatures& v1);

protected:
    //! Parameter affecting sensitivity of the measure.
    double m_dInvRho;
};


} // namespace seg
} // namespace mds

#endif // MDS_TRIHOMOGENEITYMEASURE_H

