//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2005/10/17                          \n
 *
 * $Id: mdsTriIterator.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Triangle pixels iterator.
 */

#ifndef MDS_TRIITERATOR_H
#define MDS_TRIITERATOR_H

#include "mdsTriBase.h"

// MDSTk
#include <MDSTk/Base/mdsMemory.h>
#include <MDSTk/Base/mdsIterator.h>
#include <MDSTk/Math/mdsBase.h>

// VCTL
#include <VectorEntity/mcentity.h>


namespace mds
{

// Declare iterator traits first...
MDS_ITERATOR_TRAITS(seg, CTriIterator, seg::CTriBase::tPixel);

namespace seg
{

//==============================================================================
/*!
 * Iterator used to traverse pixels inside a triangle.
 * - TODO: forEach() concept.
 */
class CTriIterator
    : public mds::base::CIteratorBase<CTriIterator>
    , public mds::base::CContainerHolder<CTriBase::tImage, CTriBase::tPixel>
{
public:
    //! Iterator base.
    typedef mds::base::CContainerHolder<CTriBase::tImage, CTriBase::tPixel> base;

    //! Base types.
    typedef CTriBase::tImage tImage;
    typedef CTriBase::tPixel tPixel;

public:
    //! Constructor.
    CTriIterator(tImage *pImage, vctl::MCTri *pTriangle)
    {
        init(pImage, pTriangle);
    }

    //! Copy constructor
    CTriIterator(const CTriIterator& It)
        : base(It)
    {
        mds::base::memCopy(&m_Impl, &It.m_Impl, 1);
    }

    //! Destructor
    ~CTriIterator() {}

    //! Assignment operator
    CTriIterator& operator=(const CTriIterator& It)
    {
        base::m_pContainer = It.m_pContainer;
        base::m_pItem = It.m_pItem;
        mds::base::memCopy(&m_Impl, &It.m_Impl, 1);
        return *this;
    }


    //! Returns current iterator position.
    mds::tSize getX() const { return m_Impl.m_X; }
    mds::tSize getY() const { return m_Impl.m_Y; }

    //! Returns true if iterator points after the last triangle
    //! pixel.
    bool isEnd() const { return (m_Impl.m_Y > m_Impl.m_iMaxY); }

protected:
    //! All data members.
    struct SDataMembers
    {
        //! Bounding rectangle.
        int m_iMinX, m_iMaxX, m_iMinY, m_iMaxY;

        //! Deltas.
        int m_iFDX1, m_iFDX2, m_iFDX3;
        int m_iFDY1, m_iFDY2, m_iFDY3;

        //! Flags.
        int m_iCY1, m_iCY2, m_iCY3;
        int m_iCX1, m_iCX2, m_iCX3;

        //! Iterator position.
        mds::tSize m_X, m_Y;
    };

    //! Data members.
    SDataMembers m_Impl;

public:
    //! Moves iterator to the next triangle pixel.
    void advance()
    {
        do {
            if( isEnd() )
            {
                break;
            }
            else
            {
                next();
            }
        } while( !isInner() );
    }

protected:
    //! Returns true if the current pixel is inside the triangle.
    bool isInner()
    {
        return (m_Impl.m_iCX1 <= 0 && m_Impl.m_iCX2 <= 0 && m_Impl.m_iCX3 <= 0);
    }

    //! Initializes the triangle iterator.
    void init(tImage *pImage, vctl::MCTri *pTriangle)
    {
        MDS_ASSERT(pTriangle);

        // Get vertices of the triangle
        vctl::MCVertex *pVertices[3];
        pTriangle->GetVerticeS(pVertices);

        // Fixed-point coordinates
        int iY1 = mds::math::round2Int(16.0 * pVertices[0]->GetY());
        int iY2 = mds::math::round2Int(16.0 * pVertices[1]->GetY());
        int iY3 = mds::math::round2Int(16.0 * pVertices[2]->GetY());

        int iX1 = mds::math::round2Int(16.0 * pVertices[0]->GetX());
        int iX2 = mds::math::round2Int(16.0 * pVertices[1]->GetX());
        int iX3 = mds::math::round2Int(16.0 * pVertices[2]->GetX());

        // Deltas
        int iDX1 = iX2 - iX1;
        int iDX2 = iX3 - iX2;
        int iDX3 = iX1 - iX3;

        int iDY1 = iY2 - iY1;
        int iDY2 = iY3 - iY2;
        int iDY3 = iY1 - iY3;

        // Fixed-point deltas
        m_Impl.m_iFDX1 = iDX1 << 4;
        m_Impl.m_iFDX2 = iDX2 << 4;
        m_Impl.m_iFDX3 = iDX3 << 4;

        m_Impl.m_iFDY1 = iDY1 << 4;
        m_Impl.m_iFDY2 = iDY2 << 4;
        m_Impl.m_iFDY3 = iDY3 << 4;

        // Bounding rectangle
        m_Impl.m_iMinX = mds::math::getMin(iX1, iX2, iX3) >> 4;
        m_Impl.m_iMaxX = mds::math::getMax(iX1, iX2, iX3) >> 4;
        m_Impl.m_iMinY = mds::math::getMin(iY1, iY2, iY3) >> 4;
        m_Impl.m_iMaxY = mds::math::getMax(iY1, iY2, iY3) >> 4;

        // Half-edge constants
        int iC1 = iDX1 * iY1 - iDY1 * iX1;
        int iC2 = iDX2 * iY2 - iDY2 * iX2;
        int iC3 = iDX3 * iY3 - iDY3 * iX3;

        // Initialization
        m_Impl.m_iCY1 = iC1 - iDX1 * (m_Impl.m_iMinY << 4) + iDY1 * (m_Impl.m_iMinX << 4);
        m_Impl.m_iCY2 = iC2 - iDX2 * (m_Impl.m_iMinY << 4) + iDY2 * (m_Impl.m_iMinX << 4);
        m_Impl.m_iCY3 = iC3 - iDX3 * (m_Impl.m_iMinY << 4) + iDY3 * (m_Impl.m_iMinX << 4);

        m_Impl.m_iCX1 = m_Impl.m_iCY1;
        m_Impl.m_iCX2 = m_Impl.m_iCY2;
        m_Impl.m_iCX3 = m_Impl.m_iCY3;

        m_Impl.m_X = m_Impl.m_iMinX;
        m_Impl.m_Y = m_Impl.m_iMinY;

        m_pContainer = pImage;
        m_pItem = pImage->getPtr(m_Impl.m_X, m_Impl.m_Y);

        // Find the first pixel
        if( !isInner() )
        {
            advance();
        }
    }

    //! Moves iterator to the next pixel.
    void next()
    {
        ++m_pItem;
        if( ++m_Impl.m_X > m_Impl.m_iMaxX )
        {
            m_Impl.m_X = m_Impl.m_iMinX;
            ++m_Impl.m_Y;
            m_Impl.m_iCY1 -= m_Impl.m_iFDX1;
            m_Impl.m_iCY2 -= m_Impl.m_iFDX2;
            m_Impl.m_iCY3 -= m_Impl.m_iFDX3;
            m_Impl.m_iCX1 = m_Impl.m_iCY1;
            m_Impl.m_iCX2 = m_Impl.m_iCY2;
            m_Impl.m_iCX3 = m_Impl.m_iCY3;
            m_pItem = m_pContainer->getPtr(m_Impl.m_X, m_Impl.m_Y);
        }
        else
        {
            m_Impl.m_iCX1 += m_Impl.m_iFDY1;
            m_Impl.m_iCX2 += m_Impl.m_iFDY2;
            m_Impl.m_iCX3 += m_Impl.m_iFDY3;
        }
    }
};


} // namespace seg
} // namespace mds

#endif // MDS_TRIITERATOR_H

