//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2006/11/29                          \n
 *
 * $Id: mdsTriMembershipFCM2.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Membership function based on fuzzy c-Means clustering algorithm.
 * - Modified approach which classifies predefined sub-regions instead of
 *   triangles.   
 */

#ifndef MDS_TRIMEMBERSHIPFCM2_H
#define MDS_TRIMEMBERSHIPFCM2_H

#define FCM_LOGGING_ENABLED

#include "mdsTriMembershipFunc.h"

// MDSTk
#include <MDSTk/Base/mdsSTLContainer.h>
#include <MDSTk/Math/Algorithm/mdsFuzzyCMeans.h>

// STL
#include <vector>


namespace mds
{
namespace seg
{

//==============================================================================
/*!
 * Region membership function.
 * - Fuzzy c-Means (FCM) clustering algorithm.
 * - Modified approach which classifies predefined sub-regions instead of
 *   triangles.
 */
class CTriMembershipFCM2 : public CTriMembershipFunc
{
public:
    //! Vector of image features extracted for every sub-region.
    typedef std::vector<CTriFeatures> tSubRegions;

public:
    //! Default constructor.
    CTriMembershipFCM2(tSubRegions& SubRegions, double dWeight);

    //! Virtual destructor.
    virtual ~CTriMembershipFCM2() {}

    //! Computes membership function of all triangles in a given list.
    //! - Return false on failure.
    virtual bool operator()(vctl::MCTri *pFirstTriangle);

    //! Computes membership function of all triangles in a given list.
    //! - Predefined number of clusters.
    //! - Return false on failure.
    virtual bool operator()(vctl::MCTri *pFirstTriangle,
                            mds::tSize NumOfClasses
                            );

protected:
    //! Wrapped STL container.
    typedef mds::base::CSTLContainer<tSubRegions> tContainer;

protected:
    //! Used clustering algorithm.
    mds::math::CFuzzyCMeans<tContainer, NUM_OF_FEATURES> m_FCM;

    //! Reference to the vector of image features.
    tSubRegions& m_SubRegions;
};


} // namespace seg
} // namespace mds

#endif // MDS_TRIMEMBERSHIPFCM2_H

