//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2005/07/26                          \n
 *
 * $Id: mdsTriMesh.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Adaptive triangular mesh based on 2D Delaunay triangulation.
 */

#ifndef MDS_TRIMESH_H
#define MDS_TRIMESH_H

#include "mdsDelaunayTri.h"
#include "mdsTriFeatures.h"
#include "mdsTriSimilarityMeasure.h"
#include "mdsTriHomogeneityMeasure.h"

// STL
#include <vector>


namespace mds
{
namespace seg
{

//==============================================================================
/*!
 * Class representing adaptive triangular mesh.
 */
class CTriMesh : public CDelaunayTri, public CTriBase
{
public:
    //! Smart pointer type.
    //! - Declares type tSmartPtr.
    MDS_SHAREDPTR(CTriMesh);

    //! Standard method getEntityName().
    MDS_ENTITY_NAME("TriMesh");

    //! Standard method getEntityCompression().
    MDS_ENTITY_COMPRESSION(mds::mod::CC_RAW);

    //! Standard method getEntityBlockSize().
    MDS_ENTITY_BLOCK_SIZE(4096);

    //! Vector of image features.
    typedef std::vector<CTriFeatures> tSegments;

    //! Default parameters.
    static const tSize DEFAULT_CELL;
    static const tSize DEFAULT_CELL_MARGIN;
    static const double DEFAULT_INIT_THRESHOLD;
    static const double DEFAULT_SPLITTING_THRESHOLD;
    static const double DEFAULT_DISTANCE_THRESHOLD;

public:
    //! Default constructor.
    CTriMesh();

    //! Constructor.
    CTriMesh(mds::img::CDImage *pEdgeImage);

    //! Virtual destructor.
    virtual ~CTriMesh();


    //! Returns pointer to the underlying image.
    mds::img::CDImage *getImagePtr() { return m_spImage; }

    //! Selection of the image for segmentation.
    void setImage(mds::img::CDImage *pImage) { m_spImage = pImage; }

    //! Returns pointer to the underlying edge image.
    mds::img::CDImage *getEdgeImagePtr() { return m_spEdgeImage; }

    //! Selection of the image for segmentation.
    void setEdgeImage(mds::img::CDImage *pEdgeImage) { m_spEdgeImage = pEdgeImage; }


    //! Returns number of regions.
    int getNumOfSegments() const { return m_iNumOfSegments; }

    //! Sets the number of regions.
    void setNumOfSegments(int i) { m_iNumOfSegments = i; }

    //! Returns vector of image features.
    tSegments& getSegments() { return m_Segments; }


    //! Initializes the Delaunay triangulation.
    //! - Uses vertex detector to initialize the list of vertices.
    //! - Removes all triangles.
    //! - Underlying edge image must be set.
    bool init(double dThreshold = DEFAULT_INIT_THRESHOLD,
              tSize CellSize = DEFAULT_CELL,
              tSize NumOfNodes = DEFAULT_NUMBER_OF_NODES,
              double dMinEdgeLength = DEFAULT_MIN_EDGE_LENGTH
              );

    //! Evaluates image features of all triangles in the mesh.
    //! - Underlying image must be set!
    bool evaluateFeatures(bool bSkipSmall = true);

    //! Evaluates features of all segments in the mesh.
    //! - Underlying image must be set!
    //! - Triangles in the mesh must be classified to regions!
    bool evaluateSegmentFeatures(bool bSkipSmall = true);

    //! Adaptation of the triangular mesh to the underlying image.
    //! - Edge splitting algorithm.
    //! - Both images must be set!
    bool edgeSplitting(double dMinLength = DEFAULT_MIN_EDGE_LENGTH,
                       double dOrientThreshold = DEFAULT_SPLITTING_THRESHOLD
                       );

    //! Adaptation of the triangular mesh to the underlying image.
    //! - Triangle splitting algorithm.
    //! - Both images must be set!
    bool triangleSplitting(CTriHomogeneityMeasure& Measure,
                           double dMinLength = DEFAULT_MIN_EDGE_LENGTH
                           );

    //! Simple region growing segmentation. Every region grows while a given
    //! similarity predicate holds and boundary edge is not touched.
    //! - Image features must be evaluated!
    //! - Both images must be set!
    bool regionGrowing(CTriSimilarityMeasure& Measure,
                       double dBThreshold = DEFAULT_DISTANCE_THRESHOLD
                       );

    //! Aglomerative merging of regions in the mesh.
    //! - Evaluates image features.
    //! - Both images must be set!
    bool regionMerging(CTriSimilarityMeasure& Measure,
                       double dBThreshold = DEFAULT_DISTANCE_THRESHOLD,
                       tSize NumOfRegions = -1
                       );

    //! Classify all non-classified triangles with respect to the neighbours.
    //! - This method doesn't use image features for classification!
    //! - Both images must be set!
    bool classifyAdjacent(double dBThreshold = DEFAULT_DISTANCE_THRESHOLD);

    //! Removes noisy (wrongly classified) tetrahedrons in the mesh.
    //! - Principle of the noise removal algorithm is similar to the median
    //!   filtering approach which is well known in image processing.
    //! - This method doesn't use image features for classification!
    bool noiseReduction();

    //! Visualizes the triangular mesh.
    //! - Draws all triangles as simple lines to a given image.
    void visualize(mds::img::CDImage *pImage);


    //! Writes triangular mesh to a given output channel in STL format.
    //! - Returns false on failure.
    //! - Modified Premysl Krsek's original code.
    bool saveSTL(mds::mod::CChannel& Channel);

    //! Writes triangular mesh to a given output channel in VRML format.
    //! - Returns false on failure.
    //! - Modified Premysl Krsek's original code.
    bool saveVRML(mds::mod::CChannel& Channel);


    //! Serializes the triangular mesh.
    template <class S>
    void serialize(mds::mod::CChannelSerializer<S>& Writer)
    {
        // Serialize the Delaunay trianglation
        CDelaunayTri::serialize(Writer);
    }

    //! Deserializes the triangular mesh.
    template <class S>
    void deserialize(mds::mod::CChannelSerializer<S>& Reader)
    {
        // Deserialize the Delaunay trianglation
        CDelaunayTri::deserialize(Reader);

        // Evaluate the number of regions
        int iMaxRegion = -1;
        vctl::MCTri *pTriangle = getFirstTriangle();
        for( ; pTriangle; pTriangle = pTriangle->GetNext() )
        {
            // Get triangle region
            int iRegion = getTriangleRegion(pTriangle);
            if( iRegion > iMaxRegion )
            {
                iMaxRegion = iRegion;
            }
        }

        // Set the number of regions
        m_iNumOfSegments = iMaxRegion + 1;
    }

protected:
    //! Vector of points.
    typedef std::vector<vctl::MCPoint3D> tPoints;

protected:
    //! Pointer to the underlying image.
    mds::img::CDImagePtr m_spImage;

    //! Pointer to the underlying edge image.
    mds::img::CDImagePtr m_spEdgeImage;

    //! Number of segments in the mesh.
    int m_iNumOfSegments;

    //! Vector of image features extracted for every region.
    tSegments m_Segments;

    //! Currently used homogeneity measure.
    CTriHomogeneityMeasure *m_pHomogeneityMeasure;

protected:
//public:
    //! Checks if a given edge is a boundary edge.
    //! - Returns true if the edge and some image edge are parallel.
    bool checkBoundaryEdge(vctl::MCEdge *pEdge,
                           double dThreshold = DEFAULT_DISTANCE_THRESHOLD
                           );

    //! Eliminates unused indexes of regions shaking down all currently
    //! assigned indexes.
    void reassignRegions();

    //! Splits a given edge.
    //! - Returns true if the edge was divided.
    bool splitEdge(vctl::MCEdge *pEdge,
                   double dMinLength,
                   double dThreshold,
                   tPoints& Points
                   );

    //! Inserts triangle edges to the list of edges.
    static void edgeSplittingInsertFunc(vctl::MCTri *pTriangle,
                                        CDelaunayTri *pMesh
                                        );

    //! Splits a given triangle.
    //! - Returns true if the triangle was divided.
    bool splitTriangle(vctl::MCTri *pTriangle, double dMinLength);

    //! Evaluates triangle homogeneity and inserts it to the sorted queue.
    static void triangleSplittingInsertFunc(vctl::MCTri *pTriangle,
                                            CDelaunayTri *pMesh
                                            );

    //! Unsorts triangle from the queue.
    static void triangleSplittingRemoveFunc(vctl::MCTri *pTriangle,
                                            CDelaunayTri *pMesh
                                            );
};


//=============================================================================
/*
 * Basic template instances and type definitions.
 * - Using smart pointers.
 */

//! Triangular mesh.
typedef CTriMesh::tSmartPtr CTriMeshPtr;


} // namespace seg
} // namespace mds

#endif // MDS_TRIMESH_H

