 //==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2006/11/17                          \n
 *
 * $Id: mdsTriSimilarityMeasure.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Triangles similarity measure.
 */

#ifndef MDS_TRISIMILARITYMEASURE_H
#define MDS_TRISIMILARITYMEASURE_H

#include "mdsTriFeatures.h"

// MDSTk
#include <MDSTk/Math/mdsVectorFunctions.h>

// STL
#include <cmath>


namespace mds
{
namespace seg
{

//==============================================================================
/*!
 * Abstract class of a similarity measure.
 */
class CTriSimilarityMeasure
{
public:
    //! Default constructor.
    CTriSimilarityMeasure(double dThreshold = 0.5) : m_dThreshold(dThreshold) {}

    //! Virtual destructor.
    virtual ~CTriSimilarityMeasure() {}

    //! Computes similarity of two given triangle feature vectors.
    virtual double getSimilarity(const CTriFeatures& v1, const CTriFeatures& v2) = 0;


    //! Compares similarity of two given triangle feature vectors against
    //! the internal threshold.
    //! - Returns true if the similarity is greater than the threshold.
    bool checkSimilarity(const CTriFeatures& v1, const CTriFeatures& v2)
    {
        return (this->getSimilarity(v1, v2) > m_dThreshold);
    }

    //! Returns value of the internal threshold.
    double getThreshold() { return m_dThreshold; }

    //! Sets the internal threshold.
    void setThreshold(double dValue) { m_dThreshold = dValue; }

protected:
    //! Internal threshold.
    double m_dThreshold;
};


//==============================================================================
/*!
 * Basic similarity measure.
 * - Compares mean voxel values.
 */
class CTriMeanValueSimilarity : public CTriSimilarityMeasure
{
public:
    //! Default parameters.
    static const double DEFAULT_THRESHOLD;
    static const double DEFAULT_RHO;

public:
    //! Default constructor.
    CTriMeanValueSimilarity(double dThreshold = DEFAULT_THRESHOLD,
                            double dRho = DEFAULT_RHO
                            );

    //! Virtual destructor.
    virtual ~CTriMeanValueSimilarity() {}

    //! Computes similarity of two given feature vectors.
    virtual double getSimilarity(const CTriFeatures& v1, const CTriFeatures& v2);

protected:
    //! Parameter affecting sensitivity of the measure.
    double m_dInvRho;
};


} // namespace seg
} // namespace mds

#endif // MDS_TRISIMILARITYMEASURE_H

