//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2005/07/26                          \n
 *
 * $Id: mdsTriVertexDetector.h 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Initial vertex detector.
 */

#ifndef MDS_TRIVERTEXDETECTOR_H
#define MDS_TRIVERTEXDETECTOR_H

#include "mdsTriBase.h"

// MDSTk
#include <MDSTk/Base/mdsAssert.h>
#include <MDSTk/Base/mdsIdentifier.h>
#include <MDSTk/Image/mdsImageFunctions.h>
#include <MDSTk/Math/mdsBase.h>

// VCTL
#include <VectorEntity/mcvertices.h>

// STL
#include <sstream>


namespace mds
{
namespace seg
{

//==============================================================================
/*
 * Global definitions.
 */

//! Image/volume vertex detection algorithms.
MDS_DECLARE_ID(CELL_MAX);


//==============================================================================
/*!
 * Basic vertex detector template.
 * - Parameter 'Id' is type of the detection algorithm.
 */
template <typename Id>
class CTriVertexDetector
{
public:
    //! Detects all vertices in a given image. Returned list is sorted
    //! in order of vertex strength.
    bool operator()(const mds::img::CDImage *pImage, vctl::MCVerticeS *pVertices);
};


//==============================================================================
/*!
 * Vertex detector template specialization.
 * - Just one candidate pixel (that which has maximal value) is taken
 *   from an image cell.
 */
template <>
class CTriVertexDetector<CELL_MAX>
{
public:
    //! Default constructor
    //! - Input image is divided into cells, while just one candidate
    //!   vertex is taken from each cell. Parameter 'CellSize' represents
    //!   size of these subwindows in pixels.
    //! - Threshold affecting the overall number of pre-selected candidate
    //!   vertices.
    CTriVertexDetector(double dThreshold,
                       mds::tSize CellSize,
                       mds::tSize CellMargin
                       )
        : m_dThreshold(dThreshold)
        , m_CellSize(CellSize)
        , m_CellMargin(CellMargin)
    {}

    //! Detects all vertices in a given edge image. Returned list is sorted
    //! in order of vertex significance.
    bool operator()(const mds::img::CDImage *pImage, vctl::MCVerticeS *pVertices);

protected:
    //! Threshold.
    double m_dThreshold;

    //! Cell size and margin.
    mds::tSize m_CellSize, m_CellMargin;
};


//=============================================================================
/*
 * Methods templates.
 */

// Include file containing methods templates
#include "mdsTriVertexDetector.hxx"


} // namespace seg
} // namespace mds

#endif // MDS_TRIVERTEXDETECTOR_H

