//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2006/04/03                          \n
 *
 * $Id: mdsDelaunayTetra.cpp 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - 3D Delaunay triangulation.
 */

#include <MDSTk/VectorSegmentation/mdsDelaunayTetra.h>

// MDSTk
#include <MDSTk/Base/mdsMemory.h>
#include <MDSTk/Base/mdsGlobalLog.h>
#include <MDSTk/Math/mdsBase.h>
#include <MDSTk/Math/ver1/mdsVector.h>

// STL
#include <set>
#include <deque>
#include <algorithm>


namespace mds
{
namespace seg
{

//==============================================================================
/*
 * Data members of the class CDelaunayTri.
 */
const mds::tSize CDelaunayTetra::DEFAULT_NUMBER_OF_NODES    = 5;
const double CDelaunayTetra::DEFAULT_MIN_EDGE_LENGTH        = 1.5;
const double CDelaunayTetra::DEFAULT_MIN_DIHEDRAL_ANGLE     = 0.1;      // [degrees]
//const double CDelaunayTetra::DEFAULT_MIN_DIHEDRAL_ANGLE     = 1.0;
//const double CDelaunayTetra::MAX_MOVEMENT                   = 0.1;
const double CDelaunayTetra::MAX_MOVEMENT                   = 0.001;


//==============================================================================
/*
 * Implementation of the class CDelaunayTetra.
 */
CDelaunayTetra::CDelaunayTetra()
    : m_pVertices(m_Tetrahedra.GetVerticeS())
    , m_dMinX(0.0)
    , m_dMinY(0.0)
    , m_dMinZ(0.0)
    , m_dMaxX(0.0)
    , m_dMaxY(0.0)
    , m_dMaxZ(0.0)
    , m_dMinEdgeLength(DEFAULT_MIN_EDGE_LENGTH)
    , m_dMinDihedralAngle(DEFAULT_MIN_DIHEDRAL_ANGLE)
    , m_MinEdgeLengthFunc(checkMinEdgeLengthFunc)
    , m_ConstraintFunc(checkConstraintFunc)
{
    MDS_ASSERT(m_pVertices);

    // Enable vertex existence test
    m_pVertices->SetTestExistence(true);
}


bool CDelaunayTetra::attractVertex(vctl::MCPoint3D *pPoint)
{
    return false;
}


bool CDelaunayTetra::init(vctl::MCPoint3D *pMin,
                          vctl::MCPoint3D *pMax,
                          mds::tSize NumOfNodes
                          )
{
    MDS_ASSERT(NumOfNodes > 1);

    // Remove all tetrahedra and their vertices
    clear();

    // Initialize minimal and maximal coordinates
    m_dMinX = pMin->GetX();
    m_dMinY = pMin->GetY();
    m_dMinZ = pMin->GetZ();
    m_dMaxX = pMax->GetX();
    m_dMaxY = pMax->GetY();
    m_dMaxZ = pMax->GetZ();

    // The number of nodes
/*  mds::tSize NumOfNodesX = NumOfNodes;
    mds::tSize NumOfNodesY = NumOfNodes;
    mds::tSize NumOfNodesZ = NumOfNodes;

    // Size of the mesh
    double dSizeX = pMax->GetX() - pMin->GetX();
    double dSizeY = pMax->GetY() - pMin->GetY();
    double dSizeZ = pMax->GetZ() - pMin->GetZ();

    // Distance between two neighbouring nodes
    double dStepX = dSizeX / (NumOfNodesX - 1);
    double dStepY = dSizeY / (NumOfNodesY - 1);
    double dStepZ = dSizeZ / (NumOfNodesZ - 1);

    // Estimate the number of nodes with respect to the minimal edge length
    while( dStepX < m_dMinEdgeLength && NumOfNodesX > 2 )
    {
        dStepX = dSizeX / (--NumOfNodesX - 1);
    }
    while( dStepY < m_dMinEdgeLength && NumOfNodesY > 2 )
    {
        dStepY = dSizeY / (--NumOfNodesY - 1);
    }
    while( dStepZ < m_dMinEdgeLength && NumOfNodesZ > 2 )
    {
        dStepZ = dSizeZ / (--NumOfNodesZ - 1);
    }*/

    // The number of nodes
    mds::tSize NumOfNodesX = 2;
    mds::tSize NumOfNodesY = 2;
    mds::tSize NumOfNodesZ = 2;

    // Size of the mesh
    double dSizeX = pMax->GetX() - pMin->GetX();
    double dSizeY = pMax->GetY() - pMin->GetY();
    double dSizeZ = pMax->GetZ() - pMin->GetZ();

    // Distance between two neighbouring nodes
    double dStepX = dSizeX;
    double dStepY = dSizeY;
    double dStepZ = dSizeZ;

    // Helper values
    mds::tSize MaxNodeX = NumOfNodesX - 1;
    mds::tSize MaxNodeY = NumOfNodesY - 1;
    mds::tSize MaxNodeZ = NumOfNodesZ - 1;

    // Helper vector representing 3D array of vertices
    mds::math::ver1::CVector<vctl::MCVertex *> Nodes(NumOfNodesX * NumOfNodesY * NumOfNodesZ);
    mds::tSize YOffset = NumOfNodesX;
    mds::tSize ZOffset = NumOfNodesX * NumOfNodesY;

    // Helper macro
#ifdef NODES
#undef NODES
#endif
#define NODES(i, j, k)        Nodes((i) + (j) * YOffset + (k) * ZOffset)

    // Create nodes covering initial mesh
    double dZ = pMin->GetZ();
    mds::tSize i, j, k;
    for( k = 0; k < NumOfNodesZ; ++k, dZ += dStepZ )
    {
        int BoundaryZ = 0;
        if( k == 0 )
        {
            BoundaryZ |= ZMIN_BOUNDARY;
        }
        else if( k == MaxNodeZ )
        {
            BoundaryZ |= ZMAX_BOUNDARY;
        }
        double dY = pMin->GetY();
        for( j = 0; j < NumOfNodesY; ++j, dY += dStepY )
        {
            int BoundaryY = 0;
            if( j == 0 )
            {
                BoundaryY |= YMIN_BOUNDARY;
            }
            else if( j == MaxNodeY )
            {
                BoundaryY |= YMAX_BOUNDARY;
            }
            double dX = pMin->GetX();
            for( i = 0; i < NumOfNodesX; ++i, dX += dStepX )
            {
                int Boundary = BoundaryZ | BoundaryY;
                if( i == 0 )
                {
                    Boundary |= XMIN_BOUNDARY;
                }
                else if( i == MaxNodeX )
                {
                    Boundary |= XMAX_BOUNDARY;
                }
                if( Boundary )
                {
                    NODES(i, j, k) = newVertex(dX, dY, dZ);
                    NODES(i, j, k)->SetFlag(Boundary);
                }
                else
                {
                    NODES(i, j, k) = randomVertex(dX, dY, dZ, 3.0);
                }
            }
        }
    }

    // Insert initial tetrahedra
    addTetrahedron(NODES(0, 0, 0),
                   NODES(0, MaxNodeY, 0),
                   NODES(MaxNodeX, MaxNodeY, 0),
                   NODES(0, MaxNodeY, MaxNodeZ)
                   );
    addTetrahedron(NODES(0, 0, 0),
                   NODES(MaxNodeX, MaxNodeY, 0),
                   NODES(MaxNodeX, 0, 0),
                   NODES(MaxNodeX, 0, MaxNodeZ)
                   );
    addTetrahedron(NODES(MaxNodeX, MaxNodeY, 0),
                   NODES(MaxNodeX, 0, MaxNodeZ),
                   NODES(0, MaxNodeY, MaxNodeZ),
                   NODES(MaxNodeX, MaxNodeY, MaxNodeZ)
                   );
    addTetrahedron(NODES(0, 0, 0),
                   NODES(MaxNodeX, 0, MaxNodeZ),
                   NODES(0, 0, MaxNodeZ),
                   NODES(0, MaxNodeY, MaxNodeZ)
                   );
    addTetrahedron(NODES(0, 0, 0),
                   NODES(MaxNodeX, MaxNodeY, 0),
                   NODES(0, MaxNodeY, MaxNodeZ),
                   NODES(MaxNodeX, 0, MaxNodeZ)
                   );

    // Insert others tetrahedra
    for( k = 0; k < NumOfNodesZ; ++k )
    {
        bool bCornerZ = (k == 0 || k == MaxNodeZ);
        for( j = 0; j < NumOfNodesY; ++j )
        {
            bool bCornerY = (j == 0 || j == MaxNodeY);
            for( i = 0; i < NumOfNodesX; ++i )
            {
                // Do not insert corners
                if( bCornerZ && bCornerY && (i == 0 || i == MaxNodeX) )
                {
                    continue;
                }

                // Insert a new vertex
                addVertex(NODES(i, j, k));
            }
        }
    }

    // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
/*  MDS_LOG_NOTE("CDelaunayTetra::init():");
    vctl::MCVertex *pVertex = getFirstVertex();
    while( pVertex )
    {
        // Print vertex info
        MDS_LOG_NOTE("  Vertex (" << int(pVertex) << ") Flag:" << pVertex->GetFlag());

        // Get list of registered tetrahedra
        tTetrahedra Tetrahedra;
        pVertex->GetRegisteredTetraList(Tetrahedra);

        // Traverse the list
        tTetrahedra::iterator it = Tetrahedra.begin();
        tTetrahedra::iterator itEnd = Tetrahedra.end();
        while( it != itEnd )
        {
            MDS_LOG_NOTE("    Tetrahedron (" << int(*it) << ')');
            ++it;
        }

        // Get next one
        pVertex = pVertex->GetNext();
    }*/
#endif // MDS_SEG_LOGGING_ENABLED

    // O.K.
    return true;
}


int CDelaunayTetra::checkBoundary(vctl::MCPoint3D *pPoint)
{
    // Vertex coordinates
    double dX = pPoint->GetX();
    double dY = pPoint->GetY();
    double dZ = pPoint->GetZ();

    // Initial flag
    int iBoundary = 0;

    // Check if the vertex lies inside the mesh
    // - Solved by projection of outer points to the mesh bounding box
    // - This may produce vertices similar to some existing ones!
    if( dX <= m_dMinX )
    {
        pPoint->SetX(m_dMinX);
        iBoundary |= XMIN_BOUNDARY;
    }
    else if( dX >= m_dMaxX )
    {
        pPoint->SetX(m_dMaxX);
        iBoundary |= XMAX_BOUNDARY;
    }
    if( dY <= m_dMinY )
    {
        pPoint->SetY(m_dMinY);
        iBoundary |= YMIN_BOUNDARY;
    }
    else if( dY >= m_dMaxY )
    {
        pPoint->SetY(m_dMaxY);
        iBoundary |= YMAX_BOUNDARY;
    }
    if( dZ <= m_dMinZ )
    {
        pPoint->SetZ(m_dMinZ);
        iBoundary |= ZMIN_BOUNDARY;
    }
    else if( dZ >= m_dMaxZ )
    {
        pPoint->SetZ(m_dMaxZ);
        iBoundary |= ZMAX_BOUNDARY;
    }

    // Return the vertex boundary flag
    return iBoundary;
}


vctl::MCVertex *CDelaunayTetra::randomVertex(double x, double y, double z, double dMax)
{
    // Randomly move the vertex
    vctl::MCPoint3D Point(x + m_Random.random(-dMax, dMax),
                          y + m_Random.random(-dMax, dMax),
                          z + m_Random.random(-dMax, dMax)
                          );

    // Lies vertex on the mesh boundary?
    int iBoundary = checkBoundary(&Point);

    // Create a new vertex
    vctl::MCVertex *pVertex = newVertex(&Point);
    if( iBoundary )
    {
        pVertex->SetFlag(iBoundary);
    }

    // O.K.
    return pVertex;
}


bool CDelaunayTetra::construct()
{
    // Not sure test if the init(...) method has been called
    MDS_CHECK(getNumOfTetrahedra() >= 2, return false);

    // First version - vertices sorted by image edge magnitude
    // Sort all vertices
/*    m_pVertices->SortAll();

    // Get sorted queue
    typedef vctl::MCVerticeS::MT_QUEUE tQueue;
    tQueue *pQueue = m_pVertices->GetQueue();

    // Traverse the sorted queue in decreasing order
    tQueue::MT_RITER Iterator = pQueue->GetMaxIter();
    while( !pQueue->TestEndIter(Iterator) )
    {
        // Get pointer to the vertex
        vctl::MCVertex *pVertex = (vctl::MCVertex *)Iterator->second;

        // Lies vertex on the mesh boundary?
        int iBoundary = checkBoundary(pVertex);

        // Add vertex to the triangular mesh
        if( pVertex->GetValue() >= 0.0 )
        {
            addVertex(pVertex, iBoundary);
        }

        // Get next one
        ++Iterator;
    }

    // Clear/unsort the queue
    pQueue->Clear();*/


    // Latest functional version
/*    vctl::MCVertex *pVertex = getFirstVertex();
    while( pVertex )
    {
        // Check if the vertex is already inserted
        if( !pVertex->GetRegisteredTetra() )
        {
            // Lies vertex on the mesh boundary?
            int iBoundary = checkBoundary(pVertex);

            // Add vertex to the triangular mesh
            addVertex(pVertex, iBoundary);
//            if( !addVertex(pVertex, iBoundary) )
//            {
//                MDS_LOG_NOTE("CDelaunayTetra::construct(): Failed to add a new vertex ("
//                    << pVertex->GetX() << ", "
//                    << pVertex->GetY() << ", "
//                    << pVertex->GetZ() << ")"
//                    );
//            }
        }

        // Get next vertex
        pVertex = pVertex->GetNext();
    }

    // Remove all unused vertices
    m_Tetrahedra.EraseVerticesNoTetras();*/


    // Prepare vector of new vertices
    tVertices NewVertices;
    vctl::MCVertex *pVertex = getFirstVertex();
    while( pVertex )
    {
        // Check if the vertex is already inserted
        if( !pVertex->GetRegisteredTetra() )
        {
            NewVertices.push_back(pVertex);
        }

        // Get next vertex
        pVertex = pVertex->GetNext();
    }

    // Random shuffle
    random_shuffle(NewVertices.begin(), NewVertices.end());

    // Insert all vertices randomly
    tVertices::iterator it = NewVertices.begin();
    tVertices::iterator itEnd = NewVertices.end();
    for( ; it != itEnd; ++it )
    {
        // Lies vertex on the mesh boundary?
        int iBoundary = checkBoundary(*it);

        // Add vertex to the triangular mesh
        addVertex(*it, iBoundary);
    }

    // Remove all unused vertices
    m_Tetrahedra.EraseVerticesNoTetras();

    // O.K.
    return true;
}


bool CDelaunayTetra::optimizeByVertexAdding(double dMinQuality, bool bRandomPosition)
{
    // Set quality of all tetrahedra
    vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
    while( pTetrahedron )
    {
        // Compute and set the tetrahedron quality
        pTetrahedron->SetValue(getTetrahedronQuality(pTetrahedron));

        // Get next one
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Sort all tetrahedra
    m_Tetrahedra.SortAll();

    // Get sorted queue
    vctl::MCTetraS::MT_QUEUE *pQueue = m_Tetrahedra.GetQueue();

#ifdef MDS_SEG_LOGGING_ENABLED
    MDS_LOG_NOTE("CDelaunayTetra::optimizeByVertexAdding(): started");
#endif // MDS_SEG_LOGGING_ENABLED

    // Square mininal edge length
//    double dSquareMinEdgeLength = m_dMinEdgeLength * m_dMinEdgeLength;

    // Optimize the mesh
    for( ;; )
    {
        // Get pointer to the last item
        pTetrahedron = (vctl::MCTetra *)pQueue->GetMax();

#ifdef MDS_SEG_LOGGING_ENABLED
        MDS_LOG_NOTE("Tetrahedron: ");
        MDS_LOG_NOTE("  Quality: " << pTetrahedron->GetValue());
        MDS_LOG_NOTE("  Max edge: " << pTetrahedron->GetMaxEdgeLength());
#endif // MDS_SEG_LOGGING_ENABLED

        // Check the tetrahedron quality
        if( !pTetrahedron || pTetrahedron->GetValue() < dMinQuality )
        {
            // No optimization is needed
            break;
        }

        // Check the tetrahedron edge length
//        if( pTetrahedron->GetMaxEdgeLength() > m_dMinEdgeLength )
        {
            // Get tetrahedron circumsphere
            vctl::MCSphere *pSphere = getTetrahedronCircumsphere(pTetrahedron);

            // Circumsphere center
            vctl::MCPoint3D Center;
            pSphere->GetCenter(Center);

            // Move the vertex randomly
            if( bRandomPosition )
            {
                double dRadius = 0.5 * pSphere->GetRadius();

                // Randomly move the vertex
                Center.SetX(Center.GetX() + m_Random.random(-dRadius, dRadius));
                Center.SetY(Center.GetY() + m_Random.random(-dRadius, dRadius));
                Center.SetZ(Center.GetZ() + m_Random.random(-dRadius, dRadius));
            }

            // Move the vertex to the best position
//            this->attractVertex(&Center);

            // Lies vertex on the mesh boundary?
            int iBoundary = checkBoundary(&Center);

            // Create a new vertex
            vctl::MCVertex *pVertex = newVertex(&Center);

            // Insert vertex to the mesh
            if( !addVertex(pVertex, iBoundary, normInsertFunc, normRemoveFunc) )
            {
                // Unsort the tetrahedron, no optimization
                pQueue->UnSort(pTetrahedron->GetValue(), pTetrahedron);
                pTetrahedron->SetValue(0.0);

                // Remove the vertex
                if( !pVertex->GetRegisteredTetra() )
                {
                    m_pVertices->Erase(pVertex);
                }
            }
        }
/*        else
        {
            // Unsort the tetrahedron, no optimization
            pQueue->UnSort(pTetrahedron->GetValue(), pTetrahedron);
            pTetrahedron->SetValue(0.0);
        }*/
    }

    // Clear/unsort the queue
    pQueue->Clear();

#ifdef MDS_SEG_LOGGING_ENABLED
    MDS_LOG_NOTE("CDelaunayTetra::optimizeByVertexAdding(): finished");
#endif // MDS_SEG_LOGGING_ENABLED

    // O.K.
    return true;
}


bool CDelaunayTetra::addVertex(vctl::MCVertex *pVertex,
                               int iBoundaryVertex,
                               CDelaunayTetra::tTetrahedronFunc InsertFunc,
                               CDelaunayTetra::tTetrahedronFunc RemoveFunc
                               )
{
    MDS_ASSERT(pVertex);

    // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
    MDS_LOG_NOTE("CDelaunayTetra::addVertex(): ("
        << pVertex->GetX() << ' '
        << pVertex->GetY() << ' '
        << pVertex->GetZ() << ')'
        );
#endif // MDS_SEG_LOGGING_ENABLED

    // Check if such vertex already exists
    if( pVertex->GetRegisteredTetra() )
    {
//        MDS_LOG_NOTE("CDelaunayTetra::addVertex(): Vertex already exists!");
        return false;
    }

    // Set boundary flag of the new vertex
    if( iBoundaryVertex )
    {
        pVertex->SetFlag(iBoundaryVertex);
    }

    // Find base tetrahedron
    vctl::MCTetra *pBaseTetra = findBaseTetrahedron(pVertex);
    if( !pBaseTetra )
    {
        // Create a new randomly moved vertex
/*        pVertex = randomVertex(pVertex);

        // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
        MDS_LOG_NOTE("CDelaunayTetra::addVertex(): randomly moved vertex ("
                     << pVertex->GetX() << ' '
                     << pVertex->GetY() << ' '
                     << pVertex->GetZ() << ')'
                    );
#endif // MDS_SEG_LOGGING_ENABLED

        // Try to find the base tetrahedron again
        pBaseTetra = findBaseTetrahedron(pVertex);
        if( !pBaseTetra )
        {*/
            MDS_LOG_NOTE("CDelaunayTetra::addVertex(): Failed to find base tetrahedron!");
            return false;
//        }
    }

    // Get all vertices of the base tetrahedron
    vctl::MCVertex *pVertices[4];
    pBaseTetra->GetVerticeS(pVertices);

/*#ifdef MDS_CHECK_MIN_EDGE_LENGTH
    // Check the minimal edge length
    if( m_MinEdgeLengthFunc )
    {
        if( !m_MinEdgeLengthFunc(pVertex, pVertices[0], this)
            || !m_MinEdgeLengthFunc(pVertex, pVertices[1], this)
            || !m_MinEdgeLengthFunc(pVertex, pVertices[2], this)
            || !m_MinEdgeLengthFunc(pVertex, pVertices[3], this) )
        {
//            MDS_LOG_NOTE("CDelaunayTetra::addVertex(): Short distance between a new vertex and one of the base tetrahedron's vertices!");
            return false;
        }
    }
#endif // MDS_CHECK_MIN_EDGE_LENGTH*/

    // Find cavity of the point
/*    tTetrahedra Cavity;
    if( !findCavity(pVertex, Cavity, pBaseTetra) )
    {
        // Create a new randomly moved vertex
        pVertex = randomVertex(pVertex);

        // Clear the cavity
        Cavity.clear();

        // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
        MDS_LOG_NOTE("CDelaunayTetra::addVertex(): randomly moved vertex ("
                     << pVertex->GetX() << ' '
                     << pVertex->GetY() << ' '
                     << pVertex->GetZ() << ')'
                    );
#endif // MDS_SEG_LOGGING_ENABLED

        // Try to find the cavity again
        if( !findCavity(pVertex, Cavity) )
        {
            MDS_LOG_NOTE("CDelaunayTetra::addVertex(): Failed to find vertex cavity!");
            return false;
        }
    }*/

    // Find base of the point
    tTetrahedra Base;
    if( !findBase(pVertex, Base, pBaseTetra) )
    {
        MDS_LOG_NOTE("CDelaunayTetra::addVertex(): Failed to find vertex base!");
        return false;
    }

    // Find cavity of the point
    tTetrahedra Cavity;
    if( !findCavityFromBase(pVertex, Cavity, Base) )
    {
        // Create a new randomly moved vertex
/*        pVertex = randomVertex(pVertex);

        // Clear the cavity
        Base.clear();
        Cavity.clear();

        // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
        MDS_LOG_NOTE("CDelaunayTetra::addVertex(): randomly moved vertex ("
                     << pVertex->GetX() << ' '
                     << pVertex->GetY() << ' '
                     << pVertex->GetZ() << ')'
                    );
#endif // MDS_SEG_LOGGING_ENABLED

        // Try to find the base again
        if( !findBase(pVertex, Base, pBaseTetra) )
        {
            MDS_LOG_NOTE("CDelaunayTetra::addVertex(): Failed to find vertex base!");
            return false;
        }

        // Try to find the cavity again
        if( !findCavityFromBase(pVertex, Cavity, Base) )
        {*/
            MDS_LOG_NOTE("CDelaunayTetra::addVertex(): Failed to find vertex cavity!");
            return false;
//        }
    }

    // Find all boundary triangles of the cavity
    tVertices BoundaryPoints;
    if( !findCavityEnvelope(Cavity, BoundaryPoints) )
    {
        MDS_LOG_NOTE("CDelaunayTetra::addVertex(): Cannot find the cavity envelope!");
        return false;
    }

    // The number of boundary points
    int iNumOfPoints = (int)BoundaryPoints.size();
    if( iNumOfPoints < 9 )
//    if( iNumOfPoints < 12 )
//    if( iNumOfPoints < 3 )
    {
        MDS_LOG_NOTE("CDelaunayTetra::addVertex(): No vertices found in the cavity envelope!");
        return false;
    }
/*    if( (iNumOfPoints % 3) != 0 )
    {
        MDS_LOG_NOTE("CDelaunayTetra::addVertex(): Wrong number of vertices in the cavity envelope!");
        return false;
    }*/

#ifdef MDS_CHECK_MIN_EDGE_LENGTH
    // Square of the mininal edge length
    // Modified 2009/12/07
//    double dSquareMinEdgeLength = m_dMinEdgeLength * m_dMinEdgeLength;
    double dSquareMinEdgeLength = 0.25 * m_dMinEdgeLength * m_dMinEdgeLength;

    // Check distance between newly inserted vertex and all boundary points
    tVertices::iterator itBP = BoundaryPoints.begin();
    tVertices::iterator itBPEnd = BoundaryPoints.end();
    for( ; itBP != itBPEnd; ++itBP )
    {
        // Compute the distance
        if( pVertex->SquareDistance(**itBP) < dSquareMinEdgeLength )
        {
            MDS_LOG_NOTE("CDelaunayTetra::addVertex(): Short distance between a new vertex and the cavity envelope!");
            return false;
        }
    }

    // Latest functional version!
    // Check the distance between newly inserted vertex and all boundary points
/*    if( m_MinEdgeLengthFunc )
    {
        tVertices::iterator itBP = BoundaryPoints.begin();
        tVertices::iterator itBPEnd = BoundaryPoints.end();
        for( ; itBP != itBPEnd; ++itBP )
        {
            // Check the distance
            if( !m_MinEdgeLengthFunc(pVertex, *itBP, this) )
            {
//                MDS_LOG_NOTE("CDelaunayTetra::addVertex(): Short distance between a new vertex and the cavity envelope!");
                return false;
            }
        }
    }*/
#endif // MDS_CHECK_MIN_EDGE_LENGTH

    // Check flatness of all newly created tetrahedra (avoid creating slivers)
    for( int j = 0; j < iNumOfPoints; j += 3 )
    {
        // Check if all vertices lie on the same boundary
        int Flags = pVertex->GetFlag();
        Flags &= BoundaryPoints[j]->GetFlag();
        Flags &= BoundaryPoints[j + 1]->GetFlag();
        Flags &= BoundaryPoints[j + 2]->GetFlag();
        if( Flags & BOUNDARY_FLAGS )
        {
            continue;
        }

        // Check the flatness
        if( checkFlatness(pVertex,
                          BoundaryPoints[j],
                          BoundaryPoints[j + 1],
                          BoundaryPoints[j + 2]
                          ) )
        {
            MDS_LOG_NOTE("CDelaunayTetra::addVertex(): Flatness check failed!");
            return false;
        }
    }

    // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
/*    MDS_LOG_NOTE("CDelaunayTetra::addVertex: Found cavity envelope (" << iNumOfPoints / 3 << " triangles) ");
    for( int l = 0; l < iNumOfPoints; l += 3 )
    {
        MDS_LOG(BoundaryPoints[l]->GetX() << ' ' << BoundaryPoints[l]->GetY() << ' ' << BoundaryPoints[l]->GetZ() << ',' << std::endl
            << BoundaryPoints[l + 1]->GetX() << ' ' << BoundaryPoints[l + 1]->GetY() << ' ' << BoundaryPoints[l + 1]->GetZ() << ',' << std::endl
            << BoundaryPoints[l + 2]->GetX() << ' ' << BoundaryPoints[l + 2]->GetY() << ' ' << BoundaryPoints[l + 2]->GetZ() << ',' << std::endl
            );
    }*/
#endif // MDS_SEG_LOGGING_ENABLED

    // Remove the cavity
    tTetrahedra::iterator it = Cavity.begin();
    tTetrahedra::iterator itEnd = Cavity.end();
    for( ; it != itEnd; ++it )
    {
        // Call the remove function
        if( RemoveFunc )
        {
            RemoveFunc(*it, this);
        }

        // Remove the tetrahedron from the mesh
        removeTetrahedron(*it, false);
    }

#ifdef MDS_CHECK_TETRAHEDRONS
    // Vector of all created tetrahedra
    tTetrahedra NewTetrahedra;
#endif // MDS_CHECK_TETRAHEDRONS

    // Insert new tetrahedra
    for( int i = 0; i < iNumOfPoints; i += 3 )
    {
        // Check if all vertices lie on the same boundary
        int Flags = pVertex->GetFlag();
        Flags &= BoundaryPoints[i]->GetFlag();
        Flags &= BoundaryPoints[i + 1]->GetFlag();
        Flags &= BoundaryPoints[i + 2]->GetFlag();
        if( Flags & BOUNDARY_FLAGS )
        {
            continue;
        }

        // Insert new tetrahedron
        vctl::MCTetra *pTetrahedron = addTetrahedron(pVertex,
                                                     BoundaryPoints[i],
                                                     BoundaryPoints[i + 1],
                                                     BoundaryPoints[i + 2]
                                                     );

        // Call the insert function
        if( InsertFunc && pTetrahedron )
        {
            InsertFunc(pTetrahedron, this);
        }

#ifdef MDS_CHECK_TETRAHEDRONS
        if( pTetrahedron )
        {
            NewTetrahedra.push_back(pTetrahedron);
        }
#endif // MDS_CHECK_TETRAHEDRONS
    }

#ifdef MDS_CHECK_TETRAHEDRONS
    it = NewTetrahedra.begin();
    itEnd = NewTetrahedra.end();
    while( it != itEnd )
    {
        if( !checkTetrahedron(*it) )
        {
            MDS_LOG_NOTE("CDelaunayTetra::addVertex(): mesh coherence check failed! Tetrahedron " << *it);
        }
        ++it;
    }
#endif // MDS_CHECK_TETRAHEDRONS

    // O.K.
    return true;
}


vctl::MCTetra *CDelaunayTetra::addTetrahedron(vctl::MCVertex *pV0,
                                              vctl::MCVertex *pV1,
                                              vctl::MCVertex *pV2,
                                              vctl::MCVertex *pV3
                                              )
{
    // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
    MDS_LOG_NOTE("CDelaunayTetra::addTetrahedron: (" << pV0->GetX() << ' ' << pV0->GetY() << ' ' << pV0->GetZ() << "), "
                 << '(' << pV1->GetX() << ' ' << pV1->GetY() << ' ' << pV1->GetZ() << "), "
                 << '(' << pV2->GetX() << ' ' << pV2->GetY() << ' ' << pV2->GetZ() << "), "
                 << '(' << pV3->GetX() << ' ' << pV3->GetY() << ' ' << pV3->GetZ() << ')'
                 );
#endif // MDS_SEG_LOGGING_ENABLED

    // Create and set the tetrahedron info struct
    STetrahedronInfo *pInfo = new STetrahedronInfo;
    MDS_ASSERT(pInfo);

    // Fill tetrahedron info struct
    try {
        // Compute tetrahedron's circumpshere
        pInfo->Circumsphere.MakeSphereTetra(pV0, pV1, pV2, pV3);
    }
    catch( ... )
    {
        // Failed to compute circumsphere
        MDS_LOG_NOTE("CDelaunayTetra::addTetrahedron(): Failed to compute tetrahedron circumsphere!");
        return NULL;
    }

    // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
    double cx, cy, cz, r;
    pInfo->Circumsphere.GetSphere(cx, cy, cz, r);
    MDS_LOG_NOTE("CDelaunayTetra::addTetrahedron(): Circumsphere (" << cx << ' ' << cy << ' ' << cz << ' ' << r << ')');
#endif // MDS_SEG_LOGGING_ENABLED

    // Create new tetrahedron
//    vctl::MCTetra *pT = m_Tetrahedra.New(pV0, pV1, pV2, pV3);
    vctl::MCTetra *pT = m_Tetrahedra.New2(pV0, pV1, pV2, pV3);
    pT->SetValuePtr(pInfo);
    pT->SetObject(-1);

    // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
    MDS_LOG_NOTE("CDelaunayTetra::addTetrahedron: Created tetrahedron " << pT);
#endif // MDS_SEG_LOGGING_ENABLED

    // Compute normal vector of the tetrahedron base side
    vctl::MCVector3D Normal;
    pT->GetNormal(0, Normal);

    // Get all vertices
    vctl::MCVertex *pVertices[4];
    pT->GetVerticeS(pVertices);

    // Vector to the top of the tetrahedron
    vctl::MCVector3D Vector(pVertices[0], pVertices[3]);

    // Check the tetrahedron orientation (dot product)
    // - Normal vector must be oriented inside
    double dResult = Normal * Vector;
    if( dResult < 0.0 )
    {
        pT->InverseOrientation();
    }

    // Return pointer to the tetrahedron
    return pT;
}


bool CDelaunayTetra::checkTetrahedron(vctl::MCTetra *pTetrahedron)
{
    MDS_CHECK(pTetrahedron, return false);

    for( int i = 0; i < 4; ++i )
    {
        // Get neighbour
//        vctl::MCTetra *pNeighbour = pTetrahedron->GetNeighbour(i);
        vctl::MCTetra *pNeighbour = pTetrahedron->GetNeighbourByVertices(i);
        if( !pNeighbour )
        {
            continue;
        }

        // Get tetrahedron side
        vctl::MCVertex *pVertices[3];
        pTetrahedron->GetTri(i, pVertices);

        // Get remaining tetrahedron vertex
        vctl::MCVertex *pV1 = pTetrahedron->GetRestVertex(i);

        // Get the remaining tetrahedron vertex
        vctl::MCVertex *pV2 = pNeighbour->GetRestVertex(pVertices[0], pVertices[1], pVertices[2]);

        // Get normal of the tetrahedron side
        vctl::MCVector3D Normal;
        Normal = pTetrahedron->GetNormal(i);

        // Helper vectors
        vctl::MCVector3D Vect1(pVertices[0], pV1);
        vctl::MCVector3D Vect2(pVertices[0], pV2);

        // Check the orientation of both vectors
        double dResult1 = Vect1 * Normal;
        double dResult2 = Vect2 * Normal;
        if( dResult1 * dResult2 > 0 )
        {
            return false;
        }
    }

    // O.K.
    return true;
}


void CDelaunayTetra::removeTetrahedron(vctl::MCTetra *pTetrahedron,
                                       bool bEraseVertices
                                       )
{
    // Delete tetrahedron info struct
    delete (STetrahedronInfo *)pTetrahedron->GetValuePtr();
    pTetrahedron->SetValuePtr(NULL);

    // Get tetrahedron vertices
    vctl::MCVertex *pVertices[4];
    pTetrahedron->GetVerticeS(pVertices);

    // Remove and de-register the tetrahedron
    m_Tetrahedra.Erase(pTetrahedron);

    // Remove unused tetrahedron vertices
    if( bEraseVertices )
    {
        if( !pVertices[0]->GetRegisteredTetra() )
        {
            m_pVertices->Erase(pVertices[0]);
        }
        if( !pVertices[1]->GetRegisteredTetra() )
        {
            m_pVertices->Erase(pVertices[1]);
        }
        if( !pVertices[2]->GetRegisteredTetra() )
        {
            m_pVertices->Erase(pVertices[2]);
        }
        if( !pVertices[3]->GetRegisteredTetra() )
        {
            m_pVertices->Erase(pVertices[3]);
        }
    }
}


// Checks dihedral angle
bool CDelaunayTetra::checkFlatness(vctl::MCVertex *pV0,
                                   vctl::MCVertex *pV1,
                                   vctl::MCVertex *pV2,
                                   vctl::MCVertex *pV3
                                   )
{
//    static const double MAX_DEVIATION = 0.000000001;
//    static const double MAX_DEVIATION = 0.000001;

    static const double dInDegrees = 1.0 / 90;
    double MAX_DEVIATION = m_dMinDihedralAngle * dInDegrees;

    MDS_CHECK(pV0 && pV1 && pV2 && pV3, return false);

    // Prepare vectors
    vctl::MCVector3D v0(pV0, pV1);
    vctl::MCVector3D v1(pV0, pV2);
    vctl::MCVector3D v2(pV0, pV3);

    // Compute normals of tetrahedron sides
    vctl::MCVector3D n01, n02;
    n01.VectorProduct(v0, v1);
    n02.VectorProduct(v0, v2);

    // Normalize both normals
    n01.Normalization();
    n02.Normalization();

    // Compute angle between both normals
    double dResult = 1.0 - mds::math::getAbs(n01 * n02);

    // Check if the normals are not parallel
    return (dResult < MAX_DEVIATION);
}


vctl::MCTetra *CDelaunayTetra::findBaseTetrahedron(vctl::MCPoint3D *pPoint)
{
//    vctl::MCTetra *pTetrahedron = (vctl::MCTetra *)m_Tetrahedra.GetFirst();
    vctl::MCTetra *pTetrahedron = m_Tetrahedra.GetLast();
    if( !pTetrahedron || !pPoint )
    {
        return NULL;
    }

    return findBaseTetrahedron(pPoint, pTetrahedron);
}


vctl::MCTetra *CDelaunayTetra::findBaseTetrahedron(vctl::MCPoint3D *pPoint,
                                                   vctl::MCTetra *pTetrahedron
                                                   )
{
    static const double THIRD   = 1.0 / 3.0;
//    static const int MAX_ITERS  = 500;
//    static const int MAX_ITERS  = 1000;
    static const int MAX_ITERS  = 10000;
    static const double BOUNDARY = -0.000001;

    // Pointer to the previous tetrahedron
    vctl::MCTetra *pPrevious[3] = { NULL, NULL, NULL };
//    std::set<vctl::MCTetra *> Previous;

    // Traverse the mesh
    for( int iter = 0; pTetrahedron && iter < MAX_ITERS; ++iter )
    {
        // Loop check
        if( pTetrahedron == pPrevious[0] || pTetrahedron == pPrevious[1] )
//        if( Previous.find(pTetrahedron) != Previous.end() )
        {
            // Logging
//            MDS_LOG_NOTE("CDelaunayTetra::findBaseTetrahedron(): Loop detected!");
//            return NULL;

            // Loop detected! Let's try to do a brute force
            // search for the tetrahedron
            return findBaseTetrahedronBruteForce(pPoint);
        }
        else
        {
            pPrevious[0] = pPrevious[1];
            pPrevious[1] = pPrevious[2];
            pPrevious[2] = pTetrahedron;
//            Previous.insert(pTetrahedron);
        }

        // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
        MDS_LOG_NOTE("CDelaunayTetra::findBaseTetrahedron(): Tetrahedron " << pTetrahedron);
#endif // MDS_SEG_LOGGING_ENABLED

        // Compute the vertex position relative to faces
        double dPosition[4];
        for( int i = 0; i < 4; ++i )
        {
            // Get vertices of the one tetrahedron side
            vctl::MCVertex *pVertices[3];
            pTetrahedron->GetTri(i, pVertices, pVertices + 1, pVertices + 2);

            // Center of the triangle
            vctl::MCPoint3D TriCenter(
                THIRD * (pVertices[0]->GetX() + pVertices[1]->GetX() + pVertices[2]->GetX()),
                THIRD * (pVertices[0]->GetY() + pVertices[1]->GetY() + pVertices[2]->GetY()),
                THIRD * (pVertices[0]->GetZ() + pVertices[1]->GetZ() + pVertices[2]->GetZ())
                );

            // Compute vector oriented from the center to given vertex
            vctl::MCVector3D TestVector(&TriCenter, pPoint);
            TestVector.Normalization();

            // Compute normal of the triangle
            vctl::MCVector3D TriNormal;
            pTetrahedron->GetNormal(i, TriNormal);
//            pTetrahedron->GetNormal2(i, TriNormal);

            // Dot product of the test vector and edge normal
            dPosition[i] = TestVector * TriNormal;
        }

        // Check the vertex position
//        if( dPosition[0] >= 0.0 && dPosition[1] >= 0.0 && dPosition[2] >= 0.0 && dPosition[3] >= 0.0 )
        if( dPosition[0] >= BOUNDARY && dPosition[1] >= BOUNDARY && dPosition[2] >= BOUNDARY && dPosition[3] >= BOUNDARY )
        {
            return pTetrahedron;
        }

        // Get the most forward edge
        int iMin1 = (dPosition[0] < dPosition[1]) ? 0 : 1;
        int iMin2 = (dPosition[2] < dPosition[3]) ? 2 : 3;
        int iMin = (dPosition[iMin1] < dPosition[iMin2]) ? iMin1 : iMin2;

        // Get neighbouring tetrahedron
        vctl::MCTetra *pNeighbour = pTetrahedron->GetNeighbour(iMin);
//        vctl::MCTetra *pNeighbour = pTetrahedron->GetNeighbourByVertices(iMin);
        if( !pNeighbour )
        {
            // Failed to find base tetrahedron
            return NULL;
//            return findBaseTetrahedronBruteForce(pPoint);
        }
        else
        {
            // Next step
            pTetrahedron = pNeighbour;
        }
    }

    // Failed to find base tetrahedron
//    MDS_LOG_NOTE("CDelaunayTetra::findBaseTetrahedron(): Failed to find the base tetrahodron!");
//    return NULL;
    return findBaseTetrahedronBruteForce(pPoint);
}


vctl::MCTetra *CDelaunayTetra::findBaseTetrahedronBruteForce(vctl::MCPoint3D *pPoint)
{
    static const double THIRD = 1.0 / 3.0;
    static const double BOUNDARY = -0.000001;

    MDS_CHECK(pPoint, return NULL);

    // Exhaustive search for the base tetrahedron
    vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
    while( pTetrahedron )
    {
        // Compute the vertex position relative to all tetrahedron faces
        double dPosition[4];
        for( int i = 0; i < 4; ++i )
        {
            // Get vertices of the one tetrahedron side
            vctl::MCVertex *pVertices[3];
            pTetrahedron->GetTri(i, pVertices, pVertices + 1, pVertices + 2);

            // Center of the triangle
            vctl::MCPoint3D TriCenter(
                THIRD * (pVertices[0]->GetX() + pVertices[1]->GetX() + pVertices[2]->GetX()),
                THIRD * (pVertices[0]->GetY() + pVertices[1]->GetY() + pVertices[2]->GetY()),
                THIRD * (pVertices[0]->GetZ() + pVertices[1]->GetZ() + pVertices[2]->GetZ())
                );

            // Compute vector oriented from the center to given vertex
            vctl::MCVector3D TestVector(&TriCenter, pPoint);
            TestVector.Normalization();

            // Compute normal of the triangle
            vctl::MCVector3D TriNormal;
            pTetrahedron->GetNormal(i, TriNormal);
//            pTetrahedron->GetNormal2(i, TriNormal);

            // Dot product of the test vector and edge normal
            dPosition[i] = TestVector * TriNormal;
        }

        // Check the vertex position
//        if( dPosition[0] >= 0.0 && dPosition[1] >= 0.0 && dPosition[2] >= 0.0 && dPosition[3] >= 0.0 )
        if( dPosition[0] >= BOUNDARY && dPosition[1] >= BOUNDARY && dPosition[2] >= BOUNDARY && dPosition[3] >= BOUNDARY )
        {
            return pTetrahedron;
        }

        // Get the next tetrahedron
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Not found
    MDS_LOG_NOTE("CDelaunayTetra::findBaseTetrahedronBruteForce(): Failed to find the base tetrahodron!");
    return NULL;
}


// 2009/12/14 (a more precise version)
vctl::MCVertex *CDelaunayTetra::findNearestVertex(vctl::MCPoint3D *pPoint, bool bStartFromPrevious)
{
    static vctl::MCTetra *pTetra = NULL;

    // Find the base tetrahedron
    if( bStartFromPrevious && pTetra )
    {
        pTetra = findBaseTetrahedron(pPoint, pTetra);
    }
    else
    {
        pTetra = findBaseTetrahedron(pPoint);
    }
    if( !pTetra )
    {
        MDS_LOG_NOTE("CDelaunayTetra::findNearestVertex(): Failed to find base tetrahedron!");
        return NULL;
    }

    // Find the base and its cavity
    tTetrahedra Base, Cavity;
    if( !findBase(pPoint, Base, pTetra) )
    {
        MDS_LOG_NOTE("CDelaunayTetra::findNearestVertex(): Failed to find base!");
        return NULL;
    }
    if( !findCavityFromBase(pPoint, Cavity, Base) )
    {
        MDS_LOG_NOTE("CDelaunayTetra::findNearestVertex(): Failed to find cavity!");
        return NULL;
    }

    // Find closest vertex
    double dMinDist = mds::CTypeTraits<double>::getMax();
    vctl::MCVertex *pMinVertex = NULL;
    tTetrahedra::iterator itEnd = Cavity.end();
    tTetrahedra::iterator it = Cavity.begin();
    for( ; it != itEnd; ++it )
    {
        // Get tetrahedron vertices
        vctl::MCVertex *pVertices[4];
        (*it)->GetVerticeS(pVertices);

        // Calc distances
        for( int i = 0; i < 4; ++i )
        {
            double dDist = pVertices[i]->SquareDistance(*pPoint);
            if( dDist < dMinDist )
            {
                dMinDist = dDist;
                pMinVertex = pVertices[i];
            }
        }
    }

    // Return the closest vertex
    return pMinVertex;
}


// Before2009/12/14
/*vctl::MCVertex *CDelaunayTetra::findNearestVertex(vctl::MCPoint3D *pPoint, bool bStartFromPrevious)
{
    static vctl::MCTetra *pTetra = NULL;

    // Find base tetrahedron
    if( bStartFromPrevious && pTetra )
    {
        pTetra = findBaseTetrahedron(pPoint, pTetra);
    }
    else
    {
        pTetra = findBaseTetrahedron(pPoint);
    }
    if( !pTetra )
    {
        return NULL;
    }

    // Get tetrahedron vertices
    vctl::MCVertex *pVertices[4];
    pTetra->GetVerticeS(pVertices);

    // Estimate distance from all vertices
    double dMinDist = pVertices[0]->SquareDistance(*pPoint);
    int iMinIndex = 0;
    for( int i = 1; i < 4; ++i )
    {
        double dDist = pVertices[i]->SquareDistance(*pPoint);
        if( dDist < dMinDist )
        {
            dMinDist = dDist;
            iMinIndex = i;
        }
    }

    // Find the closest vertex
    return pVertices[iMinIndex];
}*/


void CDelaunayTetra::clearFlags()
{
    vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
    while( pTetrahedron )
    {
        // Clear tetrahedron flag
        pTetrahedron->MaskFlag(vctl::ENTITY_MARK_FLAG);

        // Get next one
        pTetrahedron = pTetrahedron->GetNext();
    }
}


void CDelaunayTetra::clearFlags(const tTetrahedra& Tetrahedra)
{
    tTetrahedra::const_iterator it = Tetrahedra.begin();
    tTetrahedra::const_iterator itEnd = Tetrahedra.end();
    for( ; it != itEnd; ++it )
    {
        // Clear tetrahedron flag
        (*it)->MaskFlag(vctl::ENTITY_MARK_FLAG);
    }
}


bool CDelaunayTetra::findCavity(vctl::MCPoint3D *pPoint,
                                tTetrahedra& Cavity
                                )
{
    MDS_ASSERT(pPoint);

    // Find the vertex base tetrahedron
    vctl::MCTetra *pBaseTetrahedron = findBaseTetrahedron(pPoint);
    if( !pBaseTetrahedron )
    {
        return false;
    }

    // Recursive finding of the cavity
    return findCavity(pPoint, Cavity, pBaseTetrahedron);
}


/*bool CDelaunayTetra::findCavity(vctl::MCPoint3D *pPoint,
                                tTetrahedra& Cavity,
                                vctl::MCTetra *pTetrahedron
                                )
{
    // Was the tetrahedron already tested?
    if( pTetrahedron->TestFlag(vctl::ENTITY_MARK_FLAG) )
    {
        return true;
    }

    // Check if the vertex lies in circumcircle of the tetrahedron
    int iResult = getTetrahedronCircumsphere(pTetrahedron)->TestPoint(*pPoint);
    if( iResult < 0 )
    {
        return true;
    }
    else if( iResult == 0 )
    {
        return false;
    }

    // Mark the tetrahedron
    pTetrahedron->SetFlag(vctl::ENTITY_MARK_FLAG);

    // Add pointer to the tetrahedron to the output vector
    Cavity.push_back(pTetrahedron);

    // Get neighbours
    vctl::MCTetra *pNeighbours[4];
//    pTetrahedron->GetNeighboursByVertices(pNeighbours);
    pTetrahedron->GetNeighbours(pNeighbours);

    // Recursive call of the function
    if( pNeighbours[0] )
    {
        if( !findCavity(pPoint, Cavity, pNeighbours[0]) )
        {
            return false;
        }
    }
    if( pNeighbours[1] )
    {
        if( !findCavity(pPoint, Cavity, pNeighbours[1]) )
        {
            return false;
        }
    }
    if( pNeighbours[2] )
    {
        if( !findCavity(pPoint, Cavity, pNeighbours[2]) )
        {
            return false;
        }
    }
    if( pNeighbours[3] )
    {
        if( !findCavity(pPoint, Cavity, pNeighbours[3]) )
        {
            return false;
        }
    }

    // O.K.
    return true;
}*/

// latest functional version
// - simple Delaunay empty circumsphere criterion
bool CDelaunayTetra::findCavity(vctl::MCPoint3D *pPoint,
                                tTetrahedra& Cavity,
                                vctl::MCTetra *pTetrahedron
                                )
{
    // Helper vector of marked tetrahedra
    tTetrahedra Marked;

    // Queue of tetrahedra
    typedef std::deque<vctl::MCTetra *> tQueue;
    tQueue Queue;

    // Get its neighbours
    Queue.push_back(pTetrahedron);

    // Grow the cavity
    while( !Queue.empty() )
    {
        // Get pointer to the first element in the queue
        vctl::MCTetra *pT = Queue.front();
        Queue.pop_front();

        // Check the flag
        if( pT->TestFlag(vctl::ENTITY_MARK_FLAG) )
        {
            continue;
        }

        // Mark the tetrahedron
        pT->SetFlag(vctl::ENTITY_MARK_FLAG);
        Marked.push_back(pT);

        // Check if the vertex lies inside circumsphere of the tetrahedron
        int iResult = getTetrahedronCircumsphere(pT)->TestPoint(*pPoint);
        if( iResult < 0 )
        {
            continue;
        }
        else if( iResult == 0 )
        {
            clearFlags(Marked);
            return false;
        }

        // Add pointer to the tetrahedron to the output vector
        Cavity.push_back(pT);

        // Get neighbours
        vctl::MCTetra *pNeighbours[4];
//        pTetrahedron->GetNeighboursByVertices(pNeighbours);
        pT->GetNeighbours(pNeighbours);

        // Check the neighbours
        if( pNeighbours[0] && !pNeighbours[0]->TestFlag(vctl::ENTITY_MARK_FLAG) )
        {
            Queue.push_back(pNeighbours[0]);
        }
        if( pNeighbours[1] && !pNeighbours[1]->TestFlag(vctl::ENTITY_MARK_FLAG) )
        {
            Queue.push_back(pNeighbours[1]);
        }
        if( pNeighbours[2] && !pNeighbours[2]->TestFlag(vctl::ENTITY_MARK_FLAG) )
        {
            Queue.push_back(pNeighbours[2]);
        }
        if( pNeighbours[3] && !pNeighbours[3]->TestFlag(vctl::ENTITY_MARK_FLAG) )
        {
            Queue.push_back(pNeighbours[3]);
        }
    }

    // Clear marked flag
    clearFlags(Marked);

    // O.K.
    return true;
}


bool CDelaunayTetra::findBase(vctl::MCPoint3D *pPoint,
                              tTetrahedra& Base,
                              vctl::MCTetra *pTetrahedron
                              )
{
    static const double THIRD = 1.0 / 3.0;
    static const double BOUNDARY = -0.000001;
//    static const double BOUNDARY = 0.0;

    // Helper vector of marked tetrahedra
    tTetrahedra Marked;

    // Queue of tetrahedra
    typedef std::deque<vctl::MCTetra *> tQueue;
    tQueue Queue;

    // Add starting tetrahedron to the queue
    Queue.push_back(pTetrahedron);

    // Get neighbours
    vctl::MCTetra *pNeighbours[4];
//    pTetrahedron->GetNeighboursByVertices(pNeighbours);
    pTetrahedron->GetNeighbours(pNeighbours);

    // Check the neighbours
    if( pNeighbours[0] )
    {
        Queue.push_back(pNeighbours[0]);
    }
    if( pNeighbours[1] )
    {
        Queue.push_back(pNeighbours[1]);
    }
    if( pNeighbours[2] )
    {
        Queue.push_back(pNeighbours[2]);
    }
    if( pNeighbours[3] )
    {
        Queue.push_back(pNeighbours[3]);
    }

    // Grow the base
    while( !Queue.empty() )
    {
        // Get pointer to the first element in the queue
        vctl::MCTetra *pT = Queue.front();
        Queue.pop_front();

        // Check the flag
        if( pT->TestFlag(vctl::ENTITY_MARK_FLAG) )
        {
            continue;
        }

        // Mark the tetrahedron
        pT->SetFlag(vctl::ENTITY_MARK_FLAG);
        Marked.push_back(pT);

        // Compute the vertex position relative to all tetrahedron facets
        double dPosition[4];
        for( int i = 0; i < 4; ++i )
        {
            // Get vertices of the one tetrahedron side
            vctl::MCVertex *pVertices[3];
            pT->GetTri(i, pVertices, pVertices + 1, pVertices + 2);

            // Center of the triangle
            vctl::MCPoint3D TriCenter(
                THIRD * (pVertices[0]->GetX() + pVertices[1]->GetX() + pVertices[2]->GetX()),
                THIRD * (pVertices[0]->GetY() + pVertices[1]->GetY() + pVertices[2]->GetY()),
                THIRD * (pVertices[0]->GetZ() + pVertices[1]->GetZ() + pVertices[2]->GetZ())
                );

            // Compute vector oriented from the center to given vertex
            vctl::MCVector3D TestVector(&TriCenter, pPoint);
//            TestVector.Normalization();

            // Compute normal of the triangle
            vctl::MCVector3D TriNormal;
//            pT->GetNormal(i, TriNormal);
            pT->GetNormal2(i, TriNormal);

            // Dot product of the test vector and edge normal
            dPosition[i] = TestVector * TriNormal;
        }

        // Check the vertex position
        if( dPosition[0] >= BOUNDARY && dPosition[1] >= BOUNDARY && dPosition[2] >= BOUNDARY && dPosition[3] >= BOUNDARY )
        {
            // Add pointer to the tetrahedron to the output vector
            Base.push_back(pT);
        }
        else
        {
            continue;
        }

        // Get neighbours
//        pTetrahedron->GetNeighboursByVertices(pNeighbours);
        pT->GetNeighbours(pNeighbours);

        // Check the neighbours
        if( pNeighbours[0] && !pNeighbours[0]->TestFlag(vctl::ENTITY_MARK_FLAG) )
        {
            Queue.push_back(pNeighbours[0]);
        }
        if( pNeighbours[1] && !pNeighbours[1]->TestFlag(vctl::ENTITY_MARK_FLAG) )
        {
            Queue.push_back(pNeighbours[1]);
        }
        if( pNeighbours[2] && !pNeighbours[2]->TestFlag(vctl::ENTITY_MARK_FLAG) )
        {
            Queue.push_back(pNeighbours[2]);
        }
        if( pNeighbours[3] && !pNeighbours[3]->TestFlag(vctl::ENTITY_MARK_FLAG) )
        {
            Queue.push_back(pNeighbours[3]);
        }
    }

//    MDS_LOG_NOTE("  Number of tetrahedra in the base = " << Base.size());

    // Clear all flags
    clearFlags(Marked);

    // O.K.
    return true;
}


// - Cavity correction
// - Constrained Delaunay triangulation criterion
bool CDelaunayTetra::findCavityFromBase(vctl::MCPoint3D *pPoint,
                                        tTetrahedra& Cavity,
                                        const tTetrahedra& Base
                                        )
{
    static const double THIRD = 1.0 / 3;
    static const double BOUNDARY = 0.000001;
//    static const double BOUNDARY = 0.0;

    // Helper vector of marked tetrahedra
    tTetrahedra Marked;

    // Queue of tetrahedra
    typedef std::deque<vctl::MCTetra *> tQueue;
    tQueue Queue;

    // Put all neighbours of base tetrahedra to the queue
    tTetrahedra::const_iterator it = Base.begin();
    tTetrahedra::const_iterator itEnd = Base.end();
    for( ; it != itEnd; ++it )
    {
        // Current tetrahedron
        vctl::MCTetra *pT = *it;

        // Get neighbours
        vctl::MCTetra *pNeighbours[4];
//        pTetrahedron->GetNeighboursByVertices(pNeighbours);
        pT->GetNeighbours(pNeighbours);

        // Mark the base tetrahedron
        pT->SetFlag(vctl::ENTITY_MARK_FLAG);
        Marked.push_back(pT);

        // Check the neighbours
        for( int i = 0; i < 4; ++i )
        {
            if( !pNeighbours[i] || pNeighbours[i]->TestFlag(vctl::ENTITY_MARK_FLAG) )
            {
                continue;
            }

            // Insert tetrahedron to the cavity
            Queue.push_back(pNeighbours[i]);
        }
    }

    // Grow the cavity
    while( !Queue.empty() )
    {
        // Get pointer to the first element in the queue
        vctl::MCTetra *pT = Queue.front();
        Queue.pop_front();

        // Check the flag
        if( pT->TestFlag(vctl::ENTITY_MARK_FLAG) )
        {
            continue;
        }

        // Mark the tetrahedron
        pT->SetFlag(vctl::ENTITY_MARK_FLAG);
        Marked.push_back(pT);

        // Check if the vertex lies inside circumsphere of the tetrahedron
        int iResult = getTetrahedronCircumsphere(pT)->TestPoint(*pPoint);
        if( iResult < 0 )
        {
            continue;
        }
/*        else if( iResult == 0 )
        {
            clearFlags(Marked);
            return false;
        }*/

        // Add pointer to the tetrahedron to the output vector
        Cavity.push_back(pT);

        // Get neighbours
        vctl::MCTetra *pNeighbours[4];
//        pTetrahedron->GetNeighboursByVertices(pNeighbours);
        pT->GetNeighbours(pNeighbours);

        // Check the neighbours
        for( int i = 0; i < 4; ++i )
        {
            if( !pNeighbours[i] || pNeighbours[i]->TestFlag(vctl::ENTITY_MARK_FLAG) )
            {
                continue;
            }

            // Insert tetrahedron to the cavity
            Queue.push_back(pNeighbours[i]);
        }
    }

//    MDS_LOG_NOTE("  Number of tetrahedra in the cavity = " << Cavity.size());

    // Clear flags
    clearFlags(Marked);
    Marked.clear();

    // Set of all internal vertices of the cavity
    typedef std::map<vctl::MCVertex *, vctl::MCTetra *> tInternals;
    tInternals Internal;

    // Mark all tetrahedra in the base
    it = Base.begin();
    itEnd = Base.end();
    for( ; it != itEnd; ++it )
    {
        vctl::MCTetra *pTetrahedron = *it;
        pTetrahedron->SetFlag(vctl::ENTITY_MARK_FLAG);
        Marked.push_back(pTetrahedron);

        // Get vertices
        vctl::MCVertex *pVertices[4];
        pTetrahedron->GetVerticeS(pVertices);
        Internal[pVertices[0]] = pTetrahedron;
        Internal[pVertices[1]] = pTetrahedron;
        Internal[pVertices[2]] = pTetrahedron;
        Internal[pVertices[3]] = pTetrahedron;
    }

    // Mark all tetrahedra in the cavity
    it = Cavity.begin();
    itEnd = Cavity.end();
    for( ; it != itEnd; ++it )
    {
        vctl::MCTetra *pTetrahedron = *it;
        pTetrahedron->SetFlag(vctl::ENTITY_MARK_FLAG);
        Marked.push_back(pTetrahedron);

        // Get vertices
        vctl::MCVertex *pVertices[4];
        pTetrahedron->GetVerticeS(pVertices);
        Internal[pVertices[0]] = pTetrahedron;
        Internal[pVertices[1]] = pTetrahedron;
        Internal[pVertices[2]] = pTetrahedron;
        Internal[pVertices[3]] = pTetrahedron;
    }

    // Traverse all tetrahedra in the cavity
    it = Cavity.begin();
    itEnd = Cavity.end();
    for( ; it != itEnd; ++it )
    {
        vctl::MCTetra *pTetrahedron = *it;

        // Skip already removed tetrahedra
        if( !pTetrahedron->TestFlag(vctl::ENTITY_MARK_FLAG) )
        {
            continue;
        }

        // Get its neighbours
        vctl::MCTetra *pNeighbours[4];
//        pTetrahedron->GetNeighboursByVertices(pNeighbours);
        pTetrahedron->GetNeighbours(pNeighbours);

        // Remove all boundary vertices
        for( int i = 0; i < 4; ++i )
        {
            if( !pNeighbours[i] || !pNeighbours[i]->TestFlag(vctl::ENTITY_MARK_FLAG) )
            {
                // Get vertices of the tetrahedron facet
                vctl::MCVertex *pVertices[3];
                pTetrahedron->GetTri(i, pVertices, pVertices + 1, pVertices + 2);

                // Remove them from the cavity
                Internal.erase(pVertices[0]);
                Internal.erase(pVertices[1]);
                Internal.erase(pVertices[2]);
            }
            else
            {
                // Get vertices of the tetrahedron facet
                vctl::MCVertex *pVertices[4];
                pTetrahedron->GetTri(i, pVertices, pVertices + 1, pVertices + 2);

                // Check if the facet forms boundary between two regions
                if( !m_ConstraintFunc(pVertices[0], pVertices[1], pVertices[2], this) )
                {
                    continue;
                }

                // Center of the facet
                vctl::MCPoint3D TriCenter(
                    THIRD * (pVertices[0]->GetX() + pVertices[1]->GetX() + pVertices[2]->GetX()),
                    THIRD * (pVertices[0]->GetY() + pVertices[1]->GetY() + pVertices[2]->GetY()),
                    THIRD * (pVertices[0]->GetZ() + pVertices[1]->GetZ() + pVertices[2]->GetZ())
                    );

                // Compute vector oriented from the center to given vertex
                vctl::MCVector3D TestVector1(&TriCenter, pPoint);

                // Compute vector oriented from the center to remaining tetrahedron vertex
                vctl::MCVector3D TestVector2(&TriCenter, pTetrahedron->GetRestVertex(i));

                // Compute normal of the triangle
                vctl::MCVector3D TriNormal;
                pTetrahedron->GetNormal2(i, TriNormal);

                // Dot product of the test vector and edge normal
                double d1 = TestVector1 * TriNormal;
                double d2 = TestVector2 * TriNormal;

                // Tetra to remove
                vctl::MCTetra *pT = (d1 * d2 < 0.0) ? pTetrahedron : pNeighbours[i];

                // Remove tetra from the cavity
                pT->MaskFlag(vctl::ENTITY_MARK_FLAG);
                pT->GetVerticeS(pVertices);
                Internal.erase(pVertices[0]);
                Internal.erase(pVertices[1]);
                Internal.erase(pVertices[2]);
                Internal.erase(pVertices[3]);
                break;
            }
        }
    }

    // Cavity correction step
    bool bChanged1 = true, bChanged2 = true;
    while( bChanged1 || bChanged2 )
    {
        // Initialize the flag
        bChanged1 = false;
        bChanged2 = false;

        // Remove internal vertices
        if( Internal.size() > 0 )
        {
            // Get an internal vertex and its tetrahedron
            vctl::MCVertex *pVertex = Internal.begin()->first;
            vctl::MCTetra *pTetrahedron = Internal.begin()->second;

            // Remove the tetrahedron
            pTetrahedron->MaskFlag(vctl::ENTITY_MARK_FLAG);
            bChanged1 = true;

            vctl::MCVertex *pVertices[4];
            pTetrahedron->GetVerticeS(pVertices);
            Internal.erase(pVertices[0]);
            Internal.erase(pVertices[1]);
            Internal.erase(pVertices[2]);
            Internal.erase(pVertices[3]);
        }

        // Traverse all tetrahedra in the cavity
        tTetrahedra::iterator it2 = Cavity.begin();
        tTetrahedra::iterator it2End = Cavity.end();
        for( ; it2 != it2End; ++it2 )
        {
            vctl::MCTetra *pTetrahedron = *it2;

            // Skip already removed tetrahedra
            if( !pTetrahedron->TestFlag(vctl::ENTITY_MARK_FLAG) )
            {
                continue;
            }

            // Get neighbours
            vctl::MCTetra *pNeighbours[4];
//            pTetrahedron->GetNeighboursByVertices(pNeighbours);
            pTetrahedron->GetNeighbours(pNeighbours);

            // Check all boundary facets
            for( int i = 0; i < 4; ++i )
            {
                if( !pNeighbours[i] || !pNeighbours[i]->TestFlag(vctl::ENTITY_MARK_FLAG) )
                {
                    // Get vertices of the tetrahedron side
                    vctl::MCVertex *pVertices[3];
                    pTetrahedron->GetTri(i, pVertices, pVertices + 1, pVertices + 2);

                    // Center of the triangle
                    vctl::MCPoint3D TriCenter(
                        THIRD * (pVertices[0]->GetX() + pVertices[1]->GetX() + pVertices[2]->GetX()),
                        THIRD * (pVertices[0]->GetY() + pVertices[1]->GetY() + pVertices[2]->GetY()),
                        THIRD * (pVertices[0]->GetZ() + pVertices[1]->GetZ() + pVertices[2]->GetZ())
                        );

                    // Compute vector oriented from the center to given vertex
                    vctl::MCVector3D TestVector(&TriCenter, pPoint);
//                    TestVector.Normalization();

                    // Compute normal of the triangle
                    vctl::MCVector3D TriNormal;
//                    pTetrahedron->GetNormal(i, TriNormal);
                    pTetrahedron->GetNormal2(i, TriNormal);

                    // Dot product of the test vector and edge normal
                    double dPosition = TestVector * TriNormal;

                    // Check if the facet is not visible
                    if( dPosition <= BOUNDARY )
                    {
                        // Remove the tetrahedron from the cavity
                        pTetrahedron->MaskFlag(vctl::ENTITY_MARK_FLAG);
                        bChanged2 = true;

                        vctl::MCVertex *pVertices[4];
                        pTetrahedron->GetVerticeS(pVertices);
                        Internal.erase(pVertices[0]);
                        Internal.erase(pVertices[1]);
                        Internal.erase(pVertices[2]);
                        Internal.erase(pVertices[3]);
                        break;
                    }
                }
            }
            if( bChanged2 )
            {
                break;
            }
        }
    }

    // Remove all not marked tetrahedra
    tTetrahedra FinalCavity;
    it = Cavity.begin();
    itEnd = Cavity.end();
    for( ; it != itEnd; ++it )
    {
        if( (*it)->TestFlag(vctl::ENTITY_MARK_FLAG) )
        {
            FinalCavity.push_back(*it);
        }
    }
    Cavity.swap(FinalCavity);

//    MDS_LOG_NOTE("  After correction = " << Cavity.size());

    // Add all base tetrahedra to the final cavity
    Cavity.insert(Cavity.end(), Base.begin(), Base.end());

    // Clear flags
    clearFlags(Marked);

    // O.K.
    return true;
}


void CDelaunayTetra::clear()
{
    if( getNumOfTetrahedra() == 0 )
    {
        return;
    }

    // Delete info structs
    vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
    while( pTetrahedron )
    {
        // Delete tetrahedron info struct
        delete (STetrahedronInfo *)pTetrahedron->GetValuePtr();
        pTetrahedron->SetValuePtr(NULL);

        // Get next one
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Clear tetrahedra and their vertices
    m_Tetrahedra.ClearAll();
    m_pVertices->ClearAll();
}


bool CDelaunayTetra::findCavityEnvelope(const tTetrahedra& Cavity,
                                        tVertices& Envelope
                                        )
{
    // Clear the cavity envelope
    Envelope.clear();

    // Mark all tetrahedra in the cavity
    tTetrahedra::const_iterator it = Cavity.begin();
    tTetrahedra::const_iterator itEnd = Cavity.end();
    for( ; it != itEnd; ++it )
    {
        (*it)->SetFlag(vctl::ENTITY_MARK_FLAG);
    }

    // Find boundary edges
    it = Cavity.begin();
    itEnd = Cavity.end();
    for( ; it != itEnd; ++it )
    {
        // Get pointer to the tetrahedron
        vctl::MCTetra *pTetrahedron = *it;

        // Get neighbours
        vctl::MCTetra *pNeighbours[4];
//        pTetrahedron->GetNeighboursByVertices(pNeighbours);
        pTetrahedron->GetNeighbours(pNeighbours);

        // Check if neighbouring tetrahedra are marked
        for( int i = 0; i < 4; ++i )
        {
            if( !pNeighbours[i] || !pNeighbours[i]->TestFlag(vctl::ENTITY_MARK_FLAG) )
            {
                // Get vertices of the tetrahedron side
                vctl::MCVertex *pVertices[3];
                pTetrahedron->GetTri(i, pVertices);

                // Add edge end-points to the list
                Envelope.push_back(pVertices[0]);
                Envelope.push_back(pVertices[1]);
                Envelope.push_back(pVertices[2]);
            }
        }
    }

    // Clear flags of all tetrahedra
    clearFlags(Cavity);

    // O.K.
    return true;
}


bool CDelaunayTetra::getMinCoordinates(double& dX, double& dY, double& dZ)
{
    vctl::MCVertex *pVertex = getFirstVertex();
    if( !pVertex )
    {
        dX = 0.0;
        dY = 0.0;
        dZ = 0.0;
        return false;
    }

    dX = pVertex->GetX();
    dY = pVertex->GetY();
    dZ = pVertex->GetZ();

    pVertex = pVertex->GetNext();
    while( pVertex )
    {
        double dTemp = pVertex->GetX();
        if( dTemp < dX )
        {
            dX = dTemp;
        }

        dTemp = pVertex->GetY();
        if( dTemp < dY )
        {
            dY = dTemp;
        }

        dTemp = pVertex->GetZ();
        if( dTemp < dZ )
        {
            dZ = dTemp;
        }

        // Get next one
        pVertex = pVertex->GetNext();
    }

    // O.K.
    return true;
}


bool CDelaunayTetra::getMaxCoordinates(double& dX, double& dY, double& dZ)
{
    vctl::MCVertex *pVertex = getFirstVertex();
    if( !pVertex )
    {
        dX = 0.0;
        dY = 0.0;
        dZ = 0.0;
        return false;
    }

    dX = pVertex->GetX();
    dY = pVertex->GetY();
    dZ = pVertex->GetZ();

    pVertex = pVertex->GetNext();
    while( pVertex )
    {
        double dTemp = pVertex->GetX();
        if( dTemp > dX )
        {
            dX = dTemp;
        }

        dTemp = pVertex->GetY();
        if( dTemp > dY )
        {
            dY = dTemp;
        }

        dTemp = pVertex->GetZ();
        if( dTemp > dZ )
        {
            dZ = dTemp;
        }

        // Get next one
        pVertex = pVertex->GetNext();
    }

    // O.K.
    return true;
}


double CDelaunayTetra::getTetrahedronQuality(vctl::MCTetra *pTetrahedron)
{
    static const double NORM_FACTOR = mds::math::SQRT6 / 12.0;

    MDS_ASSERT(pTetrahedron);

    // Get tetrahedron circuit and area
    double dArea = pTetrahedron->GetArea();
    double dVolume = pTetrahedron->GetVolume();
    MDS_ASSERT(dArea != 0.0 && dVolume != 0.0);

    // Compute radius of the inscribed circumsphere
//    double dR = 3.0 * dVolume / dArea;
    double dInvR = dArea / (3.0 * dVolume);

    // Get length of the longest edge
    double dL = pTetrahedron->GetMaxEdgeLength();

    // Compute the tetrahedron quality
//    return (NORM_FACTOR * dL / dR);
    return (NORM_FACTOR * dL * dInvR);
}


/*double CDelaunayTetra::getTetrahedronQuality(vctl::MCTetra *pTetrahedron)
{
    MDS_ASSERT(pTetrahedron);

    // Compute radius of the circumsphere
    double dR = getTetrahedronCircumsphere(pTetrahedron)->GetRadius();

    // Get length of the shortest edge
    double dL = pTetrahedron->GetMinEdgeLength();

    // Compute the tetrahedron quality
    if( dL > 0.0 )
    {
        return dR / dL;
    }
    else
    {
        return dR;
    }
}*/


void CDelaunayTetra::normInsertFunc(vctl::MCTetra *pTetrahedron, CDelaunayTetra *pMesh)
{
    MDS_CHECK(pTetrahedron && pMesh, return);

    // Compute and set the tetrahedron quality
    pTetrahedron->SetValue(getTetrahedronQuality(pTetrahedron));

    // Get sorted queue
    vctl::MCTetraS::MT_QUEUE *pQueue = pMesh->m_Tetrahedra.GetQueue();

    // Sort the tetrahedron
    pQueue->Sort(pTetrahedron->GetValue(), pTetrahedron);
}


void CDelaunayTetra::normRemoveFunc(vctl::MCTetra *pTetrahedron, CDelaunayTetra *pMesh)
{
    MDS_CHECK(pTetrahedron && pMesh, return);

    // Get sorted queue
    vctl::MCTetraS::MT_QUEUE *pQueue = pMesh->m_Tetrahedra.GetQueue();

    // Unsort the tetrahedron
    if( pTetrahedron->GetValue() > 0.0 )
    {
        pQueue->UnSort(pTetrahedron->GetValue(), pTetrahedron);
    }
}


bool CDelaunayTetra::checkMinEdgeLengthFunc(vctl::MCPoint3D *p1,
                                            vctl::MCPoint3D *p2,
                                            CDelaunayTetra* pMesh
                                            )
{
    MDS_CHECK(p1 && p2 && pMesh, return false);

    // Square mininal edge length
    // Modified 2009/12/07
//    double dSquareMinEdgeLength = pMesh->m_dMinEdgeLength * pMesh->m_dMinEdgeLength;
    double dSquareMinEdgeLength = 0.25 * pMesh->m_dMinEdgeLength * pMesh->m_dMinEdgeLength;

    // Check the distance
    return (p1->SquareDistance(*p2) > dSquareMinEdgeLength);
}


bool CDelaunayTetra::checkConstraintFunc(vctl::MCPoint3D *p1,
                                         vctl::MCPoint3D *p2,
                                         vctl::MCPoint3D *p3,
                                         CDelaunayTetra* pMesh
                                         )
{
//    MDS_CHECK(p1 && p2 && p3 && pMesh, return false);

    return false;
}


} // namespace seg
} // namespace mds

