//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2005/07/26                          \n
 *
 * $Id: mdsDelaunayTri.cpp 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - 2D Delaunay triangulation.
 */

#include <MDSTk/VectorSegmentation/mdsDelaunayTri.h>

// MDSTk
#include <MDSTk/Base/mdsMemory.h>
#include <MDSTk/Base/mdsGlobalLog.h>
#include <MDSTk/Math/mdsBase.h>
#include <MDSTk/Math/ver1/mdsMatrix.h>


namespace mds
{
namespace seg
{

//==============================================================================
/*
 * Data members of the class CDelaunayTri.
 */
const tSize CDelaunayTri::DEFAULT_NUMBER_OF_NODES       = 7;
//const double CDelaunayTri::DEFAULT_MIN_EDGE_LENGTH      = 3.0;
const double CDelaunayTri::DEFAULT_MIN_EDGE_LENGTH      = 1.0;
const double CDelaunayTri::MAX_MOVEMENT                 = 0.5;


//==============================================================================
/*
 * Implementation of the class CDelaunayTri.
 */
CDelaunayTri::CDelaunayTri()
    : m_pVertices(m_Triangles.GetVerticeS())
    , m_dMinX(0.0)
    , m_dMinY(0.0)
    , m_dMaxX(0.0)
    , m_dMaxY(0.0)
{
    MDS_ASSERT(m_pVertices);

    // Enable vertex existence test
    m_pVertices->SetTestExistence(true);
}


bool CDelaunayTri::init(vctl::MCPoint3D *pMin,
                        vctl::MCPoint3D *pMax,
                        mds::tSize NumOfNodes,
                        double dMinEdgeLength
                        )
{
    MDS_ASSERT(NumOfNodes > 1);

    // Remove all triangles and their vertices
    clear();

    // Initialize minimal and maximal coordinates
    m_dMinX = pMin->GetX();
    m_dMinY = pMin->GetY();
    m_dMaxX = pMax->GetX();
    m_dMaxY = pMax->GetY();

    // Helper value
    mds::tSize MaxNode = NumOfNodes - 1;

    // Distance between two neighbouring nodes
    double dStepX = (pMax->GetX() - pMin->GetX()) / MaxNode;
    double dStepY = (pMax->GetY() - pMin->GetY()) / MaxNode;

    // Create nodes covering initial mesh
    mds::math::ver1::CMatrix<vctl::MCVertex *> Nodes(NumOfNodes, NumOfNodes);
    double dY = pMin->GetY();
    mds::tSize i, j;
    for( j = 0; j < NumOfNodes; ++j, dY += dStepY )
    {
        int BoundaryY = 0;
        if( j == 0 )
        {
            BoundaryY |= YMIN_BOUNDARY;
        }
        else if( j == MaxNode )
        {
            BoundaryY |= YMAX_BOUNDARY;
        }
        double dX = pMin->GetX();
        for( i = 0; i < NumOfNodes; ++i, dX += dStepX )
        {
            int Boundary = BoundaryY;
            if( i == 0 )
            {
                Boundary |= XMIN_BOUNDARY;
            }
            else if( i == MaxNode )
            {
                Boundary |= XMAX_BOUNDARY;
            }
            if( Boundary )
            {
                Nodes(i, j) = m_pVertices->New(dX, dY, 0.0);
                Nodes(i, j)->SetFlag(Boundary);
            }
            else
            {
                Nodes(i, j) = randomVertex(dX, dY, 3.0);
            }
        }
    }

    // Insert initial triangles
    for( j = 0; j < MaxNode; ++j )
    {
        for( i = 0; i < MaxNode; ++i )
        {
            // Add cell triangles
            addTriangle(Nodes(i, j), Nodes(i+1, j), Nodes(i, j+1));
            addTriangle(Nodes(i+1, j), Nodes(i+1, j + 1), Nodes(i, j+1));
        }
    }

    // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
/*    MDS_LOG_NOTE("CDelaunayTri::init():");
    vctl::MCVertex *pVertex = getFirstVertex();
    while( pVertex )
    {
        // Print vertex info
        MDS_LOG_NOTE("  Vertex (" << int(pVertex) << ") Flag:" << pVertex->GetFlag());

        // Get list of registered triangles
        tTriangles Triangles;
        pVertex->GetRegisteredTriList(Triangles);

        // Traverse the list
        tTriangles::iterator it = Triangles.begin();
        tTriangles::iterator itEnd = Triangles.end();
        while( it != itEnd )
        {
            MDS_LOG_NOTE("    Triangle (" << int(*it) << ')');
            ++it;
        }

        // Get next one
        pVertex = pVertex->GetNext();
    }*/
#endif // MDS_SEG_LOGGING_ENABLED

    // O.K.
    return true;
}


int CDelaunayTri::checkBoundary(vctl::MCPoint3D *pPoint)
{
    // Vertex coordinates
    double dX = pPoint->GetX();
    double dY = pPoint->GetY();

    // Initial flag
    int iBoundary = 0;

    // Check if the vertex lies inside the mesh
    // - Solved by projection of outer points to the mesh bounding box
    // - This may produce vertices similar to some existing ones!
    if( dX <= m_dMinX )
    {
        pPoint->SetX(m_dMinX);
        iBoundary |= XMIN_BOUNDARY;
    }
    else if( dX >= m_dMaxX )
    {
        pPoint->SetX(m_dMaxX);
        iBoundary |= XMAX_BOUNDARY;
    }
    if( dY <= m_dMinY )
    {
        pPoint->SetY(m_dMinY);
        iBoundary |= YMIN_BOUNDARY;
    }
    else if( dY >= m_dMaxY )
    {
        pPoint->SetY(m_dMaxY);
        iBoundary |= YMAX_BOUNDARY;
    }

    // Return the vertex boundary flag
    return iBoundary;
}


vctl::MCVertex *CDelaunayTri::randomVertex(double x, double y, double dMax)
{
    // Randomly move the vertex
    vctl::MCPoint3D Point(x + m_Random.random(-dMax, dMax),
                          y + m_Random.random(-dMax, dMax),
                          0.0
                          );

    // Lies vertex on the mesh boundary?
    int iBoundary = checkBoundary(&Point);

    // Create a new vertex
    vctl::MCVertex *pVertex = newVertex(&Point);
    if( iBoundary )
    {
        pVertex->SetFlag(iBoundary);
    }

    // O.K.
    return pVertex;
}


bool CDelaunayTri::construct()
{
    // Not sure test if the init(...) method has been called
    MDS_CHECK(getNumOfTriangles() >= 2, return false);

    // Sort all vertices
    m_pVertices->SortAll();

    // Get sorted queue
    typedef vctl::MCVerticeS::MT_QUEUE tQueue;
    tQueue *pQueue = m_pVertices->GetQueue();

    // Traverse the sorted queue in decreasing order
    tQueue::MT_RITER Iterator = pQueue->GetMaxIter();
    while( !pQueue->TestEndIter(Iterator) )
    {
        // Get pointer to the vertex
        vctl::MCVertex *pVertex = (vctl::MCVertex *)Iterator->second;

        // Lies vertex on the mesh boundary?
        int iBoundary = checkBoundary(pVertex);

        // Add vertex to the triangular mesh
        if( pVertex->GetValue() > 0.0 )
        {
            addVertex(pVertex, iBoundary);
        }

        // Get next one
        ++Iterator;
    }

    // Clear/unsort the queue
    pQueue->Clear();

    // O.K.
    return true;
}


bool CDelaunayTri::normalize(double dMinQuality, double dMinEdgeLength)
{
    // Set quality of all triangles
    vctl::MCTri *pTriangle = getFirstTriangle();
    while( pTriangle )
    {
        // Compute and set the triangle quality
        pTriangle->SetValue(getTriangleQuality(pTriangle));

        // Get next one
        pTriangle = pTriangle->GetNext();
    }

    // Sort all triangles
    m_Triangles.SortAll();

    // Get sorted queue
    vctl::MCTriS::MT_QUEUE *pQueue = m_Triangles.GetQueue();

#ifdef MDS_SEG_LOGGING_ENABLED
    MDS_LOG_NOTE("CDelaunayTri::normalize(): started");
#endif // MDS_SEG_LOGGING_ENABLED

    // Optimize the mesh
    for( ;; )
    {
        // Get pointer to the last item
        pTriangle = pQueue->GetMax();

#ifdef MDS_SEG_LOGGING_ENABLED
        MDS_LOG_NOTE("Triangle: ");
        MDS_LOG_NOTE("  Quality: " << pTriangle->GetValue());
        MDS_LOG_NOTE("  Max edge: " << pTriangle->GetMaxEdgeLength());
#endif // MDS_SEG_LOGGING_ENABLED

        // Check the triangle quality
        if( !pTriangle || pTriangle->GetValue() < dMinQuality )
        {
            // No optimization is needed
            break;
        }

        // Check the triangle edge length
        if( pTriangle->GetMaxEdgeLength() > dMinEdgeLength )
        {
            // Get triangle circumsphere
            vctl::MCSphere *pSphere = getTriangleCircumsphere(pTriangle);

            // Circumsphere center
            vctl::MCPoint3D Center;
            pSphere->GetCenter(Center);

            // Lies vertex on the mesh boundary?
            int iBoundary = checkBoundary(&Center);

            // Create a new vertex
            vctl::MCVertex *pVertex = newVertex(&Center);

            // Insert vertex to the mesh
            if( !addVertex(pVertex, iBoundary, normInsertFunc, normRemoveFunc) )
            {
                // Unsort the triangle, no optimization
                pQueue->UnSort(pTriangle->GetValue(), pTriangle);
                pTriangle->SetValue(0.0);

                // Remove the vertex
                if( !pVertex->GetRegisteredTri() )
                {
                    m_pVertices->Erase(pVertex);
                }
            }
        }
        else
        {
            // Unsort the triangle, no optimization
            pQueue->UnSort(pTriangle->GetValue(), pTriangle);
            pTriangle->SetValue(0.0);
        }
    }

    // Clear/unsort the queue
    pQueue->Clear();

#ifdef MDS_SEG_LOGGING_ENABLED
    MDS_LOG_NOTE("CDelaunayTri::normalize(): finished");
#endif // MDS_SEG_LOGGING_ENABLED

    // O.K.
    return true;
}


bool CDelaunayTri::addVertex(vctl::MCVertex *pVertex,
                             int iBoundaryVertex,
                             CDelaunayTri::tTriangleFunc InsertFunc,
                             CDelaunayTri::tTriangleFunc RemoveFunc
                             )
{
    MDS_ASSERT(pVertex);

    // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
    MDS_LOG_NOTE("CDelaunayTri::addVertex(): ("
        << pVertex->GetX() << ','
        << pVertex->GetY() << ','
        << pVertex->GetZ() << ')'
        );
#endif // MDS_SEG_LOGGING_ENABLED

    // Set boundary flag of the new vertex
    if( iBoundaryVertex )
    {
        pVertex->SetFlag(iBoundaryVertex);
    }

    // Find cavity of the point
    tTriangles Cavity;
    if( !findCavity(pVertex, Cavity) )
    {
        // Create a new randomly moved vertex
        pVertex = randomVertex(pVertex);

        // Clear the cavity
        clearFlags(Cavity);
        Cavity.clear();

        // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
        MDS_LOG_NOTE("CDelaunayTri::addVertex(): randomly moved vertex ("
                     << pVertex->GetX() << ' '
                     << pVertex->GetY() << ' '
                     << pVertex->GetZ() << ')'
                    );
#endif // MDS_SEG_LOGGING_ENABLED

        // Try to find the cavity again
        if( !findCavity(pVertex, Cavity) )
        {
            clearFlags(Cavity);
            return false;
        }
    }

    // Find boundary edges of the cavity
    tVertices BoundaryPoints;
    if( !findCavityEnvelope(Cavity, BoundaryPoints) )
    {
        MDS_LOG_NOTE("CDelaunayTri::addVertex(): Cannot find the cavity envelope!");
        clearFlags(Cavity);
        return false;
    }

    // The number of boundary points
    int iNumOfPoints = (int)BoundaryPoints.size();
    if( iNumOfPoints < 2 )
    {
        MDS_LOG_NOTE("CDelaunayTri::addVertex(): No vertices found in the cavity envelope!");
        clearFlags(Cavity);
        return false;
    }

    // Check the distance between the inserted vertex and all boundary points
    tVertices::iterator itBP = BoundaryPoints.begin();
    tVertices::iterator itBPEnd = BoundaryPoints.end();
    while( itBP != itBPEnd )
    {
        // Compute the distance
        if( pVertex->Distance(**itBP) < DEFAULT_MIN_EDGE_LENGTH )
        {
            MDS_LOG_NOTE("CDelaunayTri::addVertex(): Short distance between a new vertex and the cavity envelope!");
            return false;
        }

        // Next boundary point
        ++itBP;
    }

    // Check flatness of all newly created tetrahedrons (avoid creating slivers)
/*    for( int j = 0; j < iNumOfPoints; j += 3 )
    {
        // Check if all vertices lie on the same boundary
        int Flags = pVertex->GetFlag();
        Flags &= BoundaryPoints[j]->GetFlag();
        Flags &= BoundaryPoints[j + 1]->GetFlag();
        if( Flags & BOUNDARY_FLAGS )
        {
            continue;
        }

        // Check the flatness
        if( checkFlatness(pVertex,
                          BoundaryPoints[j],
                          BoundaryPoints[j + 1]
                          ) )
        {
            MDS_LOG_NOTE("CDelaunayTri::addVertex(): Flatness check failed!");
            return false;
        }
    }*/

    // Remove the cavity
    tTriangles::iterator it = Cavity.begin();
    tTriangles::iterator itEnd = Cavity.end();
    while( it != itEnd )
    {
        // Call the remove function
        if( RemoveFunc )
        {
            RemoveFunc(*it, this);
        }

        // Remove the triangle from the mesh
        removeTriangle(*it, false);

        // Next one
        ++it;
    }

    // Insert new triangles
    for( int i = 0; i < iNumOfPoints; i += 2 )
    {
        // Check if all vertices lie on the same boundary
        int Flags = pVertex->GetFlag();
        Flags &= BoundaryPoints[i]->GetFlag();
        Flags &= BoundaryPoints[i + 1]->GetFlag();
        if( Flags & BOUNDARY_FLAGS )
        {
            continue;
        }

        // Insert new triangle
        vctl::MCTri *pTriangle = addTriangle(pVertex,
                                             BoundaryPoints[i],
                                             BoundaryPoints[i + 1]
                                             );

        // Call the insert function
        if( InsertFunc && pTriangle )
        {
            InsertFunc(pTriangle, this);
        }
    }

    // O.K.
    return true;
}


vctl::MCTri *CDelaunayTri::addTriangle(vctl::MCVertex *pV0,
                                       vctl::MCVertex *pV1,
                                       vctl::MCVertex *pV2
                                       )
{
    // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
    MDS_LOG_NOTE("CDelaunayTri::addTriangle: (" << pV0->GetX() << ' ' << pV0->GetY() << "), "
        << '(' << pV1->GetX() << ' ' << pV1->GetY() << "), "
        << '(' << pV2->GetX() << ' ' << pV2->GetY() << "), "
        );
#endif // MDS_SEG_LOGGING_ENABLED

    // Create and set the triangle info struct
    STriangleInfo *pInfo = new STriangleInfo;
    MDS_ASSERT(pInfo);

    // Fill triangle info struct
    try {
        // Compute triangle's circumpshere
        pInfo->Circumsphere.MakeSphereTri(pV0, pV1, pV2);
    }
    catch( ... )
    {
        // Failed to compute circumsphere
        MDS_LOG_NOTE("CDelaunayTri::addTriangle(): Failed to compute triangle circumsphere!");
        return NULL;
    }

    // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
    double cx, cy, cz, r;
    pInfo->Circumsphere.GetSphere(&cx, &cy, &cz, &r);
    MDS_LOG_NOTE("CDelaunayTri::addTriangle(): Circumsphere (" << cx << ' ' << cy << ' ' << cz << ' ' << r << ')');
#endif // MDS_SEG_LOGGING_ENABLED

    // Create new triangle
    vctl::MCTri *pT = m_Triangles.New(pV0, pV1, pV2);
    pT->SetValuePtr(pInfo);
    pT->SetObject(-1);

    // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
    MDS_LOG_NOTE("CDelaunayTri::addTriangle: Created triangle " << pT);
#endif // MDS_SEG_LOGGING_ENABLED

    // Compute normal vector of the triangle
    vctl::MCVector3D Normal;
    pT->GetNormal(Normal);

    // Z-axis
//  vctl::MCVector3D ZAxis(0.0, 0.0, 1.0);

    // Check the triangle orientation (dot product)
    // - Inverse its orientation so that the vector product of the z-axis (in
    //   other words (0,0,1) vector) and any triangle edge is oriented outwards.
//  double dResult = Normal * ZAxis;
//  if( dResult < 0.0 )
    if( Normal.GetZ() < 0.0 )
    {
        pT->InverseOrientation();
    }

    // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
    vctl::MCVertex *pVertices[3];
    pT->GetVerticeS(pVertices);
    MDS_LOG_NOTE("CDelaunayTri::addTriangle: (" << pVertices[0]->GetX() << ',' << pVertices[0]->GetY() << ',' << pVertices[0]->GetZ() << "), "
                 << '(' << pVertices[1]->GetX() << ',' << pVertices[1]->GetY() << ',' << pVertices[1]->GetZ() << "), "
                 << '(' << pVertices[2]->GetX() << ',' << pVertices[2]->GetY() << ',' << pVertices[2]->GetZ() << ')'
                );
#endif // MDS_SEG_LOGGING_ENABLED

    // Return pointer to the triangle
    return pT;
}


void CDelaunayTri::removeTriangle(vctl::MCTri *pTriangle,
                                  bool bEraseVertices
                                 )
{
    // Delete triangle info struct
    delete (STriangleInfo *)pTriangle->GetValuePtr();
    pTriangle->SetValuePtr(NULL);

    // Get triangle vertices
    vctl::MCVertex *pVertices[3];
    pTriangle->GetVerticeS(pVertices);

    // Remove and de-register the triangle
    m_Triangles.Erase(pTriangle);

    // Remove unused triangle vertices
    if( bEraseVertices )
    {
        for( int i = 0; i < 3; ++i )
        {
            if( !pVertices[i]->GetRegisteredTri() )
            {
                m_pVertices->Erase(pVertices[i]);
            }
        }
    }
}


bool CDelaunayTri::checkFlatness(vctl::MCVertex *pV0,
                                 vctl::MCVertex *pV1,
                                 vctl::MCVertex *pV2
                                 )
{
    static const double MAX_DEVIATION = 0.000000001;

    MDS_CHECK(pV0 && pV1 && pV2, return false);

    // Prepare vectors
    vctl::MCVector3D v0(pV0, pV1);
    vctl::MCVector3D v1(pV0, pV2);

    // Normalize vectors
    v0.Normalization();
    v1.Normalization();

    // Compute angle between both vectors
    double dResult = 1.0 - mds::math::getAbs(v0 * v1);

    // Check if the normals are not parallel
    return (dResult < MAX_DEVIATION);
}


vctl::MCTri *CDelaunayTri::findBaseTriangle(vctl::MCPoint3D *pPoint)
{
    vctl::MCTri *pTriangle = (vctl::MCTri *)m_Triangles.GetFirst();
    if( !pTriangle || !pPoint )
    {
        return NULL;
    }

    return findBaseTriangle(pPoint, pTriangle);
}


vctl::MCTri *CDelaunayTri::findBaseTriangle(vctl::MCPoint3D *pPoint,
                                            vctl::MCTri *pTriangle
                                            )
{
    // Vector concident with the z-axis
//  static vctl::MCVector3D ZAxis(0.0, 0.0, 1.0);

    static const int MAX_ITERS  = 100;

    // Pointer to the previous triangle
    vctl::MCTri *pPrevious[3] = { NULL, NULL, NULL };

    // Traverse the mesh
    for( int iter = 0; pTriangle && iter < MAX_ITERS; ++iter )
    {
        // Loop check
        if( pTriangle == pPrevious[0] || pTriangle == pPrevious[1] )
        {
            // Logging
            MDS_LOG_NOTE("CDelaunayTri::findBaseTriangle(): Loop detected!");
            return NULL;
        }
        else
        {
            pPrevious[0] = pPrevious[1];
            pPrevious[1] = pPrevious[2];
            pPrevious[2] = pTriangle;
        }

        // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
        MDS_LOG_NOTE("CDelaunayTri::findBaseTriangle(): Triangle " << pTriangle);
#endif // MDS_SEG_LOGGING_ENABLED

        // Compute the vertex position relative to edges
        double dPosition[3];
        for( int i = 0; i < 3; ++i )
        {
            // Next end-point of the edge
            int j = (i + 1) % 3;

            // Get vertices of the triangle
            vctl::MCVertex *pVertices[3];
            pTriangle->GetVerticeS(pVertices);

            // Center of the edge
            vctl::MCPoint3D EdgeCenter(
                0.5 * (pVertices[i]->GetX() + pVertices[j]->GetX()),
                0.5 * (pVertices[i]->GetY() + pVertices[j]->GetY()),
                0.0
            );

            // Compute vector oriented from the center to given vertex
            vctl::MCVector3D TestVector(&EdgeCenter, pPoint);
//            TestVector.Normalization();

            // Compute normal of the edge
            vctl::MCVector3D EdgeVector(pVertices[i], pVertices[j]);
//            vctl::MCVector3D EdgeNormal;
//            EdgeNormal.VectorProduct(EdgeVector, ZAxis);
            vctl::MCVector3D EdgeNormal(EdgeVector.GetY(), -EdgeVector.GetX(), 0.0);
//            EdgeNormal.Normalize();

            // Dot product of the test vector and edge normal
            dPosition[i] = TestVector * EdgeNormal;
        }

        // Check the vertex position
        if( dPosition[0] <= 0.0 && dPosition[1] <= 0.0 && dPosition[2] <= 0.0 )
        {
            return pTriangle;
        }

        // Get the most forward edge
        int iMax = (dPosition[0] > dPosition[1]) ? 0 : 1;
        iMax = (dPosition[2] > dPosition[iMax]) ? 2 : iMax;

        // Get neighbouring triangle
        vctl::MCTri *pNeighbour = pTriangle->GetNeighbour(iMax);
        if( !pNeighbour )
        {
            return NULL;
        }
        else
        {
            // Next step
            pTriangle = pNeighbour;
        }
    }

    // Failed to find base triangle
    return NULL;
}


void CDelaunayTri::clearFlags()
{
    vctl::MCTri *pTriangle = getFirstTriangle();
    while( pTriangle )
    {
        // Clear triangle flag
        pTriangle->MaskFlag(vctl::ENTITY_MARK_FLAG);

        // Get next one
        pTriangle = pTriangle->GetNext();
    }
}


void CDelaunayTri::clearFlags(const tTriangles& Triangles)
{
    tTriangles::const_iterator it = Triangles.begin();
    tTriangles::const_iterator itEnd = Triangles.end();
    while( it != itEnd )
    {
        // Clear triangle flag
        (*it)->MaskFlag(vctl::ENTITY_MARK_FLAG);

        // Get next triangle
        ++it;
    }
}


bool CDelaunayTri::findCavity(vctl::MCPoint3D *pPoint,
                              tTriangles& Cavity
                             )
{
    MDS_ASSERT(pPoint);

    // Find the vertex base traingle
    vctl::MCTri *pBaseTriangle = findBaseTriangle(pPoint);
    MDS_CHECK(pBaseTriangle, return false);

/*    // Clear flags of all found triangles
    tTriangles::iterator it = Cavity.begin();
    tTriangles::iterator itEnd = Cavity.end();
    while( it != itEnd )
    {
        (*it)->ReplaceFlag(vctl::NO_FLAG);
        ++it;
    }*/

    // Recursive finding of the cavity
    return findCavity(pPoint, Cavity, pBaseTriangle);
}


bool CDelaunayTri::findCavity(vctl::MCPoint3D *pPoint,
                              tTriangles& Cavity,
                              vctl::MCTri *pTriangle
                              )
{
    // Is the triangle already tested?
    if( pTriangle->TestFlag(vctl::ENTITY_MARK_FLAG) )
    {
        return true;
    }

    // Check if the vertex lies in circumcircle of the triangle
    int iResult = getTriangleCircumsphere(pTriangle)->TestPoint(*pPoint);
    if( iResult < 0 )
    {
        return true;
    }
    else if( iResult == 0 )
    {
        return false;
    }

    // Mark the triangle
    pTriangle->SetFlag(vctl::ENTITY_MARK_FLAG);

    // Add pointer to the triangle to the output vector
    Cavity.push_back(pTriangle);

    // Get neighbours
    vctl::MCTri *pNeighbours[3];
    pTriangle->GetNeighbours(pNeighbours);

    // Recursive call of the function
    if( pNeighbours[0] )
    {
        if( !findCavity(pPoint, Cavity, pNeighbours[0]) )
        {
            return false;
        }
    }
    if( pNeighbours[1] )
    {
        if( !findCavity(pPoint, Cavity, pNeighbours[1]) )
        {
            return false;
        }
    }
    if( pNeighbours[2] )
    {
        if( !findCavity(pPoint, Cavity, pNeighbours[2]) )
        {
            return false;
        }
    }

    // O.K.
    return true;
}


void CDelaunayTri::clear()
{
    if( getNumOfTriangles() == 0 )
    {
        return;
    }

    // Delete info structs
    vctl::MCTri *pTriangle = getFirstTriangle();
    while( pTriangle )
    {
        // Delete triangle info struct
        delete (STriangleInfo *)pTriangle->GetValuePtr();
        pTriangle->SetValuePtr(NULL);

        // Get next one
        pTriangle = pTriangle->GetNext();
    }

    // Clear triangles and their vertices
    m_Triangles.ClearAll();
    m_pVertices->ClearAll();
}


bool CDelaunayTri::findCavityEnvelope(const tTriangles& Cavity,
                                      tVertices& Envelope
                                     )
{
/*    // Mark all triangles in the cavity
    tTriangles::const_iterator it = Cavity.begin();
    tTriangles::const_iterator itEnd = Cavity.end();
    while( it != itEnd )
    {
        // Mark the triangle
        (*it)->SetFlag(vctl::ENTITY_MARK_FLAG);
        ++it;
    }*/

    // Clear the cavity envelope
    Envelope.clear();

    // Find boundary edges
    tTriangles::const_iterator it = Cavity.begin();
    tTriangles::const_iterator itEnd = Cavity.end();
    while( it != itEnd )
    {
        // Get pointer to the triangle
        vctl::MCTri *pTriangle = *it;

        // Get vertices of the triangle
        vctl::MCVertex *pVertices[3];
        pTriangle->GetVerticeS(pVertices);

        // Get neighbours
        vctl::MCTri *pNeighbours[3];
        pTriangle->GetNeighbours(pNeighbours);

        // Check if neighbouring triangles are marked
        if( !pNeighbours[0] || !pNeighbours[0]->TestFlag(vctl::ENTITY_MARK_FLAG) )
        {
            // Add edge end-points to the list
            Envelope.push_back(pVertices[0]);
            Envelope.push_back(pVertices[1]);
        }
        if( !pNeighbours[1] || !pNeighbours[1]->TestFlag(vctl::ENTITY_MARK_FLAG) )
        {
            Envelope.push_back(pVertices[1]);
            Envelope.push_back(pVertices[2]);
        }
        if( !pNeighbours[2] || !pNeighbours[2]->TestFlag(vctl::ENTITY_MARK_FLAG) )
        {
            Envelope.push_back(pVertices[2]);
            Envelope.push_back(pVertices[0]);
        }

        // Next triangle
        ++it;
    }

    // Clear flags of all triangles
    it = Cavity.begin();
    while( it != itEnd )
    {
        (*it)->MaskFlag(vctl::ENTITY_MARK_FLAG);
        ++it;
    }

    // O.K.
    return true;
}


bool CDelaunayTri::getMinCoordinates(double& dX, double& dY)
{
    vctl::MCVertex *pVertex = getFirstVertex();
    if( !pVertex )
    {
        dX = 0.0;
        dY = 0.0;
        return false;
    }

    dX = pVertex->GetX();
    dY = pVertex->GetY();

    pVertex = pVertex->GetNext();
    while( pVertex )
    {
        double dTemp = pVertex->GetX();
        if( dTemp < dX )
        {
            dX = dTemp;
        }

        dTemp = pVertex->GetY();
        if( dTemp < dY )
        {
            dY = dTemp;
        }

        // Get next one
        pVertex = pVertex->GetNext();
    }

    // O.K.
    return true;
}


bool CDelaunayTri::getMaxCoordinates(double& dX, double& dY)
{
    vctl::MCVertex *pVertex = getFirstVertex();
    if( !pVertex )
    {
        dX = 0.0;
        dY = 0.0;
        return false;
    }

    dX = pVertex->GetX();
    dY = pVertex->GetY();

    pVertex = pVertex->GetNext();
    while( pVertex )
    {
        double dTemp = pVertex->GetX();
        if( dTemp > dX )
        {
            dX = dTemp;
        }

        dTemp = pVertex->GetY();
        if( dTemp > dY )
        {
            dY = dTemp;
        }

        // Get next one
        pVertex = pVertex->GetNext();
    }

    // O.K.
    return true;
}


double CDelaunayTri::getTriangleQuality(vctl::MCTri *pTriangle)
{
    static const double NORM_FACTOR = mds::math::SQRT3 / 6.0;

    MDS_ASSERT(pTriangle);

    // Get triangle circuit and area
    double dPerimeter = pTriangle->GetCircuit();
    double dArea = pTriangle->GetArea();
    MDS_ASSERT(dPerimeter != 0.0 && dArea != 0.0);

    // Compute radius of the inscribed circle
    double dR = 2.0 * dArea / dPerimeter;

    // Get length of the longest edge
    double dL = pTriangle->GetMaxEdgeLength();

    // Compute the triangle quality
    return (NORM_FACTOR * dL / dR);
}


void CDelaunayTri::normInsertFunc(vctl::MCTri *pTriangle, CDelaunayTri *pMesh)
{
    MDS_ASSERT(pTriangle && pMesh);

    // Compute and set the triangle quality
    pTriangle->SetValue(getTriangleQuality(pTriangle));

    // Get sorted queue
    vctl::MCTriS::MT_QUEUE *pQueue = pMesh->m_Triangles.GetQueue();

    // Sort the triangle
    pQueue->Sort(pTriangle->GetValue(), pTriangle);
}


void CDelaunayTri::normRemoveFunc(vctl::MCTri *pTriangle, CDelaunayTri *pMesh)
{
    MDS_ASSERT(pTriangle && pMesh);

    // Get sorted queue
    vctl::MCTriS::MT_QUEUE *pQueue = pMesh->m_Triangles.GetQueue();

    // Unsort the triangle
    if( pTriangle->GetValue() > 0.0 )
    {
        pQueue->UnSort(pTriangle->GetValue(), pTriangle);
    }
}


} // namespace seg
} // namespace mds

