//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2006/04/04                          \n
 *
 * $Id: mdsTetraFeatures.cpp 1629 2010-02-16 14:17:39Z spanel $
 *
 * Description:
 * - Tetrahedron feature vector.
 */

#include <MDSTk/VectorSegmentation/mdsTetraIterator.h>
#include <MDSTk/VectorSegmentation/mdsTetraFeatures.h>
#include <MDSTk/VectorSegmentation/mdsTetraData.h>

// MDSTk
#include <MDSTk/Base/mdsGlobalLog.h>
#include <MDSTk/Image/mdsVolumeFiltering.h>
#include <MDSTk/Math/mdsVectorFunctions.h>

// STL
#include <cmath>


namespace mds
{
namespace seg
{

//==============================================================================
/*
 * Implementation of the class CTetraFeaturesExtractor.
 */
void CTetraFeaturesExtractor::acumulate(tVolume *pVolume, vctl::MCTetra *pTetrahedron, tScale *pScale)
{
    MDS_CHECK(pVolume && pTetrahedron && pScale, return);

//    MDS_LOG_NOTE("pTetra: " << pTetrahedron << ", pVolume: " << pVolume);

#ifdef SKIP_BOUNDARY_TETRAS
    static const int DOMAIN_BOUNDARY_FLAG = 1 << 11;

    vctl::MCVertex *pVertices[4];
    pTetrahedron->GetVerticeS(pVertices);
    if( pVertices[0]->TestFlag(DOMAIN_BOUNDARY_FLAG)
        || pVertices[1]->TestFlag(DOMAIN_BOUNDARY_FLAG)
        || pVertices[2]->TestFlag(DOMAIN_BOUNDARY_FLAG)
        || pVertices[3]->TestFlag(DOMAIN_BOUNDARY_FLAG) )
    {
        return;
    }
#endif // SKIP_BOUNDARY_TETRAS

#ifdef SKIP_EDGE_VOXELS
    // Edge magnitude threshold
    static const double EDGE_THRESHOLD = 100.0;

    // Sobel filters
    mds::img::CVolumeSobelX<tVolume> SobelX;
    mds::img::CVolumeSobelY<tVolume> SobelY;
    mds::img::CVolumeSobelZ<tVolume> SobelZ;
#endif // SKIP_EDGE_VOXELS

    // Create tetrahedron iterator
    CTetraIterator it(pVolume, pTetrahedron, pScale);

    // Traverse all voxels
    int iCount = 0;
    for( ; !it.isEnd(); ++it )
    {
#ifdef SKIP_EDGE_VOXELS
        // Image derivatives
        double dGradX = SobelX.getResponse(*pVolume, it.getX(), it.getY(), it.getZ());
        double dGradY = SobelY.getResponse(*pVolume, it.getX(), it.getY(), it.getZ());
        double dGradZ = SobelZ.getResponse(*pVolume, it.getX(), it.getY(), it.getZ());
        double dMagnitude = mds::math::getAbs(dGradX) + mds::math::getAbs(dGradY) + mds::math::getAbs(dGradZ);
        if( dMagnitude > EDGE_THRESHOLD )
        {
            continue;
        }
#endif

        // Current voxel value
        double dValue = double(*it);

        // Count the number of voxels
        ++iCount;

        // BASIC TRIANGLE FEATURES
        m_dSum += dValue;
        m_dSumSqr += dValue * dValue;
    }

#ifdef SUBST_SMALL_TETRAS
    if( iCount > 0 )
    {
        // Total number of voxels
        m_iNumOfVoxels += iCount;
    }
    else
    {
        // Center of the tetrahedron
        vctl::MCVertex Center;

        // Get pointer to the tetra circumsphere
/*        vctl::MCSphere *pCircumsphere = getTetrahedronCircumsphere(pTetrahedron);
        if( pCircumsphere )
        {
            // Get center of the circumsphere
            pCircumsphere->GetCenter(Center);
        }
        else*/
        {
            pTetrahedron->GetCenter(Center);
        }

        // Convert its coordinates
        tSize x, y, z;
        pScale->conv2Volume(&Center, x, y, z);

        // Project on the boundary
/*        x = mds::math::getMin(mds::math::getMax(x, 0), pVolume->getXSize());
        y = mds::math::getMin(mds::math::getMax(y, 0), pVolume->getYSize());
        z = mds::math::getMin(mds::math::getMax(z, 0), pVolume->getZSize());*/

        // Get voxel value
        double dValue = pVolume->get(x, y, z);

/*#ifdef SKIP_EDGE_VOXELS
        // Image derivatives
        double dGradX = SobelX.getResponse(*pVolume, x, y, z);
        double dGradY = SobelY.getResponse(*pVolume, x, y, z);
        double dGradZ = SobelZ.getResponse(*pVolume, x, y, z);
        double dMagnitude = mds::math::getAbs(dGradX) + mds::math::getAbs(dGradY) + mds::math::getAbs(dGradZ);
        if( dMagnitude < EDGE_THRESHOLD )
        {
#endif*/

        // Total number of voxels
        ++m_iNumOfVoxels;

        // BASIC TRIANGLE FEATURES
        m_dSum += dValue;
        m_dSumSqr += dValue * dValue;

/*#ifdef SKIP_EDGE_VOXELS
        }
#endif*/
    }
#else
    // Total number of voxels
    m_iNumOfVoxels += iCount;
#endif // SUBST_SMALL_TETRAS
}


void CTetraFeaturesExtractor::acumulate(const CTetraFeaturesExtractor& Extractor)
{
    m_dSum += Extractor.m_dSum;
    m_dSumSqr += Extractor.m_dSumSqr;
    m_iNumOfVoxels += Extractor.m_iNumOfVoxels;
}


bool CTetraFeaturesExtractor::evaluate(tVector *pVector, bool bCheckNumOfVoxels)
{
    MDS_CHECK(pVector, return false);

    // Clear the whole feature vector
    pVector->fill(0);

    // Set the number of voxels
    pVector->setNumOfVoxels(m_iNumOfVoxels);

    // Check the number of voxels
    double dInvCount = 1.0;
    int iLimit = (bCheckNumOfVoxels) ? TetraFeatures::MIN_NUM_OF_VOXELS : 1;
    if( m_iNumOfVoxels >= iLimit )
    {
        dInvCount = 1.0 / m_iNumOfVoxels;
    }
    else
    {
        // Set the error code
        pVector->setError(CTetraFeatures::ERROR_CANNOT_EVALUATE);
        return false;
    }

    // BASIC FEATURES
    double dMean = m_dSum * dInvCount;
    double dVariance = m_dSumSqr * dInvCount - (dMean * dMean);
    if( mds::math::getAbs(dVariance) < 0.000001 )
    {
        dVariance = 0.0;
    }

    pVector->setVoxelMean(tFeature(dMean));
//    pVector->setVoxelVariance(tFeature(std::sqrt(dVariance)));
    pVector->setVoxelVariance(tFeature(dVariance));

    // O.K.
    pVector->setError(0);
    return true;
}


void CTetraFeaturesExtractor::clear()
{
    // BASIC TRIANGLE FEATURES
    m_dSum = 0.0;
    m_dSumSqr = 0.0;
    m_iNumOfVoxels = 0;
}


} // namespace seg
} // namespace mds

