//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2006/12/21                          \n
 *
 * $Id: mdsTetraMembershipEM2.cpp 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Membership function based on Expection-Maximization (EM) algorithm.
 * - Modified approach which classifies predefined sub-regions instead of
 *   tetrahedra.
 */

#include <MDSTk/VectorSegmentation/mdsTetraMembershipEM2.h>

// MDSTk
#include <MDSTk/Base/mdsGlobalLog.h>


namespace mds
{
namespace seg
{

//==============================================================================
/*
 * Implementation of the CTetraMembershipEM class.
 */
CTetraMembershipEM2::CTetraMembershipEM2(tSubRegions& SubRegions, double dMinAddChange)
    : m_EM(dMinAddChange)
    , m_SubRegions(SubRegions)
{
}


bool CTetraMembershipEM2::operator()(vctl::MCTetra *pFirstTetrahedron)
{
    MDS_CHECK(pFirstTetrahedron && m_SubRegions.size() > 0, return false);

    // Number of sub-regions
    int iNumOfSubRegions = (int)m_SubRegions.size();

    // Check that all features are correctly evaluated
    tSubRegions CheckedSubRegions;
    CheckedSubRegions.reserve(iNumOfSubRegions);
    std::vector<int> Mapping(iNumOfSubRegions);
    for( tSize i = 0; i < iNumOfSubRegions; ++i )
    {
        if( !m_SubRegions[i].isError() )
        {
            CheckedSubRegions.push_back(m_SubRegions[i]);
            Mapping[i] = (int)CheckedSubRegions.size() - 1;
        }
        else
        {
            Mapping[i] = -1;
        }
    }

    // Classify all sub-regions
    tContainer Container(CheckedSubRegions);
    if( !m_EM.execute(Container) )
    {
        return false;
    }

    // Classify all input tetrahedra
    vctl::MCTetra *pTetrahedron = pFirstTetrahedron;
    while( pTetrahedron )
    {
        // Get tetrahedron region
        int iRegion = getTetrahedronRegion(pTetrahedron);
        
        // If the tetrahedron is not yet classified
        if( iRegion >= 0 && iRegion < iNumOfSubRegions && Mapping[iRegion] != -1 )
        {
            iRegion = m_EM.getMembership(Mapping[iRegion]);
        }
        else
        {
            iRegion = -1;
        }

        // Set region index of the tetrahedron
        setTetrahedronRegion(pTetrahedron, iRegion);
        
        // Get the next tetrahedron
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Set the final number of segments
    m_NumOfSegments = m_EM.getNumOfComponents();

    // O.K.
    return true;
}


bool CTetraMembershipEM2::operator()(vctl::MCTetra *pFirstTetrahedron,
                                     mds::tSize NumOfClasses
                                     )
{
    MDS_CHECK(pFirstTetrahedron && m_SubRegions.size() > 0, return false);

    // Number of sub-regions
    int iNumOfSubRegions = (int)m_SubRegions.size();

    // Check that all features are correctly evaluated
    tSubRegions CheckedSubRegions;
    CheckedSubRegions.reserve(iNumOfSubRegions);
    std::vector<int> Mapping(iNumOfSubRegions);
    for( tSize i = 0; i < iNumOfSubRegions; ++i )
    {
        if( !m_SubRegions[i].isError() )
        {
            CheckedSubRegions.push_back(m_SubRegions[i]);
            Mapping[i] = (int)CheckedSubRegions.size() - 1;
        }
        else
        {
            Mapping[i] = -1;
        }
    }

    // Classify all sub-regions
    tContainer Container(CheckedSubRegions);
    if( !m_EM.execute(Container, NumOfClasses) )
    {
        return false;
    }

    // Classify all input tetrahedra
    vctl::MCTetra *pTetrahedron = pFirstTetrahedron;
    while( pTetrahedron )
    {
        // Get tetrahedron region
        int iRegion = getTetrahedronRegion(pTetrahedron);
        
        // If the tetrahedron is not yet classified
        if( iRegion >= 0 && iRegion < iNumOfSubRegions && Mapping[iRegion] != -1 )
        {
            iRegion = m_EM.getMembership(Mapping[iRegion]);
        }
        else
        {
            iRegion = -1;
        }

        // Set region index of the tetrahedron
        setTetrahedronRegion(pTetrahedron, iRegion);
        
        // Get the next tetrahedron
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Set the final number of segments
    m_NumOfSegments = m_EM.getNumOfComponents();

    // O.K.
    return true;
}


} // namespace seg
} // namespace mds

