//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2006/05/06                          \n
 *
 * $Id: mdsTetraMesh.cpp 1629 2010-02-16 14:17:39Z spanel $
 *
 * Description:
 * - Adaptive tetrahedral mesh based on 3D Delaunay triangulation.
 */

#include <MDSTk/VectorSegmentation/mdsTetraMesh.h>
#include <MDSTk/VectorSegmentation/mdsTetraVertexDetector.h>
#include <MDSTk/VectorSegmentation/mdsTetraEdgeIterator.h>
#include <MDSTk/VectorSegmentation/mdsTetraTriIterator.h>
#include <MDSTk/VectorSegmentation/mdsTetraIterator.h>

// MDSTk
#include <MDSTk/Base/mdsMemory.h>
#include <MDSTk/Base/mdsGlobalLog.h>
#include <MDSTk/Image/mdsLine.h>
#include <MDSTk/Image/mdsVolumeFiltering.h>
#include <MDSTk/Image/mdsPixelTraits.h>
#include <MDSTk/Image/Utilities/mdsAdjacencyGraph.h>
#include <MDSTk/Math/mdsBase.h>
#include <MDSTk/Math/mdsStaticMatrix.h>
#include <MDSTk/Math/MatrixFunctions/mdsEig.h>
#include <MDSTk/Math/MatrixFunctions/mdsGeneral.h>

// STL
#include <cmath>
#include <ctime>
#include <string>
#include <sstream>
#include <deque>
#include <list>

// Hash map
//#define USE_HASH_MAP
#ifdef USE_HASH_MAP
#   ifdef _MSC_VER
#       include <hash_map>
        namespace stdhash = stdext;
#   else
#       include <ext/hash_map>
        namespace stdhash = __gnu_cxx;
#   endif // _MSC_VER
#endif // USE_HASH_MAP


namespace mds
{
namespace seg
{

//==============================================================================
/*
 * Data members of the class CTetraMesh.
 */
const tSize CTetraMesh::DEFAULT_CELL                    = 3;
//const tSize CTetraMesh::DEFAULT_CELL_MARGIN             = 1;
//const tSize CTetraMesh::DEFAULT_CELL_MARGIN             = 3;
const tSize CTetraMesh::DEFAULT_CELL_MARGIN             = 5;
//const tSize CTetraMesh::DEFAULT_CELL_MARGIN             = 16;
const tSize CTetraMesh::DEFAULT_VERTEX_NEIGHBOURHOOD    = 3;
const double CTetraMesh::DEFAULT_INIT_THRESHOLD         = 0.5;
const double CTetraMesh::DEFAULT_SPLITTING_THRESHOLD    = 0.5;
const double CTetraMesh::DEFAULT_DISTANCE_THRESHOLD     = 1.5;
const double CTetraMesh::MIN_TETRAHEDRON_VOLUME         = 2.0;
const double CTetraMesh::DEFAULT_DELTA_THRESHOLD        = 1.5;
const double CTetraMesh::DEFAULT_CORNER_MIN_EDGE_LENGTH = 0.5;
const double CTetraMesh::DEFAULT_AVG_TETRA_SIZE         = 30.0;
const double CTetraMesh::DEFAULT_K                      = 1.0;


//==============================================================================
/*
 * Implementation of the class CTetraMesh.
 */
CTetraMesh::CTetraMesh()
    : m_iNumOfSegments(0)
    , m_VertexNeighbourhood(DEFAULT_VERTEX_NEIGHBOURHOOD)
    , m_dCornerMinEdgeLength(DEFAULT_CORNER_MIN_EDGE_LENGTH)
    , m_pHomogeneityMeasure(NULL)
{
}


CTetraMesh::~CTetraMesh()
{
}


void CTetraMesh::setVolume(mds::img::CDensityVolume *pVolume)
{
    m_spVolume = pVolume;

    setVoxelSize(pVolume->getDX(), pVolume->getDY(), pVolume->getDZ());
}


void CTetraMesh::setEdgeVolume(mds::img::CDensityVolume *pEdgeVolume)
{
    m_spEdgeVolume = pEdgeVolume;

    setVoxelSize(pEdgeVolume->getDX(), pEdgeVolume->getDY(), pEdgeVolume->getDZ());
}


void CTetraMesh::setControlSpace(mds::img::CDensityVolume *pControlSpace)
{
    m_spControlSpace = pControlSpace;

    setVoxelSize(pControlSpace->getDX(), pControlSpace->getDY(), pControlSpace->getDZ());
}


//! Helper structure representing a voxel position.
struct SVoxelInfo
{
    //! Position
    short m_x, m_y, m_z;
//    mds::tSize m_x, m_y, m_z;

    //! Default constructor.
    SVoxelInfo() {}

    //! Constructor
    SVoxelInfo(mds::tSize x, mds::tSize y, mds::tSize z) : m_x(short(x)), m_y(short(y)), m_z(short(z)) {}
//    SVoxelInfo(mds::tSize x, mds::tSize y, mds::tSize z) : m_x(x), m_y(y), m_z(z) {}
};


//! Helper structure representing a voxel position.
struct SVoxelInfo2
{
    //! Position
//    mds::tSize m_x, m_y, m_z, m_Count;
    short m_x, m_y, m_z, m_Count;

    //! Value
    mds::seg::CTetraMesh::tVoxel m_Value;

    //! Default constructor.
    SVoxelInfo2() {}

    //! Constructor
/*    SVoxelInfo2(mds::tSize x,
                mds::tSize y,
                mds::tSize z,
                mds::seg::CTetraMesh::tVoxel Value,
                mds::tSize Count = 0
                )
        : m_x(x), m_y(y), m_z(z)
        , m_Value(Value)
        , m_Count(Count)
    {}*/
    SVoxelInfo2(mds::tSize x,
                mds::tSize y,
                mds::tSize z,
                mds::seg::CTetraMesh::tVoxel Value,
                mds::tSize Count = 0
                )
        : m_x(short(x)), m_y(short(y)), m_z(short(z))
        , m_Count(short(Count))
        , m_Value(Value)
    {}
};


// Medial axis version
bool CTetraMesh::prepareControlSpace(double dK, double dAvgTetrahedronSize)
{
    static const tVoxel Max = mds::img::CPixelTraits<tVoxel>::getPixelMax();

    // Logging
    MDS_LOG_NOTE("CTetraMesh::prepareControlSpace(): started");

    // Size of the volume data
    tSize XSize = m_spEdgeVolume->getXSize();
    tSize YSize = m_spEdgeVolume->getYSize();
    tSize ZSize = m_spEdgeVolume->getZSize();
    tSize Margin = m_spEdgeVolume->getMargin();

    MDS_CHECK(XSize > 0 && YSize > 0 && ZSize > 0 && Margin > 0, return false);

    // Container of voxels
    typedef std::vector<SVoxelInfo> tVoxels;
//    typedef std::list<SVoxelInfo> tVoxels;
    tVoxels Voxels1, Voxels2;

    // Resize the control space
    m_spControlSpace->create(XSize, YSize, ZSize);
    m_spControlSpace->fillEntire(Max);
    m_spControlSpace->fillMargin(0);

    // Set voxel properties
    m_spControlSpace->setVoxel(m_spEdgeVolume->getDX(),
                               m_spEdgeVolume->getDY(),
                               m_spEdgeVolume->getDZ()
                               );

    // Process the edge volume
    tVolume::tIterator itEnd = m_spEdgeVolume->getEnd();
    tVolume::tIterator it = m_spEdgeVolume->getBegin();
    for( ; it != itEnd; ++it )
    {
        // Edge points
        if( *it > 0 )
        {
            Voxels1.push_back(SVoxelInfo(it.getX(), it.getY(), it.getZ()));
        }
    }

    MDS_LOG_NOTE("Distance transform (edge points)");

    // For every voxel compute its distance from the nearest edge
    tVoxel Dist;
    for( Dist = 0; Dist < Max && !Voxels1.empty(); ++Dist )
    {
        // Logging
        MDS_LOG_NOTE("  Number of voxels = " << Voxels1.size());

        // Clear the output container
        Voxels2.clear();

        // Iterate all voxels
        tVoxels::iterator it2End = Voxels1.end();
        tVoxels::iterator it2 = Voxels1.begin();
        for( ; it2 != it2End; ++it2 )
        {
            // Get current voxel position
            mds::tSize x = it2->m_x;
            mds::tSize y = it2->m_y;
            mds::tSize z = it2->m_z;

            // Voxel value
            tVoxel Value = m_spControlSpace->get(x, y, z);

            // Check the voxel value
            if( Value <= Dist )
            {
                continue;
            }

            // Set the voxel value
            m_spControlSpace->set(x, y, z, Dist);

            // Check voxel neighbours
            tVoxel NewDist = Dist + 1;
            Value = m_spControlSpace->get(x + 1, y, z);
            if( Value > NewDist )
            {
                Voxels2.push_back(SVoxelInfo(x + 1, y, z));
            }
            Value = m_spControlSpace->get(x - 1, y, z);
            if( Value > NewDist )
            {
                Voxels2.push_back(SVoxelInfo(x - 1, y, z));
            }
            Value = m_spControlSpace->get(x, y + 1, z);
            if( Value > NewDist )
            {
                Voxels2.push_back(SVoxelInfo(x, y + 1, z));
            }
            Value = m_spControlSpace->get(x, y - 1, z);
            if( Value > NewDist )
            {
                Voxels2.push_back(SVoxelInfo(x, y - 1, z));
            }
            Value = m_spControlSpace->get(x, y, z + 1);
            if( Value > NewDist )
            {
                Voxels2.push_back(SVoxelInfo(x, y, z + 1));
            }
            Value = m_spControlSpace->get(x, y, z - 1);
            if( Value > NewDist )
            {
                Voxels2.push_back(SVoxelInfo(x, y, z - 1));
            }
        }

        // Interchange both containers
        std::swap(Voxels1, Voxels2);
    }

    MDS_LOG_NOTE("Medial axis finding");

    // Find local maxima of the distance transform
    Voxels1.clear();
    m_spControlSpace->fillMargin(Max);
    itEnd = m_spControlSpace->getEnd();
    it = m_spControlSpace->getBegin();
    for( ; it != itEnd; ++it )
    {
        // Get current voxel position
        mds::tSize x = it.getX();
        mds::tSize y = it.getY();
        mds::tSize z = it.getZ();

        // Current value
        tVoxel Value = *it;

        // Check local maxima
        if( Value >= m_spControlSpace->get(x - 1, y, z)
            && Value >= m_spControlSpace->get(x + 1, y, z)
            && Value >= m_spControlSpace->get(x, y - 1, z)
            && Value >= m_spControlSpace->get(x, y + 1, z)
            && Value >= m_spControlSpace->get(x, y, z - 1)
            && Value >= m_spControlSpace->get(x, y, z + 1)
            )
        {
            Voxels1.push_back(SVoxelInfo(x, y, z));
        }
    }

    MDS_LOG_NOTE("Estimating local feature size");

    // Distance from medial axis
    m_spControlSpace->fillEntire(Max);
    m_spControlSpace->fillMargin(0);
    for( Dist = 0; Dist < Max && !Voxels1.empty(); ++Dist )
    {
        // Logging
        MDS_LOG_NOTE("  Number of voxels = " << Voxels1.size());

        // Clear the output container
        Voxels2.clear();

        // Iterate all voxels
        tVoxels::iterator it2End = Voxels1.end();
        tVoxels::iterator it2 = Voxels1.begin();
        for( ; it2 != it2End; ++it2 )
        {
            // Get current voxel position
            mds::tSize x = it2->m_x;
            mds::tSize y = it2->m_y;
            mds::tSize z = it2->m_z;

            // Voxel value
            tVoxel Value = m_spControlSpace->get(x, y, z);

            // Check the voxel value
            if( Value <= Dist )
            {
                continue;
            }

            // Set the voxel value
            m_spControlSpace->set(x, y, z, Dist);

            // Check voxel neighbours
            tVoxel NewDist = Dist + 1;
            Value = m_spControlSpace->get(x + 1, y, z);
            if( Value > NewDist )
            {
                Voxels2.push_back(SVoxelInfo(x + 1, y, z));
            }
            Value = m_spControlSpace->get(x - 1, y, z);
            if( Value > NewDist )
            {
                Voxels2.push_back(SVoxelInfo(x - 1, y, z));
            }
            Value = m_spControlSpace->get(x, y + 1, z);
            if( Value > NewDist )
            {
                Voxels2.push_back(SVoxelInfo(x, y + 1, z));
            }
            Value = m_spControlSpace->get(x, y - 1, z);
            if( Value > NewDist )
            {
                Voxels2.push_back(SVoxelInfo(x, y - 1, z));
            }
            Value = m_spControlSpace->get(x, y, z + 1);
            if( Value > NewDist )
            {
                Voxels2.push_back(SVoxelInfo(x, y, z + 1));
            }
            Value = m_spControlSpace->get(x, y, z - 1);
            if( Value > NewDist )
            {
                Voxels2.push_back(SVoxelInfo(x, y, z - 1));
            }
        }

        // Interchange both containers
        std::swap(Voxels1, Voxels2);
    }

    Voxels1.clear();
    Voxels2.clear();
    Voxels1.reserve(0);
    Voxels2.reserve(0);
//    Voxels1.resize(0);
//    Voxels2.resize(0);

    // Priority queue
#ifdef USE_HASH_MAP
    typedef stdhash::hash_multimap<tVoxel, SVoxelInfo2> tQueue;
#else
    typedef std::multimap<tVoxel, SVoxelInfo2> tQueue;
#endif // USE_HASH_MAP
    tQueue Queue;

    // Estimate local feature size in place of image edges
    itEnd = m_spEdgeVolume->getEnd();
    it = m_spEdgeVolume->getBegin();
    for( ; it != itEnd; ++it )
    {
        // Edge points
        if( *it > 0 )
        {
            mds::tSize x = it.getX();
            mds::tSize y = it.getY();
            mds::tSize z = it.getZ();

            tVoxel Value = m_spControlSpace->get(x, y, z);

            Queue.insert(tQueue::value_type(Value, SVoxelInfo2(x, y, z, Value, 0)));

//            MDS_LOG_NOTE("LFS = " << Value);
        }
//        else
//        {
//            m_spControlSpace->set(it.getX(), it.getY(), it.getZ(), 0);
//        }
    }

    MDS_LOG_NOTE("Generating control space");

    // Propagate control space value
    m_spControlSpace->fillEntire(Max);
    m_spControlSpace->fillMargin(0);
    while( !Queue.empty() )
    {
        // Pick voxel at the begining
        tQueue::iterator itItem = Queue.begin();
        tVoxel ItemValue = itItem->first;
        SVoxelInfo2 Item = itItem->second;
        Queue.erase(itItem);

        // Get current voxel position
        mds::tSize x = Item.m_x;
        mds::tSize y = Item.m_y;
        mds::tSize z = Item.m_z;

        // Voxel value
        tVoxel Value = m_spControlSpace->get(x, y, z);

        // Check the voxel value
        if( Value <= ItemValue )
        {
            continue;
        }

        // Set the voxel value
        m_spControlSpace->set(x, y, z, ItemValue);

        // Estimate further control space value
        ++(Item.m_Count);
        tVoxel NewValue = tVoxel(dK * Item.m_Count + Item.m_Value);
        if( NewValue >= Max )
        {
            break;
        }

        // Check voxel neighbours
        Value = m_spControlSpace->get(x + 1, y, z);
        if( Value > NewValue )
        {
            Queue.insert(tQueue::value_type(NewValue, SVoxelInfo2(x + 1, y, z, Item.m_Value, Item.m_Count)));
        }
        Value = m_spControlSpace->get(x - 1, y, z);
        if( Value > NewValue )
        {
            Queue.insert(tQueue::value_type(NewValue, SVoxelInfo2(x - 1, y, z, Item.m_Value, Item.m_Count)));
        }
        Value = m_spControlSpace->get(x, y + 1, z);
        if( Value > NewValue )
        {
            Queue.insert(tQueue::value_type(NewValue, SVoxelInfo2(x, y + 1, z, Item.m_Value, Item.m_Count)));
        }
        Value = m_spControlSpace->get(x, y - 1, z);
        if( Value > NewValue )
        {
            Queue.insert(tQueue::value_type(NewValue, SVoxelInfo2(x, y - 1, z, Item.m_Value, Item.m_Count)));
        }
        Value = m_spControlSpace->get(x, y, z + 1);
        if( Value > NewValue )
        {
            Queue.insert(tQueue::value_type(NewValue, SVoxelInfo2(x, y, z + 1, Item.m_Value, Item.m_Count)));
        }
        Value = m_spControlSpace->get(x, y, z - 1);
        if( Value > NewValue )
        {
            Queue.insert(tQueue::value_type(NewValue, SVoxelInfo2(x, y, z - 1, Item.m_Value, Item.m_Count)));
        }
    }

    // Final normalization
    double dNorm = dAvgTetrahedronSize / mds::img::getMean<double>(*m_spControlSpace);
    itEnd = m_spControlSpace->getEnd();
    it = m_spControlSpace->getBegin();
    for( ; it != itEnd; ++it )
    {
//        *it = tVoxel(Norm * double(*it));
        *it = tVoxel(dNorm * double(*it));
    }

    // O.K.
    return true;
}

// New medial axis version without queue
// (see SVN history before 2009/11/30)
// prepareControlSpace()

// Lower byte of voxel value represents distance from the closest edge
// Higher byte represents distance from the closest corner
// (see SVN history before 2009/11/30)
// prepareControlSpace()

//void CTetraMesh::updateControlSpace(double dMaxDist)
// Version 1
//void CTetraMesh::updateControlSpace(vctl::MCPoint3D *pPoint, double dMaxDist)
// Version 0 - "slow"
//void CTetraMesh::updateControlSpace(vctl::MCPoint3D *pPoint, double dMaxDist)

// local edge finding
bool CTetraMesh::attractVertex(vctl::MCPoint3D *pPoint)
{
    // Round the vertex coordinates
    mds::tSize X, Y, Z;
    conv2Volume(pPoint, X, Y, Z);

    // Volume size
    tSize XSize = m_spEdgeVolume->getXSize();
    tSize YSize = m_spEdgeVolume->getYSize();
    tSize ZSize = m_spEdgeVolume->getZSize();

    // Check the coordinates
    mds::math::limit(X, 0, XSize - 1);
    mds::math::limit(Y, 0, YSize - 1);
    mds::math::limit(Z, 0, ZSize - 1);

    // Check if there is an image edge
    if( m_spEdgeVolume->get(X, Y, Z) > 0 )
    {
        // Convert coordinates
        conv2Real(X, Y, Z, pPoint);
        return false;
    }

    // Size of the neighbourhood
    mds::tSize MinI = mds::math::getMax(X - m_VertexNeighbourhood, 0);
    mds::tSize MinJ = mds::math::getMax(Y - m_VertexNeighbourhood, 0);
    mds::tSize MinK = mds::math::getMax(Z - m_VertexNeighbourhood, 0);
    mds::tSize MaxI = mds::math::getMin(X + m_VertexNeighbourhood, XSize - 1);
    mds::tSize MaxJ = mds::math::getMin(Y + m_VertexNeighbourhood, YSize - 1);
    mds::tSize MaxK = mds::math::getMin(Z + m_VertexNeighbourhood, ZSize - 1);

    // Initialization
    static const mds::tSize MAX = mds::CTypeTraits<mds::tSize>::getMax();
    mds::tSize MinDist = MAX;

    // Find the closest edge point in the neighbourhood
    mds::tSize NewX, NewY, NewZ;
    for( mds::tSize k = MinK; k < MaxK; ++k )
    {
        mds::tSize TempK = (Z - k) * (Z - k);
        for( mds::tSize j = MinJ; j < MaxJ; ++j )
        {
            mds::tSize TempJ = (Y - j) * (Y - j);
            for( mds::tSize i = MinI; i < MaxI; ++i )
            {
                mds::tSize Dist = (X - i) * (X - i) + TempJ + TempK;
                if( m_spEdgeVolume->get(i, j, k) > 0 && Dist < MinDist )
                {
                    MinDist = Dist;
                    NewX = i;
                    NewY = j;
                    NewZ = k;
                }
            }
        }
    }

    if( MinDist < MAX )
    {
        // Convert coordinates
        conv2Real(NewX, NewY, NewZ, pPoint);
        return true;
    }

    // No change of the vertex position
    return false;
}


bool CTetraMesh::subpixelAttractVertex(vctl::MCPoint3D *pPoint)
{
    static const int MAX_ITERS = 3;

    // Get the volume coordinates
    mds::img::CPoint3D Center;
    conv2Volume(pPoint, Center.x(), Center.y(), Center.z());
    tSize x = mds::math::round2Int(Center.x());
    tSize y = mds::math::round2Int(Center.y());
    tSize z = mds::math::round2Int(Center.z());

    // Check if there is an image edge
    if( m_spEdgeVolume->get(x, y, z) == 0 )
    {
        return false;
    }

    // Volume size
    tSize XSize = m_spVolume->getXSize() - m_VertexNeighbourhood;
    tSize YSize = m_spVolume->getYSize() - m_VertexNeighbourhood;
    tSize ZSize = m_spVolume->getZSize() - m_VertexNeighbourhood;

    // Normalization
//    double m_dNorm = 1.0 / m_VertexNeighbourhood;
    int iCount = 0;
    for( int c = 1; c <= m_VertexNeighbourhood; ++c )
    {
        iCount += c;
    }
    double m_dNorm = 1.0 / iCount;

    // Sobel operators
    mds::img::CVolumeSobelX<tVolume> SobelX;
    mds::img::CVolumeSobelY<tVolume> SobelY;
    mds::img::CVolumeSobelZ<tVolume> SobelZ;

    // Image derivatives
    double dGradX = SobelX.getResponse(*m_spVolume, x, y, z);
    double dGradY = SobelY.getResponse(*m_spVolume, x, y, z);
    double dGradZ = SobelZ.getResponse(*m_spVolume, x, y, z);

    // Edge normal
    mds::img::CVector3D Normal(dGradX, dGradY, dGradZ);
    Normal.normalize();

    // Find subpixel edge position
    for( int iter = 0; iter < MAX_ITERS; ++iter )
    {
        // Current value
        double dValue = m_spVolume->interpolate(Center);

        // Helper points
        mds::img::CPoint3D Left(Center);
        mds::img::CPoint3D Right(Center);

        // Estimate new position
        double dSum = 0.0;
        for( int i = 1; i <= m_VertexNeighbourhood; ++i )
        {
            // Neighbours
            Left.x() -= Normal.x();
            Left.y() -= Normal.y();
            Left.z() -= Normal.z();
            Right.x() += Normal.x();
            Right.y() += Normal.y();
            Right.z() += Normal.z();

            // Image values
            double dLValue = m_spVolume->interpolate(Left) - dValue;
            double dRValue = m_spVolume->interpolate(Right) - dValue;
            double dDiff = dRValue - dLValue;
            if( mds::math::getAbs(dDiff) < 50 )
            {
                return false;
            }

            // Estimate change of edge center position
//            dSum += (dRValue + dLValue) / dDiff;
            dSum += (dRValue + dLValue) / dDiff * i;
        }

        // Normalization
        dSum *= m_dNorm;
        if( dSum > m_VertexNeighbourhood )
        {
            return false;
        }

        // Calculate new center position
        Center.x() += dSum * Normal.x();
        Center.y() += dSum * Normal.y();
        Center.z() += dSum * Normal.z();
        x = mds::math::round2Int(Center.x());
        y = mds::math::round2Int(Center.y());
        z = mds::math::round2Int(Center.z());

//        MDS_LOG_NOTE("  " << dSum);

        // Check the position
        if( Center.x() <= m_VertexNeighbourhood || Center.x() >= XSize
            || Center.y() <= m_VertexNeighbourhood || Center.y() >= YSize
            || Center.z() <= m_VertexNeighbourhood || Center.z() >= ZSize )
        {
            return false;
        }
    }

//    MDS_LOG_NOTE("New position: " << Center.x << ", " << Center.y << ", " << Center.z);

    // Convert final coordinates
    conv2Real(Center.x(), Center.y(), Center.z(), pPoint);
    return true;
}


// Control space gradient descent method
// edge + corner distance
// (see SVN history)
// bool CTetraMesh::isotropicAttractVertex(vctl::MCPoint3D *pPoint)


// using medial axis
// - simple control space
bool CTetraMesh::isotropicAttractVertex(vctl::MCPoint3D *pPoint)
{
    MDS_CHECK(pPoint, return false);

    // Sobel oparators
    mds::img::CVolumeSobelX<tVolume> SobelX;
    mds::img::CVolumeSobelY<tVolume> SobelY;
    mds::img::CVolumeSobelZ<tVolume> SobelZ;

    // Round the vertex coordinates
    mds::tSize x, y, z;
    conv2Volume(pPoint, x, y, z);

    // Volume size
    tSize XSize = m_spControlSpace->getXSize();
    tSize YSize = m_spControlSpace->getYSize();
    tSize ZSize = m_spControlSpace->getZSize();

    // Check the coordinates
    mds::math::limit(x, 0, XSize - 1);
    mds::math::limit(y, 0, YSize - 1);
    mds::math::limit(z, 0, ZSize - 1);

    // Gradient descent method
    for( int i = 0; i < m_VertexNeighbourhood; ++i )
    {
        // Edge found?
        tVoxel Value = m_spEdgeVolume->get(x, y, z);
        if( Value > 0 )
        {
            // Convert coordinates
            conv2Real(x, y, z, pPoint);
            return true;
        }

        // Control space derivatives
        double dGradX = SobelX.getResponse(*m_spControlSpace, x, y, z);
        double dGradY = SobelY.getResponse(*m_spControlSpace, x, y, z);
        double dGradZ = SobelZ.getResponse(*m_spControlSpace, x, y, z);

        // Move the vertex along the gradient
        double dMagX = mds::math::getAbs(dGradX);
        double dMagY = mds::math::getAbs(dGradY);
        double dMagZ = mds::math::getAbs(dGradZ);
        if( dMagX > dMagY )
        {
            if( dMagX > dMagZ )
            {
                x += (dGradX > 0.0) ? -1 : 1;
            }
            else
            {
                z += (dGradZ > 0.0) ? -1 : 1;
            }
        }
        else
        {
            if( dMagY > dMagZ )
            {
                y += (dGradY > 0.0) ? -1 : 1;
            }
            else
            {
                z += (dGradZ > 0.0) ? -1 : 1;
            }
        }

        // Limit the coordinates
        if( !m_spControlSpace->checkPosition(x, y, z) )
        {
            // Out of range
            return false;
        }
    }

    // No change of the vertex position
    return false;
}


bool CTetraMesh::init(double dThreshold,
                      mds::tSize CellSize,
                      mds::tSize NumOfNodes
                      )
{
    // Image size and margin
    mds::tSize XSize = m_spEdgeVolume->getXSize();
    mds::tSize YSize = m_spEdgeVolume->getYSize();
    mds::tSize ZSize = m_spEdgeVolume->getZSize();
    mds::tSize Margin = m_spEdgeVolume->getMargin();

    // Check the image size
    MDS_CHECK(XSize > 0 && YSize > 0 && ZSize > 0 && Margin > 0, return false);

    // Init the triangulation
    vctl::MCPoint3D UpperLeft(0.0, 0.0, 0.0);
    vctl::MCPoint3D BottomRight(XSize - 1, YSize - 1, ZSize - 1);
    conv2Real(&UpperLeft);
    conv2Real(&BottomRight);
    if( !CDelaunayTetra::init(&UpperLeft, &BottomRight, NumOfNodes) )
    {
        // Initialization has failed
        return false;
    }

    // Use vertex detector to prepare list of candidate vertices
    CTetraVertexDetector<CELL_MAX> Detector(dThreshold, CellSize, DEFAULT_CELL_MARGIN);
    if( !Detector(m_spEdgeVolume, m_pVertices, this) )
    {
        // Can't detect vertices
        return false;
    }

    // O.K.
    return true;
}


void CTetraMesh::visualize(mds::img::CDVolume *pVolume)
{
    MDS_ASSERT(pVolume);

    // Make all edges of the mesh
    m_Tetrahedra.MakeAllTetrasEdges();

    // Get pointer to the container
    vctl::MCEdgeS *pEdges = m_Tetrahedra.GetEdgeS();

    // Use the white color
    const tVoxel White = pVolume->color2Voxel(mds::img::C_WHITE);

    // Visualize all edges
    vctl::MCEdge *pEdge = (vctl::MCEdge *)pEdges->GetFirst();
    while( pEdge )
    {
        // Get edge vertices
/*        vctl::MCVertex *pVertices[2];
        pEdge->GetVerticeS(pVertices);

        // Scale the edges
        int x[2], y[2], z[2];
        conv2Volume(pVertices[0], x[0], y[0], z[0]);
        conv2Volume(pVertices[1], x[1], y[1], z[1]);

        // Create edge
        mds::img::CLine Edge(float(x[0]), float(y[0]), float(z[0]),
                             float(x[1]), float(y[1]), float(z[1]),
                             mds::img::C_WHITE
                             );

        // Draw the edge
        Edge.draw(*pVolume);*/

        // Iterate throuhg all edge voxels
        CTetraEdgeIterator It(pVolume, pEdge, this);
        while( !It.isEnd() )
        {
            // Set the voxel value
            *It = White;

            // Move to the next voxel
            ++It;
        }

        // Get next edge
        pEdge = pEdge->GetNext();
    }

    // Clear the list of edges
    pEdges->ClearAll();
}


bool CTetraMesh::evaluateFeatures(bool bSkipSmall)
{
    // Volume size and margin
    mds::tSize XSize = m_spVolume->getXSize();
    mds::tSize YSize = m_spVolume->getYSize();
    mds::tSize ZSize = m_spVolume->getZSize();

    // Check the volume size
    MDS_CHECK(XSize > 0 && YSize > 0 && ZSize > 0, return false);

    // 2009/12/14
    if( !markBoundaryVertices() )
    {
        MDS_LOG_NOTE("CTetraMesh::evaluateFeatures(): Cannot evaluate boundary vertices. The edge volume must be available!");
    }

    // Image feature extraction
    CTetraFeaturesExtractor FeatureExtraction;

    // Re-compute features of all tetrahedra
    vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
    while( pTetrahedron )
    {
        // Get pointer to the tetrahedron features
        CTetraFeatures *pFeatures = getTetrahedronFeatures(pTetrahedron);

//        MDS_LOG_NOTE("pTetra: " << pTetrahedron << ", pFeatures: " << pFeatures);

        // Clear extracted features (evaluate every tetrahedron separately)
        FeatureExtraction.clear();

        // Acumulate features from just a one tetrahedron
        FeatureExtraction.acumulate(m_spVolume, pTetrahedron, this);

        // Compute the feature vector
        if( !FeatureExtraction.evaluate(pFeatures, bSkipSmall) )
        {
            // Logging
            MDS_LOG_NOTE("CTetraMesh::evaluateFeatures(): Cannot evaluate tetrahedron features!");
        }

        // Get next one
        pTetrahedron = pTetrahedron->GetNext();
    }

    // O.K.
    return true;
}


bool CTetraMesh::evaluateSegmentFeatures(bool bSkipSmall)
{
    // Volume size and margin
    mds::tSize XSize = m_spVolume->getXSize();
    mds::tSize YSize = m_spVolume->getYSize();
    mds::tSize ZSize = m_spVolume->getZSize();

    // Check the image size
    MDS_CHECK(XSize > 0 && YSize > 0 && ZSize > 0, return false);

    // Check the number of segments
    MDS_CHECK(m_iNumOfSegments > 0, return false);

    // Image feature extractors
    typedef std::vector<CTetraFeaturesExtractor> tExtractors;
    tExtractors Extractors(m_iNumOfSegments);

    // Evaluate image features of all regions
    vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
    while( pTetrahedron )
    {
        // Get tetrahedron region
        int iRegion = getTetrahedronRegion(pTetrahedron);

        // If the tetrahedron is not yet classified
        if( iRegion >= 0 && iRegion < m_iNumOfSegments )
        {
            // Acumulate image features
            Extractors[iRegion].acumulate(m_spVolume, pTetrahedron, this);
        }

        // Get the next tetrahedron
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Resize the vector of image features
    m_Segments.resize(m_iNumOfSegments);

    // Evaluate image features of all regions
    for( int i = 0; i < m_iNumOfSegments; ++i )
    {
        Extractors[i].evaluate(&(m_Segments[i]), bSkipSmall);
    }

    // O.K.
    return true;
}


bool CTetraMesh::optimizeByVertexMoving(double dMinQuality)
{
    // Remove all currently unused vertices
    m_Tetrahedra.EraseVerticesNoTetras();

    // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
    MDS_LOG_NOTE("CTetraMesh::optimizeByVertexMoving(): started");
    MDS_LOG_NOTE("  Initial number of vertices = " << getNumOfVertices());
#endif // MDS_SEG_LOGGING_ENABLED

    // Set quality of all tetrahedra
    vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
    while( pTetrahedron )
    {
        // Compute and set the tetrahedron quality
        pTetrahedron->SetValue(getTetrahedronQuality(pTetrahedron));

        // Clear flags
        pTetrahedron->MaskFlag(vctl::ENTITY_MARK_FLAG);

        // Get next tetra
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Sort all tetrahedra
    m_Tetrahedra.SortAll();

    // Get the sorted queue
    vctl::MCTetraS::MT_QUEUE *pQueue = m_Tetrahedra.GetQueue();

    // Vector of moved vertices
    tVertices MovedVertices;

    // Vector of the new positions
    tPoints NewPositions;

    // Optimize the mesh
    int iCount = 0;
    for( ;; )
    {
        // Get pointer to the last item
        pTetrahedron = (vctl::MCTetra *)pQueue->GetMax();

#ifdef MDS_SEG_LOGGING_ENABLED
        MDS_LOG_NOTE("  Tetrahedron quality = " << pTetrahedron->GetValue());
#endif // MDS_SEG_LOGGING_ENABLED

        // Check the tetrahedron quality
        if( !pTetrahedron || pTetrahedron->GetValue() < dMinQuality )
        {
            // No optimization is needed
            break;
        }

        // Check if the tetrahedron is marked
        if( !pTetrahedron->TestFlag(vctl::ENTITY_MARK_FLAG) )
        {
            // Get the shortest tetrahedron edge
            vctl::MCVertex *pVertices[2];
            pTetrahedron->GetMinEdge(&pVertices[0], &pVertices[1]);

            // Get index of the vertex in current tetrahedron
            int iIndex = pTetrahedron->IsVertex(pVertices[0]);

            // Get opposite triangle of the vertex
            vctl::MCVertex *pV0, *pV1, *pV2;
            pTetrahedron->GetRestTri(iIndex, &pV0, &pV1, &pV2);

            // Compute the center
            vctl::MCPoint3D Center[2];
            Center[0].SetX(0.33 * (pV0->GetX() + pV1->GetX() + pV2->GetX()));
            Center[0].SetY(0.33 * (pV0->GetY() + pV1->GetY() + pV2->GetY()));
            Center[0].SetZ(0.33 * (pV0->GetZ() + pV1->GetZ() + pV2->GetZ()));

            // Get index of the vertex in current tetrahedron
            iIndex = pTetrahedron->IsVertex(pVertices[1]);

            // Get opposite triangle of the vertex
            pTetrahedron->GetRestTri(iIndex, &pV0, &pV1, &pV2);

            // Compute the center
            Center[1].SetX(0.33 * (pV0->GetX() + pV1->GetX() + pV2->GetX()));
            Center[1].SetY(0.33 * (pV0->GetY() + pV1->GetY() + pV2->GetY()));
            Center[1].SetZ(0.33 * (pV0->GetZ() + pV1->GetZ() + pV2->GetZ()));

            // Evaluate distances
            double dDist0 = pVertices[0]->SquareDistance(Center[0]);
            double dDist1 = pVertices[1]->SquareDistance(Center[1]);

            // Add vertex to the list of moved vertices
            iIndex = (dDist0 < dDist1) ? 0 : 1;
            if( (pVertices[iIndex]->GetFlag() & BOUNDARY_FLAGS) == 0 )
            {
                MovedVertices.push_back(pVertices[iIndex]);
                NewPositions.push_back(Center[iIndex]);

                // Mark all tetrahedra around the moved vertex
                vctl::MCTetra *pHelperTetra = pVertices[iIndex]->GetRegisteredTetra();
                while( pHelperTetra )
                {
                    // Mark the tetrahedron
                    pHelperTetra->SetFlag(vctl::ENTITY_MARK_FLAG);

                    // Move to the next tetrahedron around the vertex
                    pHelperTetra = pHelperTetra->GetVertexTetra(pVertices[iIndex]);
                }

                ++iCount;
            }
        }

        // Unsort the tetrahedron
        pQueue->UnSort(pTetrahedron->GetValue(), pTetrahedron);
        pTetrahedron->SetValue(0.0);
    }

    // Clear/unsort the queue
    pQueue->Clear();

#ifdef MDS_SEG_LOGGING_ENABLED
    MDS_LOG_NOTE("  Number of vertices to move = " << iCount);
    MDS_LOG_NOTE("  Re-constructing the mesh ...");
#endif // MDS_SEG_LOGGING_ENABLED

    // Delete all info structures
    pTetrahedron = getFirstTetrahedron();
    while( pTetrahedron )
    {
        // Delete tetrahedron info struct
        delete (STetrahedronInfo *)pTetrahedron->GetValuePtr();
        pTetrahedron->SetValuePtr(NULL);

        // Get next one
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Destroy all tetras in the mesh
    m_Tetrahedra.DeRegistrationAll();
    m_Tetrahedra.ClearAll();

    // Erase all moved vertices
    tVertices::iterator itEnd = MovedVertices.end();
    for( tVertices::iterator it = MovedVertices.begin(); it != itEnd; ++it )
    {
        m_pVertices->Erase(*it);
    }

    // Create vertices at new positions
    tPoints::iterator it2End = NewPositions.end();
    for( tPoints::iterator it2 = NewPositions.begin(); it2 != it2End; ++it2 )
    {
        // Attract the vertex to its best position
//        attractVertex(&(*it2));

        // Create the vertex
        m_pVertices->New(*it2);
    }

    // Create corner vertices
    vctl::MCVertex *pVertex000 = m_pVertices->New(m_dMinX, m_dMinY, m_dMinZ);
    vctl::MCVertex *pVertex100 = m_pVertices->New(m_dMaxX, m_dMinY, m_dMinZ);
    vctl::MCVertex *pVertex010 = m_pVertices->New(m_dMinX, m_dMaxY, m_dMinZ);
    vctl::MCVertex *pVertex001 = m_pVertices->New(m_dMinX, m_dMinY, m_dMaxZ);
    vctl::MCVertex *pVertex110 = m_pVertices->New(m_dMaxX, m_dMaxY, m_dMinZ);
    vctl::MCVertex *pVertex011 = m_pVertices->New(m_dMinX, m_dMaxY, m_dMaxZ);
    vctl::MCVertex *pVertex101 = m_pVertices->New(m_dMaxX, m_dMinY, m_dMaxZ);
    vctl::MCVertex *pVertex111 = m_pVertices->New(m_dMaxX, m_dMaxY, m_dMaxZ);

    // Insert initial tetrahedra
    addTetrahedron(pVertex000, pVertex010, pVertex110, pVertex011);
    addTetrahedron(pVertex000, pVertex110, pVertex100, pVertex101);
    addTetrahedron(pVertex110, pVertex101, pVertex011, pVertex111);
    addTetrahedron(pVertex000, pVertex101, pVertex001, pVertex011);
    addTetrahedron(pVertex000, pVertex110, pVertex011, pVertex101);

    setDefaultMinEdgeLengthFunc();

    // Construct the triangulation
    if( !construct() )
    {
        MDS_LOG_NOTE("CTetraMesh::optimizeByVertexMoving(): Cannot construct the triangulation!");
        return false;
    }

#ifdef MDS_SEG_LOGGING_ENABLED
    MDS_LOG_NOTE("  Final number of vertices = " << getNumOfVertices());
    MDS_LOG_NOTE("CTetraMesh::optimizeByVertexMoving(): finished successfuly");
#endif // MDS_SEG_LOGGING_ENABLED

    // O.K.
    return true;
}


bool CTetraMesh::edgeSplitting(double dThreshold)
{
    static const int MAX_ITERS = 5;

    MDS_CHECK(dThreshold > 0.0, return false);

    MDS_LOG_NOTE("CTetraMesh::edgeSplitting(): begin");

    // Set default edge length checking function
    setDefaultMinEdgeLengthFunc();

    // Construct container of all edges in the mesh
    m_Tetrahedra.MakeAllTetrasEdges();

    // Get pointer to the container of edges
    vctl::MCEdgeS *pEdges = m_Tetrahedra.GetEdgeS();

    // Helper vector of newly created vertices
    tPoints Points;

    // Iterate
    for( int iter = 0; pEdges->GetNumber() > 0 && iter < MAX_ITERS; ++iter )
    {
        // Logging the number of edges
        MDS_LOG_TIME("Iteration begin");
        MDS_LOG_NOTE("  Iteration = " << iter);
        MDS_LOG_NOTE("  Number of edges = " << pEdges->GetNumber());

        // Clear the helper vector of vertices
        Points.clear();

        // Evaluate and split all edges
        // - Construct vector of new vertices
        // - Remove edges which were not divided
        vctl::MCEdge *pEdge = (vctl::MCEdge *)pEdges->GetFirst();
        while( pEdge )
        {
            // Split the edge
            splitEdge(pEdge, dThreshold, Points);

            // Get next edge
//            vctl::MCEdge *pPrevious = pEdge;
            pEdge = pEdge->GetNext();

            // Remove already processed edge from container
//            pEdges->Erase(pPrevious);
        }

        // Number of new points
        MDS_LOG_TIME("After splitting");
        MDS_LOG_NOTE("  Number of newly created points = " << Points.size());

        // Insert all points to the mesh
        tPoints::iterator it = Points.begin();
        tPoints::iterator itEnd = Points.end();
        int iCount = 0;
        while( it != itEnd )
        {
            // Get pointer to the point
            vctl::MCPoint3D *pPoint = &(*it);

            // Check if the point lies on the mesh boundary
            int iBoundaryFlags = checkBoundary(pPoint);

            // Create a new vertex
            vctl::MCVertex *pVertex = newVertex(pPoint);

            // Insert it to the mesh
            if( !addVertex(pVertex,
                           iBoundaryFlags,
                           edgeSplittingInsertFunc,
                           edgeSplittingRemoveFunc
                           ) )
            {
                // Remove the vertex
                if( !pVertex->GetRegisteredTetra() )
                {
                    m_pVertices->Erase(pVertex);
                }
            }
            else
            {
                ++iCount;
            }

            // Get next point
            ++it;
        }

        MDS_LOG_TIME("After adding vertices");
        MDS_LOG_NOTE("  Number of successfuly added points = " << iCount);

        // Check the list of edges and remove edges which are not used
//        m_Tetrahedra.EraseEdgesNoTetras();
    }

    // Clear the list of edges
    pEdges->DeRegistrationAll();
    pEdges->ClearAll();

    MDS_LOG_NOTE("CTetraMesh::edgeSplitting(): end");

    // O.K.
    return true;
}


// Last functional version
// (see SVN history before 2009/11/30)
// bool CTetraMesh::edgeSplitting(double dThreshold)
// bool CTetraMesh::splitEdge(vctl::MCEdge *pEdge, double dOrientThreshold)


bool CTetraMesh::splitEdge(vctl::MCEdge *pEdge, double dOrientThreshold, tPoints& Points)
{
    MDS_ASSERT(pEdge);

    // Sobel oparators
    mds::img::CVolumeSobelX<tVolume> SobelX;
    mds::img::CVolumeSobelY<tVolume> SobelY;
    mds::img::CVolumeSobelZ<tVolume> SobelZ;

    // Vertices of the edge
    vctl::MCVertex *pVertices[2];
    pEdge->GetVerticeS(pVertices);

    // Create edge pixels iterator
    CTetraEdgeIterator It(m_spEdgeVolume, pEdge, this);
    if( It.isEnd() )
    {
        return false;
    }

    // Pointer to the starting point
    vctl::MCPoint3D *pStartPoint = pVertices[0];

    // Compute square of the minimal edge length
    double dThreshold = m_dMinEdgeLength * m_dMinEdgeLength;

    // Tetrahedron edge
    vctl::MCVector3D TetraEdge(pVertices[0], pVertices[1]);
    TetraEdge.Normalization();

    // Value at the beginning of the edge
    mds::img::CDVolume::tVoxel LastValue = *It;
    ++It;

    // Traverse all pixels and find pixel lying on an image edge having maximal magnitude
    bool bResult = false;
    for( ; !It.isEnd(); ++It )
    {
        // Get current pixel value
        mds::img::CDVolume::tVoxel Value = *It;

        // Is there an image edge?
//        if( mds::math::getAbs(LastValue - Value) > 0 )
        if( mds::math::getAbs(LastValue - Value) > 50 )
//        if( Value > 0 )
        {
            // Splitting point
            vctl::MCPoint3D Point(It.getRealX(), It.getRealY(), It.getRealZ());

            // Check size of produced edges
            double dLength = pStartPoint->SquareDistance(Point);
            if( dLength < dThreshold )
            {
                continue;
            }
            dLength = pVertices[1]->SquareDistance(Point);
            if( dLength < dThreshold )
            {
                continue;
            }

            // Volume derivatives
            double dGradX = SobelX.getResponse(*m_spVolume, It.getX(), It.getY(), It.getZ());
            double dGradY = SobelY.getResponse(*m_spVolume, It.getX(), It.getY(), It.getZ());
            double dGradZ = SobelZ.getResponse(*m_spVolume, It.getX(), It.getY(), It.getZ());

            // Volume edge gradient
            vctl::MCVector3D VolumeEdgeNormal(dGradX, dGradY, dGradZ);

            // Convert it to the real coordinates :)
            conv2Volume(&VolumeEdgeNormal);
            VolumeEdgeNormal.Normalization();

            // Compare orientation of the volume edge gradient and tetrahedron edge
            // - Dot product of both vectors
            // - Value must be greater than a given threshold. Optimal case is
            //   that both vectors are parallel (volume edge and tetrahedron edge are
            //   ortogonal), then -1 or 1 is obtained.
            if( mds::math::getAbs(VolumeEdgeNormal * TetraEdge) > dOrientThreshold )
            {
                Points.push_back(Point);
                pStartPoint = &Points.back();
//                LastValue = Value;
//                bResult = true;
                return true;
            }
        }

        LastValue = Value;
    }

    // No splitting
    return bResult;
}


void CTetraMesh::edgeSplittingInsertFunc(vctl::MCTetra *pTetrahedron, CDelaunayTetra *pMesh)
{
    MDS_ASSERT(pTetrahedron && pMesh);

    // Get pointer to the triangular mesh
    CTetraMesh *pTetraMesh = (CTetraMesh *)pMesh;

    // Insert edges of the triangle to the list
    pTetraMesh->m_Tetrahedra.MakeTetraEdges(pTetrahedron);
}


void CTetraMesh::edgeSplittingRemoveFunc(vctl::MCTetra *pTetrahedron, CDelaunayTetra *pMesh)
{
    MDS_ASSERT(pTetrahedron && pMesh);

    // Get pointer to the triangular mesh
    CTetraMesh *pTetraMesh = (CTetraMesh *)pMesh;

    // Get tetrahedron vertices
    vctl::MCVertex *pVertices[4];
    pTetrahedron->GetVerticeS(pVertices);

    // Get container of edges
    vctl::MCEdgeS *pEdges = pTetraMesh->m_Tetrahedra.GetEdgeS();

    // Unsort all edges of the tetrahedron
    vctl::MCEdge *pEdge = pEdges->TestExistence(pVertices[0], pVertices[1]);
    if( pEdge )
    {
        pEdges->Erase(pEdge);
    }

    pEdge = pEdges->TestExistence(pVertices[1], pVertices[2]);
    if( pEdge )
    {
        pEdges->Erase(pEdge);
    }

    pEdge = pEdges->TestExistence(pVertices[2], pVertices[0]);
    if( pEdge )
    {
        pEdges->Erase(pEdge);
    }

    pEdge = pEdges->TestExistence(pVertices[0], pVertices[3]);
    if( pEdge )
    {
        pEdges->Erase(pEdge);
    }

    pEdge = pEdges->TestExistence(pVertices[1], pVertices[3]);
    if( pEdge )
    {
        pEdges->Erase(pEdge);
    }

    pEdge = pEdges->TestExistence(pVertices[2], pVertices[3]);
    if( pEdge )
    {
        pEdges->Erase(pEdge);
    }
}


// Latest functional version
// (see SVN history before 2009/11/30)
//void CTetraMesh::edgeSplittingInsertFunc(vctl::MCTetra *pTetrahedron, CDelaunayTetra *pMesh)
//void CTetraMesh::edgeSplittingRemoveFunc(vctl::MCTetra *pTetrahedron, CDelaunayTetra *pMesh)
//bool CTetraMesh::edgeSplitting(double dThreshold)
//bool CTetraMesh::splitEdge(vctl::MCEdge *pEdge, double dOrientThreshold, tPoints& Points)


bool CTetraMesh::isotropicVertexMoving()
{
    // Remove all currently unused vertices
    m_Tetrahedra.EraseVerticesNoTetras();

    // Logging
    MDS_LOG_NOTE("CTetraMesh::isotropicVertexMoving(): started");
    MDS_LOG_NOTE("  Initial number of vertices = " << getNumOfVertices());

    // Vector of moved vertices
    tVertices MovedVertices;

    // Vector of the new positions
    tPoints NewPositions;

    // Number of moved vertices
    int iCount = 0;

    // Attract all vertices in the mesh
    vctl::MCVertex *pVertex = getFirstVertex();
    while( pVertex )
    {
        // Skip all boundary vertices
        if( (pVertex->GetFlag() & BOUNDARY_FLAGS) == 0 )
        {
            // Create copy of the vertex coordinates
            vctl::MCPoint3D Point(pVertex);

            // Try to attract the vertex
//            if( isotropicAttractVertex(&Point) )
            if( attractVertex(&Point) )
            {
                MovedVertices.push_back(pVertex);
                NewPositions.push_back(Point);
                ++iCount;
            }
        }

        // Get next vertex
        pVertex = pVertex->GetNext();
    }

    MDS_LOG_NOTE("  Number of vertices to move = " << iCount);
    MDS_LOG_NOTE("  Re-constructing the mesh ...");

    // Delete all info structures
    vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
    while( pTetrahedron )
    {
        // Delete tetrahedron info struct
        delete (STetrahedronInfo *)pTetrahedron->GetValuePtr();
        pTetrahedron->SetValuePtr(NULL);

        // Get next one
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Destroy all tetras in the mesh
    m_Tetrahedra.DeRegistrationAll();
    m_Tetrahedra.ClearAll();

    // Erase all moved vertices
    tVertices::iterator itEnd = MovedVertices.end();
    for( tVertices::iterator it = MovedVertices.begin(); it != itEnd; ++it )
    {
        m_pVertices->Erase(*it);
    }

    // Create vertices at the new positions
    tPoints::iterator it2End = NewPositions.end();
    for( tPoints::iterator it2 = NewPositions.begin(); it2 != it2End; ++it2 )
    {
        pVertex = m_pVertices->New(*it2);
    }

    // Create corner vertices
    vctl::MCVertex *pVertex000 = m_pVertices->New(m_dMinX, m_dMinY, m_dMinZ);
    vctl::MCVertex *pVertex100 = m_pVertices->New(m_dMaxX, m_dMinY, m_dMinZ);
    vctl::MCVertex *pVertex010 = m_pVertices->New(m_dMinX, m_dMaxY, m_dMinZ);
    vctl::MCVertex *pVertex001 = m_pVertices->New(m_dMinX, m_dMinY, m_dMaxZ);
    vctl::MCVertex *pVertex110 = m_pVertices->New(m_dMaxX, m_dMaxY, m_dMinZ);
    vctl::MCVertex *pVertex011 = m_pVertices->New(m_dMinX, m_dMaxY, m_dMaxZ);
    vctl::MCVertex *pVertex101 = m_pVertices->New(m_dMaxX, m_dMinY, m_dMaxZ);
    vctl::MCVertex *pVertex111 = m_pVertices->New(m_dMaxX, m_dMaxY, m_dMaxZ);

    // Insert initial tetrahedra
    addTetrahedron(pVertex000, pVertex010, pVertex110, pVertex011);
    addTetrahedron(pVertex000, pVertex110, pVertex100, pVertex101);
    addTetrahedron(pVertex110, pVertex101, pVertex011, pVertex111);
    addTetrahedron(pVertex000, pVertex101, pVertex001, pVertex011);
    addTetrahedron(pVertex000, pVertex110, pVertex011, pVertex101);

    // Set edge length checking function
    setMinEdgeLengthFunc(checkIsotropicMinEdgeLengthFunc);
//    setDefaultMinEdgeLengthFunc();

    // Construct the triangulation
    if( !construct() )
    {
        MDS_LOG_NOTE("CTetraMesh::isotropicVertexMoving(): Cannot construct the triangulation!");
        return false;
    }

    // Set default edge length checking function
    setDefaultMinEdgeLengthFunc();

    MDS_LOG_NOTE("  Final number of vertices = " << getNumOfVertices());
    MDS_LOG_NOTE("CTetraMesh::isotropicVertexMoving(): finished successfuly");

    // O.K.
    return true;
}


bool CTetraMesh::subpixelVertexMoving()
{
    // Remove all currently unused vertices
    m_Tetrahedra.EraseVerticesNoTetras();

    // Logging
    MDS_LOG_NOTE("CTetraMesh::subpixelVertexMoving(): started");
    MDS_LOG_NOTE("  Initial number of vertices = " << getNumOfVertices());

    // Vector of moved vertices
    tVertices MovedVertices;

    // Vector of the new positions
    tPoints NewPositions;

    // Number of moved vertices
    int iCount = 0;

    // Attract all vertices in the mesh
    vctl::MCVertex *pVertex = getFirstVertex();
    while( pVertex )
    {
        // Skip all boundary vertices
        if( (pVertex->GetFlag() & BOUNDARY_FLAGS) == 0 )
        {
            // Create copy of the vertex coordinates
            vctl::MCPoint3D Point(pVertex);

            // Try to attract the vertex
            if( subpixelAttractVertex(&Point) )
            {
                MovedVertices.push_back(pVertex);
                NewPositions.push_back(Point);
                ++iCount;
            }
        }

        // Get next vertex
        pVertex = pVertex->GetNext();
    }

    MDS_LOG_NOTE("  Number of vertices to move = " << iCount);
    MDS_LOG_NOTE("  Re-constructing the mesh ...");

    // Delete all info structures
    vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
    while( pTetrahedron )
    {
        // Delete tetrahedron info struct
        delete (STetrahedronInfo *)pTetrahedron->GetValuePtr();
        pTetrahedron->SetValuePtr(NULL);

        // Get next one
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Destroy all tetras in the mesh
    m_Tetrahedra.DeRegistrationAll();
    m_Tetrahedra.ClearAll();

    // Erase all moved vertices
    tVertices::iterator itEnd = MovedVertices.end();
    for( tVertices::iterator it = MovedVertices.begin(); it != itEnd; ++it )
    {
        m_pVertices->Erase(*it);
    }

    // Create vertices at the new positions
    tPoints::iterator it2End = NewPositions.end();
    for( tPoints::iterator it2 = NewPositions.begin(); it2 != it2End; ++it2 )
    {
        pVertex = m_pVertices->New(*it2);
    }

    // Create corner vertices
    vctl::MCVertex *pVertex000 = m_pVertices->New(m_dMinX, m_dMinY, m_dMinZ);
    vctl::MCVertex *pVertex100 = m_pVertices->New(m_dMaxX, m_dMinY, m_dMinZ);
    vctl::MCVertex *pVertex010 = m_pVertices->New(m_dMinX, m_dMaxY, m_dMinZ);
    vctl::MCVertex *pVertex001 = m_pVertices->New(m_dMinX, m_dMinY, m_dMaxZ);
    vctl::MCVertex *pVertex110 = m_pVertices->New(m_dMaxX, m_dMaxY, m_dMinZ);
    vctl::MCVertex *pVertex011 = m_pVertices->New(m_dMinX, m_dMaxY, m_dMaxZ);
    vctl::MCVertex *pVertex101 = m_pVertices->New(m_dMaxX, m_dMinY, m_dMaxZ);
    vctl::MCVertex *pVertex111 = m_pVertices->New(m_dMaxX, m_dMaxY, m_dMaxZ);

    // Insert initial tetrahedra
    addTetrahedron(pVertex000, pVertex010, pVertex110, pVertex011);
    addTetrahedron(pVertex000, pVertex110, pVertex100, pVertex101);
    addTetrahedron(pVertex110, pVertex101, pVertex011, pVertex111);
    addTetrahedron(pVertex000, pVertex101, pVertex001, pVertex011);
    addTetrahedron(pVertex000, pVertex110, pVertex011, pVertex101);

    // Set edge length checking function
//    setMinEdgeLengthFunc(checkIsotropicMinEdgeLengthFunc);
    setDefaultMinEdgeLengthFunc();

    // Construct the triangulation
    if( !construct() )
    {
        MDS_LOG_NOTE("CTetraMesh::subpixelVertexMoving(): Cannot construct the triangulation!");
        return false;
    }

    // Set default edge length checking function
//    setDefaultMinEdgeLengthFunc();

    MDS_LOG_NOTE("  Final number of vertices = " << getNumOfVertices());
    MDS_LOG_NOTE("CTetraMesh::subpixelVertexMoving(): finished successfuly");

    // O.K.
    return true;
}


//! Helper vertex position accumulator structure.
struct SAccumulator2
{
    //! Optimal position.
    double m_x, m_y, m_z;

    //! Minimal distance to an image edge.
    double m_dMinLength;

    //! Default constructor.
    SAccumulator2() : m_x(0.0), m_y(0.0), m_z(0.0), m_dMinLength(1000000.0) {}
};


bool CTetraMesh::approachEdges(bool bCheckIsotropicMinEdgeLength)
{
    static const int MAX_ITERS = 3;

    // Remove all currently unused vertices
    m_Tetrahedra.EraseVerticesNoTetras();

    // Turn off the "vertex existence test"
//    m_pVertices->SetTestExistence(false);

    // Logging
    MDS_LOG_NOTE("CTetraMesh::approachEdges(): started");

    // Vector of the new positions
    tPoints NewPositions;
    NewPositions.reserve(100000);

    // Allocate and clear accumulators
    std::vector<SAccumulator2> Accumulators;
    Accumulators.reserve(100000);

    // Iterate
    for( int iter = 0; iter < MAX_ITERS; ++iter )
    {
        MDS_LOG_NOTE("  Iteration: " << iter);
        MDS_LOG_NOTE("    Initial number of vertices = " << getNumOfVertices());

        // Clear the vertices
        NewPositions.clear();

        // Phase 1) find boundary vertices

        // Clear accumulators
        Accumulators.clear();
        Accumulators.push_back(SAccumulator2());

        // Assign accumulators to vertices
        vctl::MCVertex *pVertex = getFirstVertex();
        for( ; pVertex; pVertex = pVertex->GetNext() )
        {
            pVertex->SetIndex(0);

            // Clear vertex flag
            pVertex->MaskFlag(MARKED);
        }

        // Traverse all edge points and accumulate coordinates in closest vertices
        bool bInitialized = false;
        tVolume::tIterator itEnd = m_spEdgeVolume->getEnd();
        for( tVolume::tIterator it = m_spEdgeVolume->getBegin(); it != itEnd; ++it )
        {
            // Current value
            tVoxel EdgeValue = *it;

            // Lies the vertex on an image edge?
            if( EdgeValue > 0 )
            {
                // Current point
                vctl::MCPoint3D Point(it.getX(), it.getY(), it.getZ());

                // Find the closest vertex
                vctl::MCPoint3D Vertex;
                conv2Real(Point.GetX(), Point.GetY(), Point.GetZ(), &Vertex);
                vctl::MCVertex *pVertex = findNearestVertex(&Vertex, bInitialized);
                if( !pVertex )
                {
                    continue;
                }
                else
                {
                    bInitialized = true;
                }

                // Accumulator index
                unsigned int uIndex = pVertex->GetIndex();
                if( uIndex == 0 )
                {
                    Accumulators.push_back(SAccumulator2());
                    uIndex = Accumulators.size() - 1;
                    pVertex->SetIndex(uIndex);
                }

                // Current accumulator
                SAccumulator2& Accumulator = Accumulators[uIndex];

                // Square distance to the closest vertex
                double dDistance = pVertex->SquareDistance(Vertex);

                // Accumulate coordinates
                if( dDistance < Accumulator.m_dMinLength )
                {
                    Accumulator.m_x = Vertex.GetX();
                    Accumulator.m_y = Vertex.GetY();
                    Accumulator.m_z = Vertex.GetZ();
                    Accumulator.m_dMinLength = dDistance;
                }
            }
        }

        MDS_LOG_NOTE("    Num of accumulators = " << Accumulators.size());

        // Check all affected vertices
        int iCount = 0;
        for( pVertex = getFirstVertex(); pVertex; pVertex = pVertex->GetNext() )
        {
            // Accumulator index
            unsigned int uIndex = pVertex->GetIndex();
            if( uIndex == 0 )
            {
                continue;
            }

            // Current accumulator
            SAccumulator2& Accumulator = Accumulators[uIndex];

            // Check if the vertex actually lies on an image edge?
            tSize x, y, z;
            conv2Volume(pVertex, x, y, z);
            if( m_spEdgeVolume->get(x, y, z) > 0 )
            {
                continue;
            }

            // Calculate the new position
            vctl::MCPoint3D Point(Accumulator.m_x, Accumulator.m_y, Accumulator.m_z);

            // Limit the position
            checkBoundary(&Point);

            // Store the new position
            NewPositions.push_back(Point);
            ++iCount;
        }

        // Logging
        MDS_LOG_NOTE("    Number of newly created boundary vertices = " << iCount);

        // Set edge length checking function
        if( bCheckIsotropicMinEdgeLength )
        {
//            setMinEdgeLengthFunc(checkIsotropicMinEdgeLengthFunc);
            setMinEdgeLengthFunc(checkIsotropicMinEdgeLengthFunc2);
        }
        else
        {
            setDefaultMinEdgeLengthFunc();
        }

        // Insert all edge points to the mesh
        tPoints::iterator it2 = NewPositions.begin();
        tPoints::iterator it2End = NewPositions.end();
        iCount = 0;
        for( ; it2 != it2End; ++it2 )
        {
            // Get pointer to the point
            vctl::MCPoint3D *pPoint = &(*it2);

            // Check if the point lies on the mesh boundary
            int iBoundaryFlags = checkBoundary(pPoint);

            // Create a new vertex
            vctl::MCVertex *pVertex = newVertex(pPoint);

            // Insert it to the mesh
            if( !addVertex(pVertex, iBoundaryFlags) )
            {
                // Remove the vertex
                if( !pVertex->GetRegisteredTetra() )
                {
                    m_pVertices->Erase(pVertex);
                }
            }
            else
            {
                // Update the control space
//                updateControlSpace(pVertex, m_dMinEdgeLength);
                ++iCount;
            }
        }

        MDS_LOG_NOTE("  Number of successfuly added points = " << iCount);
    }

    MDS_LOG_NOTE("CTetraMesh::approachEdges(): finished successfuly");

    // Turn on the "vertex existence test"
//    m_pVertices->SetTestExistence(true);

    // Set default edge length checking function
    setDefaultMinEdgeLengthFunc();

    // O.K.
    return true;
}


// TRADITTIONAL VARIATIONAL MESHING (Alliez05)

//! Helper vertex position accumulator structure used by variational meshing.
struct SAccumulator
{
    //! Optimal vertex position.
    double m_x, m_y, m_z;

    //! Normalization
    double m_dNorm;

    //! Default constructor.
    SAccumulator() : m_x(0.0), m_y(0.0), m_z(0.0), m_dNorm(0.0) {}
};


bool CTetraMesh::markBoundaryVertices()
{
    if( !m_spEdgeVolume.get() )
    {
        return false;
    }

    // Clear vertex flags
    vctl::MCVertex *pVertex = getFirstVertex();
    for( ; pVertex; pVertex = pVertex->GetNext() )
    {
        pVertex->MaskFlag(DOMAIN_BOUNDARY);
    }

    // Traverse all edge points and find closest vertices
    bool bInitialized = false;
    int iCount = 0;
    tVolume::tIterator itEnd = m_spEdgeVolume->getEnd();
    for( tVolume::tIterator it = m_spEdgeVolume->getBegin(); it != itEnd; ++it )
    {
        // Current value
        tVoxel EdgeValue = *it;

        // Lies the vertex on an image edge?
        if( EdgeValue > 0 )
        {
            // The current position in volume space
            mds::img::CPoint3D Point(it.getX(), it.getY(), it.getZ());

            // Find the closest vertex
            vctl::MCPoint3D Vertex;
            conv2Real(Point.x(), Point.y(), Point.z(), &Vertex);
            vctl::MCVertex *pVertex = findNearestVertex(&Vertex, bInitialized);
            if( !pVertex )
            {
                continue;
            }
            else
            {
                // Start next finding here
                bInitialized = true;
            }

            // Mark the vertex
            pVertex->SetFlag(DOMAIN_BOUNDARY);
            ++iCount;
        }
    }

    // Logging
    MDS_LOG_NOTE("CTetraMesh::markBoundaryVertices(): found " << iCount << " boundary vertices");

    // O.K.
    return true;
}


bool CTetraMesh::isBoundaryTriangle(vctl::MCVertex *p0,
                                    vctl::MCVertex *p1,
                                    vctl::MCVertex *p2
                                    )
{
    static const double BOUNDARY_THRESHOLD = 0.2;

    // First, check if all tetrahedron's vertices form domain boundary?
    if( !p0->TestFlag(DOMAIN_BOUNDARY)
        || !p1->TestFlag(DOMAIN_BOUNDARY)
        || !p2->TestFlag(DOMAIN_BOUNDARY) )
    {
        return false;
    }

    // Test image edges
    return checkBoundaryTriangle(p0, p1, p2, BOUNDARY_THRESHOLD);

/*    return (p0->TestFlag(DOMAIN_BOUNDARY)
            && p1->TestFlag(DOMAIN_BOUNDARY)
            && p2->TestFlag(DOMAIN_BOUNDARY));*/
}


// Basic two step version "vertex optimization + boundary optimization"
bool CTetraMesh::variationalMeshing(bool bCheckIsotropicMinEdgeLength)
{
    static const int MAX_ITERS = 3;
//    static const double CORNER_WEIGHT = 1000.0;
    static const double CORNER_WEIGHT = 1.0;
    static const double EDGE_WEIGHT = 0.8;
//    static const double EDGE_WEIGHT = 1.0;

    // Remove all currently unused vertices
    m_Tetrahedra.EraseVerticesNoTetras();

    // Turn off the "vertex existence test"
    m_pVertices->SetTestExistence(false);

    // Logging
    MDS_LOG_NOTE("CTetraMesh::variationalMeshing(): started");

    // Vector of the new positions
    tPoints NewPositions, NewPositions2;
    NewPositions.reserve(100000);
    NewPositions2.reserve(100000);

    // Allocate and clear accumulators
    std::vector<SAccumulator> Accumulators;
    Accumulators.reserve(100000);

    // Iterate
    for( int iter = 0; iter < MAX_ITERS; ++iter )
    {
        MDS_LOG_NOTE("  Iteration: " << iter);
        MDS_LOG_NOTE("    Initial number of vertices = " << getNumOfVertices());

        // Clear the vertices
        NewPositions.clear();
        NewPositions2.clear();

        // Phase 1) boundary optimization

        // Clear accumulators
        Accumulators.clear();
        Accumulators.push_back(SAccumulator());

        // Assign accumulators to vertices
        vctl::MCVertex *pVertex = getFirstVertex();
        for( ; pVertex; pVertex = pVertex->GetNext() )
        {
            pVertex->SetIndex(0);

            // Clear vertex flag
            pVertex->MaskFlag(DOMAIN_BOUNDARY);
        }

        // Minimal edge length
        double dMin = getMinVolumeSize(m_dMinEdgeLength);

        // Traverse all edge points and accumulate weighted coordinates
        // in closest vertices
        bool bInitialized = false;
        tVolume::tIterator itEnd = m_spEdgeVolume->getEnd();
        for( tVolume::tIterator it = m_spEdgeVolume->getBegin(); it != itEnd; ++it )
        {
            // Current value
            tVoxel EdgeValue = *it;

            // Lies the vertex on an image edge?
            if( EdgeValue > 0 )
            {
                // Control space value
                mds::img::CPoint3D Point(it.getX(), it.getY(), it.getZ());
                double Value = getControlSpaceValue(Point);
                double dH1 = (Value < dMin) ? dMin : Value;

                // Find the closest vertex
                vctl::MCPoint3D Vertex;
                conv2Real(Point.x(), Point.y(), Point.z(), &Vertex);
                vctl::MCVertex *pVertex = findNearestVertex(&Vertex, bInitialized);
                if( !pVertex )
                {
                    continue;
                }
                else
                {
                    bInitialized = true;
                }

                // Accumulator index
                unsigned int uIndex = pVertex->GetIndex();
                if( uIndex == 0 )
                {
                    Accumulators.push_back(SAccumulator());
                    uIndex = Accumulators.size() - 1;
                    pVertex->SetIndex(uIndex);
                }

                // Current accumulator
                SAccumulator& Accumulator = Accumulators[uIndex];

                // Is it an image corner or an image edge?
                double dWeight = 1.0;
                double dControl = dH1 * dH1 * dH1;
//                if( EdgeValue == mds::img::CPixelTraits<tVoxel>::getPixelMax() )
//                {
//                    dWeight = CORNER_WEIGHT;
////                    dControl *= dH1;
//                }
//                else
                if( EdgeValue != mds::img::CPixelTraits<tVoxel>::getPixelMax() )
                {
                    dControl *= dH1;
                }

                // Accumulate coordinates
                double dTemp = dWeight / dControl;
                Accumulator.m_x += Vertex.GetX() * dTemp;
                Accumulator.m_y += Vertex.GetY() * dTemp;
                Accumulator.m_z += Vertex.GetZ() * dTemp;
                Accumulator.m_dNorm += dTemp;
            }
        }

        MDS_LOG_NOTE("    Num of accumulators = " << Accumulators.size());

        // Move all affected vertices
        int iCount = 0;
        for( pVertex = getFirstVertex(); pVertex; pVertex = pVertex->GetNext() )
        {
            // Accumulator index
            unsigned int uIndex = pVertex->GetIndex();
            if( uIndex == 0 )
            {
                continue;
            }

            // Current accumulator
            SAccumulator& Accumulator = Accumulators[uIndex];

            // Should be the vertex position modified?
            if( Accumulator.m_dNorm > 0.0 )
            {
                // Mark the vertex
                pVertex->SetFlag(DOMAIN_BOUNDARY);
                ++iCount;

                // Calculate the new position
                double dInvNorm = 1.0 / Accumulator.m_dNorm;
                vctl::MCPoint3D Point(Accumulator.m_x * dInvNorm,
                                      Accumulator.m_y * dInvNorm,
                                      Accumulator.m_z * dInvNorm
                                      );

                // Limit the position
                checkBoundary(&Point);

                // Store the new position
                NewPositions2.push_back(Point);
            }
        }

        // Logging
        MDS_LOG_NOTE("    Number of moved boundary vertices = " << iCount);

        // Phase 2) Move all non-boundary vertices to their optimal position

        // Move all vertices in the mesh
        for( pVertex = getFirstVertex(); pVertex; pVertex = pVertex->GetNext() )
        {
            // Check if vertex has been already moved
            if( pVertex->TestFlag(DOMAIN_BOUNDARY) )
            {
                continue;
            }

            // Calculate new vertex position
            double x = 0.0, y = 0.0, z = 0.0, dNorm = 0.0;

            // Traverse all tetrahedra around the current vertex
            vctl::MCTetra *pTetra = pVertex->GetRegisteredTetra();
            while( pTetra )
            {
                // Centroid of the tetrahedra
                vctl::MCPoint3D Center;
                pTetra->GetCenter(Center);

                // Get control space value
                mds::img::CPoint3D Point;
                conv2Volume(&Center, Point.x(), Point.y(), Point.z());
                double Value = getControlSpaceValue(Point);
                double dH1 = (Value < dMin) ? dMin : Value;

                // Tetrahedra volume
                double dVolume = pTetra->GetVolume();

                // Circumcenter
                vctl::MCPoint3D Circumcenter;
                getTetrahedronCircumsphere(pTetra)->GetCenter(Circumcenter);

                // Get vertices
                vctl::MCVertex *pVertices[4];
                pTetra->GetVerticeS(pVertices);
                double dEdgeWeight = 1.0;
                if( pVertices[0]->TestFlag(DOMAIN_BOUNDARY) )
                {
                    dEdgeWeight *= EDGE_WEIGHT;
                }
                if( pVertices[1]->TestFlag(DOMAIN_BOUNDARY) )
                {
                    dEdgeWeight *= EDGE_WEIGHT;
                }
                if( pVertices[2]->TestFlag(DOMAIN_BOUNDARY) )
                {
                    dEdgeWeight *= EDGE_WEIGHT;
                }
                if( pVertices[3]->TestFlag(DOMAIN_BOUNDARY) )
                {
                    dEdgeWeight *= EDGE_WEIGHT;
                }

                // Evaluate a new position
                double dTemp = dEdgeWeight * dVolume / (dH1 * dH1 * dH1);
                x += Circumcenter.GetX() * dTemp;
                y += Circumcenter.GetY() * dTemp;
                z += Circumcenter.GetZ() * dTemp;
                dNorm += dTemp;

                // Get next tetrahedron
                pTetra = pTetra->GetVertexTetra(pVertex);
            }

            // Calculate the new position
            double dInvNorm = 1.0 / dNorm;
            vctl::MCPoint3D Point(x * dInvNorm, y * dInvNorm, z * dInvNorm);

            // Limit the position
            checkBoundary(&Point);

            // Attract vertices to image edges
//            mds::seg::CTetraEdgeDDAIterator It(m_spEdgeVolume, pVertex, &Point, this);
//            ++It;
//            for( ; !It.isEnd(); ++It )
//            {
//                // Is there an image edge?
//                if( *It > 0 )
//                {
//                    Point.SetX(It.getRealX());
//                    Point.SetY(It.getRealY());
//                    Point.SetZ(It.getRealZ());
//                    break;
//                }
//            }

            // Store the new position
            NewPositions.push_back(Point);
        }

        MDS_LOG_NOTE("    Re-constructing the mesh ...");

        // Delete all info structures
        vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
        while( pTetrahedron )
        {
            // Delete tetrahedron info struct
            delete (STetrahedronInfo *)pTetrahedron->GetValuePtr();
            pTetrahedron->SetValuePtr(NULL);

            // Get next one
            pTetrahedron = pTetrahedron->GetNext();
        }

        // Destroy all tetras in the mesh
        m_Tetrahedra.DeRegistrationAll();
        m_Tetrahedra.ClearAll();

        // Erase all vertices
        m_pVertices->ClearAll();

        // Create corner vertices
        vctl::MCVertex *pVertex000 = m_pVertices->New(m_dMinX, m_dMinY, m_dMinZ);
        vctl::MCVertex *pVertex100 = m_pVertices->New(m_dMaxX, m_dMinY, m_dMinZ);
        vctl::MCVertex *pVertex010 = m_pVertices->New(m_dMinX, m_dMaxY, m_dMinZ);
        vctl::MCVertex *pVertex001 = m_pVertices->New(m_dMinX, m_dMinY, m_dMaxZ);
        vctl::MCVertex *pVertex110 = m_pVertices->New(m_dMaxX, m_dMaxY, m_dMinZ);
        vctl::MCVertex *pVertex011 = m_pVertices->New(m_dMinX, m_dMaxY, m_dMaxZ);
        vctl::MCVertex *pVertex101 = m_pVertices->New(m_dMaxX, m_dMinY, m_dMaxZ);
        vctl::MCVertex *pVertex111 = m_pVertices->New(m_dMaxX, m_dMaxY, m_dMaxZ);

        // Insert initial tetrahedra
        addTetrahedron(pVertex000, pVertex010, pVertex110, pVertex011);
        addTetrahedron(pVertex000, pVertex110, pVertex100, pVertex101);
        addTetrahedron(pVertex110, pVertex101, pVertex011, pVertex111);
        addTetrahedron(pVertex000, pVertex101, pVertex001, pVertex011);
        addTetrahedron(pVertex000, pVertex110, pVertex011, pVertex101);

        // Create vertices at the new positions
        tPoints::iterator it2End = NewPositions.end();
        tPoints::iterator it2 = NewPositions.begin();
        for( ; it2 != it2End; ++it2 )
        {
            m_pVertices->New(*it2);
        }

        // Create boundary vertices at the new positions
        it2End = NewPositions2.end();
        it2 = NewPositions2.begin();
        for( ; it2 != it2End; ++it2 )
        {
            pVertex = m_pVertices->New(*it2);
        }

        // Set edge length checking function
        if( bCheckIsotropicMinEdgeLength )
        {
//            setMinEdgeLengthFunc(checkIsotropicMinEdgeLengthFunc);
            setMinEdgeLengthFunc(checkIsotropicMinEdgeLengthFunc2);
        }
        else
        {
            setDefaultMinEdgeLengthFunc();
        }

        // Construct the triangulation
        if( !construct() )
        {
            MDS_LOG_NOTE("CTetraMesh::variationalMeshing(): Cannot construct the triangulation!");
            return false;
        }

        MDS_LOG_NOTE("    Final number of vertices = " << getNumOfVertices());
    }

    MDS_LOG_NOTE("CTetraMesh::variationalMeshing(): finished successfuly");

    // Turn on the "vertex existence test"
    m_pVertices->SetTestExistence(true);

    // Set default edge length checking function
    setDefaultMinEdgeLengthFunc();

    // O.K.
    return true;
}


// NEW VERSION
// - internal and boundary vertices treated together
// - boundary vertex projected on plane fitting accumulated edge points

//! Helper vertex position accumulator structure used by variational meshing.
/*struct SAccumulator
{
    typedef std::vector<vctl::MCPoint3D> tEdgePoints;

    //! Edge points.
    tEdgePoints m_EdgePoints;

    //! Parameters of the fitted plane.
    vctl::MCVector3D m_Normal;
    double m_D;

    //! Center of gravity.
    double m_x, m_y, m_z;

    //! Is the plane valid?
    bool m_bValid;

    //! Default constructor.
    SAccumulator() : m_D(0.0), m_x(0.0), m_y(0.0), m_z(0.0), m_bValid(false) {}

    //! 3D plane fitting.
    //! - Returns false on failure.
    bool findPlane()
    {
        // Any edge point?
        if( m_EdgePoints.empty() )
        {
            return false;
        }

        // Find center of gravity
        m_x = m_y = m_z = 0.0;
        tEdgePoints::iterator itEnd = m_EdgePoints.end();
        tEdgePoints::iterator it = m_EdgePoints.begin();
        for( ; it != itEnd; ++it )
        {
            m_x += it->GetX();
            m_y += it->GetY();
            m_z += it->GetZ();
        }
        double dNorm = 1.0 / double(m_EdgePoints.size());
        m_x *= dNorm;
        m_y *= dNorm;
        m_z *= dNorm;

        // Check the minimal number of vertices
        if( m_EdgePoints.size() < 3 )
        {
            return false;
        }

//        MDS_LOG_NOTE("findPlane():");

        // Estimate covariance matrix
        typedef mds::math::CStaticMatrix<double,3,3> tCovMatrix;
        tCovMatrix CovMatrix;
        CovMatrix.fill(0.0);

        // Fill covariance matrix
        itEnd = m_EdgePoints.end();
        it = m_EdgePoints.begin();
        for( ; it != itEnd; ++it )
        {
            vctl::MCPoint3D& Point = *it;

            double dx = Point.getX() - m_x;
            double dy = Point.getY() - m_y;
            double dz = Point.getZ() - m_z;

            CovMatrix(0,0) += dx * dx;
            CovMatrix(1,1) += dy * dy;
            CovMatrix(2,2) += dz * dz;

            CovMatrix(0,1) += dx * dy;
            CovMatrix(0,2) += dx * dz;
            CovMatrix(1,2) += dy * dz;

//            MDS_LOG_NOTE("  " << Point.GetX() << " " << Point.GetY() << " " << Point.GetZ());
        }
        CovMatrix(1,0) = CovMatrix(0,1);
        CovMatrix(2,0) = CovMatrix(0,2);
        CovMatrix(2,1) = CovMatrix(1,2);

        // Normalization using N-1
        if( m_EdgePoints.size() > 1 )
        {
            double dNorm = 1.0 / double(m_EdgePoints.size() - 1);
        }
        CovMatrix *= dNorm;

//        MDS_LOG_NOTE(CovMatrix);

        // Calculate eigen values and vectors
        mds::math::CStaticVector<double, 3> EigenValues;
        EigenValues.fill(0.0);
        try {
            mds::math::eig(CovMatrix, EigenValues);
        }
        catch( ... )
        {
            MDS_LOG_NOTE("Error: Cannot calculate eigen vectors!");
            return false;
        }

//        MDS_LOG_NOTE(CovMatrix);
//        MDS_LOG_NOTE(EigenValues);

        // Find most significant eigen values
        int max0 = 0, max1 = 1;
        if( EigenValues(0) < EigenValues(1) )
        {
            max0 = 1;
            max1 = (EigenValues(0) < EigenValues(2)) ? 2 : 0;
        }
        else
        {
            max0 = 0;
            max1 = (EigenValues(1) < EigenValues(2)) ? 2 : 1;
        }

        // Plane normal (vector product of the two most significant eigen vectors)
        m_Normal.getX() = CovMatrix(max0,1) * CovMatrix(max1,2) - CovMatrix(max0,2) * CovMatrix(max1,1);
        m_Normal.getY() = CovMatrix(max0,2) * CovMatrix(max1,0) - CovMatrix(max0,0) * CovMatrix(max1,2);
        m_Normal.getZ() = CovMatrix(max0,0) * CovMatrix(max1,1) - CovMatrix(max0,1) * CovMatrix(max1,0);

        // Normalization
        m_Normal.Normalization();

        // Plane position
        m_D = -m_x * m_Normal.GetX() - m_y * m_Normal.GetY() - m_z * m_Normal.GetZ();

//        MDS_LOG_NOTE("  Plane: " << m_Normal.GetX() << " " << m_Normal.GetY() << " " << m_Normal.GetZ() << " " << m_D);

        // O.K.
        return (m_bValid = true);
    }
};


// Basic two step version "vertex optimization + boundary optimization"
bool CTetraMesh::variationalMeshing(bool bCheckIsotropicMinEdgeLength)
{
    static const int MAX_ITERS = 3;
//    static const double CORNER_WEIGHT = 1000.0;
    static const double CORNER_WEIGHT = 1.0;
    static const double EDGE_WEIGHT = 0.8;
//    static const double EDGE_WEIGHT = 1.0;

    // Remove all currently unused vertices
    m_Tetrahedra.EraseVerticesNoTetras();

    // Turn off the "vertex existence test"
    m_pVertices->SetTestExistence(false);

    // Logging
    MDS_LOG_NOTE("CTetraMesh::variationalMeshing(): started");

    // Vector of the new positions
    tPoints NewPositions;
    NewPositions.reserve(100000);

    // Allocate and clear accumulators
    typedef std::vector<SAccumulator> tAccumulators;
    tAccumulators Accumulators;
    Accumulators.reserve(10000);

    // Iterate
    for( int iter = 0; iter < MAX_ITERS; ++iter )
    {
        MDS_LOG_NOTE("  Iteration: " << iter);
        MDS_LOG_NOTE("    Initial number of vertices = " << getNumOfVertices());

        // Clear the vertices
        NewPositions.clear();

        // Phase 1) finding boundary vertices

        // Clear accumulators
        Accumulators.clear();
        Accumulators.push_back(SAccumulator());

        // Assign accumulators to vertices
        vctl::MCVertex *pVertex = getFirstVertex();
        for( ; pVertex; pVertex = pVertex->GetNext() )
        {
            pVertex->SetIndex(0);

            // Clear vertex flag
            pVertex->MaskFlag(VARMESH_BOUNDARY);
        }

        // Minimal edge length
        double dMin = getMinVolumeSize(m_dMinEdgeLength);

        // Traverse all edge points and accumulate coordinates
        // in closest vertices
        bool bInitialized = false;
        tVolume::tIterator itEnd = m_spEdgeVolume->getEnd();
        for( tVolume::tIterator it = m_spEdgeVolume->getBegin(); it != itEnd; ++it )
        {
            // Current value
            tVoxel EdgeValue = *it;

            // Lies the vertex on an image edge?
            if( EdgeValue > 0 )
            {
                // Control space value
                mds::img::CPoint3D Point(it.getX(), it.getY(), it.getZ());

                // Find the closest vertex
                vctl::MCPoint3D Vertex;
                conv2Real(Point.x, Point.y, Point.z, &Vertex);
                vctl::MCVertex *pVertex = findNearestVertex(&Vertex, bInitialized);
                if( !pVertex )
                {
                    continue;
                }
                else
                {
                    bInitialized = true;
                }

                // Mark the vertex
                pVertex->SetFlag(VARMESH_BOUNDARY);

                // Accumulator index
                unsigned int uIndex = pVertex->GetIndex();
                if( uIndex == 0 )
                {
                    Accumulators.push_back(SAccumulator());
                    uIndex = Accumulators.size() - 1;
                    pVertex->SetIndex(uIndex);
                }

                // Current accumulator
                SAccumulator& Accumulator = Accumulators[uIndex];

                // Is it an image corner or an image edge?
//                if( EdgeValue == mds::img::CPixelTraits<tVoxel>::getPixelMax() )
//                {
//                }
//                else
//                {
//                }

                // Accumulate coordinates
                Accumulator.m_EdgePoints.push_back(Vertex);
            }
        }

        MDS_LOG_NOTE("    Num of accumulators = " << Accumulators.size());

        // Estimate local planes
//        tAccumulators::iterator atEnd = Accumulators.end();
//        tAccumulators::iterator at = Accumulators.begin();
//        for( ; at != atEnd; ++at )
//        {
//            if( !at->findPlane() )
//            {
//                MDS_LOG_NOTE("    Cannot estimate local plane!");
//            }
//        }


        // Estimate local planes
        tAccumulators::iterator atEnd = Accumulators.end();
        tAccumulators::iterator at = Accumulators.begin();
        for( ; at != atEnd; ++at )
        {
            SAccumulator& Acc = *at;

            // Any edge point?
            if( Acc.m_EdgePoints.empty() )
            {
                continue;
            }

//            MDS_LOG_NOTE("findPlane():");

            // Find center of gravity
            Acc.m_x = Acc.m_y = Acc.m_z = 0.0;
            SAccumulator::tEdgePoints::iterator itEnd = Acc.m_EdgePoints.end();
            SAccumulator::tEdgePoints::iterator it = Acc.m_EdgePoints.begin();
            for( ; it != itEnd; ++it )
            {
                Acc.m_x += it->GetX();
                Acc.m_y += it->GetY();
                Acc.m_z += it->GetZ();
            }
            double dNorm = 1.0 / double(Acc.m_EdgePoints.size());
            Acc.m_x *= dNorm;
            Acc.m_y *= dNorm;
            Acc.m_z *= dNorm;

//            MDS_LOG_NOTE("  Center: " << Acc.m_x << " " << Acc.m_y << " " << Acc.m_z);

            // Sobel operators
            mds::img::CVolumeSobelX<tVolume> SobelX;
            mds::img::CVolumeSobelY<tVolume> SobelY;
            mds::img::CVolumeSobelZ<tVolume> SobelZ;

            // Estimate normal using image derivatives
            Acc.m_Normal.getX() = 0.0;
            Acc.m_Normal.getY() = 0.0;
            Acc.m_Normal.getZ() = 0.0;
            itEnd = Acc.m_EdgePoints.end();
            it = Acc.m_EdgePoints.begin();
            for( ; it != itEnd; ++it )
            {
                vctl::MCPoint3D& Point = *it;

                // Estimate current volume position
                tSize x, y, z;
                conv2Volume(&Point, x, y, z);

                // Image derivatives
                double dGradX = SobelX.getResponse(*m_spVolume, x, y, z);
                double dGradY = SobelY.getResponse(*m_spVolume, x, y, z);
                double dGradZ = SobelZ.getResponse(*m_spVolume, x, y, z);

                // Control space gradient
                vctl::MCVector3D VolumeEdgeNormal(dGradX, dGradY, dGradZ);

                // Convert to real coordinates :)
                conv2Volume(&VolumeEdgeNormal);
//                VolumeEdgeNormal.Normalization();

                // Compute average normal
                Acc.m_Normal += VolumeEdgeNormal;

//                MDS_LOG_NOTE("  " << Point.GetX() << " " << Point.GetY() << " " << Point.GetZ());
            }

            // Normalization
            Acc.m_Normal.Normalization();

            // Plane position
            Acc.m_D = -Acc.m_x * Acc.m_Normal.GetX() - Acc.m_y * Acc.m_Normal.GetY() - Acc.m_z * Acc.m_Normal.GetZ();

//            MDS_LOG_NOTE("  Plane: " << Acc.m_Normal.GetX() << " " << Acc.m_Normal.GetY() << " " << Acc.m_Normal.GetZ() << " " << Acc.m_D);

            // Plane found
            Acc.m_bValid = true;
        }

        // Phase 2) Move all vertices to their optimal position

        // Move all vertices in the mesh
        int iCount = 0;
        for( pVertex = getFirstVertex(); pVertex; pVertex = pVertex->GetNext() )
        {
            // Calculate new vertex position
            double x = 0.0, y = 0.0, z = 0.0, dNorm = 0.0;

            // Traverse all tetrahedra around the current vertex
            vctl::MCTetra *pTetra = pVertex->GetRegisteredTetra();
            while( pTetra )
            {
                // Centroid of the tetrahedra
                vctl::MCPoint3D Center;
                pTetra->GetCenter(Center);

                // Get control space value
                mds::img::CPoint3D Point;
                conv2Volume(&Center, Point.x, Point.y, Point.z);
                double Value = getControlSpaceValue(Point);
                double dH1 = (Value < dMin) ? dMin : Value;

                // Tetrahedra volume
                double dVolume = pTetra->GetVolume();

                // Circumcenter
                vctl::MCPoint3D Circumcenter;
                getTetrahedronCircumsphere(pTetra)->GetCenter(Circumcenter);

                // Get vertices
//                vctl::MCVertex *pVertices[4];
//                pTetra->GetVerticeS(pVertices);
//                double dEdgeWeight = 1.0;
//                if( pVertices[0]->TestFlag(MARKED) )
//                {
//                    dEdgeWeight *= EDGE_WEIGHT;
//                }
//                if( pVertices[1]->TestFlag(MARKED) )
//                {
//                    dEdgeWeight *= EDGE_WEIGHT;
//                }
//                if( pVertices[2]->TestFlag(MARKED) )
//                {
//                    dEdgeWeight *= EDGE_WEIGHT;
//                }
//                if( pVertices[3]->TestFlag(MARKED) )
//                {
//                    dEdgeWeight *= EDGE_WEIGHT;
//                }

                // Evaluate a new position
//                double dTemp = dEdgeWeight * dVolume / (dH1 * dH1 * dH1);
                double dTemp = dVolume / (dH1 * dH1 * dH1);
                x += Circumcenter.GetX() * dTemp;
                y += Circumcenter.GetY() * dTemp;
                z += Circumcenter.GetZ() * dTemp;
                dNorm += dTemp;

                // Get next tetrahedron
                pTetra = pTetra->GetVertexTetra(pVertex);
            }

            // Calculate the new position
            double dInvNorm = 1.0 / dNorm;
            vctl::MCPoint3D Point(x * dInvNorm, y * dInvNorm, z * dInvNorm);

            // Check if vertex approximates boundary
            if( pVertex->TestFlag(VARMESH_BOUNDARY) )
            {
                // Accumulator index
                unsigned int uIndex = pVertex->GetIndex();
                if( uIndex > 0 )
                {
                    // Current accumulator
                    SAccumulator& Accumulator = Accumulators[uIndex];
                    if( Accumulator.m_bValid )
                    {
                        // Project vertex on a plane approximating edges locally
                        double dDistance = Accumulator.m_Normal.getX() * pVertex->getX();
                        dDistance += Accumulator.m_Normal.getY() * pVertex->getY();
                        dDistance += Accumulator.m_Normal.getZ() * pVertex->getZ();
                        dDistance += Accumulator.m_D;
                        Point.getX() -= dDistance * Accumulator.m_Normal.getX();
                        Point.getY() -= dDistance * Accumulator.m_Normal.getY();
                        Point.getZ() -= dDistance * Accumulator.m_Normal.getZ();

                        // Increase number of vertices
                        ++iCount;
                    }
                    else
                    {
                        // Center of gravity
//                        Point.getX() = Accumulator.m_x;
//                        Point.getY() = Accumulator.m_y;
//                        Point.getZ() = Accumulator.m_z;

                        MDS_LOG_NOTE("Error: cannot approximate local plane!");
                    }
                }
            }

            // Limit the position
            checkBoundary(&Point);

            // Store the new position
            NewPositions.push_back(Point);
        }

        MDS_LOG_NOTE("    Number of boundary vertices projected on a local plane = " << iCount);
        MDS_LOG_NOTE("    Re-constructing the mesh ...");

        // Delete all info structures
        vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
        while( pTetrahedron )
        {
            // Delete tetrahedron info struct
            delete (STetrahedronInfo *)pTetrahedron->GetValuePtr();
            pTetrahedron->SetValuePtr(NULL);

            // Get next one
            pTetrahedron = pTetrahedron->GetNext();
        }

        // Destroy all tetras in the mesh
        m_Tetrahedra.DeRegistrationAll();
        m_Tetrahedra.ClearAll();

        // Erase all vertices
        m_pVertices->ClearAll();

        // Create corner vertices
        vctl::MCVertex *pVertex000 = m_pVertices->New(m_dMinX, m_dMinY, m_dMinZ);
        vctl::MCVertex *pVertex100 = m_pVertices->New(m_dMaxX, m_dMinY, m_dMinZ);
        vctl::MCVertex *pVertex010 = m_pVertices->New(m_dMinX, m_dMaxY, m_dMinZ);
        vctl::MCVertex *pVertex001 = m_pVertices->New(m_dMinX, m_dMinY, m_dMaxZ);
        vctl::MCVertex *pVertex110 = m_pVertices->New(m_dMaxX, m_dMaxY, m_dMinZ);
        vctl::MCVertex *pVertex011 = m_pVertices->New(m_dMinX, m_dMaxY, m_dMaxZ);
        vctl::MCVertex *pVertex101 = m_pVertices->New(m_dMaxX, m_dMinY, m_dMaxZ);
        vctl::MCVertex *pVertex111 = m_pVertices->New(m_dMaxX, m_dMaxY, m_dMaxZ);

        // Insert initial tetrahedra
        addTetrahedron(pVertex000, pVertex010, pVertex110, pVertex011);
        addTetrahedron(pVertex000, pVertex110, pVertex100, pVertex101);
        addTetrahedron(pVertex110, pVertex101, pVertex011, pVertex111);
        addTetrahedron(pVertex000, pVertex101, pVertex001, pVertex011);
        addTetrahedron(pVertex000, pVertex110, pVertex011, pVertex101);

        // Create vertices at the new positions
        tPoints::iterator it2End = NewPositions.end();
        tPoints::iterator it2 = NewPositions.begin();
        for( ; it2 != it2End; ++it2 )
        {
            m_pVertices->New(*it2);
        }

        // Set edge length checking function
        if( bCheckIsotropicMinEdgeLength )
        {
//            setMinEdgeLengthFunc(checkIsotropicMinEdgeLengthFunc);
            setMinEdgeLengthFunc(checkIsotropicMinEdgeLengthFunc2);
        }
        else
        {
            setDefaultMinEdgeLengthFunc();
        }

        // Construct the triangulation
        if( !construct() )
        {
            MDS_LOG_NOTE("CTetraMesh::variationalMeshing(): Cannot construct the triangulation!");
            return false;
        }

        MDS_LOG_NOTE("    Final number of vertices = " << getNumOfVertices());
    }

    MDS_LOG_NOTE("CTetraMesh::variationalMeshing(): finished successfuly");

    // Turn on the "vertex existence test"
    m_pVertices->SetTestExistence(true);

    // Set default edge length checking function
    setDefaultMinEdgeLengthFunc();

    // O.K.
    return true;
}*/


// Version 2 "all at once"
// - Latest functional version!
bool CTetraMesh::isotropicEdgeSplitting(double dThreshold, bool bBoundary)
{
//    static const int MAX_ITERS = 3;
//    static const int MAX_ITERS = 5;
    static const int MAX_ITERS = 1;

    MDS_CHECK(dThreshold > 0.0, return false);

    MDS_LOG_NOTE("CTetraMesh::isotropicEdgeSplitting(): begin");

    // Update the control space
//    updateControlSpace(m_dMinEdgeLength);

    // Set edge length checking function
//    setMinEdgeLengthFunc(checkIsotropicMinEdgeLengthFunc);

    // Construct container of all edges in the mesh
//    m_Tetrahedra.MakeAllTetrasEdges();

    // Get pointer to the container of edges
//    vctl::MCEdgeS *pEdges = m_Tetrahedra.GetEdgeS();

    // Helper vector of newly created vertices
    tPoints Points, EdgePoints;

    // Iterate
//    for( int iter = 0; pEdges->GetNumber() > 0 && iter < MAX_ITERS; ++iter )
    for( int iter = 0; iter < MAX_ITERS; ++iter )
    {
        // Logging the number of edges
        MDS_LOG_TIME("Iteration begin");
        MDS_LOG_NOTE("  Iteration = " << iter);
//        MDS_LOG_NOTE("  Number of edges = " << pEdges->GetNumber());
        MDS_LOG_NOTE("  Number of vertices = " << getNumOfVertices());

        // Clear the helper vector of vertices
        EdgePoints.clear();
        Points.clear();

        // Clear vertices flags
        vctl::MCVertex *pVertex = getFirstVertex();
        for( ; pVertex; pVertex = pVertex->GetNext() )
        {
            pVertex->MaskFlag(MARKED);
            pVertex->SetIndex(0);
        }

        // Initial color
        unsigned int uiColor = 0;

        // Evaluate and split all edges
        // - Construct vector of new vertices
        // - Remove edges which were not divided
//        vctl::MCEdge *pEdge = (vctl::MCEdge *)pEdges->GetFirst();
//        while( pEdge )
        for( pVertex = getFirstVertex(); pVertex; pVertex = pVertex->GetNext() )
        {
            // Mark the vertex
            pVertex->SetFlag(MARKED);

            // Getting tetrahedra around the current vertex
            // is used to process all edges around the vertex.
            // - Index coloring prevents multiple occurences.
            ++uiColor;
            vctl::MCTetra *pTetra = pVertex->GetRegisteredTetra();
            for( ; pTetra; pTetra = pTetra->GetVertexTetra(pVertex) )
            {
                // Get tetrahedron vertices
                vctl::MCVertex *pVertices[4];
                pTetra->GetVerticeS(pVertices);

                // Process vertices
                if( !pVertices[0]->TestFlag(MARKED) && pVertices[0]->GetIndex() < uiColor )
                {
                    isotropicSplitEdge(pVertex, pVertices[0], dThreshold, EdgePoints, Points, bBoundary);
                    pVertices[0]->SetIndex(uiColor);
                }
                if( !pVertices[1]->TestFlag(MARKED) && pVertices[1]->GetIndex() < uiColor )
                {
                    isotropicSplitEdge(pVertex, pVertices[1], dThreshold, EdgePoints, Points, bBoundary);
                    pVertices[1]->SetIndex(uiColor);
                }
                if( !pVertices[2]->TestFlag(MARKED) && pVertices[2]->GetIndex() < uiColor )
                {
                    isotropicSplitEdge(pVertex, pVertices[2], dThreshold, EdgePoints, Points, bBoundary);
                    pVertices[2]->SetIndex(uiColor);
                }
                if( !pVertices[3]->TestFlag(MARKED) && pVertices[3]->GetIndex() < uiColor )
                {
                    isotropicSplitEdge(pVertex, pVertices[3], dThreshold, EdgePoints, Points, bBoundary);
                    pVertices[3]->SetIndex(uiColor);
                }
            }

            // Check the edge
//            if( pEdge->TestFlag(vctl::ENTITY_MARK_FLAG) )
//            {
//                // Get next edge
//                pEdge = pEdge->GetNext();
//                continue;
//            }

            // Split the edge
//            if( !isotropicSplitEdge(pEdge, dThreshold, Points) )
//            {
//                // Mark the edge
//                pEdge->SetFlag(vctl::ENTITY_MARK_FLAG);
//            }
//            isotropicSplitEdge(pEdge, dThreshold, EdgePoints, Points, bBoundary);

            // Mark the edge
//            pEdge->SetFlag(vctl::ENTITY_MARK_FLAG);

            // Get next edge
//            vctl::MCEdge *pPrevious = pEdge;
//            pEdge = pEdge->GetNext();

            // Remove already processed edge from container
//            pEdges->Erase(pPrevious);
        }

        // Number of new points
        MDS_LOG_TIME("After splitting");
        MDS_LOG_NOTE("  Number of newly created edge vertices = " << EdgePoints.size());
        MDS_LOG_NOTE("  Number of newly created vertices = " << Points.size());

        // No min edge length checking
//        setMinEdgeLengthFunc(NULL);
//        setMinEdgeLengthFunc(checkNoImageEdgeIntersectionFunc);
//        setMinEdgeLengthFunc(checkIsotropicMinEdgeLengthFunc);
//        setMinEdgeLengthFunc(checkIsotropicMinEdgeLengthFunc2);
        setDefaultMinEdgeLengthFunc();

        // Insert all edge points to the mesh
        tPoints::iterator it = EdgePoints.begin();
        tPoints::iterator itEnd = EdgePoints.end();
        int iCount = 0;
        while( it != itEnd )
        {
            // Get pointer to the point
            vctl::MCPoint3D *pPoint = &(*it);

            // Check if the point lies on the mesh boundary
            int iBoundaryFlags = checkBoundary(pPoint);

            // Create a new vertex
            vctl::MCVertex *pVertex = newVertex(pPoint);

            // Insert it to the mesh
            if( !addVertex(pVertex,
                           iBoundaryFlags//,
//                           isotropicEdgeSplittingInsertFunc,
//                           isotropicEdgeSplittingRemoveFunc
                           ) )
            {
                // Remove the vertex
                if( !pVertex->GetRegisteredTetra() )
                {
                    m_pVertices->Erase(pVertex);
                }
            }
            else
            {
                // Update the control space
//                updateControlSpace(pVertex, m_dMinEdgeLength);
                ++iCount;
            }

            // Get next point
            ++it;
        }

        MDS_LOG_TIME("After adding vertices");
        MDS_LOG_NOTE("  Number of successfuly added edge points = " << iCount);

        // Set edge length checking function
//        setDefaultMinEdgeLengthFunc();
//        setMinEdgeLengthFunc(NULL);
        setMinEdgeLengthFunc(checkIsotropicMinEdgeLengthFunc);
//        setMinEdgeLengthFunc(checkNoImageEdgeIntersectionFunc);

        // Insert all points to the mesh
//        tPoints::iterator it = Points.begin();
//        tPoints::iterator itEnd = Points.end();
        iCount = 0;
        it = Points.begin();
        itEnd = Points.end();
        while( it != itEnd )
        {
            // Get pointer to the point
            vctl::MCPoint3D *pPoint = &(*it);

            // Check if the point lies on the mesh boundary
            int iBoundaryFlags = checkBoundary(pPoint);

            // Create a new vertex
            vctl::MCVertex *pVertex = newVertex(pPoint);

            // Insert it to the mesh
            if( !addVertex(pVertex,
                           iBoundaryFlags//,
//                           isotropicEdgeSplittingInsertFunc,
//                           isotropicEdgeSplittingRemoveFunc
                           ) )
            {
                // Remove the vertex
                if( !pVertex->GetRegisteredTetra() )
                {
                    m_pVertices->Erase(pVertex);
                }
            }
            else
            {
                // Update the control space
//                updateControlSpace(pVertex, m_dMinEdgeLength);
                ++iCount;
            }

            // Get next point
            ++it;
        }

        MDS_LOG_NOTE("  Number of successfuly added points = " << iCount);

        // Check the list of edges and remove edges which are not used
//        m_Tetrahedra.EraseEdgesNoTetras();
    }

    // Clear the list of edges
//    pEdges->DeRegistrationAll();
//    pEdges->ClearAll();

    // Set default edge length checking function
    setDefaultMinEdgeLengthFunc();

    MDS_LOG_NOTE("CTetraMesh::isotropicEdgeSplitting(): end");

    // O.K.
    return true;
}


// Version 3 "one by one"
// (see SVN history before 2009/11/30)
//bool CTetraMesh::isotropicEdgeSplitting(double dThreshold)

// Version "all at once"
// (see SVN history before 2009/11/30)
//bool CTetraMesh::isotropicEdgeSplitting(double dThreshold)


// Latest functional version!
bool CTetraMesh::isotropicSplitEdge(vctl::MCVertex *pV0,
                                    vctl::MCVertex *pV1,
                                    double dThreshold,
                                    tPoints& EdgePoints,
                                    tPoints& Points,
                                    bool bBoundaryEdges
                                    )
{
//    static const double dRestrictedZone = 5.0;
//    static const double dOrientThreshold = 0.2;
    static const double dOrientThreshold = 0.3;

//    MDS_CHECK(pEdge, return false);
    MDS_CHECK(pV0 && pV1, return false);

//    MDS_LOG_NOTE("isotropicSplitEdge(): begin");

    // Compute square of the minimal edge length
//    double dSquareMinEdgeLength = 0.1;
    double dSquareMinEdgeLength = m_dMinEdgeLength * m_dMinEdgeLength;
//    double dSquareMinEdgeLength = m_dCornerMinEdgeLength * m_dCornerMinEdgeLength;

    // Vertices of the edge
/*    vctl::MCVertex *pVertices[2];
    pEdge->GetVerticeS(pVertices);

    // Check the edge length
    if( pVertices[0]->SquareDistance(*pVertices[1]) < 4.0 * dSquareMinEdgeLength )
    {
        return false;
    }*/
    if( pV0->SquareDistance(*pV1) < 4.0 * dSquareMinEdgeLength )
    {
        return false;
    }

    // Sobel oparators
    mds::img::CVolumeSobelX<tVolume> SobelX;
    mds::img::CVolumeSobelY<tVolume> SobelY;
    mds::img::CVolumeSobelZ<tVolume> SobelZ;

    // Helper container of vertices
//    typedef std::list<vctl::MCPoint3D> tList;
    typedef std::vector<vctl::MCPoint3D> tList;
    tList Midpoints;
//    Midpoints.push_back(*pVertices[0]);
    Midpoints.push_back(*pV0);

    // Any point added?
    bool bResult = false;

    // Create edge pixels iterator
//    CTetraEdgeIterator It(m_spEdgeVolume, pEdge, this);
//    CTetraEdgeIterator It(m_spEdgeVolume, pV0, pV1, this);
//    CTetraEdgeDDAIterator It(m_spEdgeVolume, pEdge, this);
    CTetraEdgeDDAIterator It(m_spEdgeVolume, pV0, pV1, this);

    if( It.isEnd() )
    {
        return false;
    }

    // Value at the beginning of the edge
    tVoxel LastValue = *It;
    ++It;

    // Tetrahedron edge
//    vctl::MCVector3D TetraEdge(pVertices[0], pVertices[1]);
    vctl::MCVector3D TetraEdge(pV0, pV1);
    TetraEdge.Normalization();

    // Construct the midpoints
    for( ; !It.isEnd(); ++It )
    {
        // Get current control space value
        tVoxel Value = *It;
//        tVoxel Value1 = Value & 0x7f;
//        tVoxel Value2 = (Value & 0x7f00) >> 8;

        if( LastValue > 0 && Value > 0 )
        {
            continue;
        }

        // Is there an image edge?
//        if( Value > 0 )
//        if( Value1 == 0 )
//        if( mds::math::getAbs(LastValue - Value) > 50 )
//        if( mds::math::getAbs(LastValue - Value) > 100 )
        if( mds::math::getAbs(LastValue - Value) > 0 )
        {
            // Splitting point
            vctl::MCPoint3D Point(It.getRealX(), It.getRealY(), It.getRealZ());

            // Evaluate its subpixel position
/*            if( !subpixelAttractVertex(&Point) )
            {
                continue;
            }*/

            // Check size of produced edges
            double dLength = Point.SquareDistance(Midpoints.back());
            if( dLength < dSquareMinEdgeLength )
            {
                continue;
            }
//            dLength = Point.SquareDistance(*pVertices[1]);
            dLength = Point.SquareDistance(*pV1);
            if( dLength < dSquareMinEdgeLength )
            {
                continue;
            }

            // Estimate current volume position
//            tSize x, y, z;
//            conv2Volume(&Point, x, y, z);
            tSize x = It.getX();
            tSize y = It.getY();
            tSize z = It.getZ();

            // Control space derivatives
            double dGradX = SobelX.getResponse(*m_spVolume, x, y, z);
            double dGradY = SobelY.getResponse(*m_spVolume, x, y, z);
            double dGradZ = SobelZ.getResponse(*m_spVolume, x, y, z);
/*            double dGradX = SobelX.getResponse(*m_spVolume, It.getX(), It.getY(), It.getZ());
            double dGradY = SobelY.getResponse(*m_spVolume, It.getX(), It.getY(), It.getZ());
            double dGradZ = SobelZ.getResponse(*m_spVolume, It.getX(), It.getY(), It.getZ());*/
//            double dGradX = SobelX.getResponse(*m_spVolume, mds::math::round2Int(It.getX()), mds::math::round2Int(It.getY()), mds::math::round2Int(It.getZ()));
//            double dGradY = SobelY.getResponse(*m_spVolume, mds::math::round2Int(It.getX()), mds::math::round2Int(It.getY()), mds::math::round2Int(It.getZ()));
//            double dGradZ = SobelZ.getResponse(*m_spVolume, mds::math::round2Int(It.getX()), mds::math::round2Int(It.getY()), mds::math::round2Int(It.getZ()));

            // Control space gradient
            vctl::MCVector3D VolumeEdgeNormal(dGradX, dGradY, dGradZ);

            // Convert to real coordinates :)
            conv2Volume(&VolumeEdgeNormal);
            VolumeEdgeNormal.Normalization();

            // Compare orientation of the control space gradient and tetrahedron edge
            // - Dot product of both vectors.
            // - Value must be greater than a given threshold. Optimal case is
            //   that both vectors are parallel (volume edge and tetrahedron edge are
            //   ortogonal), then -1 or 1 is obtained.
            if( mds::math::getAbs(VolumeEdgeNormal * TetraEdge) > dOrientThreshold )
            {
                // Insert point to the vector
                Midpoints.push_back(Point);
//                Points.push_back(Point);
                EdgePoints.push_back(Point);
                bResult = true;
            }
        }

        LastValue = Value;
    }

    // Insert the second end-point of the edge
//    Midpoints.push_back(*pVertices[1]);
    Midpoints.push_back(*pV1);

    // Check number of subedges
    if( bBoundaryEdges && Midpoints.size() < 3 )
    {
        return false;
    }

//    MDS_LOG_NOTE("  Number of sub-edges = " << Midpoints.size() - 1);

    // Split all the sub-edges in the list
//    tList::iterator itEnd = Midpoints.end();
    tList::iterator it = Midpoints.begin();
    tList::iterator itNext = Midpoints.begin();
    ++itNext;
    for( ; itNext != Midpoints.end(); ++it, ++itNext )
    {
        // Try to split the sub-edge
//        if( isotropicSplitSubEdge(*it, *itNext, dThreshold, dRestrictedZone, Points) )
        if( isotropicSplitSubEdge(*it, *itNext, dThreshold, m_dMinEdgeLength, Points) )
        // Latest functional version!
//        if( isotropicSplitSubEdge(*it, *itNext, dThreshold, m_dCornerMinEdgeLength, Points) )
//        if( isotropicSplitSubEdge(*it, *itNext, dThreshold, 2 * m_dMinEdgeLength, Points) )
        {
            // Some new points were added
            bResult = true;
        }
    }

    return bResult;
}


bool CTetraMesh::isotropicSplitSubEdge(vctl::MCPoint3D &p1,
                                       vctl::MCPoint3D &p2,
                                       double dThreshold,
                                       double dRestrictedZone,
                                       tPoints& Points
                                       )
{
    static const std::size_t MaxMidpoints = 1000;

//    MDS_LOG_NOTE("isotropicSplitSubEdge():");

    // Check the edge length
    if( getControlSpaceLength(p1, p2) < dThreshold )
//    if( getControlSpaceLength(p1, p2) < 1.0 )
    {
        return false;
    }

    // Helper container of vertices
    typedef std::list<vctl::MCPoint3D> tList;
    tList Midpoints;
    Midpoints.push_back(p1);
    Midpoints.push_back(p2);

    // Construct the midpoints
    tList::iterator it = Midpoints.begin();
    tList::iterator itNext = Midpoints.begin();
    ++itNext;
    while( itNext != Midpoints.end() && Midpoints.size() < MaxMidpoints )
    {
        // Evaluate the edge length
        double dLength = getControlSpaceLength(*it, *itNext);
        if( dLength > dThreshold )
        {
            // Create a new point in the middle of the edge
            vctl::MCPoint3D NewPoint(0.5 * (itNext->GetX() + it->GetX()),
                                     0.5 * (itNext->GetY() + it->GetY()),
                                     0.5 * (itNext->GetZ() + it->GetZ())
                                     );

            // Insert it to the list
            itNext = Midpoints.insert(itNext, NewPoint);
        }
        else
        {
            ++it;
            ++itNext;
        }
    }

//    MDS_LOG_NOTE("  Number of midpoints = " << Midpoints.size());

    // Minimal allowed distance from an image edge
    // - Enabled in the latest functional version!
/*    double dMinDistance = getMinVolumeSize(dRestrictedZone);

    // Remove all midpoints which are too close to an image edge
    itEnd = Midpoints.end();
    --itEnd;
    it = Midpoints.begin();
    ++it;
    while( it != itEnd )
    {
        // Get the midpoint coordinates
        mds::tSize x, y, z;
        conv2Volume(&(*it), x, y, z);

        // Check value of the control space
        tVoxel Value = m_spControlSpace->get(x, y, z);
        tVoxel Value1 = Value & 0x7f;
        if( Value1 > 0 && Value1 < dMinDistance )
        {
            // Remove point form the list
            tList::iterator it2Del = it;
            ++it;
            Midpoints.erase(it2Del);
        }
        else
        {
            ++it;
        }
    }*/

//    MDS_LOG_NOTE("  Final number of midpoints = " << Midpoints.size());

    // Check the number of points
    if( Midpoints.size() < 3 )
    {
        return false;
    }

    // Compute square of the minimal edge length
    double dSquareMinEdgeLength = m_dMinEdgeLength * m_dMinEdgeLength;
//    double dSquareMinEdgeLength = m_dCornerMinEdgeLength * m_dCornerMinEdgeLength;

    // Number of added points
    int iCount = 0;

    // Create new splitting points
    it = Midpoints.begin();
    itNext = Midpoints.begin();
    ++itNext;
    double dSum = 0.0;
    vctl::MCPoint3D StartPoint = *it;
    while( itNext != Midpoints.end() )
    {
        // Get the control space length
        double dLength = getControlSpaceLength(*it, *itNext);

        // Check real size of produced edges
        // - Disabled in the latest functional version!
        double dL1, dL2;
        dL1 = it->SquareDistance(StartPoint);
        dL2 = it->SquareDistance(Midpoints.back());
        if( dL1 < dSquareMinEdgeLength || dL2 < dSquareMinEdgeLength )
        {
            // Move to the next midpoint
            dSum += dLength;
            ++it;
            ++itNext;
            continue;
        }

        // Check the control space length
        if( (dSum + dLength) > 1.0 )
        {
            // Find the best position
            // - Disabled in the latest functional version!
/*            dL1 = mds::math::getAbs(1.0 - dSum);
            dL2 = mds::math::getAbs(1.0 - dSum - dLength);
            if( dL1 < dL2 )*/
            {
                // Insert a new point to the output list
                Points.push_back(*it);
                StartPoint = *it;
            }
/*            else
            {
                // Insert a new point to the output list
                Points.push_back(*itNext);
                StartPoint = *itNext;

                // Move to the next midpoint
                ++it;
                if( it == itEnd )
                {
                    break;
                }

                // Next point
                itNext = it;
                ++itNext;

                // Get length of the edge segment
                dLength = getControlSpaceLength(*it, *itNext);
            }*/

            // New point was created
            ++iCount;

            // Clear the sum
            dSum = 0.0;
        }

        // Move to the next midpoint
        dSum += dLength;
        ++it;
        ++itNext;
    }

//    MDS_LOG_NOTE("  Number of added points = " << iCount);

    return (iCount > 0);
}


// Version "all at once"
void CTetraMesh::isotropicEdgeSplittingInsertFunc(vctl::MCTetra *pTetrahedron, CDelaunayTetra *pMesh)
{
    MDS_ASSERT(pTetrahedron && pMesh);

    // Get pointer to the triangular mesh
    CTetraMesh *pTetraMesh = (CTetraMesh *)pMesh;

    // Insert edges of the triangle to the list
    pTetraMesh->m_Tetrahedra.MakeTetraEdges(pTetrahedron);
}


void CTetraMesh::isotropicEdgeSplittingRemoveFunc(vctl::MCTetra *pTetrahedron, CDelaunayTetra *pMesh)
{
    MDS_ASSERT(pTetrahedron && pMesh);

    // Get pointer to the triangular mesh
    CTetraMesh *pTetraMesh = (CTetraMesh *)pMesh;

    // Get tetrahedron vertices
    vctl::MCVertex *pVertices[4];
    pTetrahedron->GetVerticeS(pVertices);

    // Get container of edges
    vctl::MCEdgeS *pEdges = pTetraMesh->m_Tetrahedra.GetEdgeS();

    // Unsort all edges of the tetrahedron
    vctl::MCEdge *pEdge = pEdges->TestExistence(pVertices[0], pVertices[1]);
    if( pEdge )
    {
        pEdges->Erase(pEdge);
    }

    pEdge = pEdges->TestExistence(pVertices[1], pVertices[2]);
    if( pEdge )
    {
        pEdges->Erase(pEdge);
    }

    pEdge = pEdges->TestExistence(pVertices[2], pVertices[0]);
    if( pEdge )
    {
        pEdges->Erase(pEdge);
    }

    pEdge = pEdges->TestExistence(pVertices[0], pVertices[3]);
    if( pEdge )
    {
        pEdges->Erase(pEdge);
    }

    pEdge = pEdges->TestExistence(pVertices[1], pVertices[3]);
    if( pEdge )
    {
        pEdges->Erase(pEdge);
    }

    pEdge = pEdges->TestExistence(pVertices[2], pVertices[3]);
    if( pEdge )
    {
        pEdges->Erase(pEdge);
    }
}


// Version 3 "one by one"
// (see SVN history before 2009/11/30)
//void CTetraMesh::isotropicEdgeSplittingInsertFunc(vctl::MCTetra *pTetrahedron, CDelaunayTetra *pMesh)
//void CTetraMesh::isotropicEdgeSplittingRemoveFunc(vctl::MCTetra *pTetrahedron, CDelaunayTetra *pMesh)

// Version 1 "one by one"
// (see SVN history before 2009/11/30)
//void CTetraMesh::isotropicEdgeSplittingInsertFunc(vctl::MCTetra *pTetrahedron, CDelaunayTetra *pMesh)
//double CTetraMesh::getControlSpaceLength(vctl::MCPoint3D& p1, vctl::MCPoint3D& p2)
// ...


// Using medial axis, log() version
double CTetraMesh::getControlSpaceLength(vctl::MCPoint3D& p1, vctl::MCPoint3D& p2)
{
    // Get value at the first end-point
    mds::img::CPoint3D P1;
    conv2Volume(&p1, P1.x(), P1.y(), P1.z());
    double Value = getControlSpaceValue(P1);
    double dMin = getMinVolumeSize(m_dMinEdgeLength);
    double dH1 = (Value < dMin) ? dMin : Value;

    // Get value at the second point
    mds::img::CPoint3D P2;
    conv2Volume(&p2, P2.x(), P2.y(), P2.z());
    Value = getControlSpaceValue(P2);
    double dH2 = (Value < dMin) ? dMin : Value;

    // Usual length of the edge
    double dDist = P1.getDistance(P2);

    // Normalization
    if( dH2 < dH1 )
    {
        mds::math::swap(dH1, dH2);
    }

    // Check the special case
    double dDiff = dH2 - dH1;
    if( dDiff < 1.0 )
    {
        return (dDist / dH2);
    }

    // Estimate length in the transformed space
    double dLength = dDist * dDiff / (dH1 * dH2 * std::log(dH2 / dH1));
/*    if( dLength <= 0.0 )
    {
        MDS_LOG_NOTE("Control space length lower then zero!");
    }*/
    return dLength;
}


// Two values in one (distance form edge and distance from corner)
// (see SVN history before 2009/11/30)
//void CTetraMesh::getControlSpaceValue(const mds::img::CPoint3D& Point, double& dV1, double& dV2)


// using medial axis
double CTetraMesh::getControlSpaceValue(const mds::img::CPoint3D& Point)
{
    int iX = int(Point.x());
    int iY = int(Point.y());
    int iZ = int(Point.z());

    double dX = Point.x() - iX;
    double dY = Point.y() - iY;
    double dZ = Point.z() - iZ;

    double dInvX = 1.0 - dX;
    double dInvY = 1.0 - dY;
    double dInvZ = 1.0 - dZ;

    double dV = dInvX * dInvY * dInvZ * m_spControlSpace->get(iX, iY, iZ);
    dV += dX * dInvY * dInvZ * m_spControlSpace->get(iX + 1, iY, iZ);
    dV += dInvX * dY * dInvZ * m_spControlSpace->get(iX, iY + 1, iZ);
    dV += dX * dY * dInvZ * m_spControlSpace->get(iX + 1, iY + 1, iZ);
    dV += dInvX * dInvY * dZ * m_spControlSpace->get(iX, iY, iZ + 1);
    dV += dX * dInvY * dZ * m_spControlSpace->get(iX + 1, iY, iZ + 1);
    dV += dInvX * dY * dZ * m_spControlSpace->get(iX, iY + 1, iZ + 1);
    dV += dX * dY * dZ * m_spControlSpace->get(iX + 1, iY + 1, iZ + 1);

    return dV;
}


bool CTetraMesh::checkIsotropicMinEdgeLengthFunc(vctl::MCPoint3D *p1,
                                                 vctl::MCPoint3D *p2,
                                                 CDelaunayTetra* pMesh
                                                 )
{
    MDS_CHECK(p1 && p2 && pMesh, return false);

    // Get pointer to the triangular mesh
    CTetraMesh *pTetraMesh = (CTetraMesh *)pMesh;

    // Evaluate length of the edge
    double dLength = pTetraMesh->getControlSpaceLength(*p1, *p2);

    // Check the edge length
//    return (mds::math::getAbs(dLength) > 0.9);
//  Latest functional version!
    return (dLength > 0.9);
//    return (dLength > 0.5);
//    return (dLength > 0.1);
}


bool CTetraMesh::checkIsotropicMinEdgeLengthFunc2(vctl::MCPoint3D *p1,
                                                  vctl::MCPoint3D *p2,
                                                  CDelaunayTetra* pMesh
                                                  )
{
    MDS_CHECK(p1 && p2 && pMesh, return false);

    // Get pointer to the triangular mesh
    CTetraMesh *pTetraMesh = (CTetraMesh *)pMesh;

    // Evaluate length of the edge
    double dLength = pTetraMesh->getControlSpaceLength(*p1, *p2);

    // Check the edge length
//    return (mds::math::getAbs(dLength) > 0.2);
//    return (dLength > 0.1);
    return (dLength > 0.5);
//    return (dLength > 0.3);
}


bool CTetraMesh::checkNoImageEdgeIntersectionFunc(vctl::MCPoint3D *p1,
                                                  vctl::MCPoint3D *p2,
                                                  CDelaunayTetra* pMesh
                                                  )
{
    static const double dOrientThreshold = 0.5;
    static const tVoxel VoxelMax = mds::img::CPixelTraits<tVoxel>::getPixelMax();

    MDS_CHECK(p1 && p2 && pMesh, return false);

    // Get pointer to the mesh
    CTetraMesh *pTetraMesh = (CTetraMesh *)pMesh;

    // Create edge pixels iterator
    CTetraEdgeIterator It(pTetraMesh->m_spEdgeVolume, p1, p2, pTetraMesh);
//    CTetraEdgeDDAIterator It(pTetraMesh->m_spEdgeVolume, p1, p2, pTetraMesh);
    if( It.isEnd() )
    {
//        return true;
        return false;
    }

    // Sobel oparators
    mds::img::CVolumeSobelX<tVolume> SobelX;
    mds::img::CVolumeSobelY<tVolume> SobelY;
    mds::img::CVolumeSobelZ<tVolume> SobelZ;

    // Tetrahedron edge
    vctl::MCVector3D TetraEdge(p1, p2);
    TetraEdge.Normalization();

    // Value at the beginning of the edge
    mds::img::CDVolume::tVoxel LastValue = *It;

/*    MDS_LOG_NOTE("  " << LastValue
        << ", x = " << It.getX()
        << ", y = " << It.getY()
        << ", z = " << It.getZ()
        );*/

    ++It;

    // Check if the edge intersects any image edge
    for( ; !It.isEnd(); ++It )
    {
        // Get current pixel value
        mds::img::CDVolume::tVoxel Value = *It;

/*        MDS_LOG_NOTE("  " << Value
            << ", x = " << It.getX()
            << ", y = " << It.getY()
            << ", z = " << It.getZ()
            );*/

        // Is there an image edge?
        if( mds::math::getAbs(LastValue - Value) > 500 )
//        if( Value > 0 )
        {
            // Control space derivatives
            double dGradX = SobelX.getResponse(*(pTetraMesh->m_spVolume), It.getX(), It.getY(), It.getZ());
            double dGradY = SobelY.getResponse(*(pTetraMesh->m_spVolume), It.getX(), It.getY(), It.getZ());
            double dGradZ = SobelZ.getResponse(*(pTetraMesh->m_spVolume), It.getX(), It.getY(), It.getZ());

            // Control space gradient
            vctl::MCVector3D VolumeEdgeNormal(dGradX, dGradY, dGradZ);

            // Convert to the real world :)
            pTetraMesh->conv2Volume(&VolumeEdgeNormal);
            VolumeEdgeNormal.Normalization();

            // Compare orientation of the image gradient and tetrahedron edge
            // - Dot product of both vectors.
            // - Value must be greater than a given threshold. Optimal case is
            //   that both vectors are parallel (volume edge and tetrahedron edge are
            //   ortogonal), then -1 or 1 is obtained.
            if( mds::math::getAbs(VolumeEdgeNormal * TetraEdge) > dOrientThreshold )
            {
/*                MDS_LOG_NOTE("  Image edge found: orient = " << mds::math::getAbs(VolumeEdgeNormal * TetraEdge)
                    << ", delta = " << mds::math::getAbs(LastValue - Value)
                    );*/
                return false;
            }
        }

        LastValue = Value;
    }

//    MDS_LOG_NOTE("  -");

    // Both points are visible
    return true;
}


bool CTetraMesh::checkBoundaryFacetFunc(vctl::MCPoint3D *p1,
                                        vctl::MCPoint3D *p2,
                                        vctl::MCPoint3D *p3,
                                        CDelaunayTetra* pMesh
                                        )
{
    static const double dBoundaryThreshold = 0.1;

    MDS_CHECK(p1 && p2 && p3 && pMesh, return false);

    // Get pointer to the mesh
    CTetraMesh *pTetraMesh = (CTetraMesh *)pMesh;

    // Check boundary triangle
    return pTetraMesh->checkBoundaryTriangle(p1, p2, p3, dBoundaryThreshold);
}


bool CTetraMesh::tetrahedronSplitting(CTetraHomogeneityMeasure& Measure)
{
    // Logging
    MDS_LOG_NOTE("CTetraMesh::tetrahedronSplitting(): begin");

    // Compute homogeneity of all tetrahedra
    vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
    while( pTetrahedron )
    {
        // Get tetrahedron features
        CTetraFeatures *pFeatures = getTetrahedronFeatures(pTetrahedron);

        // Compute and set the tetrahedron homogeneity
        if( !pFeatures->isError() )
        {
            pTetrahedron->SetValue(Measure.getHomogeneity(*pFeatures));
        }
        else
        {
//            pTetrahedron->SetValue(0.0);
            pTetrahedron->SetValue(1.0);
        }

#ifdef MDS_SEG_LOGGING_ENABLED
        MDS_LOG_NOTE("CTetraMesh::tetrahedronSplitting(): Tetrahedron");
        MDS_LOG_NOTE("     Variance: " << pFeatures->getVoxelVariance());
        MDS_LOG_NOTE("  Homogeneity: " << pTetrahedron->GetValue());
#endif // MDS_SEG_LOGGING_ENABLED

        // Get next one
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Sort all tetrahedra
    m_Tetrahedra.SortAll();

    // Get sorted queue
    vctl::MCTetraS::MT_QUEUE *pQueue = m_Tetrahedra.GetQueue();

    // Number of splitted tetrahedra
    int iCount = 0;

    // Set the homogeneity measure
    m_pHomogeneityMeasure = &Measure;

    // Iterate while the container of edges is not empty
    for( ;; )
    {
        // Get pointer to the last item
        pTetrahedron = pQueue->GetMin();

        // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
        MDS_LOG_NOTE("CTetraMesh::tetrahedronSplitting(): Tetrahedron");
        MDS_LOG_NOTE("  Homogeneity: " << pTetrahedron->GetValue());
#endif // MDS_SEG_LOGGING_ENABLED
        MDS_LOG_NOTE("  Homogeneity: " << pTetrahedron->GetValue());

        // Check the tetrahedron homogeneity
        if( !pTetrahedron || pTetrahedron->GetValue() > Measure.getThreshold() )
        {
            // No optimization is needed
            break;
        }

        // Check the tetrahedron edge length
        if( pTetrahedron->GetValue() > 0.0 /*|| pTetrahedron->GetMaxEdgeLength() > 2.0 * m_dMinEdgeLength*/ )
        {
            // Split the tetrahedron
            if( !splitTetrahedron(pTetrahedron) )
            {
                // Unsort the tetrahedron, no optimization
                try
                {
                    pQueue->UnSort(pTetrahedron->GetValue(), pTetrahedron);
                }
                catch( ... )
                {
                }
                pTetrahedron->SetValue(0.0);
            }
            else
            {
                MDS_LOG_NOTE("CTetraMesh::tetrahedronSplitting(): Tetrahedron splitted");
                ++iCount;
            }
        }
        else
        {
            // Unsort the tetrahedron
            try
            {
                pQueue->UnSort(pTetrahedron->GetValue(), pTetrahedron);
            }
            catch( ... )
            {
            }
            pTetrahedron->SetValue(0.0);
        }
    }

    // Clear/unsort the queue
    pQueue->Clear();

    // Clear the homogeneity measure
    m_pHomogeneityMeasure = NULL;

    // Logging
    MDS_LOG_NOTE("CTetraMesh::tetrahedronSplitting(): finished");
    MDS_LOG_NOTE("CTetraMesh::tetrahedronSplitting(): Number of splitted tetrahedra: " << iCount);

    // O.K.
    return true;
}


bool CTetraMesh::splitTetrahedron(vctl::MCTetra *pTetrahedron)
{
    MDS_ASSERT(pTetrahedron);

    // Get tetrahedron vertices
    vctl::MCVertex *pVertices[4];
    pTetrahedron->GetVerticeS(pVertices);

    // Compute square of the minimal edge length
    double dThreshold = m_dMinEdgeLength * m_dMinEdgeLength;

    // Create tetrahedron voxels iterator
    CTetraIterator It(m_spEdgeVolume, pTetrahedron, this);

    // Helper variables
    mds::img::CDVolume::tVoxel Max = 0;
    double dMaxX = 0, dMaxY = 0, dMaxZ = 0;

    // Traverse all voxels and find best position for a new vertex
    for( ; !It.isEnd(); ++It )
    {
        // Get current voxel value
        mds::img::CDVolume::tVoxel Value = *It;

        // Is there an volume edge?
        if( Value > Max )
        {
            // Splitting point
            vctl::MCPoint3D Point(It.getRealX(), It.getRealY(), It.getRealZ());

            // Check size of produced edges
            double dLength = pVertices[0]->SquareDistance(Point);
            if( dLength < dThreshold )
            {
                continue;
            }
            dLength = pVertices[1]->SquareDistance(Point);
            if( dLength < dThreshold )
            {
                continue;
            }
            dLength = pVertices[2]->SquareDistance(Point);
            if( dLength < dThreshold )
            {
                continue;
            }
            dLength = pVertices[3]->SquareDistance(Point);
            if( dLength < dThreshold )
            {
                continue;
            }

            // Save the current position
            Max = Value;
            dMaxX = Point.GetX();
            dMaxY = Point.GetY();
            dMaxZ = Point.GetZ();
        }
    }

    // Check the maximum
    if( Max > 0 )
    {
        // Check the boudary
        vctl::MCPoint3D Point(dMaxX, dMaxY, dMaxZ);
        int iBoundaryFlags = checkBoundary(&Point);

        // Create a new vertex
        vctl::MCVertex *pVertex = newVertex(&Point);

        // Insert vertex to the mesh
        if( !addVertex(pVertex,
                       iBoundaryFlags,
                       tetrahedronSplittingInsertFunc,
                       tetrahedronSplittingRemoveFunc
                       ) )
        {
            // Remove the vertex
            if( !pVertex->GetRegisteredTetra() )
            {
                try
                {
                    m_pVertices->Erase(pVertex);
                }
                catch( ... )
                {
                }
            }

            // Cannot insert a new vertex
            return false;
        }
        else
        {
            // Tetrahedron has been splitted
            return true;
        }
    }

    // No splitting
    return false;
}


void CTetraMesh::tetrahedronSplittingInsertFunc(vctl::MCTetra *pTetrahedron, CDelaunayTetra *pMesh)
{
    MDS_CHECK(pTetrahedron && pMesh, return);

    // Get pointer to the mesh
    CTetraMesh *pTetraMesh = (CTetraMesh *)pMesh;

    // Check if the measure is set
    if( !pTetraMesh->m_pHomogeneityMeasure )
    {
        pTetrahedron->SetValue(0.0);
        return;
    }

    // Get pointer to the tetrahedron features
    CTetraFeatures *pFeatures = getTetrahedronFeatures(pTetrahedron);

    // Feature extractor
    CTetraFeaturesExtractor FeatureExtraction;

    // Acumulate features from just a one triangle
    FeatureExtraction.acumulate(pTetraMesh->m_spVolume, pTetrahedron, pTetraMesh);

    // Compute the feature vector
    if( FeatureExtraction.evaluate(pFeatures, false) )
    {
        // Compute and set the tetrahedron quality
        pTetrahedron->SetValue(pTetraMesh->m_pHomogeneityMeasure->getHomogeneity(*pFeatures));

        // Get sorted queue
        vctl::MCTetraS::MT_QUEUE *pQueue = pTetraMesh->m_Tetrahedra.GetQueue();

        // Sort the tetrahedron
        pQueue->Sort(pTetrahedron->GetValue(), pTetrahedron);
    }
    else
    {
        pTetrahedron->SetValue(0.0);
    }
}


void CTetraMesh::tetrahedronSplittingRemoveFunc(vctl::MCTetra *pTetrahedron, CDelaunayTetra *pMesh)
{
    MDS_CHECK(pTetrahedron && pMesh, return);

    // Get pointer to the mesh
    CTetraMesh *pTetraMesh = (CTetraMesh *)pMesh;

    // Get sorted queue
    vctl::MCTetraS::MT_QUEUE *pQueue = pTetraMesh->m_Tetrahedra.GetQueue();

    // Unsort the tetrahedron
    if( pTetrahedron->GetValue() > 0.0 )
    {
        pQueue->UnSort(pTetrahedron->GetValue(), pTetrahedron);
    }
}


bool CTetraMesh::checkTetrahedronVolume(vctl::MCTetra *pTetra)
{
    if( !pTetra )
    {
        return false;
    }

    // Convert an element (cube of size 1) to the volume coordinates
    vctl::MCVector3D Cube(1.0, 1.0, 1.0);
    conv2Volume(&Cube);

    // Volume of the cube
    double dCubeVolume = Cube.GetX() * Cube.GetY() * Cube.GetZ();

    // TODO: It's not neccessary to compute parameters of the conversion
    // every time.

    // Convert the tetrahedron volume
    double dTetraVolume = pTetra->GetVolume() * dCubeVolume;

    MDS_LOG_NOTE("Tetrahedron volume = " << dTetraVolume);

    // Check the volume
    return (dTetraVolume > MIN_TETRAHEDRON_VOLUME);
}


bool CTetraMesh::regionGrowing(CTetraSimilarityMeasure& Measure, double dBThreshold)
{
    // Added 2009/12/11
    if( !markBoundaryVertices() )
    {
        return false;
    }

    // Region and tetrahedron features
    CTetraFeaturesExtractor RegionExtractor, TetraExtractor;
    CTetraFeatures RegionFeatures, TetraFeatures;

    // Logging
    MDS_LOG_NOTE("CTetraMesh::regionGrowing()");

    // Initialize index of the region
    int iRegionIndex = m_iNumOfSegments;

    // Evaluate volume of all tetrahedra
    vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
    while( pTetrahedron  )
    {
        // Evaluate volume
        pTetrahedron->SetValue(pTetrahedron->GetVolume());

        // Get the next tetra
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Sort all tetrahedra according to the volume
    m_Tetrahedra.SortAll();

    // Get sorted queue
    vctl::MCTetraS::MT_QUEUE *pQueue = m_Tetrahedra.GetQueue();

    // Iterate while the queue is not empty
    for( ;; )
    {
        // Get pointer to the first item
        pTetrahedron = pQueue->GetMax();
//        pTetrahedron = pQueue->GetMin();
        if( !pTetrahedron )
        {
            break;
        }

        // Unsort the tetrahedron
        try
        {
            pQueue->UnSort(pTetrahedron->GetValue(), pTetrahedron);
        }
        catch( ... )
        {
        }
        pTetrahedron->SetValue(0.0);

        // Check the tetrahedron region
        if( getTetrahedronRegion(pTetrahedron) >= 0 )
        {
            continue;
        }

        // Create a new region
        ++iRegionIndex;

        // Clear extracted region features
        RegionExtractor.clear();
        RegionExtractor.acumulate(m_spVolume, pTetrahedron, this);
        RegionExtractor.evaluate(&RegionFeatures, false);
        if( RegionFeatures.isError() )
        {
            continue;
        }

        // Change region of the tetrahedron
        setTetrahedronRegion(pTetrahedron, iRegionIndex - 1);

        // Queue of tetrahedra
        typedef std::deque<vctl::MCTetra *> tQueue;
        tQueue Queue;

        // Grow the region
        Queue.push_back(pTetrahedron);
        while( !Queue.empty() )
        {
            // Get pointer to the first element in the queue
            vctl::MCTetra *pT = Queue.front();
            Queue.pop_front();

            // Get neighbours
            vctl::MCTetra *pNeighbours[4];
            pT->GetNeighbours(pNeighbours);

            // Get regions of all neighbours
            for( int i = 0; i < 4; ++i )
            {
                if( pNeighbours[i] && getTetrahedronRegion(pNeighbours[i]) == -1 )
                {
                    // Get the tetrahedron face
                    vctl::MCVertex *pV1, *pV2, *pV3;
                    pT->GetTri(i, &pV1, &pV2, &pV3);

                    // Check that there is no image edge
//                    if( checkBoundaryTriangle(pV1, pV2, pV3, dBThreshold) )
                    // Modified 2009/12/11
                    if( dBThreshold > 0.0 && isBoundaryTriangle(pV1, pV2, pV3) )
                    {
                        continue;
                    }

                    // Evaluate tetrahedron features
                    TetraExtractor.clear();
                    TetraExtractor.acumulate(m_spVolume, pNeighbours[i], this);
                    TetraExtractor.evaluate(&TetraFeatures, false);
                    if( TetraFeatures.isError() )
                    {
                        continue;
                    }

                    // Check similarity
                    if( Measure.checkSimilarity(RegionFeatures, TetraFeatures) )
                    {
                        RegionExtractor.acumulate(TetraExtractor);
                        RegionExtractor.evaluate(&RegionFeatures, false);

                        setTetrahedronRegion(pNeighbours[i], iRegionIndex - 1);
                        Queue.push_back(pNeighbours[i]);
                    }
                }
            }
        }
    }

    // Clear/unsort the queue
    pQueue->Clear();

    // Set the number of regions in the mesh
    m_iNumOfSegments = iRegionIndex;

    // Logging
    MDS_LOG_NOTE("  Final number of regions = " << m_iNumOfSegments);

    // O.K.
    return true;
}


// Faster RAG
bool CTetraMesh::regionMerging(CTetraSimilarityMeasure& Measure,
                               double dBThreshold,
                               double dDeltaThreshold,
                               tSize NumOfRegions,
                               bool bMergeAdjacent
                               )
{
//    MDS_CHECK(m_iNumOfSegments > 0, return false);

    // Added 2009/12/11
    if( !markBoundaryVertices() )
    {
        return false;
    }

    // Logging
    MDS_LOG_NOTE("CTetraMesh::merging()");

    // Assign unique regions to all non-classified tetrahedra
    if( bMergeAdjacent )
    {
        vctl::MCTetra *pT = getFirstTetrahedron();
        while( pT )
        {
            // Get tetrahedron region
            if( getTetrahedronRegion(pT) == -1 && checkTetrahedronVolume(pT) )
            {
                setTetrahedronRegion(pT, m_iNumOfSegments++);
            }

            // Get the next triangle
            pT = pT->GetNext();
        }
    }

    // Shake down used indexes of regions
    reassignRegions();

    // Logging
    MDS_LOG_NOTE("  Number of regions = " << m_iNumOfSegments);

    // Create the RAG (Region Adjacency Graph)
    typedef mds::img::CAdjacencyGraph<double> tRAG;
    tRAG RAG(m_iNumOfSegments);

    // Initial mapping of regions
    mds::math::CVector<int> Mapping(m_iNumOfSegments);
    for( int i = 0; i < m_iNumOfSegments; ++i )
    {
        Mapping(i) = i;
    }

    // Image feature extractors
    typedef std::vector<CTetraFeaturesExtractor> tExtractors;
    tExtractors Extractors(m_iNumOfSegments);

    // Evaluate image features of all regions and create the RAG graph
    vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
    while( pTetrahedron )
    {
        // Get tetrahedron region
        int iRegion = getTetrahedronRegion(pTetrahedron);

        // If the tetrahedron is classified
        if( iRegion >= 0 && iRegion < m_iNumOfSegments )
        {
            // Acumulate image features
            Extractors[iRegion].acumulate(m_spVolume, pTetrahedron, this);

            // Initialize the RAG
            if( bMergeAdjacent )
            {
                // Get tetrahedron neighbours
                vctl::MCTetra *pNeighbours[4];
                pTetrahedron->GetNeighbours(pNeighbours);

                // Get regions of all neighbours and fill the RAG
                for( int i = 0; i < 4; ++i )
                {
                    if( pNeighbours[i] )
                    {
                        int iRegion2 = getTetrahedronRegion(pNeighbours[i]);
                        if( iRegion2 >= 0 && iRegion2 < m_iNumOfSegments && iRegion2 != iRegion )
                        {
                            // Check if the RAG is already filled or not
                            if( RAG.get(iRegion, iRegion2, -1.0) >= 0.0 )
                            {
                                continue;
                            }

                            // Get the shared triangle
                            vctl::MCVertex *pV1, *pV2, *pV3;
                            pTetrahedron->GetTri(i, &pV1, &pV2, &pV3);

                            // Check that there is no image edge
//                            if( checkBoundaryTriangle(pV1, pV2, pV3, dBThreshold) )
                            // Modified 2009/12/11
                            if( dBThreshold > 0.0 && isBoundaryTriangle(pV1, pV2, pV3) )
                            {
                                continue;
                            }

                            // Set the RAG
                            RAG.set(iRegion, iRegion2, 0.0);
                        }
                    }
                }
            }
        }

        // Get the next tetrahedron
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Initialize the RAG
    if( !bMergeAdjacent )
    {
        // Any region can be merged with all others
        for( int j = 0; j < m_iNumOfSegments; ++j )
        {
            for( int i = j + 1; i < m_iNumOfSegments; ++i )
            {
                RAG.set(i, j, 0.0);
            }
        }
    }

    // Evaluate image features of all regions
    int iCount = m_iNumOfSegments;
    m_Segments.resize(m_iNumOfSegments);
    for( int r = 0; r < m_iNumOfSegments; ++r )
    {
        Extractors[r].evaluate(&(m_Segments[r]), false);
//        Extractors[r].evaluate(&(m_Segments[r]), true);
        if( m_Segments[r].isError() )
        {
            RAG.clear(r);
            Mapping(r) = -1;
            --iCount;
        }
        else
        {
            // Logging
//            MDS_LOG(m_Segments[r].getVoxelMean() << " " << m_Segments[r].getVoxelVariance() << "," << std::endl);
        }
    }

    // Logging
    MDS_LOG_NOTE("  Number of correctly evaluated regions = " << iCount);

    // Evaluate similarity of all regions
//    int iNumOfPossibleMerges = 0;
    for( tSize j = 0; j < RAG.getNumOfRegions(); ++j )
    {
        tRAG::tIterator itEnd = RAG.getEnd(j);
        tRAG::tIterator it = RAG.getBegin(j);
        for( ; it != itEnd; ++it )
        {
            // Get region index and value
            tSize i = RAG.getRegion(it);
            double dValue = RAG.getValue(it);

            // Evaluate the similarity
            if( dValue <= 0.0 )
            {
                RAG.set(j, i, Measure.getSimilarity(m_Segments[j], m_Segments[i]));
//                ++iNumOfPossibleMerges;
            }
        }
    }

    // Minimal number of regions
//    iNumOfPossibleMerges /= 10;
    int iNumOfPossibleMerges = iCount / 10;
    int iMinCount = (iNumOfPossibleMerges > 5) ? iCount - iNumOfPossibleMerges : iCount;

    // Helper mapping of regions
    bool bRestoreMapping = false;
    mds::math::CVector<int> SavedMapping(Mapping);

    // While any two regions can be merged
    double dMinSimilarity = 1.0e6;
    for( ; iCount > NumOfRegions; )
    {
        // Find maximum similarity
        tSize iMax1, iMax2;
        double dMaxSimilarity = RAG.getMax(iMax1, iMax2, -1.0);

        // Logging
        MDS_LOG_NOTE("  Maximal similarity = " << dMaxSimilarity << ", RAG(" << iMax1 << ", " << iMax2 << ")");

        // Check the maximal similarity
//        if( dMaxSimilarity <= 0.0 )
        if( dMaxSimilarity <= Measure.getThreshold() )
        {
            break;
        }

        // Check the termination criterion
        if( NumOfRegions == -1 && iCount <= iMinCount )
        {
            double dDelta = dMaxSimilarity / dMinSimilarity;
            MDS_LOG_NOTE("  Delta = " << dDelta);
            if( dDelta > dDeltaThreshold )
//            if( dDelta > 1.5 )
//            if( dDelta > 2.0 )
//            if( dDelta > 5.0 )
//            if( dDelta > 10.0 )
            {
                bRestoreMapping = true;
                break;
            }
        }

        // Save the minimal similarity
        dMinSimilarity = mds::math::getMin(dMaxSimilarity, dMinSimilarity);

//        if( NumOfRegions == -1 && dMaxSimilarity < Measure.getThreshold() )
//        {
//            break;
//        }

        // Choose the larger region
        if( m_Segments[iMax1].getNumOfVoxels() < m_Segments[iMax2].getNumOfVoxels() )
        {
            mds::math::swap(iMax1, iMax2);
        }

        // Update the RAG
        RAG.merge(iMax2, iMax1);

        // Update feature vectors
        Extractors[iMax1].acumulate(Extractors[iMax2]);
        Extractors[iMax1].evaluate(&(m_Segments[iMax1]), false);

        // Re-compute similarity of the regions
        tRAG::tIterator itEnd = RAG.getEnd(iMax1);
        tRAG::tIterator it = RAG.getBegin(iMax1);
        for( ; it != itEnd; ++it )
        {
            // Get region
            tSize i = RAG.getRegion(it);

            // Evaluate the similarity
            RAG.set(iMax1, i, Measure.getSimilarity(m_Segments[iMax1], m_Segments[i]));
        }

        SavedMapping = Mapping;

        // Update the mapping of regions
        for( int m = 0; m < m_iNumOfSegments; ++m )
        {
            if( Mapping(m) == iMax2 )
            {
                Mapping(m) = iMax1;
            }
        }

        // Decrease the number of regions
        --iCount;
    }

    if( bRestoreMapping )
    {
        Mapping = SavedMapping;
    }

    // Re-assign all tetrahedra
    pTetrahedron = getFirstTetrahedron();
    while( pTetrahedron )
    {
        // Get tetrahedron region
        int iRegion = getTetrahedronRegion(pTetrahedron);

        // Assign index of the new region
        if( iRegion >= 0 && iRegion < m_iNumOfSegments )
        {
            // Set tetrahedron region
            setTetrahedronRegion(pTetrahedron, Mapping(iRegion));
        }

        // Get the next tetrahedron
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Shake down used indexes of regions
    reassignRegions();

    // Logging
    MDS_LOG_NOTE("  Final number of regions = " << m_iNumOfSegments);

    // O.K.
    return true;
}


// Slow RAG
// (see SVN history before 2009/11/30)
/*bool CTetraMesh::regionMerging(CTetraSimilarityMeasure& Measure,
                               double dBThreshold,
                               tSize NumOfRegions,
                               bool bMergeAdjacent
                               )*/


void CTetraMesh::findIsolatedRegions()
{
    MDS_CHECK(m_iNumOfSegments > 0, return);

    // Logging
    MDS_LOG_NOTE("CTetraMesh::findIsolatedRegions():");
    MDS_LOG_NOTE("  Initial number of regions = " << m_iNumOfSegments);

    // Clear all tetrahedron flags
    clearFlags();

    // Queue of tetrahedra
    typedef std::deque<vctl::MCTetra *> tQueue;
    tQueue Queue;

    // Clear the number of regions
    m_iNumOfSegments = 0;

    // Traverse all tetrahedra in the mesh
    vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
    while( pTetrahedron )
    {
        // Get tetrahedron region
        int iTetraRegion = getTetrahedronRegion(pTetrahedron);

        // If the tetrahedron is not yet classified or marked as already processed
        if( iTetraRegion == -1 || pTetrahedron->TestFlag(vctl::ENTITY_MARK_FLAG) )
        {
            // Get the next tetrahedron
            pTetrahedron = pTetrahedron->GetNext();
            continue;
        }

        // Clear the container of tetrahedra
        Queue.clear();

        // Increment the number of regions
        ++m_iNumOfSegments;

        // Try to find an isolated region (growing from the seed)
        Queue.push_back(pTetrahedron);
        while( !Queue.empty() )
        {
            // Get pointer to the first element in the queue
            vctl::MCTetra *pT = Queue.front();
            Queue.pop_front();

            // Mark the tetrahedron
            if( pT->TestFlag(vctl::ENTITY_MARK_FLAG) )
            {
                continue;
            }

            // Mark the tetrahedron
            pT->SetFlag(vctl::ENTITY_MARK_FLAG);

            // Set tetrahedron region
            setTetrahedronRegion(pT, m_iNumOfSegments - 1);

            // Get neighbours
            vctl::MCTetra *pNeighbours[4];
            pT->GetNeighbours(pNeighbours);

            // Check all neighbours
            for( int i = 0; i < 4; ++i )
            {
                if( pNeighbours[i] && getTetrahedronRegion(pNeighbours[i]) == iTetraRegion )
                {
                    Queue.push_back(pNeighbours[i]);
                }
            }
        }

        // Get the next tetrahedron
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Logging
    MDS_LOG_NOTE("CTetraMesh::findIsolatedRegions():");
    MDS_LOG_NOTE("  Final number of regions = " << m_iNumOfSegments);

    // Clear all tetrahedron flags
    clearFlags();
}


void CTetraMesh::reassignRegions()
{
    // Check the number of regions
    if( m_iNumOfSegments <= 0 )
    {
        return;
    }

    // Mapping of regions
    mds::math::CVector<int> Mapping(m_iNumOfSegments);
    Mapping.fill(-1);

    // Re-assign all tetrahedra
    vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
    int iNewIndex = 0;
    while( pTetrahedron )
    {
        // Get tetrahedron region
        int iRegion = getTetrahedronRegion(pTetrahedron);

        // Assign index of the new region
        if( iRegion >= 0 && iRegion < m_iNumOfSegments )
        {
            // Mapping of the region
            if( Mapping(iRegion) == -1 )
            {
                Mapping(iRegion) = iNewIndex++;
            }

            // Set tetrahedron region
            setTetrahedronRegion(pTetrahedron, Mapping(iRegion));
        }
        else if( iRegion != -1 )
        {
            // Clear tetrahedron region
            setTetrahedronRegion(pTetrahedron, -1);

            // Logging
            MDS_LOG_NOTE("CTriMesh::reassignRegions(): Bad region index!");
        }

        // Get the next triangle
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Final number of regions
    m_iNumOfSegments = iNewIndex;
}


void CTetraMesh::changeRegion(int iMinIndex, int iMaxIndex, int iFinalRegion)
{
    MDS_CHECK(m_iNumOfSegments > 0, return);
    MDS_CHECK(iFinalRegion >= -1 && iFinalRegion < m_iNumOfSegments, return);

    // Check the maximal region index
    if( iMaxIndex == -1 )
    {
        iMaxIndex = iMinIndex;
    }

    // Traverse all tetrahedra in the mesh
    vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
    while( pTetrahedron )
    {
        // Get tetrahedron region
        int iTetraRegion = getTetrahedronRegion(pTetrahedron);

        // Override the classification
        if( iTetraRegion >= iMinIndex && iTetraRegion <= iMaxIndex )
        {
            setTetrahedronRegion(pTetrahedron, iFinalRegion);
        }

        // Get the next tetrahedron
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Re-assign region indexes
    reassignRegions();
}


bool CTetraMesh::classifyAdjacent(double dBThreshold)
{
    MDS_CHECK(m_iNumOfSegments > 0, return false);

    // Added 2009/12/11
    if( !markBoundaryVertices() )
    {
        return false;
    }

    // Logging
    MDS_LOG_NOTE("CTetraMesh::classifyAdjacent():");
    MDS_LOG_NOTE("  Number of regions = " << m_iNumOfSegments);

    // List of all non-classified tetrahedra
    typedef std::vector<vctl::MCTetra *> tTetrahedra;
    tTetrahedra NonClassifiedTetrahedra;

    // Create the list
    vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
    while( pTetrahedron )
    {
        // If the tetrahedron is not yet classified
        if( getTetrahedronRegion(pTetrahedron) == -1 )
        {
            NonClassifiedTetrahedra.push_back(pTetrahedron);
        }

        // Get the next tetrahedron
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Logging
    MDS_LOG_NOTE("  Number of non-classified tetrahedra = " << (int)NonClassifiedTetrahedra.size());

    // Classify all tetrahedra with respect to its neighbours
    int iNumOfChanges = 1;
    for( int iters = 0; iters < 100 && iNumOfChanges > 0; ++iters )
    {
        // Reset the number of re-assigned tetrahedra
        iNumOfChanges = 0;

        // Traverse all tetrahedra
        tTetrahedra::iterator itEnd = NonClassifiedTetrahedra.end();
        tTetrahedra::iterator it = NonClassifiedTetrahedra.begin();
        for( ; it != itEnd; ++it )
        {
            // Get neighbours of the tetrahedron
            vctl::MCTetra *pNeighbours[4];
            (*it)->GetNeighbours(pNeighbours);

            // Get regions
            int iRegions[4]= { -1, -1 , -1, -1 };
            double dAreas[4] = { 0.0, 0.0, 0.0, 0.0 };
            int iNumOfNeighbours = 0;
            for( int i = 0; i < 4; ++i )
            {
                if( pNeighbours[i] )
                {
                    // Get the region
                    int iRegion = getTetrahedronRegion(pNeighbours[i]);
                    if( iRegion != -1 )
                    {
                        // Get the shared triangle
                        vctl::MCVertex *pV1, *pV2, *pV3;
                        (*it)->GetTri(i, &pV1, &pV2, &pV3);

                        // Check that there is no image edge
//                        if( checkBoundaryTriangle(pV1, pV2, pV3, dBThreshold) )
                        // Modified 2009/12/11
                        if( dBThreshold > 0.0 && isBoundaryTriangle(pV1, pV2, pV3) )
                        {
                            continue;
                        }

                        // Count the number of neighbours
                        ++iNumOfNeighbours;

                        // Compute area of the triangle
                        vctl::MCTri Tri(pV1, pV2, pV3);
                        double dTetraFaceArea = Tri.GetArea();

                        // Find corresponding region
                        if( iRegion == iRegions[0] )
                        {
                            dAreas[0] += dTetraFaceArea;
                        }
                        else if( iRegion == iRegions[1] )
                        {
                            dAreas[1] += dTetraFaceArea;
                        }
                        else if( iRegion == iRegions[2] )
                        {
                            dAreas[2] += dTetraFaceArea;
                        }
                        else if( iRegion == iRegions[3] )
                        {
                            dAreas[3] += dTetraFaceArea;
                        }
                        else
                        {
                            // New region
                            iRegions[i] = iRegion;
                            dAreas[i] = dTetraFaceArea;
                        }
                    }
                }
            }

            // Index of the region
            int iNewRegion = -1;

            // Evaluate the region index
//            if( iNumOfNeighbours >= 2 )
            if( iNumOfNeighbours >= 1 )
            {
                // Find maximum shared area
                double dMaxArea = 0.0;
                for( int j = 0; j < 4; ++j )
                {
                    if( dAreas[j] > dMaxArea )
                    {
                        dMaxArea = dAreas[j];
                        iNewRegion = iRegions[j];
                    }
                }
            }

            // Re-assign the tetrahedron
            if( iNewRegion != -1 && iNewRegion != getTetrahedronRegion(*it) )
            {
                setTetrahedronRegion(*it, iNewRegion);
                ++iNumOfChanges;
            }
        }

        // Logging
        MDS_LOG_NOTE("  Number of changes = " << iNumOfChanges);
    }

    // O.K.
    return true;
}


bool CTetraMesh::noiseReduction()
{
//    static const int MAX_ITERS = 10;
    static const int MAX_ITERS = 2;
    static const double MULTIPLIER = 2.0;

    MDS_CHECK(m_iNumOfSegments > 0, return false);

    // Number of re-classified tetrahedra
    int iNumOfChanges = 1;

    // Iterate
    for( int iter = 0; iNumOfChanges > 0 && iter < MAX_ITERS; ++iter )
    {
        // Clear the number of re-classified tetrahedra
        iNumOfChanges = 0;

        // Traverse all tetrahedra in the mesh
        vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
        while( pTetrahedron )
        {
            // Get tetrahedron segment
            int iTetraRegion = getTetrahedronRegion(pTetrahedron);

            // Get its neighbours
            vctl::MCTetra *pNeighbours[4];
    //        pTetrahedron->GetNeighboursByVertices(pNeighbours);
            pTetrahedron->GetNeighbours(pNeighbours);

            // NOTE: First version
            // Get indexes of segments of all neighbours
    /*        int iRegions[4];
            for( int i = 0; i < 4; ++i )
            {
                if( pNeighbours[i] )
                {
                    iRegions[i] = getTetrahedronRegion(pNeighbours[i]);
                }
                else
                {
                    iRegions[i] = -1;
                }
            }

            // Check if tetrahedron is connected to its segment
            if( iRegions[0] == iTetraRegion
                    || iRegions[1] == iTetraRegion
                    || iRegions[2] == iTetraRegion
                    || iRegions[3] == iTetraRegion )
            {
                // Continue with the next tetrahedron
                pTetrahedron = pTetrahedron->GetNext();
                continue;
            }

            // Determine new segment
            int iNewRegion = -1;
            if( iRegions[0] == iRegions[1] )
            {
                if( iRegions[0] == iRegions[2] || iRegions[0] == iRegions[3] )
                {
                    iNewRegion = iRegions[0];
                }
            }
            if( iRegions[2] == iRegions[3] )
            {
                if( iRegions[2] == iRegions[0] || iRegions[2] == iRegions[1] )
                {
                    iNewRegion = iRegions[2];
                }
            }

            // Re-classify the current tetrahedron
            if( iNewRegion != -1 )
            {
                setTetrahedronRegion(pTetrahedron, iNewRegion);
                ++iNumOfChanges;

                // Logging
                MDS_LOG_NOTE("Re-classified tetrahedron " << pTetrahedron);
            }*/

            // Check that all neighbours are assigned to the same segment
    /*        if( iRegions[0] == iRegions[1]
                && iRegions[1] == iRegions[2]
                && iRegions[2] == iRegions[3] )
            {
                // Re-classify the current tetrahedron
                if( iTetraRegion != iRegions[0] )
                {
                    setTetrahedronRegion(pTetrahedron, iRegions[0]);
                    ++iNumOfChanges;

                    // Logging
                    MDS_LOG_NOTE("Re-classified tetrahedron " << pTetrahedron);
                }
            }*/

            // NOTE: Second version
            // Get regions
            int iRegions[4]= { -1, -1 , -1, -1 };
            double dAreas[4] = { 0.0, 0.0, 0.0, 0.0 };
            int iNumOfNeighbours = 0;
            for( int i = 0; i < 4; ++i )
            {
                if( pNeighbours[i] )
                {
                    // Get the region
                    int iRegion = getTetrahedronRegion(pNeighbours[i]);
                    if( iRegion != -1 )
                    {
                        // Get the shared triangle
                        vctl::MCVertex *pV1, *pV2, *pV3;
                        pTetrahedron->GetTri(i, &pV1, &pV2, &pV3);

                        // Count the number of neighbours
                        ++iNumOfNeighbours;

                        // Compute area of the triangle
                        vctl::MCTri Tri(pV1, pV2, pV3);
                        double dTetraFaceArea = Tri.GetArea();

                        // Find corresponding region
                        if( iRegion == iRegions[0] )
                        {
                            dAreas[0] += dTetraFaceArea;
                        }
                        else if( iRegion == iRegions[1] )
                        {
                            dAreas[1] += dTetraFaceArea;
                        }
                        else if( iRegion == iRegions[2] )
                        {
                            dAreas[2] += dTetraFaceArea;
                        }
                        else if( iRegion == iRegions[3] )
                        {
                            dAreas[3] += dTetraFaceArea;
                        }
                        else
                        {
                            // New region
                            iRegions[i] = iRegion;
                            dAreas[i] = dTetraFaceArea;
                        }
                    }
                }
            }

            // Index of the region
            int iNewRegion = -1;
            double dMaxArea = 0.0;

            // Evaluate the region index
            if( iNumOfNeighbours >= 3 )
            {
                // Find maximum shared area
                for( int j = 0; j < 4; ++j )
                {
                    if( dAreas[j] > dMaxArea )
                    {
                        dMaxArea = dAreas[j];
                        iNewRegion = iRegions[j];
                    }
                }
            }

            // Check the maximum value
            if( iTetraRegion != -1 )
            {
                // Find the area
                double dArea = 0.0;
                if( iTetraRegion == iRegions[0] )
                {
                    dArea = dAreas[0];
                }
                else if( iTetraRegion == iRegions[1] )
                {
                    dArea = dAreas[1];
                }
                else if( iTetraRegion == iRegions[2] )
                {
                    dArea = dAreas[2];
                }
                else if( iTetraRegion == iRegions[3] )
                {
                    dArea = dAreas[3];
                }

                // Compare the values
                if( dMaxArea < MULTIPLIER * dArea )
                {
                    iNewRegion = -1;
                }
            }

            // Re-assign the tetrahedron
            if( iNewRegion != -1 && iNewRegion != iTetraRegion )
            {
                setTetrahedronRegion(pTetrahedron, iNewRegion);
                ++iNumOfChanges;
            }

            // Get next tetrahedron
            pTetrahedron = pTetrahedron->GetNext();
        }

        // Logging
        MDS_LOG_NOTE("CTetraMesh::noiseReduction(): ");
        MDS_LOG_NOTE("  Iteration = " << iter);
        MDS_LOG_NOTE("  Total number of re-classified tetrahedra = " << iNumOfChanges);
    }

    // O.K.
    return true;
}


// Old version
// (see SVN history before 2009/11/30)
//bool CTetraMesh::checkBoundaryTriangle(vctl::MCTri *pTriangle, double dThreshold)


int CTetraMesh::evalEdgeDistancePoint(vctl::MCPoint3D *pPoint,
                                      mds::img::CPoint3D& Normal,
                                      int iMaxDistance
                                      )
{
    static const int PENALTY = 3;
    static const tVoxel EDGE_THRESHOLD = 50;

    // Volume coordinates
    tSize x, y, z;
    conv2Volume(pPoint, x, y, z);

    // Minimal distance to an image edge
    int iDistance = 0;

    // Find an image edge
    if( m_spEdgeVolume->get(x, y, z) <= EDGE_THRESHOLD )
    {
        // Pixel neighbours
        mds::img::CPoint3D Left(x, y, z);
        mds::img::CPoint3D Right(x, y, z);

        // Evaluate minimal distance to an image edge
        iDistance = 1;
        for( ; iDistance < iMaxDistance; ++iDistance )
        {
            // Step along the normal vector
            Left -= Normal;
            Right += Normal;

            // Interpolate voxel value
            tVoxel LeftValue = m_spEdgeVolume->interpolate(Left);
            if( LeftValue > EDGE_THRESHOLD )
            {
                break;
            }

            // Interpolate voxel value
            tVoxel RightValue = m_spEdgeVolume->interpolate(Right);
            if( RightValue > EDGE_THRESHOLD )
            {
                break;
            }
        }
    }

    // No edge found
    if( iDistance == iMaxDistance )
    {
//        iDistance = iMaxDistance + 1;
        iDistance = PENALTY * iMaxDistance;
    }

    return iDistance;
}


// new "approximate" version
bool CTetraMesh::checkBoundaryTriangle(vctl::MCPoint3D *p0,
                                       vctl::MCPoint3D *p1,
                                       vctl::MCPoint3D *p2,
                                       double dThreshold
                                       )
{
    static const int MAX_DISTANCE = 3;
    static const double THIRD = 1.0 / 3;
    static const double NORM = 1.0 / (10 * (MAX_DISTANCE + 1));

    MDS_CHECK(p0 && p1 && p2, return true);

    // Normal vector of the triangle
    vctl::MCVector3D v01(p0, p1);
    vctl::MCVector3D v02(p0, p2);
    vctl::MCVector3D TriNormal(v01 % v02);

    // Convert it to the volume coordinates :)
    conv2Real(&TriNormal);
    TriNormal.Normalization();
    mds::img::CPoint3D Normal(TriNormal.GetX(), TriNormal.GetY(), TriNormal.GetZ());

    // Helper variables
    double dDistSum = 0.0;

    // Evaluate distance at the vertices
    dDistSum += evalEdgeDistancePoint(p0, Normal, MAX_DISTANCE);
    dDistSum += evalEdgeDistancePoint(p1, Normal, MAX_DISTANCE);
    dDistSum += evalEdgeDistancePoint(p2, Normal, MAX_DISTANCE);

    // Evaluate distance in the middle of edges
    vctl::MCPoint3D m0(0.5 * (p0->GetX() + p1->GetX()),
                       0.5 * (p0->GetY() + p1->GetY()),
                       0.5 * (p0->GetZ() + p1->GetZ())
                       );
    vctl::MCPoint3D m1(0.5 * (p1->GetX() + p2->GetX()),
                       0.5 * (p1->GetY() + p2->GetY()),
                       0.5 * (p1->GetZ() + p2->GetZ())
                       );
    vctl::MCPoint3D m2(0.5 * (p2->GetX() + p0->GetX()),
                       0.5 * (p2->GetY() + p0->GetY()),
                       0.5 * (p2->GetZ() + p0->GetZ())
                       );
    dDistSum += evalEdgeDistancePoint(&m0, Normal, MAX_DISTANCE);
    dDistSum += evalEdgeDistancePoint(&m1, Normal, MAX_DISTANCE);
    dDistSum += evalEdgeDistancePoint(&m2, Normal, MAX_DISTANCE);

    // Evaluate distance in the middle of edge centers
    vctl::MCPoint3D c0(0.5 * (m0.GetX() + m1.GetX()),
                       0.5 * (m0.GetY() + m1.GetY()),
                       0.5 * (m0.GetZ() + m1.GetZ())
                       );
    vctl::MCPoint3D c1(0.5 * (m1.GetX() + m2.GetX()),
                       0.5 * (m1.GetY() + m2.GetY()),
                       0.5 * (m1.GetZ() + m2.GetZ())
                       );
    vctl::MCPoint3D c2(0.5 * (m2.GetX() + m0.GetX()),
                       0.5 * (m2.GetY() + m0.GetY()),
                       0.5 * (m2.GetZ() + m0.GetZ())
                       );
    dDistSum += evalEdgeDistancePoint(&c0, Normal, MAX_DISTANCE);
    dDistSum += evalEdgeDistancePoint(&c1, Normal, MAX_DISTANCE);
    dDistSum += evalEdgeDistancePoint(&c2, Normal, MAX_DISTANCE);

    // Evaluate distance in the middle of the facet
    vctl::MCPoint3D Center(THIRD * (p0->GetX() + p1->GetX() + p2->GetX()),
                           THIRD * (p0->GetY() + p1->GetY() + p2->GetY()),
                           THIRD * (p0->GetZ() + p1->GetZ() + p2->GetZ())
                           );
    dDistSum += evalEdgeDistancePoint(&Center, Normal, MAX_DISTANCE);

    // Evaluate if the triangle lies on an image edge (region boundary)
    double dRatio = dDistSum * NORM;
    return (dRatio < dThreshold);
}


bool CTetraMesh::checkBoundaryEdge(vctl::MCEdge *pEdge,
                                   vctl::MCVector3D *pNormal,
                                   double dThreshold
                                   )
{
    static const int MAX_DISTANCE = 3;
    static const int MAX_DIFFERENCE = 1;
    static const tVoxel EDGE_THRESHOLD = 0;

    MDS_CHECK(pEdge && pNormal, return true);

    // Sobel filters
    mds::img::CVolumeSobelX<tVolume> SobelX;
    mds::img::CVolumeSobelY<tVolume> SobelY;
    mds::img::CVolumeSobelZ<tVolume> SobelZ;

    // Vertices of the edge
    vctl::MCVertex *pVertices[2];
    pEdge->GetVerticeS(pVertices);

    // Triangle edge
//    vctl::MCVector3D TriEdge(pVertices[0], pVertices[1]);
//    TriEdge.Normalization();

    // Triangle normal
    mds::img::CPoint3D Normal(mds::img::tCoordinate(pNormal->GetX()),
                              mds::img::tCoordinate(pNormal->GetY()),
                              mds::img::tCoordinate(pNormal->GetZ())
                              );

    // Create edge pixels iterator
    CTetraEdgeIterator it(m_spEdgeVolume, pEdge, this);

    // Helper variables
    double dDistSum = 0.0;
    double dOrientSum = 0.0;

    // Traverse all pixels and check if there is some parallel volume edge
    int iCount = 0;
    int iLastDist = 0;
    for( ; !it.isEnd(); ++it, ++iCount )
    {
        // Current position
        tSize x = it.getX();
        tSize y = it.getY();
        tSize z = it.getZ();

        // Minimal distance to an image edge
        int iDistance = 0;

        // Find an image edge
        if( *it <= EDGE_THRESHOLD )
        {
            // Pixel neighbours
            mds::img::CPoint3D Left(x, y, z);
            mds::img::CPoint3D Right(x, y, z);

            // Evaluate minimal distance to an image edge
            iDistance = 1;
            for( ; iDistance < MAX_DISTANCE; ++iDistance )
            {
                // Step along the normal vector
                Left -= Normal;
                Right += Normal;

                // Interpolate voxel value
                tVoxel LeftValue = m_spEdgeVolume->interpolate(Left);
                if( LeftValue > EDGE_THRESHOLD )
                {
                    x = mds::math::round2Int(Left.x());
                    y = mds::math::round2Int(Left.y());
                    z = mds::math::round2Int(Left.z());
                    break;
                }

                // Interpolate voxel value
                tVoxel RightValue = m_spEdgeVolume->interpolate(Right);
                if( RightValue > EDGE_THRESHOLD )
                {
                    x = mds::math::round2Int(Right.x());
                    y = mds::math::round2Int(Right.y());
                    z = mds::math::round2Int(Right.z());
                    break;
                }
            }
        }

        // Check the 1st derivation
        if( mds::math::getAbs(iLastDist - iDistance) > MAX_DIFFERENCE )
        {
            return false;
        }
        iLastDist = iDistance;

        // No edge found
        if( iDistance == MAX_DISTANCE )
        {
            dDistSum += MAX_DISTANCE + 1;
            continue;
        }

        // Image derivatives
        double dGradX = SobelX.getResponse(*m_spVolume, x, y, z);
        double dGradY = SobelY.getResponse(*m_spVolume, x, y, z);
        double dGradZ = SobelZ.getResponse(*m_spVolume, x, y, z);

        // Image edge gradient
        vctl::MCVector3D ImageEdgeNormal(dGradX, dGradY, dGradZ);
        ImageEdgeNormal.Normalization();

        // Compare orientation of the image edge normal and triangle edge
        // - Compute dot product of both vectors
//        double dOrient = ImageEdgeNormal * TriEdge;
        double dOrient = *pNormal * ImageEdgeNormal;

        // Cumulate weighted distances
//        dDistSum += (iDistance + 1) * (1.0 - mds::math::getAbs(dOrient));
//        dDistSum += (iDistance + 1) * mds::math::getAbs(dOrient);
        dDistSum += iDistance + 1;
        dOrientSum += 1.0 - mds::math::getAbs(dOrient);
    }

    // Check the distance at the end
    if( mds::math::getAbs(iLastDist) > MAX_DIFFERENCE )
    {
        return false;
    }

    // Evaluate if the edge lies on an image edge (region boundary)
    if( iCount > 0 )
    {
        double dRatio = dDistSum / double(iCount);
        double dOrientRatio = dOrientSum / double(iCount);

//        return (dRatio < dThreshold);
        return (dRatio < dThreshold) && (dOrientRatio < 0.2);
    }

    // There is boundary edge
    return true;
}


void CTetraMesh::createTriangles(int iRegion, bool bBoundary)
{
    // Get pointer to the list of triangles
    vctl::MCTriS *pTriangles = m_Tetrahedra.GetTriS();

    // Clear flags
    clearFlags();

    // Traverse all tetrahedra in the mesh
    vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
    while (pTetrahedron)
    {
        // Get tetrahedron segment
        int iTetraRegion = getTetrahedronRegion(pTetrahedron);

        // Check the segment
        if( iRegion == -1 || iRegion == iTetraRegion )
        {
            // Create triangles
            for( int i = 0; i < 4; ++i )
            {
                // Get neighbouring tetrahedron
//                vctl::MCTetra *pNeighbour = pTetrahedron->GetNeighbour(i);
                vctl::MCTetra *pNeighbour = pTetrahedron->GetNeighbourByVertices(i);

                // Check if the triangle forms boundary between two segments
                if( bBoundary && iRegion != -1 )
                {
                    // Check neighbouring tetrahedron
                    if( pNeighbour )
                    {
                        // Get segment of the neighbour
                        int iNeighbourRegion = getTetrahedronRegion(pNeighbour);

                        // Compare both indexes
                        if( iTetraRegion == iNeighbourRegion )
                        {
                            continue;
                        }
                    }
                }

                // Check if the triangle already exists
                if( pNeighbour && pTetrahedron->TestFlag(vctl::ENTITY_MARK_FLAG) )
                {
                    continue;
                }

                // Get triangle vertices
                vctl::MCVertex *pVertices[3];
                pTetrahedron->GetTri(i, pVertices);

                // Create a new triangle
//                pTriangles->New(pVertices[0], pVertices[1], pVertices[2]);
                pTriangles->New(pVertices[0], pVertices[2], pVertices[1]);
            }

            // Mark the tetrahedron
            pTetrahedron->SetFlag(vctl::ENTITY_MARK_FLAG);
        }

        // Ziskani nasledujiciho tetra site
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Clear flags
    clearFlags();
}


void CTetraMesh::createTriangles(double dA, double dB, double dC, double dD, int iRegion)
{
    // Get pointer to the list of triangles
    vctl::MCTriS *pTriangles = m_Tetrahedra.GetTriS();

    // Clear flags
    clearFlags();

    // Traverse all tetrahedra in the mesh
    vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
    for( ; pTetrahedron; pTetrahedron = pTetrahedron->GetNext() )
    {
        // Get tetrahedron segment
        int iTetraRegion = getTetrahedronRegion(pTetrahedron);

        // Check the segment
        if( iRegion != -1 && iRegion != iTetraRegion )
        {
            continue;
        }

        // Get tetrahedron vertices
        vctl::MCVertex *pVertices[4];
        pTetrahedron->GetVerticeS(pVertices);

        // Check position of all vertices with respect to the plane
        int iPositive = 0, iNegative = 0;
        for( int i = 0; i < 4; ++i )
        {
            double dValue = dA * pVertices[i]->GetX();
            dValue += dB * pVertices[i]->GetY();
            dValue += dC * pVertices[i]->GetZ();
            dValue += dD;
            if( dValue > 0.0 )
            {
                ++iPositive;
            }
            else if( dValue < 0.0 )
            {
                --iNegative;
            }
        }

        // Check the tetrahedron position
//        if( iPositive > 3 )
        if( iPositive > 2 )
        {
            // Generate tetrahedron faces
            for( int j = 0; j < 4; ++j )
            {
                // Get neighbouring tetrahedron
//                vctl::MCTetra *pNeighbour = pTetrahedron->GetNeighbour(j);
                vctl::MCTetra *pNeighbour = pTetrahedron->GetNeighbourByVertices(j);

                // Check if the triangle already exists
                if( pNeighbour && pTetrahedron->TestFlag(vctl::ENTITY_MARK_FLAG) )
                {
                    continue;
                }

                // Get triangle vertices
                vctl::MCVertex *pVertices[3];
                pTetrahedron->GetTri(j, pVertices);

                // Create a new triangle
//                pTriangles->New(pVertices[0], pVertices[1], pVertices[2]);
                pTriangles->New(pVertices[0], pVertices[2], pVertices[1]);
            }

            // Mark the tetrahedron
            pTetrahedron->SetFlag(vctl::ENTITY_MARK_FLAG);
        }
    }

    // Clear flags
    clearFlags();
}


void CTetraMesh::createTriangles(double dQuality, int iRegion)
{
    // Get pointer to the list of triangles
    vctl::MCTriS *pTriangles = m_Tetrahedra.GetTriS();

    // Clear flags
    clearFlags();

    // Traverse all tetrahedra in the mesh
    vctl::MCTetra *pTetrahedron = getFirstTetrahedron();
    for( ; pTetrahedron; pTetrahedron = pTetrahedron->GetNext() )
    {
        // Get tetrahedron segment
        int iTetraRegion = getTetrahedronRegion(pTetrahedron);

        // Check the segment
        if( iRegion != -1 && iRegion != iTetraRegion )
        {
            continue;
        }

        // Check tetrahedron quality
        if( getTetrahedronQuality(pTetrahedron) > dQuality )
        {
            // Generate tetrahedron faces
            for( int j = 0; j < 4; ++j )
            {
                // Get neighbouring tetrahedron
//                vctl::MCTetra *pNeighbour = pTetrahedron->GetNeighbour(j);
                vctl::MCTetra *pNeighbour = pTetrahedron->GetNeighbourByVertices(j);

                // Check if the triangle already exists
                if( pNeighbour && pTetrahedron->TestFlag(vctl::ENTITY_MARK_FLAG) )
                {
                    continue;
                }

                // Get triangle vertices
                vctl::MCVertex *pVertices[3];
                pTetrahedron->GetTri(j, pVertices);

                // Create a new triangle
//                pTriangles->New(pVertices[0], pVertices[1], pVertices[2]);
                pTriangles->New(pVertices[0], pVertices[2], pVertices[1]);
            }

            // Mark the tetrahedron
            pTetrahedron->SetFlag(vctl::ENTITY_MARK_FLAG);
        }
    }

    // Clear flags
    clearFlags();
}


void CTetraMesh::clearTriangles()
{
    // Get pointer to the list of triangles
    vctl::MCTriS *pTriangles = m_Tetrahedra.GetTriS();

    // Clear the list
    pTriangles->ClearAll();
}


//==============================================================================
/*
 * Implementation of the class CTetraMesh.
 * - Saving to the STL and VRML format.
 * - Modified Premysl Krsek's original code.
 */

//! Structure containing vertex coordinates
struct SFloatCoordinates
{
    float x, y, z;
};


bool CTetraMesh::saveSTL(mds::mod::CChannel& Channel)
{
    // Hlavicka stl souboru
    char pcHeader[90];
    mds::base::memSet<char>(pcHeader, 0, 85);

    // Nastaveni textu hlavicky stl souboru
    strcpy(pcHeader, "Binary STL file created by MDSTk");

    // Ulozeni hlavicky stl souboru
    if( !Channel.write(pcHeader, 80) )
    {
        return false;
    }

    // Get the number of triangles
    unsigned long ulNumOfTriangles = (unsigned long )getNumOfTriangles();

    // Ulozeni poctu tri site
    if( !Channel.write((char *)&ulNumOfTriangles, sizeof(unsigned long)) )
    {
        return false;
    }

    // STL atribut trianglu musi byt 0
    unsigned short int usAttribute = 0;

    // Cyklus ukladani
    vctl::MCTri *pTriangle = getFirstTriangle();
    while (pTriangle)
    {
        // Ziskani vrcholu
        vctl::MCVertex *pTriangleNode[3];
        pTriangle->GetVerticeS(pTriangleNode);

        // Ziskani normaly
        vctl::MCVector3D Normal = pTriangle->GetNormal();
        SFloatCoordinates Temp;
        Temp.x = (float)Normal.GetX();
        Temp.y = (float)Normal.GetY();
        Temp.z = (float)Normal.GetZ();

        // Zapis normaly
        if( !Channel.write((char *)&Temp, sizeof(SFloatCoordinates)) )
        {
            return false;
        }

        // Cyklus uzlu aktualniho tri
        for (int i = 0; i < 3; ++i)
        {
            // Ziskani souradnic node tri
            Temp.x = (float)pTriangleNode[i]->GetX();
            Temp.y = (float)pTriangleNode[i]->GetY();
            Temp.z = (float)pTriangleNode[i]->GetZ();

            // Zapis souradnic node
            if( !Channel.write((char *)&Temp, sizeof(SFloatCoordinates)) )
            {
                return false;
            }
        }

        // Zapis stl atributu tri
        if( !Channel.write((char *)&usAttribute, sizeof(unsigned short int)) )
        {
            return false;
        }

        // Ziskani nasledujiciho tri site
        pTriangle = pTriangle->GetNext();
    }

    // O.K.
    return true;
}


bool CTetraMesh::saveVRML(mds::mod::CChannel& Channel)
{
    // Vystupni proud
    std::stringstream Stream;

    // Ziskani aktualniho casu
    time_t lTime;
    time(&lTime);

    // Zapsani hlavicky vrml souboru
    Stream << "#VRML V2.0 utf8" << std::endl << std::endl;
    Stream << "# Created by MDSTk." << std::endl;
    Stream << "# Date: " << ctime(&lTime) << std::endl << std::endl;

    // Zapis casti Transformace do vrml souboru
    Stream << "Transform {" << std::endl << "children [" << std::endl;

    // Zapis tvaru
    Stream << "Shape {" << std::endl;

    // Zapis materialovych vlastnosti
    Stream << "appearance Appearance { material Material { diffuseColor 0.9882 0.9882 0.9882 } }" << std::endl;

    // Zapis geometrie
    Stream << "geometry IndexedFaceSet {" << std::endl;

    // Zapis parametru site
//    Stream << "ccw TRUE" << std::endl << "solid TRUE" << std::endl << "creaseAngle 1.5708" << std::endl;
    Stream << "ccw TRUE" << std::endl << "solid FALSE" << std::endl << "creaseAngle 1.5708" << std::endl;

    // Zapsani bodu site
    Stream << "coord Coordinate { point [" << std::endl;

    // Cyklus uzlu site pro jejich ulozeni do VRML
    vctl::MCVertex *pTriangleNode = getFirstVertex();
    unsigned uIndex = 0;
    while (pTriangleNode)
    {
        // Nastaveni cisla aktualniho node
        pTriangleNode->SetIndex(uIndex);
        ++uIndex;

        // Zapsani souradnic node do souboru
        Stream << pTriangleNode->GetX() << " "
            << pTriangleNode->GetY() << " "
            << pTriangleNode->GetZ() << ","
            << std::endl;

        // Ziskani nasledujiciho node
        pTriangleNode = pTriangleNode->GetNext();
    }

    // Uzavreni casti bodu site
    Stream << "] }" << std::endl;

    // Zapsani tri site
    Stream << "coordIndex [" << std::endl;

    // Cyklus tri site pro jejich ulozeni do VRML
    vctl::MCTri *pTriangle = getFirstTriangle();
    while (pTriangle)
    {
        // Zapsani indexu node tri do souboru
        Stream << pTriangle->GetVertex(0)->GetIndex() << ", "
            << pTriangle->GetVertex(1)->GetIndex() << ", "
            << pTriangle->GetVertex(2)->GetIndex() << ", -1,"
            << std::endl;

        // Ziskani nasledujiciho tri
        pTriangle = pTriangle->GetNext();
    }

    // Uzavreni casti tri site
    Stream << "]" << std::endl;

    // Uzavreni casti geometrie
    Stream << "}" << std::endl;

    // Uzavreni casti tvaru
    Stream << "}" << std::endl;

    // Uzavreni casti transformace
    Stream << "] }" << std::endl;

    // Zapis obsahu string streamu do kanalu
    if( !Channel.write(Stream.str().c_str(), Stream.str().length()) )
    {
        return false;
    }

    // O.K
    return true;
}


} // namespace seg
} // namespace mds

