//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2006/12/21                          \n
 *
 * $Id: mdsTetraSimilarityMeasure.cpp 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Tetrahedra similarity measures.
 */

#include <MDSTk/VectorSegmentation/mdsTetraSimilarityMeasure.h>


namespace mds
{
namespace seg
{

//==============================================================================
/*
 * Implementation of the class CTetraMeanValueSimilarity.
 */

const double CTetraMeanValueSimilarity::DEFAULT_THRESHOLD     = 0.1;
const double CTetraMeanValueSimilarity::DEFAULT_RHO           = 500.0;


CTetraMeanValueSimilarity::CTetraMeanValueSimilarity(double dThreshold, double dRho)
    : CTetraSimilarityMeasure(dThreshold)
{
    MDS_ASSERT(dRho > 0.0);
    
    m_dInvRho = -0.5 / (dRho * dRho);
}


double CTetraMeanValueSimilarity::getSimilarity(const CTetraFeatures& v1, const CTetraFeatures& v2)
{
//    static const int MAX_NUM = 4096 * 4096;
//    int iCount = (v1.getNumOfVoxels() < v2.getNumOfVoxels()) ? v1.getNumOfVoxels() : v2.getNumOfVoxels();
/*    double dCount = (v1.getNumOfVoxels() < v2.getNumOfVoxels()) ? v1.getNumOfVoxels() : v2.getNumOfVoxels();
    if( dCount > 0.0 )
    {
        dCount = 1.0 / dCount;
    }*/
    
    double dTemp = v1.getVoxelMean() - v2.getVoxelMean();
//    return (MAX_NUM - iCount) * std::exp(m_dInvRho * dTemp * dTemp);
//    return dCount * std::exp(m_dInvRho * dTemp * dTemp);
    return std::exp(m_dInvRho * dTemp * dTemp);
}


//==============================================================================
/*
 * Implementation of the class CTetraVarianceSimilarity.
 */

const double CTetraVarianceSimilarity::DEFAULT_THRESHOLD     = 0.8;


CTetraVarianceSimilarity::CTetraVarianceSimilarity(double dThreshold)
    : CTetraSimilarityMeasure(dThreshold)
{
}


double CTetraVarianceSimilarity::getSimilarity(const CTetraFeatures& v1, const CTetraFeatures& v2)
{
    int N1 = v1.getNumOfVoxels();
    int N2 = v2.getNumOfVoxels();
    int N = N1 + N2;
    if( N == 0 )
    {
        return 0.0;
    }
    
    double dInvN = 1.0 / N;
    double dM0 = (N1 * v1.getVoxelMean() + N2 * v2.getVoxelMean()) * dInvN;
    
    double dTemp = dM0 - v1.getVoxelMean();
    double dV10 = v1.getVoxelVariance() + dTemp * dTemp;
    
    dTemp = dM0 - v2.getVoxelMean();
    double dV20 = v2.getVoxelVariance() + dTemp * dTemp;
    
    double dV0 = (N1 * dV10 + N2 * dV20) * dInvN;

//    double dValue = v1.getVoxelVariance() * v2.getVoxelVariance();
//    return (dValue + 0.0001) / (dV0 * dV0 + 0.0001);

    double dValue = std::sqrt(v1.getVoxelVariance()) * std::sqrt(v2.getVoxelVariance());
//    return (dValue + 0.0001) / (dV0 * dV0 + 0.0001);
    return (dValue + 0.0001) / (dV0 + 0.0001);
}


//==============================================================================
/*
 * Implementation of the class CTetraMixtureSimilarity.
 */

CTetraMixtureSimilarity::CTetraMixtureSimilarity(CTetraSimilarityMeasure& M1,
                                                 CTetraSimilarityMeasure& M2,
                                                 double dW1,
                                                 double dW2
                                                 )
    : CTetraSimilarityMeasure(dW1 * M1.getThreshold() + dW2 * M2.getThreshold())
    , m_Measure1(M1)
    , m_Measure2(M2)
    , m_dW1(dW1)
    , m_dW2(dW2)
{
}


double CTetraMixtureSimilarity::getSimilarity(const CTetraFeatures& v1, const CTetraFeatures& v2)
{
    double dValue1 = m_dW1 * m_Measure1.getSimilarity(v1, v2);
    double dValue2 = m_dW2 * m_Measure2.getSimilarity(v1, v2);

    // Weighted sum of both similarities
    return dValue1 + dValue2;
}


bool CTetraMixtureSimilarity::checkSimilarity(const CTetraFeatures& v1, const CTetraFeatures& v2)
{
    double dValue1 = m_dW1 * m_Measure1.getSimilarity(v1, v2);
    double dValue2 = m_dW2 * m_Measure2.getSimilarity(v1, v2);

    // Weighted sum of both similarities
    return (dValue1 + dValue2) > m_dThreshold;
}


} // namespace seg
} // namespace mds

