//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2005/10/17                          \n
 *
 * $Id: mdsTriFeatures.cpp 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Triangle feature vector.
 */

#include <MDSTk/VectorSegmentation/mdsTriIterator.h>
#include <MDSTk/VectorSegmentation/mdsTriFeatures.h>

// MDSTk
#include <MDSTk/Base/mdsGlobalLog.h>
#include <MDSTk/Image/mdsImageFiltering.h>
#include <MDSTk/Math/mdsBase.h>
#include <MDSTk/Math/mdsVectorFunctions.h>

// STL
#include <cmath>


namespace mds
{
namespace seg
{

//==============================================================================
/*
 * Global functions.
 */

//! Safe computation of the log(x) function.
double getLog(double x)
{
    static const double LOG_OF_ZERO = -1.0e10;
    static const double LOG_SMALL = -0.5e10;

    if( x == 0.0 )
    {
        return LOG_OF_ZERO;
    }

    double dResult = std::log(x);

    if( dResult < LOG_SMALL )
    {
        return LOG_OF_ZERO;
    }

    return dResult;
}


//==============================================================================
/*
 * Implementation of the class CTriFeaturesExtractor.
 */
void CTriFeaturesExtractor::acumulate(tImage *pImage, vctl::MCTri *pTriangle)
{
    MDS_ASSERT(pImage && pTriangle);

    // Create triangle pixels iterator
    CTriIterator it(pImage, pTriangle);

    // Traverse all pixels
    for( ; !it.isEnd(); ++it )
    {
        // Current pixel value
        double dValue = double(*it);

        // Count the number of pixels
        ++m_iNumOfPixels;

        // BASIC TRIANGLE FEATURES
        m_dSum += dValue;
        m_dSumSqr += dValue * dValue;

#ifdef USE_LOCAL_MOMENTS
        // Filters used to compute local moments
        static mds::img::CImageFilter<tImage, mds::img::IF_LOCAL_MOMENT_00> Moment00;
        static mds::img::CImageFilter<tImage, mds::img::IF_LOCAL_MOMENT_10> Moment10;
        static mds::img::CImageFilter<tImage, mds::img::IF_LOCAL_MOMENT_01> Moment01;
        static mds::img::CImageFilter<tImage, mds::img::IF_LOCAL_MOMENT_11> Moment11;
        static mds::img::CImageFilter<tImage, mds::img::IF_LOCAL_MOMENT_20> Moment20;
        static mds::img::CImageFilter<tImage, mds::img::IF_LOCAL_MOMENT_02> Moment02;
        static mds::img::CImageFilter<tImage, mds::img::IF_LOCAL_MOMENT_22> Moment22;
        
        // Evaluate local moments        
        dValue = tFeature(Moment00.getResponse(*pImage, it.getX(), it.getY()));
        m_Moments(0) = mds::math::getMin(m_Moments(0), dValue);
        m_Moments(1) += dValue;
        m_Moments(2) = mds::math::getMax(m_Moments(2), dValue);

        dValue = tFeature(Moment01.getResponse(*pImage, it.getX(), it.getY()));
        m_Moments(3) = mds::math::getMin(m_Moments(3), dValue);
        m_Moments(4) += dValue;
        m_Moments(5) = mds::math::getMax(m_Moments(5), dValue);

        dValue = tFeature(Moment10.getResponse(*pImage, it.getX(), it.getY()));
        m_Moments(6) = mds::math::getMin(m_Moments(6), dValue);
        m_Moments(7) += dValue;
        m_Moments(8) = mds::math::getMax(m_Moments(8), dValue);

        dValue = tFeature(Moment11.getResponse(*pImage, it.getX(), it.getY()));
        m_Moments(9) = mds::math::getMin(m_Moments(9), dValue);
        m_Moments(10) += dValue;
        m_Moments(11) = mds::math::getMax(m_Moments(11), dValue);

        dValue = tFeature(Moment20.getResponse(*pImage, it.getX(), it.getY()));
        m_Moments(12) = mds::math::getMin(m_Moments(12), dValue);
        m_Moments(13) += dValue;
        m_Moments(14) = mds::math::getMax(m_Moments(14), dValue);

        dValue = tFeature(Moment02.getResponse(*pImage, it.getX(), it.getY()));
        m_Moments(15) = mds::math::getMin(m_Moments(15), dValue);
        m_Moments(16) += dValue;
        m_Moments(17) = mds::math::getMax(m_Moments(17), dValue);

        dValue = tFeature(Moment22.getResponse(*pImage, it.getX(), it.getY()));
        m_Moments(18) = mds::math::getMin(m_Moments(18), dValue);
        m_Moments(19) += dValue;
        m_Moments(20) = mds::math::getMax(m_Moments(20), dValue);
#endif // USE_LOCAL_MOMENTS

#ifdef USE_HARALICK_FEATURES
        // Predefined constants
        static const tSize MIN_LEVEL = 0;
        static const tSize MAX_LEVEL = TriFeatures::NUM_OF_LEVELS - 1;

        // Compute normalization factor
        static double dMin = mds::img::CPixelTraits<tPixel>::getPixelMin();
        static double dMax = mds::img::CPixelTraits<tPixel>::getPixelMax();
        static double dNorm = (MAX_LEVEL - MIN_LEVEL) / (dMax - dMin);

        // Pixel values
        tPixel Pixel0 = *it;
        tPixel Pixel1 = pImage->get(it.getX() + 1, it.getY());
        tPixel Pixel2 = pImage->get(it.getX(), it.getY() + 1);
        tPixel Pixel3 = pImage->get(it.getX() + 1, it.getY() + 1);
        tPixel Pixel4 = pImage->get(it.getX() + 1, it.getY() - 1);

        // Normalized pixel values
        // - round2Int(dNorm * (Pixel - dMin)) + MIN_LEVEL
        // - dMin = 0 , MIN_LEVEL = 0
        tSize Value0 = mds::math::round2Int(dNorm * Pixel0);
        tSize Value1 = mds::math::round2Int(dNorm * Pixel1);
        tSize Value2 = mds::math::round2Int(dNorm * Pixel2);
        tSize Value3 = mds::math::round2Int(dNorm * Pixel3);
        tSize Value4 = mds::math::round2Int(dNorm * Pixel4);

        // Fill co-occurrency matrixes
        m_Matrix1(Value0, Value1) += 1;
        m_Matrix2(Value0, Value2) += 1;
        m_Matrix3(Value0, Value3) += 1;
        m_Matrix4(Value0, Value4) += 1;       
#endif // USE_HARALICK_FEATURES
    }
}


void CTriFeaturesExtractor::acumulate(const CTriFeaturesExtractor& Extractor)
{
    m_dSum += Extractor.m_dSum;
    m_dSumSqr += Extractor.m_dSumSqr;
    m_iNumOfPixels += Extractor.m_iNumOfPixels;

#ifdef USE_LOCAL_MOMENTS
    m_Moments(0) = mds::math::getMin(m_Moments(0), Extractor.m_Moments(0));
    m_Moments(1) += Extractor.m_Moments(1);
    m_Moments(2) = mds::math::getMax(m_Moments(2), Extractor.m_Moments(2));

    m_Moments(3) = mds::math::getMin(m_Moments(3), Extractor.m_Moments(3));
    m_Moments(4) += Extractor.m_Moments(4);
    m_Moments(5) = mds::math::getMax(m_Moments(5), Extractor.m_Moments(5));

    m_Moments(6) = mds::math::getMin(m_Moments(6), Extractor.m_Moments(6));
    m_Moments(7) += Extractor.m_Moments(7);
    m_Moments(8) = mds::math::getMax(m_Moments(8), Extractor.m_Moments(8));

    m_Moments(9) = mds::math::getMin(m_Moments(9), Extractor.m_Moments(9));
    m_Moments(10) += Extractor.m_Moments(10);
    m_Moments(11) = mds::math::getMax(m_Moments(11), Extractor.m_Moments(11));

    m_Moments(12) = mds::math::getMin(m_Moments(12), Extractor.m_Moments(12));
    m_Moments(13) += Extractor.m_Moments(13);
    m_Moments(14) = mds::math::getMax(m_Moments(14), Extractor.m_Moments(14));

    m_Moments(15) = mds::math::getMin(m_Moments(15), Extractor.m_Moments(15));
    m_Moments(16) += Extractor.m_Moments(16);
    m_Moments(17) = mds::math::getMax(m_Moments(17), Extractor.m_Moments(17));

    m_Moments(18) = mds::math::getMin(m_Moments(18), Extractor.m_Moments(18));
    m_Moments(19) += Extractor.m_Moments(19);
    m_Moments(20) = mds::math::getMax(m_Moments(20), Extractor.m_Moments(20));
#endif // USE_LOCAL_MOMENTS

#ifdef USE_HARALICK_FEATURES
    m_Matrix1 += Extractor.m_Matrix1;
    m_Matrix2 += Extractor.m_Matrix2;
    m_Matrix3 += Extractor.m_Matrix3;
    m_Matrix4 += Extractor.m_Matrix4;
#endif // USE_HARALICK_FEATURES
}


bool CTriFeaturesExtractor::evaluate(tVector *pVector, bool bCheckNumOfPixels)
{
    MDS_CHECK(pVector, return false);

    // Clear the whole feature vector
    pVector->fill(0);

    // Set the number of pixels
    pVector->setNumOfPixels(m_iNumOfPixels);

    // Check the number of pixels
    double dInvCount = 1.0;
    int iLimit = (bCheckNumOfPixels) ? TriFeatures::MIN_NUM_OF_PIXELS : 0;
    if( m_iNumOfPixels > iLimit )
    {
        dInvCount /= m_iNumOfPixels;
    }
    else
    {
        // Set the error code
        pVector->setError(CTriFeatures::ERROR_CANNOT_EVALUATE);
        return false;
    }

    // BASIC FEATURES
    double dMean = m_dSum * dInvCount;
    double dVariance = m_dSumSqr * dInvCount - (dMean * dMean);
    pVector->setPixelMean(tFeature(dMean));
    pVector->setPixelVariance(tFeature(std::sqrt(dVariance)));

#ifdef USE_LOCAL_MOMENTS
    // Estimate mean local moments
    pVector->setLocalMoment(0, tFeature(m_Moments(0)));
    pVector->setLocalMoment(1, tFeature(m_Moments(1) * dInvCount));
    pVector->setLocalMoment(2, tFeature(m_Moments(2)));
    pVector->setLocalMoment(3, tFeature(m_Moments(3)));
    pVector->setLocalMoment(4, tFeature(m_Moments(4) * dInvCount));
    pVector->setLocalMoment(5, tFeature(m_Moments(5)));
    pVector->setLocalMoment(6, tFeature(m_Moments(6)));
    pVector->setLocalMoment(7, tFeature(m_Moments(7) * dInvCount));
    pVector->setLocalMoment(8, tFeature(m_Moments(8)));
    pVector->setLocalMoment(9, tFeature(m_Moments(9)));
    pVector->setLocalMoment(10, tFeature(m_Moments(10) * dInvCount));
    pVector->setLocalMoment(11, tFeature(m_Moments(11)));
    pVector->setLocalMoment(12, tFeature(m_Moments(12)));
    pVector->setLocalMoment(13, tFeature(m_Moments(13) * dInvCount));
    pVector->setLocalMoment(14, tFeature(m_Moments(14)));
    pVector->setLocalMoment(15, tFeature(m_Moments(15)));
    pVector->setLocalMoment(16, tFeature(m_Moments(16) * dInvCount));
    pVector->setLocalMoment(17, tFeature(m_Moments(17)));
    pVector->setLocalMoment(18, tFeature(m_Moments(18)));
    pVector->setLocalMoment(19, tFeature(m_Moments(19) * dInvCount));
    pVector->setLocalMoment(20, tFeature(m_Moments(20)));
#endif // USE_LOCAL_MOMENTS

#ifdef USE_HARALICK_FEATURES
    // Compute Haralick's texture features
    tSize i, j;
    double dH0 = 0.0, dH1 = 0.0, dH2 = 0.0, dH3 = 0.0;
    for( i = 0; i < TriFeatures::NUM_OF_LEVELS; ++i )
    {
        for( j = 0; j < TriFeatures::NUM_OF_LEVELS; ++j )
        {
            double dValue = dInvCount * m_Matrix1(i,j);
            dH0 += dValue * dValue;
            dH1 -= dValue * getLog(dValue);
            dH2 += (i - j) * (i - j) * dValue;
            dH3 += dValue / (1 + mds::math::getAbs(i - j));
        }
    }
    pVector->setHaralickFeature(0, tFeature(dH0));
    pVector->setHaralickFeature(1, tFeature(dH1));
    pVector->setHaralickFeature(2, tFeature(dH2));
    pVector->setHaralickFeature(3, tFeature(dH3));

    dH0 = 0.0, dH1 = 0.0, dH2 = 0.0, dH3 = 0.0;
    for( i = 0; i < TriFeatures::NUM_OF_LEVELS; ++i )
    {
        for( j = 0; j < TriFeatures::NUM_OF_LEVELS; ++j )
        {
            double dValue = dInvCount * m_Matrix2(i,j);
            dH0 += dValue * dValue;
            dH1 -= dValue * getLog(dValue);
            dH2 += (i - j) * (i - j) * dValue;
            dH3 += dValue / (1 + mds::math::getAbs(i - j));
        }
    }
    pVector->setHaralickFeature(4, tFeature(dH0));
    pVector->setHaralickFeature(5, tFeature(dH1));
    pVector->setHaralickFeature(6, tFeature(dH2));
    pVector->setHaralickFeature(7, tFeature(dH3));

    dH0 = 0.0, dH1 = 0.0, dH2 = 0.0, dH3 = 0.0;
    for( i = 0; i < TriFeatures::NUM_OF_LEVELS; ++i )
    {
        for( j = 0; j < TriFeatures::NUM_OF_LEVELS; ++j )
        {
            double dValue = dInvCount * m_Matrix3(i,j);
            dH0 += dValue * dValue;
            dH1 -= dValue * getLog(dValue);
            dH2 += (i - j) * (i - j) * dValue;
            dH3 += dValue / (1 + mds::math::getAbs(i - j));
        }
    }
    pVector->setHaralickFeature(8, tFeature(dH0));
    pVector->setHaralickFeature(9, tFeature(dH1));
    pVector->setHaralickFeature(10, tFeature(dH2));
    pVector->setHaralickFeature(11, tFeature(dH3));

    dH0 = 0.0, dH1 = 0.0, dH2 = 0.0, dH3 = 0.0;
    for( i = 0; i < TriFeatures::NUM_OF_LEVELS; ++i )
    {
        for( j = 0; j < TriFeatures::NUM_OF_LEVELS; ++j )
        {
            double dValue = dInvCount * m_Matrix4(i,j);
            dH0 += dValue * dValue;
            dH1 -= dValue * getLog(dValue);
            dH2 += (i - j) * (i - j) * dValue;
            dH3 += dValue / (1 + mds::math::getAbs(i - j));
        }
    }
    pVector->setHaralickFeature(12, tFeature(dH0));
    pVector->setHaralickFeature(13, tFeature(dH1));
    pVector->setHaralickFeature(14, tFeature(dH2));
    pVector->setHaralickFeature(15, tFeature(dH3));
#endif // USE_HARALICK_FEATURES

    // O.K.
    pVector->setError(0);
    return true;
}


void CTriFeaturesExtractor::clear()
{
    // BASIC TRIANGLE FEATURES
    m_dSum = 0.0;
    m_dSumSqr = 0.0;
    m_iNumOfPixels = 0;

#ifdef USE_LOCAL_MOMENTS
    m_Moments.fill(0.0);
#endif // USE_LOCAL_MOMENTS

#ifdef USE_HARALICK_FEATURES
    m_Matrix1.fill(0);
    m_Matrix2.fill(0);
    m_Matrix3.fill(0);
    m_Matrix4.fill(0);
#endif // USE_HARALICK_FEATURES
}


} // namespace seg
} // namespace mds

