//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2010 by Michal Spanel     \n
 *
 * Authors: Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2005/07/26                          \n
 *
 * $Id: mdsTriMesh.cpp 1619 2010-02-11 08:35:23Z spanel $
 *
 * Description:
 * - Adaptive triangular mesh based on 2D Delaunay triangulation.
 */

#include <MDSTk/VectorSegmentation/mdsTriMesh.h>
#include <MDSTk/VectorSegmentation/mdsTriIterator.h>
#include <MDSTk/VectorSegmentation/mdsTriEdgeIterator.h>
#include <MDSTk/VectorSegmentation/mdsTriVertexDetector.h>

// MDSTk
#include <MDSTk/Base/mdsMemory.h>
#include <MDSTk/Base/mdsGlobalLog.h>
#include <MDSTk/Math/mdsMatrixFunctions.h>
#include <MDSTk/Image/mdsLine.h>
#include <MDSTk/Image/mdsImageFiltering.h>

// STL
#include <string>
#include <sstream>
#include <ctime>
#include <vector>
#include <list>
#include <deque>


namespace mds
{
namespace seg
{

//==============================================================================
/*
 * Data members of the class CTriMesh.
 */
const tSize CTriMesh::DEFAULT_CELL                      = 5;
const tSize CTriMesh::DEFAULT_CELL_MARGIN               = 1;
const double CTriMesh::DEFAULT_INIT_THRESHOLD           = 0.5;
const double CTriMesh::DEFAULT_SPLITTING_THRESHOLD      = 0.5;
const double CTriMesh::DEFAULT_DISTANCE_THRESHOLD       = 1.5;


//==============================================================================
/*
 * Implementation of the class CTriMesh.
 */
CTriMesh::CTriMesh()
    : m_iNumOfSegments(0)
    , m_pHomogeneityMeasure(NULL)
{
}


CTriMesh::~CTriMesh()
{
}


bool CTriMesh::init(double dThreshold,
                    mds::tSize CellSize,
                    mds::tSize NumOfNodes,
                    double dMinEdgeLength
                    )
{
    // Image size and margin
    mds::tSize XSize = m_spEdgeImage->getXSize();
    mds::tSize YSize = m_spEdgeImage->getYSize();
    mds::tSize Margin = m_spEdgeImage->getMargin();

    // Check the image size
    MDS_CHECK(XSize > 0 && YSize > 0 && Margin > 0, return false);

    // Init the triangulation
//    vctl::MCPoint3D UpperLeft(-Margin, -Margin, 0.0);
//    vctl::MCPoint3D BottomRight(XSize + Margin, YSize + Margin, 0.0);
    vctl::MCPoint3D UpperLeft(0.0, 0.0, 0.0);
    vctl::MCPoint3D BottomRight(XSize - 1, YSize - 1, 0.0);
    if( !CDelaunayTri::init(&UpperLeft, &BottomRight, NumOfNodes) )
    {
        // Initialization has failed
        return false;
    }

    // Use vertex detector to prepare list of candidate vertices
    CTriVertexDetector<CELL_MAX> Detector(dThreshold, CellSize, DEFAULT_CELL_MARGIN);
    if( !Detector(m_spEdgeImage, m_pVertices) )
    {
        // Can't detect vertices
        return false;
    }

    // O.K.
    return true;
}


void CTriMesh::visualize(mds::img::CDImage *pImage)
{
    MDS_ASSERT(pImage);

    // Make all edges of the mesh
    m_Triangles.MakeAllTrisEdges();

    // Get pointer to the container
    vctl::MCEdgeS *pEdges = m_Triangles.GetEdgeS();

    // Visualize all edges
    vctl::MCEdge *pEdge = (vctl::MCEdge *)pEdges->GetFirst();
    while( pEdge )
    {
        // Get edge vertices
        vctl::MCVertex *pVertices[2];
        pEdge->GetVerticeS(pVertices);

        // Create edge
        mds::img::CLine Edge((float)pVertices[0]->GetX(),
                             (float)pVertices[0]->GetY(),
                             (float)pVertices[1]->GetX(),
                             (float)pVertices[1]->GetY(),
                             mds::img::C_WHITE
                             );

        // Draw the edge
        Edge.draw(*pImage);

        // Get next edge
        pEdge = pEdge->GetNext();
    }

    // Clear the list of edges
    pEdges->ClearAll();

/*    vctl::MCTri *pTriangle = getFirstTriangle();
    while( pTriangle )
    {
        // Get triangle vertices
        vctl::MCVertex *pVertices[3];
        pTriangle->GetVerticeS(pVertices);

        // Draw the triangle
        for( int i = 0; i < 3; ++i)
        {
            // Next vertex
            int j = (i + 1) % 3;

            // Create edge
            mds::img::CLine Edge((float)pVertices[i]->GetX(),
                                 (float)pVertices[i]->GetY(),
                                 (float)pVertices[j]->GetX(),
                                 (float)pVertices[j]->GetY(),
                                 mds::img::C_WHITE
                                 );

            // Draw the edge
            Edge.draw(*pImage);
        }

        // Get next one
        pTriangle = pTriangle->GetNext();
    }*/
}


bool CTriMesh::evaluateFeatures(bool bSkipSmall)
{
    // Image size and margin
    mds::tSize XSize = m_spImage->getXSize();
    mds::tSize YSize = m_spImage->getYSize();

    // Check the image size
    MDS_CHECK(XSize > 0 && YSize > 0, return false);

    // Image feature extraction
    CTriFeaturesExtractor FeatureExtraction;

    // Re-compute features of all triangles
    vctl::MCTri *pTriangle = getFirstTriangle();
    while( pTriangle )
    {
        // Get pointer to the triangle features
        CTriFeatures *pFeatures = getTriangleFeatures(pTriangle);

        // Clear extracted features (evaluate every triangle separately)
        FeatureExtraction.clear();

        // Acumulate features from just a one tiangke
        FeatureExtraction.acumulate(m_spImage, pTriangle);

        // Compute the feature vector
        if( !FeatureExtraction.evaluate(pFeatures, bSkipSmall) )
        {
            // Logging
            MDS_LOG_NOTE("CTriMesh::evaluateFeatures(): Cannot evaluate triangle features!");
        }

        // Get next one
        pTriangle = pTriangle->GetNext();
    }

    // O.K.
    return true;
}


bool CTriMesh::evaluateSegmentFeatures(bool bSkipSmall)
{
    // Image size and margin
    mds::tSize XSize = m_spImage->getXSize();
    mds::tSize YSize = m_spImage->getYSize();

    // Check the image size
    MDS_CHECK(XSize > 0 && YSize > 0, return false);

    // Check the number of segments
    MDS_CHECK(m_iNumOfSegments > 0, return false);

    // Image feature extractors
    typedef std::vector<CTriFeaturesExtractor> tExtractors;
    tExtractors Extractors(m_iNumOfSegments);

    // Evaluate image features of all regions
    vctl::MCTri *pTriangle = getFirstTriangle();
    while( pTriangle )
    {
        // Get triangle region
        int iRegion = getTriangleRegion(pTriangle);

        // If the triangle is not yet classified
        if( iRegion >= 0 && iRegion < m_iNumOfSegments )
        {
            // Acumulate image features
            Extractors[iRegion].acumulate(m_spImage, pTriangle);
        }

        // Get the next triangle
        pTriangle = pTriangle->GetNext();
    }

    // Resize the vector of image features
    m_Segments.resize(m_iNumOfSegments);

    // Evaluate image features of all regions
    for( int i = 0; i < m_iNumOfSegments; ++i )
    {
        Extractors[i].evaluate(&(m_Segments[i]), bSkipSmall);
    }

    // O.K.
    return true;
}


bool CTriMesh::edgeSplitting(double dMinLength, double dThreshold)
{
    static const int MAX_ITERS      = 5;
    static const int MIN_CHANGE     = 10;

    MDS_ASSERT(dMinLength > 0.0 && dThreshold > 0.0);

    // Construct container of all edges in the mesh
    m_Triangles.MakeAllTrisEdges();

    // Get pointer to the container
    vctl::MCEdgeS *pEdges = m_Triangles.GetEdgeS();

    // Helper vector of newly created vertices
    tPoints Points;

    // Iterate while the container of edges is not empty
    int OldNumber = 0;
    for( int iter = 0; mds::math::getAbs(OldNumber - pEdges->GetNumber()) > MIN_CHANGE && iter < MAX_ITERS; ++iter )
    {
        // Logging of the number of edges
        MDS_LOG_NOTE("Iteration: " << iter);
        MDS_LOG_NOTE("Begin: " << pEdges->GetNumber());

        // Save the number of edges
        OldNumber = pEdges->GetNumber();

        // Clear the helper vector of vertices
        Points.clear();

        // Evaluate and split all edges
        // - Construct vector of new vertices
        // - Remove edges which were not divided from the container
        vctl::MCEdge *pEdge = (vctl::MCEdge *)pEdges->GetFirst();
        while( pEdge )
        {
            // Check the edge length
            if( pEdge->GetLength() > 2.0 * dMinLength )
            {
                // Split the edge
                if( !splitEdge(pEdge, dMinLength, dThreshold, Points) )
                {
                    // Remove the edge
                    vctl::MCEdge *pTemp = pEdge;
                    pEdge = pEdge->GetNext();
                    pEdges->Erase(pTemp);
                }
                else
                {
                    // Get next edge
                    pEdge = pEdge->GetNext();
                }
            }
            else
            {
                // Remove the edge
                vctl::MCEdge *pTemp = pEdge;
                pEdge = pEdge->GetNext();
                pEdges->Erase(pTemp);
            }
        }

        MDS_LOG_NOTE("After splitting: " << pEdges->GetNumber());

        // Insert all points to the mesh
        tPoints::iterator it = Points.begin();
        tPoints::iterator itEnd = Points.end();
        while( it != itEnd )
        {
            // Get pointer to the point
            vctl::MCPoint3D *pPoint = &(*it);

            // Check if the point lies on the mesh boundary
            int iBoundaryFlags = checkBoundary(pPoint);

            // Create a new vertex
            vctl::MCVertex *pVertex = newVertex(pPoint);

            // Insert it to the mesh
            if( !addVertex(pVertex, iBoundaryFlags, edgeSplittingInsertFunc) )
            {
                // Remove the vertex
                if( !pVertex->GetRegisteredTri() )
                {
                    m_pVertices->Erase(pVertex);
                }
            }

            // Get next point
            ++it;
        }

        MDS_LOG_NOTE("End: " << pEdges->GetNumber());

        // Check the list of edges and remove edges which are not used
        m_Triangles.EraseEdgesNoTris();
    }

    // Clear the list of edges
    pEdges->ClearAll();

    // O.K.
    return true;
}


bool CTriMesh::splitEdge(vctl::MCEdge *pEdge,
                         double dMinLength,
                         double dOrientThreshold,
                         tPoints& Points
                         )
{
    MDS_ASSERT(pEdge);

    // Sobel oparators
    mds::img::CSobelX<tImage> SobelX;
    mds::img::CSobelY<tImage> SobelY;

    // Vertices of the edge
    vctl::MCVertex *pVertices[2];
    pEdge->GetVerticeS(pVertices);

    // Create edge pixels iterator
    CTriEdgeIterator It(m_spEdgeImage, pEdge);

    // Helper variables
    mds::img::CDImage::tPixel Max = 0;
    double dMaxX = 0, dMaxY = 0;

    // Traverse all pixels and find pixel lying on an image edge having maximal magnitude
    for( ; !It.isEnd(); ++It )
    {
        // Get current pixel value
        mds::img::CDImage::tPixel Value = *It;

        // Is there an image edge?
        if( Value > Max )
        {
            // Splitting point
            vctl::MCPoint3D Point(It.getX(), It.getY(), 0.0);

            // Check size of produced edges
            double dLength1 = pVertices[0]->Distance(Point);
            double dLength2 = pVertices[1]->Distance(Point);
            if( dLength1 < dMinLength || dLength2 < dMinLength )
            {
                continue;
            }

            // Image derivatives
            double dGradX = SobelX.getResponse(*m_spImage, It.getX(), It.getY());
            double dGradY = SobelY.getResponse(*m_spImage, It.getX(), It.getY());

            // Image edge gradient
            vctl::MCVector3D ImageEdgeNormal(dGradX, dGradY, 0.0);
            ImageEdgeNormal.Normalization();

            // Triangle edge
            vctl::MCVector3D TriEdge(pVertices[0], pVertices[1]);
            TriEdge.Normalization();

            // Compare orientation of the image edge gradient and triangle edge
            // - Dot product of both vectors
            // - Value must be greater than a given threshold. Optimal case is
            //   that both vectors are parallel (image edge and triangle edge are
            //   ortogonal), then -1 or 1 is obtained.
            if( mds::math::getAbs(ImageEdgeNormal * TriEdge) > dOrientThreshold )
            {
                Max = Value;
                dMaxX = Point.GetX();
                dMaxY = Point.GetY();
            }
        }
    }

    // Check the maximum
    if( Max > 0 )
    {
        // Create a new point
        vctl::MCPoint3D NewPoint(dMaxX, dMaxY, 0.0);

        // Insert it to the vector
        Points.push_back(NewPoint);

        // Edge was divided
        return true;
    }

    // No splitting
    return false;
}


void CTriMesh::edgeSplittingInsertFunc(vctl::MCTri *pTriangle, CDelaunayTri *pMesh)
{
    MDS_ASSERT(pTriangle && pMesh);

    // Get pointer to the triangular mesh
    CTriMesh *pTriMesh = (CTriMesh *)pMesh;

    // Insert edges of the triangle to the list
    pTriMesh->m_Triangles.MakeTriEdges(pTriangle);
}


bool CTriMesh::triangleSplitting(CTriHomogeneityMeasure& Measure,
                                 double dMinEdgeLength
                                 )
{
    MDS_CHECK(dMinEdgeLength > 0.0, return false);

    // Logging
    MDS_LOG_NOTE("CTriMesh::triangleSplitting(): begin");

    // Compute homogeneity of all triangles
    vctl::MCTri *pTriangle = getFirstTriangle();
    while( pTriangle )
    {
        // Get triangle features
        CTriFeatures *pFeatures = getTriangleFeatures(pTriangle);

        // Compute and set the triangle homogeneity
        if( !pFeatures->isError() )
        {
            pTriangle->SetValue(Measure.getHomogeneity(*pFeatures));
        }
        else
        {
            pTriangle->SetValue(0.0);
        }

#ifdef MDS_SEG_LOGGING_ENABLED
        MDS_LOG_NOTE("CTriMesh::triangleSplitting(): Triangle");
        MDS_LOG_NOTE("     Variance: " << pFeatures->getPixelVariance());
        MDS_LOG_NOTE("  Homogeneity: " << pTriangle->GetValue());
#endif // MDS_SEG_LOGGING_ENABLED

        // Get next one
        pTriangle = pTriangle->GetNext();
    }

    // Sort all triangles
    m_Triangles.SortAll();

    // Get sorted queue
    vctl::MCTriS::MT_QUEUE *pQueue = m_Triangles.GetQueue();

    // Number of splitted triangles
    int iCount = 0;

    // Set the homogeneity measure
    m_pHomogeneityMeasure = &Measure;

    // Iterate while the container of edges is not empty
    for( ;; )
    {
        // Get pointer to the last item
        pTriangle = pQueue->GetMin();

        // Logging
#ifdef MDS_SEG_LOGGING_ENABLED
        MDS_LOG_NOTE("CTriMesh::triangleSplitting(): Triangle");
        MDS_LOG_NOTE("  Homogeneity: " << pTriangle->GetValue());
#endif // MDS_SEG_LOGGING_ENABLED

        // Check the triangle homogeneity
        if( !pTriangle || pTriangle->GetValue() > Measure.getThreshold() )
        {
            // No optimization is needed
            break;
        }

        // Check the triangle edge length
        if( pTriangle->GetValue() > 0.0 || pTriangle->GetMaxEdgeLength() > 2.0 * dMinEdgeLength )
        {
            // Split the triangle
            if( !splitTriangle(pTriangle, dMinEdgeLength) )
            {
                // Unsort the triangle, no optimization
                try
                {
                    pQueue->UnSort(pTriangle->GetValue(), pTriangle);
                }
                catch( ... )
                {
                }
                pTriangle->SetValue(0.0);
            }
            else
            {
                MDS_LOG_NOTE("CTriMesh::triangleSplitting(): Triangle splitted");
                ++iCount;
            }
        }
        else
        {
            // Unsort the triangle
            try
            {
                pQueue->UnSort(pTriangle->GetValue(), pTriangle);
            }
            catch( ... )
            {
            }
            pTriangle->SetValue(0.0);
        }
    }

    // Clear/unsort the queue
    pQueue->Clear();

    // Clear the homogeneity measure
    m_pHomogeneityMeasure = NULL;

    // Logging
    MDS_LOG_NOTE("CTriMesh::triangleSplitting(): finished");
    MDS_LOG_NOTE("CTriMesh::triangleSplitting(): Number of splitted triangles: " << iCount);

    // O.K.
    return true;
}


bool CTriMesh::splitTriangle(vctl::MCTri *pTriangle, double dMinEdgeLength)
{
    MDS_ASSERT(pTriangle);

    // Get triangle vertices
    vctl::MCVertex *pVertices[3];
    pTriangle->GetVerticeS(pVertices);

    // Compute square of the minimal edge length
    double dThreshold = dMinEdgeLength * dMinEdgeLength;

    // Create triangle pixels iterator
    CTriIterator It(m_spEdgeImage, pTriangle);

    // Helper variables
    mds::img::CDImage::tPixel Max = 0;
    double dMaxX = 0, dMaxY = 0;

    // Traverse all pixels and find best position for a new vertex
    for( ; !It.isEnd(); ++It )
    {
        // Get current pixel value
        mds::img::CDImage::tPixel Value = *It;

        // Is there an image edge?
        if( Value > Max )
        {
            // Splitting point
            vctl::MCPoint3D Point(It.getX(), It.getY(), 0);

            // Check size of produced edges
            double dLength = pVertices[0]->SquareDistance(Point);
            if( dLength < dThreshold )
            {
                continue;
            }
            dLength = pVertices[1]->SquareDistance(Point);
            if( dLength < dThreshold )
            {
                continue;
            }
            dLength = pVertices[2]->SquareDistance(Point);
            if( dLength < dThreshold )
            {
                continue;
            }

            // Save the current position
            Max = Value;
            dMaxX = Point.GetX();
            dMaxY = Point.GetY();
        }
    }

    // Check the maximum
    if( Max > 0 )
    {
        // Create a new vertex
        vctl::MCVertex *pVertex = newVertex(dMaxX, dMaxY);

        // Check the boudary
        int iBoundaryFlags = checkBoundary(pVertex);

        // Insert vertex to the mesh
        if( !addVertex(pVertex,
                       iBoundaryFlags,
                       triangleSplittingInsertFunc,
                       triangleSplittingRemoveFunc
                       ) )
        {
            // Remove the vertex
            if( !pVertex->GetRegisteredTri() )
            {
                try
                {
                    m_pVertices->Erase(pVertex);
                }
                catch( ... )
                {
                }
            }

            // Cannot insert a new vertex
            return false;
        }
        else
        {
            // Triangle has been splitted
            return true;
        }
    }

    // No splitting
    return false;
}


void CTriMesh::triangleSplittingInsertFunc(vctl::MCTri *pTriangle, CDelaunayTri *pMesh)
{
    MDS_CHECK(pTriangle && pMesh, return);

    // Get pointer to the mesh
    CTriMesh *pTriMesh = (CTriMesh *)pMesh;

    // Check if the measure is set
    if( !pTriMesh->m_pHomogeneityMeasure )
    {
        pTriangle->SetValue(0.0);
        return;
    }

    // Get pointer to the triangle features
    CTriFeatures *pFeatures = getTriangleFeatures(pTriangle);

    // Feature extractor
    CTriFeaturesExtractor FeatureExtraction;

    // Acumulate features from just a one triangle
    FeatureExtraction.acumulate(pTriMesh->m_spImage, pTriangle);

    // Compute the feature vector
    if( FeatureExtraction.evaluate(pFeatures, false) )
    {
        // Compute and set the triangle quality
        pTriangle->SetValue(pTriMesh->m_pHomogeneityMeasure->getHomogeneity(*pFeatures));

        // Get sorted queue
        vctl::MCTriS::MT_QUEUE *pQueue = pTriMesh->m_Triangles.GetQueue();

        // Sort the triangle
        pQueue->Sort(pTriangle->GetValue(), pTriangle);
    }
    else
    {
        pTriangle->SetValue(0.0);
    }
}


void CTriMesh::triangleSplittingRemoveFunc(vctl::MCTri *pTriangle, CDelaunayTri *pMesh)
{
    MDS_CHECK(pTriangle && pMesh, return);

    // Get pointer to the mesh
    CTriMesh *pTriMesh = (CTriMesh *)pMesh;

    // Get sorted queue
    vctl::MCTriS::MT_QUEUE *pQueue = pTriMesh->m_Triangles.GetQueue();

    // Unsort the triangle
    if( pTriangle->GetValue() > 0.0 )
    {
        pQueue->UnSort(pTriangle->GetValue(), pTriangle);
    }
}


bool CTriMesh::regionGrowing(CTriSimilarityMeasure& Measure, double dBThreshold)
{
    // Logging
    MDS_LOG_NOTE("CTriMesh::regionGrowing()");

    // Initialize index of the region
    int iRegionIndex = m_iNumOfSegments;

    // Traverse all triangles
    vctl::MCTri *pTriangle = getFirstTriangle();
    while( pTriangle )
    {
        // Get triangle features
        CTriFeatures *pFeatures = getTriangleFeatures(pTriangle);

        // Check the current region
        if( getTriangleRegion(pTriangle) == -1 )
        {
            // Create a new region
            ++iRegionIndex;

            // Queue of triagnles
            typedef std::deque<vctl::MCTri *> tQueue;
            tQueue Queue;

            // Grow the region
            Queue.push_back(pTriangle);
            while( !Queue.empty() )
            {
                // Get pointer to the first element in the queue
                vctl::MCTri *pT = Queue.front();
                Queue.pop_front();

                // Check the triangle
                if( getTriangleRegion(pT) != -1 )
                {
                    continue;
                }

                // Change region of the triangle
                setTriangleRegion(pT, iRegionIndex - 1);

                // Get neighbours of the triangle
                vctl::MCTri *pNeighbours[3];
                pT->GetNeighbours(pNeighbours);

                // Get regions of all neighbours
                for( int i = 0; i < 3; ++i )
                {
                    if( pNeighbours[i] && getTriangleRegion(pNeighbours[i]) == -1 )
                    {
                        // Get the current edge
                        vctl::MCVertex *pV1, *pV2;
                        pT->GetEdge(i, &pV1, &pV2);

                        // Check that there is no image edge
                        vctl::MCEdge Edge(pV1, pV2);
                        if( checkBoundaryEdge(&Edge, dBThreshold) )
                        {
                            continue;
                        }

                        // Get triangle features
                        CTriFeatures *pFeatures2 = getTriangleFeatures(pNeighbours[i]);

                        // Check similarity of both triangles
                        if( Measure.checkSimilarity(*pFeatures, *pFeatures2) )
                        {
                            Queue.push_back(pNeighbours[i]);
                        }
                    }
                }
            }
        }

        // Get the next triangle
        pTriangle = pTriangle->GetNext();
    }

    // Set the number of regions in the mesh
    m_iNumOfSegments = iRegionIndex;

    // Logging
    MDS_LOG_NOTE("  Final number of regions = " << m_iNumOfSegments);

    // O.K.
    return true;
}


bool CTriMesh::regionMerging(CTriSimilarityMeasure& Measure,
                             double dBThreshold,
                             tSize NumOfRegions
                             )
{
//    MDS_CHECK(m_iNumOfSegments > 0, return false);

    // Logging
    MDS_LOG_NOTE("CTriMesh::merging()");

    // Assign unique regions to all non-classified triangles
    vctl::MCTri *pT = getFirstTriangle();
    while( pT )
    {
        // Get triangle region
        if( getTriangleRegion(pT) == -1 )
        {
            setTriangleRegion(pT, m_iNumOfSegments++);
        }

        // Get the next triangle
        pT = pT->GetNext();
    }

    // Shake down used indexes of regions
    reassignRegions();

    MDS_LOG_NOTE("  Number of regions = " << m_iNumOfSegments);

    // Region Adjacency Graph (RAG)
    mds::math::CMatrix<double> RAG(m_iNumOfSegments, m_iNumOfSegments);
    RAG.fill(-1.0);

    // Initial mapping of regions
    mds::math::CVector<int> Mapping(m_iNumOfSegments);
    for( int i = 0; i < m_iNumOfSegments; ++i )
    {
        Mapping(i) = i;
    }

    // Image feature extractors
    typedef std::vector<CTriFeaturesExtractor> tExtractors;
    tExtractors Extractors(m_iNumOfSegments);

    // Evaluate image features of all regions and create the RAG graph
    vctl::MCTri *pTriangle = getFirstTriangle();
    while( pTriangle )
    {
        // Get triangle region
        int iRegion = getTriangleRegion(pTriangle);

        // If the triangle is not yet classified
        if( iRegion >= 0 || iRegion < m_iNumOfSegments )
        {
            // Acumulate image features
            Extractors[iRegion].acumulate(m_spImage, pTriangle);
        }

        // Get triangle neighbours
        vctl::MCTri *pNeighbours[3];
        pTriangle->GetNeighbours(pNeighbours);

        // Get regions of all neighbours and fill the RAG
        for( int i = 0; i < 3; ++i )
        {
            if( pNeighbours[i] )
            {
                int iRegion2 = getTriangleRegion(pNeighbours[i]);
                if( iRegion2 != -1 && iRegion2 != iRegion )
                {
                    // Get the current edge
                    vctl::MCVertex *pV1, *pV2;
                    pTriangle->GetEdge(i, &pV1, &pV2);

                    // Check that there is no image edge
                    vctl::MCEdge Edge(pV1, pV2);
                    if( checkBoundaryEdge(&Edge, dBThreshold) )
                    {
                        continue;
                    }

                    // Set the RAG
                    RAG(iRegion, iRegion2) = 0.0;
                    RAG(iRegion2, iRegion) = 0.0;
                }
            }
        }

        // Get the next triangle
        pTriangle = pTriangle->GetNext();
    }

    // Evaluate image features of all regions
    m_Segments.resize(m_iNumOfSegments);
    for( int r = 0; r < m_iNumOfSegments; ++r )
    {
        Extractors[r].evaluate(&(m_Segments[r]), false);
    }

    // Evaluate similarity of all regions
    for( int j = 0; j < m_iNumOfSegments; ++j )
    {
        for( int i = j + 1; i < m_iNumOfSegments; ++i )
        {
            if( RAG(j,i) == 0.0 )
            {
                // Evaluate the similarity
                RAG(j,i) = Measure.getSimilarity(m_Segments[j], m_Segments[i]);
            }
        }
    }

    // Logging
//    MDS_LOG_NOTE("  Region Adjacency Graph = ");
//    mds::math::CMatrix<double> Part(RAG, 0, 0, 100, 100, mds::REFERENCE);
//    MDS_LOG_NOTE(Part);

    // While any two regions can be merged
    int iCount = m_iNumOfSegments;
    for( ; iCount > NumOfRegions; )
    {
        // Find maximum similarity
        double dMaxSimilarity = -1.0;
        int iMax1 = -1, iMax2 = -1;
        for( int k = 0; k < m_iNumOfSegments; ++k )
        {
            for( int l = k + 1; l < m_iNumOfSegments; ++l )
            {
                if( RAG(k,l) > dMaxSimilarity )
                {
                    dMaxSimilarity = RAG(k,l);
                    iMax1 = k;
                    iMax2 = l;
                }
            }
        }

        // Logging
        MDS_LOG_NOTE("  Maximal similarity = " << dMaxSimilarity);
//        MDS_LOG_NOTE("  Regions " << iMax1 << ", " << iMax2);

        // Check the maximal similarity
        if( dMaxSimilarity < 0.0 )
        {
            break;
        }
        if( NumOfRegions == -1 && dMaxSimilarity < Measure.getThreshold() )
        {
            break;
        }

        // Choose the larger region
        if( m_Segments[iMax1].getNumOfPixels() < m_Segments[iMax2].getNumOfPixels() )
        {
            mds::math::swap(iMax1, iMax2);
        }

        // Update the RAG
        for( int i = 0; i < m_iNumOfSegments; ++i )
        {
            if( RAG(i,iMax2) >= 0.0 )
            {
                RAG(i,iMax2) = -1.0;
                RAG(i,iMax1) = 0.0;
            }
            if( RAG(iMax2,i) >= 0.0 )
            {
                RAG(iMax2,i) = -1.0;
                RAG(iMax1,i) = 0.0;
            }
        }

        // Update feature vectors
        Extractors[iMax1].acumulate(Extractors[iMax2]);
        Extractors[iMax1].evaluate(&(m_Segments[iMax1]), false);

        // Re-compute similarity of the regions
        for( int i = iMax1 + 1; i < m_iNumOfSegments; ++i )
        {
            // Re-compute similarity of the first regions
            if( RAG(iMax1, i) >= 0.0 )
            {
                RAG(iMax1, i) = Measure.getSimilarity(m_Segments[iMax1], m_Segments[i]);
            }
        }

        // Update the mapping of regions
        for( int r = 0; r < m_iNumOfSegments; ++r )
        {
            if( Mapping(r) == iMax2 )
            {
                Mapping(r) = iMax1;
            }
        }

        // Decrease the number of regions
        --iCount;
    }

    // Re-compute the number of segments and modify the mapping table
    MDS_LOG_NOTE("  Final number of regions = " << iCount);

    // Re-assign all triangles
    pTriangle = getFirstTriangle();
    while( pTriangle )
    {
        // Get triangle region
        int iRegion = getTriangleRegion(pTriangle);

        // Assign index of the new region
        if( iRegion >= 0 && iRegion < m_iNumOfSegments )
        {
            // Set triangle region
            setTriangleRegion(pTriangle, Mapping(iRegion));
        }

        // Get the next triangle
        pTriangle = pTriangle->GetNext();
    }

    // Shake down used indexes of regions
    reassignRegions();

    // O.K.
    return true;
}


/*bool CTriMesh::merging(CTriSimilarityMeasure& Measure,
                       double dBThreshold
                       )
{
    MDS_CHECK(m_iNumOfSegments > 0, return false);

    // Logging
    MDS_LOG_NOTE("CTriMesh::merging()");
    MDS_LOG_NOTE("  Number of regions = " << m_iNumOfSegments);

    // List of non-classified triangles
    typedef std::list<vctl::MCTri *> tTriangles;
    tTriangles NonClassifiedTriangles;

    // Image feature extractors
    typedef std::vector<CTriFeaturesExtractor> tExtractors;
    tExtractors Extractors(m_iNumOfSegments + 1);

    // Create a list of non-classified triangles and evaluate image features
    // of all regions
    vctl::MCTri *pTriangle = getFirstTriangle();
    while( pTriangle )
    {
        // Get triangle region
        int iRegion = getTriangleRegion(pTriangle);

        // If the triangle is not yet classified
        if( iRegion < 0 || iRegion >= m_iNumOfSegments )
        {
            // Get pointer to the triangle features
            CTriFeatures *pFeatures = getTriangleFeatures(pTriangle);

            // Evaluate triangle features
            Extractors[m_iNumOfSegments].clear();
            Extractors[m_iNumOfSegments].acumulate(m_spImage, pTriangle);

            // Check if the features are correctly evaluated
            if( Extractors[m_iNumOfSegments].evaluate(pFeatures, false) )
            {
                NonClassifiedTriangles.push_back(pTriangle);
            }
        }
        else
        {
            // Acumulate image features
            Extractors[iRegion].acumulate(m_spImage, pTriangle);
        }

        // Get the next triangle
        pTriangle = pTriangle->GetNext();
    }

    // Logging
    MDS_LOG_NOTE("  Number of non-classified triangles = " << NonClassifiedTriangles.size());

    // Evaluate image features of all regions
    typedef std::vector<CTriFeatures> tRegions;
    tRegions Regions(m_iNumOfSegments);
    for( int i = 0; i < m_iNumOfSegments; ++i )
    {
        Extractors[i].evaluate(&(Regions[i]), false);
    }

    // The number of changes
    while( !NonClassifiedTriangles.empty() )
    {
        // Maximum similarity
        double dMaxSimilarity = -1.0;
        int iMaxRegion = -1;
        tTriangles::iterator itMaxTriangle = NonClassifiedTriangles.end();

        // Traverse all triangles
        tTriangles::iterator itEnd = NonClassifiedTriangles.end();
        tTriangles::iterator it = NonClassifiedTriangles.begin();
        for( ; it != itEnd; ++it )
        {
            // Get pointer to the triangle features
            CTriFeatures *pFeatures = getTriangleFeatures(*it);

            // Get neighbours of the triangle
            vctl::MCTri *pNeighbours[3];
            (*it)->GetNeighbours(pNeighbours);

            // Get regions of all neighbours and evaluate similarity
            for( int i = 0; i < 3; ++i )
            {
                if( pNeighbours[i] )
                {
                    int iRegion = getTriangleRegion(pNeighbours[i]);
                    if( iRegion != -1 )
                    {
                        // Get the current edge
                        vctl::MCVertex *pV1, *pV2;
                        (*it)->GetEdge(i, &pV1, &pV2);

                        // Check that there is no image edge
                        vctl::MCEdge Edge(pV1, pV2);
                        if( checkBoundaryEdge(&Edge, dBThreshold) )
                        {
                            continue;
                        }

                        // Get pointer to the region features
                        double dSimilarity = -1.0;
                        if( !Regions[iRegion].isError() )
                        {
                            dSimilarity = Measure.getSimilarity(*pFeatures, Regions[iRegion]);
                        }

                        // Compare with the maximum value
                        if( dSimilarity > dMaxSimilarity )
                        {
                            dMaxSimilarity = dSimilarity;
                            iMaxRegion = iRegion;
                            itMaxTriangle = it;
                        }
                    }
                }
            }
        }

        // Check the maximal similarity
        if( dMaxSimilarity <= 0.0 )
        {
            break;
        }

        // Classify the triangle according to the most similar region
        if( iMaxRegion >= 0 )
        {
            // Classify the triangle
            setTriangleRegion(*itMaxTriangle, iMaxRegion);

            // Update feature vector
            Extractors[iMaxRegion].acumulate(m_spImage, *itMaxTriangle);
            Extractors[iMaxRegion].evaluate(&(Regions[iMaxRegion]), false);

            // Remove classified triangle from the list
            NonClassifiedTriangles.erase(itMaxTriangle);
        }

        // Logging
        MDS_LOG_NOTE("  Maximal similarity = " << dMaxSimilarity);
    }

    // O.K.
    return true;
}*/


/*bool CTriMesh::merging(CTriSimilarityMeasure& Measure,
                       double dBThreshold
                       )
{
    MDS_CHECK(m_iNumOfSegments > 0, return false);

    // Logging
    MDS_LOG_NOTE("CTriMesh:::merging()");
    MDS_LOG_NOTE("  Number of regions = " << m_iNumOfSegments);

    // List of non-classified triangles
    typedef std::vector<vctl::MCTri *> tTriangles;
    tTriangles NonClassifiedTriangles;

    // Create a list of non-classified triangles
    vctl::MCTri *pTriangle = getFirstTriangle();
    while( pTriangle )
    {
        // If the triangle is not yet classified
        if( getTriangleRegion(pTriangle) == -1 )
        {
            // Get pointer to the triangle features
            CTriFeatures *pFeatures = getTriangleFeatures(pTriangle);

            // Check if the features are correctly evaluated
            if( !pFeatures->isError() )
            {
                NonClassifiedTriangles.push_back(pTriangle);
            }
        }

        // Get the next triangle
        pTriangle = pTriangle->GetNext();
    }

    // Logging
    MDS_LOG_NOTE("  Number of non-classified triangles = " << NonClassifiedTriangles.size());

    // The number of changes
    int iNumOfChanges = 1;
    for( int iters = 0; iters < 100 && iNumOfChanges > 0; ++iters )
    {
        // Reset the number of re-assigned triangles
        iNumOfChanges = 0;

        // Traverse all triangles
        tTriangles::iterator itEnd = NonClassifiedTriangles.end();
        tTriangles::iterator it = NonClassifiedTriangles.begin();
        for( ; it != itEnd; ++it )
        {
            // Get pointer to the triangle features
            CTriFeatures *pFeatures = getTriangleFeatures(*it);

            // Get neighbours of the triangle
            vctl::MCTri *pNeighbours[3];
            (*it)->GetNeighbours(pNeighbours);

            // Get regions of all neighbours and evaluate similarity
            int iRegions[3]= { -1, -1 , -1 };
            double dSimilarity[3] = { -1.0, -1.0, -1.0 };
            for( int i = 0; i < 3; ++i )
            {
                if( pNeighbours[i] )
                {
                    iRegions[i] = getTriangleRegion(pNeighbours[i]);
                    if( iRegions[i] != -1 )
                    {
                        // Get the current edge
                        vctl::MCVertex *pV1, *pV2;
                        (*it)->GetEdge(i, &pV1, &pV2);

                        // Check that there is no image edge
                        vctl::MCEdge Edge(pV1, pV2);
                        if( checkBoundaryEdge(&Edge, dBThreshold) )
                        {
                            continue;
                        }

                        // Get pointer to the triangle features
                        CTriFeatures *pFeatures2 = getTriangleFeatures(pNeighbours[i]);
                        if( !pFeatures2->isError() )
                        {
                            dSimilarity[i] = Measure.getSimilarity(*pFeatures, *pFeatures2);
                        }
                    }
                }
            }

            // Classify the triangle according to the most similar triangle
            int iMax = (dSimilarity[0] > dSimilarity[1]) ? 0 : 1;
            iMax = (dSimilarity[iMax] > dSimilarity[2]) ? iMax : 2;
            if( dSimilarity[iMax] >= 0.0 )
            {
                int iTriRegion = getTriangleRegion(*it);
                int iNewRegion = getTriangleRegion(pNeighbours[iMax]);
                if( iNewRegion != iTriRegion )
                {
                    setTriangleRegion(*it, iNewRegion);
                    ++iNumOfChanges;
                }
            }
        }

        // Logging
        MDS_LOG_NOTE("  Number of changes = " << iNumOfChanges);
    }

    // O.K.
    return true;
}*/


bool CTriMesh::classifyAdjacent(double dBThreshold)
{
    MDS_CHECK(m_iNumOfSegments > 0, return false);

    // Logging
    MDS_LOG_NOTE("CTriMesh::classifyAdjacent():");
    MDS_LOG_NOTE("  Number of regions = " << m_iNumOfSegments);

    // List of non-classified triangles
    typedef std::vector<vctl::MCTri *> tTriangles;
    tTriangles NonClassifiedTriangles;

    // Create a list of non-classified triangles
    vctl::MCTri *pTriangle = getFirstTriangle();
    while( pTriangle )
    {
        // If the triangle is not yet classified
        if( getTriangleRegion(pTriangle) == -1 )
        {
            NonClassifiedTriangles.push_back(pTriangle);
        }

        // Get the next triangle
        pTriangle = pTriangle->GetNext();
    }

    // Logging
    MDS_LOG_NOTE("  Number of non-classified triangles = " << (int)NonClassifiedTriangles.size());

    // Classify all triangles with respect to the neighbours
    int iNumOfChanges = 1;
    for( int iters = 0; iters < 100 && iNumOfChanges > 0; ++iters )
    {
        // Reset the number of re-assigned triangles
        iNumOfChanges = 0;

        // Traverse all triangles
        tTriangles::iterator itEnd = NonClassifiedTriangles.end();
        tTriangles::iterator it = NonClassifiedTriangles.begin();
        for( ; it != itEnd; ++it )
        {
            // Get neighbours of the triangle
            vctl::MCTri *pNeighbours[3];
            (*it)->GetNeighbours(pNeighbours);

            // Get regions
            int iRegions[3]= { -1, -1 , -1 };
            for( int i = 0; i < 3; ++i )
            {
                if( pNeighbours[i] )
                {
                    // Get the region
                    iRegions[i] = getTriangleRegion(pNeighbours[i]);
                    if( iRegions[i] != -1 )
                    {
                        // Get the current edge
                        vctl::MCVertex *pV1, *pV2;
                        (*it)->GetEdge(i, &pV1, &pV2);

                        // Check that there is no image edge
                        vctl::MCEdge Edge(pV1, pV2);
                        if( checkBoundaryEdge(&Edge, dBThreshold) )
                        {
                            continue;
                        }
                    }
                }
            }

            // Index of the region
            int iNewRegion = -1;

            // Evaluate the region index
            vctl::MCVertex *pV1, *pV2;
            if( iRegions[0] != -1 )
            {
                if( iRegions[0] == iRegions[1] || iRegions[0] == iRegions[2] )
                {
                    iNewRegion = iRegions[0];
                }
                else if( iRegions[1] == -1 )
                {
                    if( iRegions[2] == -1 )
                    {
                        iNewRegion = iRegions[0];
                    }
                    else
                    {
                        (*it)->GetEdge(0, &pV1, &pV2);
                        double dLength0 = pV1->SquareDistance(*pV2);
                        (*it)->GetEdge(2, &pV1, &pV2);
                        double dLength2 = pV1->SquareDistance(*pV2);
                        iNewRegion = (dLength0 > dLength2) ? 0 : 2;
                    }
                }
                else if( iRegions[2] == -1 )
                {
                    (*it)->GetEdge(0, &pV1, &pV2);
                    double dLength0 = pV1->SquareDistance(*pV2);
                    (*it)->GetEdge(1, &pV1, &pV2);
                    double dLength1 = pV1->SquareDistance(*pV2);
                    iNewRegion = (dLength0 > dLength1) ? 0 : 1;
                }
                else
                {
                    int iMax = (*it)->GetMaxEdge();
                    iNewRegion = iRegions[iMax];
                }
            }
            else if( iRegions[1] != -1 )
            {
                if( iRegions[1] == iRegions[2] || iRegions[2] == -1 )
                {
                    iNewRegion = iRegions[1];
                }
                else
                {
                    (*it)->GetEdge(1, &pV1, &pV2);
                    double dLength1 = pV1->SquareDistance(*pV2);
                    (*it)->GetEdge(2, &pV1, &pV2);
                    double dLength2 = pV1->SquareDistance(*pV2);
                    iNewRegion = (dLength1 > dLength2) ? 1 : 2;
                }
            }
            else if( iRegions[2] != -1 )
            {
                iNewRegion = iRegions[2];
            }

            // Re-assign the triangle
            if( iNewRegion != -1 && iNewRegion != getTriangleRegion(*it) )
            {
                setTriangleRegion(*it, iNewRegion);
                ++iNumOfChanges;
            }
        }

        // Logging
        MDS_LOG_NOTE("  Number of changes = " << iNumOfChanges);
    }

    // O.K.
    return true;
}


bool CTriMesh::noiseReduction()
{
    MDS_CHECK(m_iNumOfSegments > 0, return false);

    // Logging
    MDS_LOG_NOTE("CTriMesh::noiseReduction():");
    MDS_LOG_NOTE("  Number of regions = " << m_iNumOfSegments);

    // Number of re-classified triangles
    int iNumOfChanges = 1;
    for( int iters = 0; iters < 100 && iNumOfChanges > 0; ++iters )
    {
        // Reset the number of re-classified triangles
        iNumOfChanges = 0;

        // Traverse all triangles in the mesh
        vctl::MCTri *pTriangle = getFirstTriangle();
        while( pTriangle )
        {
            // Get triangle region
            int iTriRegion = getTriangleRegion(pTriangle);

            // Get its neighbours
            vctl::MCTri *pNeighbours[3];
            pTriangle->GetNeighbours(pNeighbours);

            // Get regions of all neighbours
            int iRegions[3] = { -1, -1, -1 };
            for( int i = 0; i < 3; ++i )
            {
                if( pNeighbours[i] )
                {
                    iRegions[i] = getTriangleRegion(pNeighbours[i]);
                }
            }

            // Check if the triangle is connected to its region
            if( iRegions[0] == iTriRegion
                || iRegions[1] == iTriRegion
                || iRegions[2] == iTriRegion )
            {
                // Continue with the next triangle
                pTriangle = pTriangle->GetNext();
                continue;
            }

            // Determine new segment
            int iNewRegion = -1;
            if( iRegions[0] == iRegions[1] || iRegions[0] == iRegions[2] )
            {
                iNewRegion = iRegions[0];
            }
            else if( iRegions[1] == iRegions[2] )
            {
                iNewRegion = iRegions[1];
            }

            // Re-classify the current triangle
            if( iNewRegion != -1 && iNewRegion != iTriRegion )
            {
                setTriangleRegion(pTriangle, iNewRegion);
                ++iNumOfChanges;
            }

            // Check that all neighbours are assigned to the same segment
    /*        if( iRegions[0] == iRegions[1]
                && iRegions[1] == iRegions[2] )
            {
                // Re-classify the triangle
                if( iTriRegion != iRegions[0] )
                {
                    setTrihedronRegion(pTriangle, iRegions[0]);
                    ++iNumOfChanges;
                }
            }*/

            // Get next triangle
            pTriangle = pTriangle->GetNext();
        }

        // Logging
        MDS_LOG_NOTE("  Number of re-classified triangles = " << iNumOfChanges);
    }

    // O.K.
    return true;
}


/*bool CTriMesh::checkBoundaryEdge(vctl::MCEdge *pEdge, double dThreshold)
{
    static const int MAX_DISTANCE = 10;
    static const tPixel EDGE_THRESHOLD = 0;
    static const double ORIENT_THRESHOLD = 0.1;
    static const double MAX_ORIENT_CHANGE = 1.0;

    MDS_CHECK(pEdge, return true);

    // Sobel filters
    mds::img::CSobelX<tImage> SobelX;
    mds::img::CSobelY<tImage> SobelY;

    // Vertices of the edge
    vctl::MCVertex *pVertices[2];
    pEdge->GetVerticeS(pVertices);

    // Triangle edge
    vctl::MCVector3D TriEdge(pVertices[0], pVertices[1]);
    TriEdge.Normalization();

    // Edge normal
    mds::img::CPoint3D Normal(mds::img::tCoordinate(-TriEdge.GetY()),
                              mds::img::tCoordinate(TriEdge.GetX())
                              );

    // Create edge pixels iterator
    CTriEdgeIterator it(m_spEdgeImage, pEdge);

    // Helper variables
    double dDistSum = 0.0, dOrientSum = 0.0, dLastOrient = 0.0;

    // Traverse all pixels and check if there is some parallel image edge
    int iCount = 0;
    for( ; !it.isEnd(); ++it, ++iCount )
    {
        // Current position
        tSize x = it.getX();
        tSize y = it.getY();

        // Find an image edge
        if( *it <= EDGE_THRESHOLD )
        {
            // Pixel neighbours
            mds::img::CPoint3D Left(x, y);
            mds::img::CPoint3D Right(x, y);

            // Evaluate minimal distance to an image edge
            int iDistance = 1;
            for( ; iDistance < MAX_DISTANCE; ++iDistance )
            {
                // Step along the normal vector
                Left -= Normal;
                Right += Normal;

                // Interpolate pixel value
                tPixel LeftValue = m_spEdgeImage->interpolate(Left);
                if( LeftValue > EDGE_THRESHOLD )
                {
                    x = mds::math::round2Int(Left.x);
                    y = mds::math::round2Int(Left.y);
                    break;
                }

                // Interpolate pixel value
                tPixel RightValue = m_spEdgeImage->interpolate(Right);
                if( RightValue > EDGE_THRESHOLD )
                {
                    x = mds::math::round2Int(Right.x);
                    y = mds::math::round2Int(Right.y);
                    break;
                }
            }

            // Cumulate distances
            dDistSum += iDistance;

            // No edge found
            if( iDistance == MAX_DISTANCE )
            {
                dOrientSum += 1.0;
                continue;
            }
        }

        // Image derivatives
        double dGradX = SobelX.getResponse(*m_spImage, x, y);
        double dGradY = SobelY.getResponse(*m_spImage, x, y);

        // Image edge gradient
        vctl::MCVector3D ImageEdgeNormal(dGradX, dGradY, 0.0);
        ImageEdgeNormal.Normalization();

        // Compare orientation of the image edge normal and triangle edge
        // - Compute dot product of both vectors
        double dOrient = ImageEdgeNormal * TriEdge;

        // Average orientation
        dOrientSum += dOrient;

        // Check difference between the previous and the current value
        if( mds::math::getAbs(dLastOrient - dOrient) > MAX_ORIENT_CHANGE )
        {
            return false;
        }
        dLastOrient = dOrient;
    }

    // Evaluate if the edge lies on an image edge (region boundary)
    if( iCount > 0 )
    {
        // Average distance and orientation
        double dInvCount = 1.0 / double(iCount);
        double dAvgDistance = dDistSum * dInvCount;
        double dAvgOrient = dOrientSum * dInvCount;

        return (dAvgOrient < ORIENT_THRESHOLD && dAvgDistance < dThreshold);
    }

    // There is boundary edge
    return true;
}*/


/*bool CTriMesh::checkBoundaryEdge(vctl::MCEdge *pEdge, double dThreshold)
{
    static const int MAX_DISTANCE = 10;
    static const int MAX_DIFFERENCE = 1;
    static const tPixel EDGE_THRESHOLD = 0;

    MDS_CHECK(pEdge, return true);

    // Vertices of the edge
    vctl::MCVertex *pVertices[2];
    pEdge->GetVerticeS(pVertices);

    // Triangle edge
    vctl::MCVector3D TriEdge(pVertices[0], pVertices[1]);
    TriEdge.Normalization();

    // Edge normal
    mds::img::CPoint3D Normal(mds::img::tCoordinate(-TriEdge.GetY()),
                              mds::img::tCoordinate(TriEdge.GetX())
                              );

    // Create edge pixels iterator
    CTriEdgeIterator it(m_spEdgeImage, pEdge);

    // Helper variables
    double dDistSum = 0.0;

    // Traverse all pixels and check if there is some parallel image edge
    int iCount = 0;
    int iLastDist = 0;
    for( ; !it.isEnd(); ++it, ++iCount )
    {
        // Is there an image edge?
        if( *it > EDGE_THRESHOLD )
        {
            // Check the 1st derivation
            if( iLastDist > MAX_DIFFERENCE )
            {
                return false;
            }
            iLastDist = 0;

            // There is an image edge
            continue;
        }

        // Pixel neighbours
        mds::img::CPoint3D Left((float)it.getX(), (float)it.getY());
        mds::img::CPoint3D Right((float)it.getX(), (float)it.getY());

        // Evaluate minimal distance to an image edge
        int iDistance = 1;
        for( ; iDistance < MAX_DISTANCE; ++iDistance )
        {
            // Step along the normal vector
            Left -= Normal;
            Right += Normal;

            // Interpolate pixel value
            tPixel LeftValue = m_spEdgeImage->interpolate(Left);
            if( LeftValue > EDGE_THRESHOLD )
            {
                break;
            }

            // Interpolate pixel value
            tPixel RightValue = m_spEdgeImage->interpolate(Right);
            if( RightValue > EDGE_THRESHOLD )
            {
                break;
            }
        }

        // Sum of distances
        dDistSum += iDistance;

        // Check the 1st derivation
        if( mds::math::getAbs(iLastDist - iDistance) > MAX_DIFFERENCE )
        {
            return false;
        }
        iLastDist = iDistance;
    }

    // Check the distance at the end
    if( mds::math::getAbs(iLastDist) > MAX_DIFFERENCE )
    {
        return false;
    }

    // Evaluate if the edge lies on an image edge
    if( iCount > 0 )
    {
        double dRatio = dDistSum / double(iCount);

        return (dRatio < dThreshold);
    }

    // There is a boundary edge
    return true;
}*/


bool CTriMesh::checkBoundaryEdge(vctl::MCEdge *pEdge, double dThreshold)
{
    static const int MAX_DISTANCE = 5;
    static const int MAX_DIFFERENCE = 1;
    static const tPixel EDGE_THRESHOLD = 0;

    MDS_CHECK(pEdge, return true);

    // Sobel filters
    mds::img::CSobelX<tImage> SobelX;
    mds::img::CSobelY<tImage> SobelY;

    // Vertices of the edge
    vctl::MCVertex *pVertices[2];
    pEdge->GetVerticeS(pVertices);

    // Triangle edge
    vctl::MCVector3D TriEdge(pVertices[0], pVertices[1]);
    TriEdge.Normalization();

    // Edge normal
    mds::img::CPoint3D Normal(mds::img::tCoordinate(-TriEdge.GetY()),
                              mds::img::tCoordinate(TriEdge.GetX())
                              );

    // Create edge pixels iterator
    CTriEdgeIterator it(m_spEdgeImage, pEdge);

    // Helper variables
    double dDistSum = 0.0;

    // Traverse all pixels and check if there is some parallel image edge
    int iCount = 0;
    int iLastDist = 0;
    for( ; !it.isEnd(); ++it, ++iCount )
    {
        // Current position
        tSize x = it.getX();
        tSize y = it.getY();

        // Minimal distance to an image edge
        int iDistance = 0;

        // Find an image edge
        if( *it <= EDGE_THRESHOLD )
        {
            // Pixel neighbours
            mds::img::CPoint3D Left(x, y);
            mds::img::CPoint3D Right(x, y);

            // Evaluate minimal distance to an image edge
            iDistance = 1;
            for( ; iDistance < MAX_DISTANCE; ++iDistance )
            {
                // Step along the normal vector
                Left -= Normal;
                Right += Normal;

                // Interpolate pixel value
                tPixel LeftValue = m_spEdgeImage->interpolate(Left);
                if( LeftValue > EDGE_THRESHOLD )
                {
                    x = mds::math::round2Int(Left.x());
                    y = mds::math::round2Int(Left.y());
                    break;
                }

                // Interpolate pixel value
                tPixel RightValue = m_spEdgeImage->interpolate(Right);
                if( RightValue > EDGE_THRESHOLD )
                {
                    x = mds::math::round2Int(Right.x());
                    y = mds::math::round2Int(Right.y());
                    break;
                }
            }
        }

        // Check the 1st derivation
        if( mds::math::getAbs(iLastDist - iDistance) > MAX_DIFFERENCE )
        {
            return false;
        }
        iLastDist = iDistance;

        // No edge found
        if( iDistance == MAX_DISTANCE )
        {
            dDistSum += MAX_DISTANCE + 1;
            continue;
        }

        // Image derivatives
        double dGradX = SobelX.getResponse(*m_spImage, x, y);
        double dGradY = SobelY.getResponse(*m_spImage, x, y);

        // Image edge gradient
        vctl::MCVector3D ImageEdgeNormal(dGradX, dGradY, 0.0);
        ImageEdgeNormal.Normalization();

        // Compare orientation of the image edge normal and triangle edge
        // - Compute dot product of both vectors
        double dOrient = ImageEdgeNormal * TriEdge;

        // Cumulate weighted distances
        dDistSum += (iDistance + 1) * mds::math::getAbs(dOrient);
    }

    // Check the distance at the end
    if( mds::math::getAbs(iLastDist) > MAX_DIFFERENCE )
    {
        return false;
    }

    // Evaluate if the edge lies on an image edge (region boundary)
    if( iCount > 0 )
    {
        double dRatio = dDistSum / double(iCount);

        return (dRatio < dThreshold);
    }

    // There is boundary edge
    return true;
}


void CTriMesh::reassignRegions()
{
    if( m_iNumOfSegments <= 0 )
    {
        return;
    }

    // Mapping of regions
    mds::math::CVector<int> Mapping(m_iNumOfSegments);
    Mapping.fill(-1);

    // Re-assign all triangles
    vctl::MCTri *pTriangle = getFirstTriangle();
    int iNewIndex = 0;
    while( pTriangle )
    {
        // Get triangle region
        int iRegion = getTriangleRegion(pTriangle);

        // Assign index of the new region
        if( iRegion >= 0 && iRegion < m_iNumOfSegments )
        {
            // Mapping of the region
            if( Mapping(iRegion) == -1 )
            {
                Mapping(iRegion) = iNewIndex++;
            }

            // Set triangle region
            setTriangleRegion(pTriangle, Mapping(iRegion));
        }
        else
        {
            // Clear triangle region
            setTriangleRegion(pTriangle, -1);

            // Logging
            MDS_LOG_NOTE("CTriMesh::reassignRegions(): Bad region index!");
        }

        // Get the next triangle
        pTriangle = pTriangle->GetNext();
    }

    // Final number of regions
    m_iNumOfSegments = iNewIndex;
}


//==============================================================================
/*
 * Implementation of the class CTriMesh.
 * - Saving to the STL and VRML format.
 * - Modified Premysl Krsek's original code.
 */

//! Structure containing vertex coordinates
struct SFloatCoordinates
{
    float x, y, z;
};


bool CTriMesh::saveSTL(mds::mod::CChannel& Channel)
{
    // Hlavicka stl souboru
    char pcHeader[90];
    mds::base::memSet<char>(pcHeader, 0, 85);

    // Nastaveni textu hlavicky stl souboru
    strcpy(pcHeader, "Binary STL file created by MDSTk");

    // Ulozeni hlavicky stl souboru
    if( !Channel.write(pcHeader, 80) )
    {
        return false;
    }

    // Ziskani poctu tri site
    unsigned long ulNumOfTriangles = (unsigned long )getNumOfTriangles();

    // Ulozeni poctu tri site
    if( !Channel.write((char *)&ulNumOfTriangles, sizeof(unsigned long)) )
    {
        return false;
    }

    // STL atribut trianglu musi byt 0
    unsigned short int usAttribute = 0;

    // Cyklus ukladani tri
    vctl::MCTri *pTriangle = getFirstTriangle();
    while (pTriangle)
    {
        // Ziskani node aktualniho tri
        vctl::MCVertex *pTriangleNode[3];                                    // pole ukazatelu na uzly aktualniho tri
        pTriangle->GetVerticeS(pTriangleNode);

        // Ziskani normaly aktualniho tri
        vctl::MCVector3D Normal = pTriangle->GetNormal();
        SFloatCoordinates Temp;
        Temp.x = (float)Normal.GetX();
        Temp.y = (float)Normal.GetY();
        Temp.z = (float)Normal.GetZ();

        // Zapis normaly
        if( !Channel.write((char *)&Temp, sizeof(SFloatCoordinates)) )
        {
            return false;
        }

        // Cyklus uzlu aktualniho tri
        for (int i = 0; i < 3; ++i)
        {
            // Ziskani souradnic node tri
            Temp.x = (float)pTriangleNode[i]->GetX();
            Temp.y = (float)pTriangleNode[i]->GetY();
            Temp.z = (float)pTriangleNode[i]->GetZ();

            // Zapis souradnic node
            if( !Channel.write((char *)&Temp, sizeof(SFloatCoordinates)) )
            {
                return false;
            }
        }

        // Zapis stl atributu tri
        if( !Channel.write((char *)&usAttribute, sizeof(unsigned short int)) )
        {
            return false;
        }

        // Ziskani nasledujiciho tri site
        pTriangle = pTriangle->GetNext();
    }

    // O.K.
    return true;
}


bool CTriMesh::saveVRML(mds::mod::CChannel& Channel)
{
    // Vystupni proud
    std::stringstream Stream;

    // Ziskani aktualniho casu
    time_t lTime;
    time(&lTime);

    // Zapsani hlavicky vrml souboru
    Stream << "#VRML V2.0 utf8" << std::endl << std::endl;
    Stream << "# Created by MDSTk." << std::endl;
    Stream << "# Date: " << ctime(&lTime) << std::endl << std::endl;

    // Zapis casti Transformace do vrml souboru
    Stream << "Transform {" << std::endl << "children [" << std::endl;

    // Zapis tvaru
    Stream << "Shape {" << std::endl;

    // Zapis materialovych vlastnosti
    Stream << "appearance Appearance { material Material { diffuseColor 0.9882 0.9882 0.9882 } }" << std::endl;

    // Zapis geometrie
    Stream << "geometry IndexedFaceSet {" << std::endl;

    // Zapis parametru site
    Stream << "ccw TRUE" << std::endl << "solid TRUE" << std::endl << "creaseAngle 1.5708" << std::endl;

    // Zapsani bodu site
    Stream << "coord Coordinate { point [" << std::endl;

    // Cyklus uzlu site pro jejich ulozeni do VRML
    vctl::MCVertex *pTriangleNode = getFirstVertex();
    unsigned uIndex = 0;
    while (pTriangleNode)
    {
        // Nastaveni cisla aktualniho node
        pTriangleNode->SetIndex(uIndex);
        ++uIndex;

        // Zapsani souradnic node do souboru
        Stream << pTriangleNode->GetX() << " "
            << pTriangleNode->GetY() << " "
            << pTriangleNode->GetZ() << ","
            << std::endl;

        // Ziskani nasledujiciho node
        pTriangleNode = pTriangleNode->GetNext();
    }

    // Uzavreni casti bodu site
    Stream << "] }" << std::endl;

    // Zapsani tri site
    Stream << "coordIndex [" << std::endl;

    // Cyklus tri site pro jejich ulozeni do VRML
    vctl::MCTri *pTriangle = getFirstTriangle();
    while (pTriangle)
    {
        // Zapsani indexu node tri do souboru
        Stream << pTriangle->GetVertex(0)->GetIndex() << ", "
            << pTriangle->GetVertex(1)->GetIndex() << ", "
            << pTriangle->GetVertex(2)->GetIndex() << ", -1,"
            << std::endl;

        // Ziskani nasledujiciho tri
        pTriangle = pTriangle->GetNext();
    }

    // Uzavreni casti tri site
    Stream << "]" << std::endl;

    // Uzavreni casti geometrie
    Stream << "}" << std::endl;

    // Uzavreni casti tvaru
    Stream << "}" << std::endl;

    // Uzavreni casti transformace
    Stream << "] }" << std::endl;

    // Zapis obsahu string streamu do kanalu
    if( !Channel.write(Stream.str().c_str(), Stream.str().length()) )
    {
        return false;
    }

    // O.K
    return true;
}


} // namespace seg
} // namespace mds

