//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2006 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsMakeTetraMesh.h                  \n
 * Section: mMakeTetraMesh                      \n
 * Date:    2006/05/09                          \n
 *
 * $Id: mdsMakeTetraMesh.cpp 1494 2009-05-21 14:36:00Z spanel $
 *
 * Description:
 * - Initializes trianugular mesh based on input edge image.
 * - First step of a segmentation based on Adaptive Delaunay triangulation.
 */

#include "mdsMakeTetraMesh.h"

// MDSTk
#include <MDSTk/Image/mdsDensityVolume.h>

// MDSTk Extensions
#include <MDSTk/VectorSegmentation/mdsTetraMesh.h>

// STL
#include <climits>
#include <iostream>
#include <string>


//==============================================================================
/*
 * Global constants.
 */

//! Module description
const std::string MODULE_DESCRIPTION    = "Module creates and initializes tetrahedral mesh based on input edge volume.";

//! Additional command line arguments
const std::string MODULE_ARGUMENTS      = "cell:nodes:edge:t:cdt";

//! Additional arguments names
const std::string MODULE_ARG_CELL       = "cell";
const std::string MODULE_ARG_NODES      = "nodes";
const std::string MODULE_ARG_EDGE       = "edge";
const std::string MODULE_ARG_T          = "t";
const std::string MODULE_ARG_CDT        = "cdt";

//! Default argument values
const int DEFAULT_CELL                  = mds::seg::CTetraMesh::DEFAULT_CELL;
const int DEFAULT_NODES                 = mds::seg::CTetraMesh::DEFAULT_NUMBER_OF_NODES;
const double DEFAULT_EDGE               = mds::seg::CTetraMesh::DEFAULT_MIN_EDGE_LENGTH;
const double DEFAULT_T                  = mds::seg::CTetraMesh::DEFAULT_INIT_THRESHOLD;

//! Allowed argument values
const double MIN_CELL                   = 1.0;
const double MAX_CELL                   = 256.0;

const double MIN_NODES                  = 2.0;
const double MAX_NODES                  = 256.0;

const double MIN_EDGE                   = 0.1;
const double MAX_EDGE                   = 50.0;

const double MIN_T                      = 0.0;
const double MAX_T                      = 5.0;


//==============================================================================
/*
 * Implementation of the class CMakeTetraMesh.
 */
CMakeTetraMesh::CMakeTetraMesh(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CMakeTetraMesh::~CMakeTetraMesh()
{
}


bool CMakeTetraMesh::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 2 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    m_iCell = DEFAULT_CELL;
    m_Arguments.value(MODULE_ARG_CELL, m_iCell);
    if( m_iCell > MAX_CELL || m_iCell < MIN_CELL )
    {
        MDS_CERR('<' << m_sFilename << "> Bad '" << MODULE_ARG_CELL << "' parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_iNodes = DEFAULT_NODES;
    m_Arguments.value(MODULE_ARG_NODES, m_iNodes);
    if( m_iNodes > MAX_NODES || m_iNodes < MIN_NODES )
    {
        MDS_CERR('<' << m_sFilename << "> Bad '" << MODULE_ARG_NODES << "' parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_dMinEdgeLength = DEFAULT_EDGE;
    m_Arguments.value(MODULE_ARG_EDGE, m_dMinEdgeLength);
    if( m_dMinEdgeLength < MIN_EDGE || m_dMinEdgeLength > MAX_EDGE )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_EDGE << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_dThreshold = DEFAULT_T;
    m_Arguments.value(MODULE_ARG_T, m_dThreshold);
    if( m_dThreshold > MAX_T || m_dThreshold < MIN_T )
    {
        MDS_CERR('<' << m_sFilename << "> Bad '" << MODULE_ARG_T << "' parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_bCDT = m_Arguments.exists(MODULE_ARG_CDT);

    // O.K.
    return true;
}


bool CMakeTetraMesh::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Create a new volume
    mds::img::CDensityVolumePtr spEdgeVolume(new mds::img::CDensityVolume);
    mds::img::CDensityVolumePtr spVolume(new mds::img::CDensityVolume);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read underlying volume from the input channel
    if( !readInput(pIChannel, spVolume.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input volume" << std::endl);
        return false;
    }


    // I/O channels
    pIChannel = getInput(1);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read underlying edge volume from the input channel
    if( !readInput(pIChannel, spEdgeVolume.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input edge volume" << std::endl);
        return false;
    }


    // Create a new mesh
    mds::seg::CTetraMeshPtr spMesh(new mds::seg::CTetraMesh);
    spMesh->setVolume(spVolume);
    spMesh->setEdgeVolume(spEdgeVolume.get());
    spMesh->setMinEdgeLength(m_dMinEdgeLength);

    if( m_bCDT )
    {
        spMesh->setConstraintFunc(mds::seg::CTetraMesh::checkBoundaryFacetFunc);
    }

    // Initialize the mesh
    if( !spMesh->init(m_dThreshold, m_iCell, m_iNodes) )
    {
        MDS_CERR('<' << m_sFilename << "> Cannot initialize tetrahedral mesh" << std::endl);
        return false;
    }

    // Construct the mesh
    if( !spMesh->construct() )
    {
        MDS_CERR('<' << m_sFilename << "> Cannot construct initial Delaunay triangulation" << std::endl);
        return false;
    }

    // Write the result to the output channel
    if( !writeOutput(pOChannel, spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write the tetrahedral mesh" << std::endl);
    }

    // Returning 'true' means to continue processing the input channel
    return true;
}


void CMakeTetraMesh::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CMakeTetraMesh::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR(std::endl);
    MDS_CERR("Extended usage: [-t dValue] [-cell iValue] [-nodes iValue] [-edge dValue]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("  -t  Threshold that affects the number of candidate vertex points." << std::endl);
    MDS_CERR("      A value in the range " << MIN_T << ".." << MAX_T << " can be given." << std::endl);
    MDS_CERR("      Default value is " << DEFAULT_T << "." << std::endl);
    MDS_CERR(std::endl);
    MDS_CERR("  -cell  Inpute image is divided into cells. Just one candidate" << std::endl);
    MDS_CERR("         vertex is taken from each cell. This parameter is size of" << std::endl);
    MDS_CERR("         the cell in pixels. Default value is " << DEFAULT_CELL << "." << std::endl);
    MDS_CERR(std::endl);
    MDS_CERR("  -nodes  The number of outline points placed on the mesh" << std::endl);
    MDS_CERR("          boundary. You can specify values within the range" << std::endl);
    MDS_CERR("          " << MIN_NODES << ".." << MAX_NODES << ". Default value is " << DEFAULT_NODES << "." << std::endl);
    MDS_CERR(std::endl);
    MDS_CERR("  -edge  Minimal allowed length of a tetrahedron edge. Values within" << std::endl);
    MDS_CERR("         the range " << MIN_EDGE << ".." << MAX_EDGE << " can be given." << std::endl);
    MDS_CERR("         Default value is " << DEFAULT_EDGE << "." << std::endl);
    MDS_CERR("Input: edge volume." << std::endl);
    MDS_CERR("Output: tetrahedral mesh." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CMakeTetraMeshPtr spModule(new CMakeTetraMesh(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

