//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2007 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsMakeTetraMeshControlSpace.h      \n
 * Section: mMakeTetraMeshControlSpace          \n
 * Date:    2007/03/02                          \n
 *
 * $Id: mdsMakeTetraMeshControlSpace.cpp 1494 2009-05-21 14:36:00Z spanel $
 *
 * Description:
 * - Initializes control space used to construct an anisotropic tetrahedral mesh.
 * - Construction of the control space is based on input edge data.
 */

#include "mdsMakeTetraMeshControlSpace.h"

// MDSTk
#include <MDSTk/Image/mdsDensityVolume.h>

// MDSTk Extensions
#include <MDSTk/VectorSegmentation/mdsTetraMesh.h>

// STL
#include <climits>
#include <iostream>
#include <string>


//==============================================================================
/*
 * Global constants.
 */

//! Module description
const std::string MODULE_DESCRIPTION    = "Module creates a control space used to construct an anisotropic tetrahedral mesh.";

//! Additional command line arguments
const std::string MODULE_ARGUMENTS      = "minlength:tetra:k";

//! Additional arguments names
const std::string MODULE_ARG_MIN_LENGTH = "minlength";
const std::string MODULE_ARG_TETRA      = "tetra";
const std::string MODULE_K              = "k";

//! Default arguments values
const double DEFAULT_MIN_LENGTH         = mds::seg::CTetraMesh::DEFAULT_MIN_EDGE_LENGTH;
const double DEFAULT_TETRA              = mds::seg::CTetraMesh::DEFAULT_AVG_TETRA_SIZE;
const double DEFAULT_K                  = mds::seg::CTetraMesh::DEFAULT_K;

//! Allowed arguments values
const double MIN_LENGTH                 = 0.1;
const double MAX_LENGTH                 = 50.0;
const double MIN_TETRA                  = 1.0;
const double MAX_TETRA                  = 250.0;


//==============================================================================
/*
 * Implementation of the class CMakeTetraMeshControlSpace.
 */
CMakeTetraMeshControlSpace::CMakeTetraMeshControlSpace(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CMakeTetraMeshControlSpace::~CMakeTetraMeshControlSpace()
{
}


bool CMakeTetraMeshControlSpace::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 2 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // Parse command line arguments
    m_dMinLength = DEFAULT_MIN_LENGTH;
    m_Arguments.value(MODULE_ARG_MIN_LENGTH, m_dMinLength);
    if( m_dMinLength < MIN_LENGTH || m_dMinLength > MAX_LENGTH )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_MIN_LENGTH << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_dTetra = DEFAULT_TETRA;
    m_Arguments.value(MODULE_ARG_TETRA, m_dTetra);
    if( m_dTetra < MIN_TETRA || m_dTetra > MAX_TETRA )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_TETRA << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_dK = DEFAULT_K;
    m_Arguments.value(MODULE_K, m_dK);
    if( m_dK < 0.0 )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_K << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // O.K.
    return true;
}


bool CMakeTetraMeshControlSpace::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Create a new volume
    mds::img::CDensityVolumePtr spVolume, spEdgeVolume, spControlSpace;

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read underlying volume from the input channel
    if( !readInput(pIChannel, spVolume.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input volume" << std::endl);
        return false;
    }


    // I/O channels
    pIChannel = getInput(1);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read underlying edge volume from the input channel
    if( !readInput(pIChannel, spEdgeVolume.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input edge volume" << std::endl);
        return false;
    }

    // Create a new mesh
    mds::seg::CTetraMeshPtr spMesh;
    spMesh->setVolume(spVolume);
    spMesh->setEdgeVolume(spEdgeVolume.get());
    spMesh->setControlSpace(spControlSpace);
    spMesh->setMinEdgeLength(m_dMinLength);

    // Create control space
//    if( !spMesh->prepareControlSpace() )
//    if( !spMesh->prepareControlSpace(m_dTetra) )
    if( !spMesh->prepareControlSpace(m_dK, m_dTetra) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to create the control space" << std::endl);
        return false;
    }

    // Write the result to the output channel
    if( !writeOutput(pOChannel, spControlSpace.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write the output volume" << std::endl);
        return false;
    }

    // Returning 'true' means to continue processing the input channel
    return false;
}


void CMakeTetraMeshControlSpace::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CMakeTetraMeshControlSpace::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR(std::endl);
    MDS_CERR("Extended usage: [-minlength dValue] [-tetra dValue] [-k dValue]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("  -minlength  Minimal allowed length of a tetrahedron edge. Values within" << std::endl);
    MDS_CERR("              the range " << MIN_LENGTH << ".." << MAX_LENGTH << " can be given." << std::endl);
    MDS_CERR("              Default value is " << DEFAULT_MIN_LENGTH << "." << std::endl);
    MDS_CERR("Input: original volume data and edge volume." << std::endl);
    MDS_CERR("Output: volume data representing control space." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CMakeTetraMeshControlSpacePtr spModule(new CMakeTetraMeshControlSpace(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

