//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2005 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsMakeTriMesh.cpp                  \n
 * Section: mMakeTriMesh                        \n
 * Date:    2005/09/27                          \n
 *
 * $Id: mdsMakeTriMesh.cpp 1494 2009-05-21 14:36:00Z spanel $
 *
 * Description:
 * - Initializes trianugular mesh based on input edge image.
 * - First step of a segmentation based on Adaptive Delaunay triangulation.
 */

#include "mdsMakeTriMesh.h"

// MDSTk
#include <MDSTk/Image/mdsSlice.h>

// MDSTk Extensions
#include <MDSTk/VectorSegmentation/mdsTriMesh.h>

// STL
#include <climits>
#include <iostream>
#include <string>


//==============================================================================
/*
 * Global constants.
 */

//! Module description
const std::string MODULE_DESCRIPTION    = "Module creates and initializes trianugular mesh based on input edge image.";

//! Additional command line arguments
const std::string MODULE_ARGUMENTS      = "cell:nodes:t";

//! Additional arguments names
const std::string MODULE_ARG_CELL       = "cell";
const std::string MODULE_ARG_NODES      = "nodes";
const std::string MODULE_ARG_T          = "t";

//! Default argument values
const int DEFAULT_CELL                  = mds::seg::CTriMesh::DEFAULT_CELL;
const int DEFAULT_NODES                 = mds::seg::CTriMesh::DEFAULT_NUMBER_OF_NODES;
const double DEFAULT_T                  = mds::seg::CTriMesh::DEFAULT_INIT_THRESHOLD;

//! Allowed argument values
const double MIN_CELL                   = 1.0;
const double MAX_CELL                   = 32.0;

const double MIN_NODES                  = 2.0;
const double MAX_NODES                  = 256.0;

const double MIN_T                      = 0.0;
const double MAX_T                      = 5.0;


//==============================================================================
/*
 * Implementation of the class CMakeTriMesh.
 */
CMakeTriMesh::CMakeTriMesh(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CMakeTriMesh::~CMakeTriMesh()
{
}


bool CMakeTriMesh::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 1 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // Parameter 'cell'
    m_iCell = DEFAULT_CELL;
    m_Arguments.value(MODULE_ARG_CELL, m_iCell);
    if( m_iCell > MAX_CELL || m_iCell < MIN_CELL )
    {
        MDS_CERR('<' << m_sFilename << "> Bad '" << MODULE_ARG_CELL << "' parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // Parameter 'nodes'
    m_iNodes = DEFAULT_NODES;
    m_Arguments.value(MODULE_ARG_NODES, m_iNodes);
    if( m_iNodes > MAX_NODES || m_iNodes < MIN_NODES )
    {
        MDS_CERR('<' << m_sFilename << "> Bad '" << MODULE_ARG_NODES << "' parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // Parameter 't'
    m_dThreshold = DEFAULT_T;
    m_Arguments.value(MODULE_ARG_T, m_dThreshold);
    if( m_dThreshold > MAX_T || m_dThreshold < MIN_T )
    {
        MDS_CERR('<' << m_sFilename << "> Bad '" << MODULE_ARG_T << "' parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // O.K.
    return true;
}


bool CMakeTriMesh::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Create a new slice
    mds::img::CSlicePtr spSlice;

    // Wait for data
    if( pIChannel->wait(1000) )
    {
        // Read slice from the input channel
        if( readInput(pIChannel, spSlice.get()) )
        {
            // Create a new triangular mesh
            mds::seg::CTriMeshPtr spMesh;
            spMesh->setEdgeImage(spSlice);

            // Initialize the triangular mesh
            if( !spMesh->init(m_dThreshold, m_iCell, m_iNodes) )
            {
                MDS_CERR('<' << m_sFilename << "> Cannot initialize triangular mesh" << std::endl);
                return false;
            }

            // Construct the mesh
            if( !spMesh->construct() )
            {
                MDS_CERR('<' << m_sFilename << "> Cannot construct initial Delaunay triangulation" << std::endl);
                return false;
            }

            // Write the result to the output channel
            if( !writeOutput(pOChannel, spMesh.get()) )
            {
                MDS_CERR('<' << m_sFilename << "> Failed to write the triangular mesh" << std::endl);
            }
        }
        else
        {
            MDS_CERR('<' << m_sFilename << "> Failed to read input slice" << std::endl);
            return false;
        }
    }
    else
    {
        MDS_LOG_NOTE("Wait timeout");
    }

    // Returning 'true' means to continue processing the input channel
    return true;
}


void CMakeTriMesh::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CMakeTriMesh::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR(std::endl);
    MDS_CERR("Extended usage: [-t dValue] [-cell iValue] [-nodes iValue]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("  -t  Threshold that affects the number of candidate vertex points." << std::endl);
    MDS_CERR("      A value in the range " << MIN_T << ".." << MAX_T << " can be given." << std::endl);
    MDS_CERR("      Default value is " << DEFAULT_T << "." << std::endl);
    MDS_CERR(std::endl);
    MDS_CERR("  -cell  Inpute image is divided into cells. Just one candidate" << std::endl);
    MDS_CERR("         vertex is taken from each cell. This parameter is size of" << std::endl);
    MDS_CERR("         the cell in pixels. Default value is " << DEFAULT_CELL << "." << std::endl);
    MDS_CERR(std::endl);
    MDS_CERR("  -nodes  The number of outline points placed on the mesh" << std::endl);
    MDS_CERR("          boundary. You can specify values within the range" << std::endl);
    MDS_CERR("          " << MIN_NODES << ".." << MAX_NODES << ". Default value is " << DEFAULT_NODES << "." << std::endl);
    MDS_CERR("Input: edge image." << std::endl);
    MDS_CERR("Output: triangular mesh." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CMakeTriMeshPtr spModule(new CMakeTriMesh(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

