//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2006 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsSaveTetraMesh.cpp                \n
 * Section: mSaveTetraMesh                      \n
 * Date:    2006/05/09                          \n
 *
 * $Id: mdsSaveTetraMesh.cpp 1523 2009-11-24 10:11:12Z spanel $
 *
 * File description:
 * - Saves input tetrahedral mesh to the VRML or STL format.
 */

#include "mdsSaveTetraMesh.h"

// MDSTk
#include <MDSTk/VectorSegmentation/mdsTetraMesh.h>

// STL
#include <climits>
#include <iostream>
#include <string>


//==============================================================================
/*
 * Global constants.
 */

//! Module description
const std::string MODULE_DESCRIPTION    = "Module saves an input tetrahedral mesh to the VRML or STL format.";

//! Additional command line arguments
const std::string MODULE_ARGUMENTS      = "format:region:boundary:cut:slivers:tetquality";

//! Additional arguments names
const std::string MODULE_ARG_FORMAT     = "format";
const std::string MODULE_ARG_REGION     = "region";
const std::string MODULE_ARG_BOUNDARY   = "boundary";
const std::string MODULE_ARG_CUT        = "cut";
const std::string MODULE_ARG_SLIVERS    = "slivers";
const std::string MODULE_ARG_TETQUALITY = "tetquality";

//! Allowed 'format' argument values
const std::string F_VRML_NAME           = "vrml";
const std::string F_STL_NAME            = "stl";

//! Default region argument value
const int DEFAULT_REGION                = -1;


//==============================================================================
/*
 * Implementation of the class CSaveTetraMesh.
 */
CSaveTetraMesh::CSaveTetraMesh(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CSaveTetraMesh::~CSaveTetraMesh()
{
}


bool CSaveTetraMesh::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 1 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // Is 'format' parameter given?
    if( !m_Arguments.exists(MODULE_ARG_FORMAT) )
    {
        MDS_CERR('<' << m_sFilename << "> Format was not specified: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // Type of resulting data
    std::string sFormatName;
    m_Arguments.value(MODULE_ARG_FORMAT, sFormatName);
    if( sFormatName == F_VRML_NAME )
    {
        m_eDataFormat = F_VRML;
    }
    else if( sFormatName == F_STL_NAME )
    {
        m_eDataFormat = F_STL;
    }
    else
    {
        MDS_CERR('<' << m_sFilename << "> Bad '" << MODULE_ARG_FORMAT << "' parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // Parameter 'region'
    m_iRegion = DEFAULT_REGION;
    m_Arguments.value(MODULE_ARG_REGION, m_iRegion);
    if( m_iRegion < -1 )
    {
        MDS_CERR('<' << m_sFilename << "> Bad '" << MODULE_ARG_REGION << "' parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_dTetQuality = 50.0;
    m_Arguments.value(MODULE_ARG_TETQUALITY, m_dTetQuality);
    if( m_dTetQuality < 0.0 )
    {
        MDS_CERR('<' << m_sFilename << "> Bad '" << MODULE_ARG_TETQUALITY << "' parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // Parameter 'boundary'
    m_bBoundary = m_Arguments.exists(MODULE_ARG_BOUNDARY);
    m_bCut = m_Arguments.exists(MODULE_ARG_CUT);
    m_bSlivers = m_Arguments.exists(MODULE_ARG_SLIVERS);

    // O.K.
    return true;
}


bool CSaveTetraMesh::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Create a new tetrahedral mesh
    mds::seg::CTetraMeshPtr spMesh;

    // Wait for data
    if( pIChannel->wait(1000) )
    {
        // Read mesh from the input channel
        if( readInput(pIChannel, spMesh.get()) )
        {
            // Generate a list of all triangles of a specified region
            spMesh->clearTriangles();
            if( !m_bCut && !m_bSlivers )
            {
                spMesh->createTriangles(m_iRegion, m_bBoundary);
            }
            else if( m_bCut )
            {
                // Get minimal and maximal coordinates
                double dMinX = 0.0, dMinY = 0.0, dMinZ = 0.0;
                double dMaxX = 0.0, dMaxY = 0.0, dMaxZ = 0.0;
                spMesh->getMinCoordinates(dMinX, dMinY, dMinZ);
                spMesh->getMaxCoordinates(dMaxX, dMaxY, dMaxZ);

                // Generate triangles
                double dPosition = -0.4 * (dMaxX - dMinX + dMaxY - dMinY + dMaxZ - dMinZ);
                spMesh->createTriangles(1.0, 1.0, 1.0, dPosition, m_iRegion);
            }
            else if( m_bSlivers )
            {
                spMesh->createTriangles(m_dTetQuality, m_iRegion);
            }

            // Write the mesh to the output channel
            switch( m_eDataFormat )
            {
                case F_VRML:
                    if( !spMesh->saveVRML(*pOChannel) )
                    {
                        MDS_CERR('<' << m_sFilename << "> Failed to write the list of triangles" << std::endl);
                    }
                    break;

                case F_STL:
                    if( !spMesh->saveSTL(*pOChannel) )
                    {
                        MDS_CERR('<' << m_sFilename << "> Failed to write the list of triangles" << std::endl);
                    }
                    break;

                default:
/*                    if( !writeOutput(pOChannel, spMesh) )
                    {
                        MDS_CERR('<' << m_sFilename << "> Failed to write the tetrahedral mesh" << std::endl);
                    }*/
                    break;
            }
        }
        else
        {
            MDS_CERR('<' << m_sFilename << "> Failed to read input mesh" << std::endl);
            return false;
        }
    }
    else
    {
        MDS_LOG_NOTE("Wait timeout");
    }

    // Returning 'true' means to continue processing the input channel
    return true;
}


void CSaveTetraMesh::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CSaveTetraMesh::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR(std::endl);
    MDS_CERR("Extended usage: [-format strValue] [-region iValue] [-boundary]" << std::endl);
    MDS_CERR("                [-slivers] [-tetquality dValue]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("  -format  Specifies format of the output data. This module" << std::endl);
    MDS_CERR("           is able to produce a list of triangles in VRML, STL and" << std::endl);
    MDS_CERR("           format. You can choose one of those types" << std::endl);
    MDS_CERR("           by giving:" << std::endl);
    MDS_CERR("             " << F_VRML_NAME << " or" << std::endl);
    MDS_CERR("             " << F_STL_NAME << "." << std::endl);
    MDS_CERR(std::endl);
    MDS_CERR("  -region  Index of the region/segment in the mesh. The output list" << std::endl);
    MDS_CERR("           of containes all triangles which belong to this region." << std::endl);
    MDS_CERR("           Default values is " << DEFAULT_REGION << "." << std::endl);
    MDS_CERR(std::endl);
    MDS_CERR("  -boundary  Generates only tringles which forms boundary of the given" << std::endl);
    MDS_CERR("             region." << std::endl);
    MDS_CERR("Input: tetrahedral mesh." << std::endl);
    MDS_CERR("Output: VRML or STL." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CSaveTetraMeshPtr spModule(new CSaveTetraMesh(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

