//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2005 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsSaveTriMesh.cpp                  \n
 * Section: mSaveTriMesh                        \n
 * Date:    2005/09/28                          \n
 *
 * $Id: mdsSaveTriMesh.cpp 1494 2009-05-21 14:36:00Z spanel $
 *
 * File description:
 * - Saves an input triangular mesh to the VRML or STL format.
 */

#include "mdsSaveTriMesh.h"

// MDSTk
#include <MDSTk/VectorSegmentation/mdsTriMesh.h>

// STL
#include <climits>
#include <iostream>
#include <string>


//==============================================================================
/*
 * Global constants.
 */

//! Module description
const std::string MODULE_DESCRIPTION    = "Module saves an input triangular mesh to the VRML or STL format.";

//! Additional command line arguments
const std::string MODULE_ARGUMENTS      = "format";

//! Additional arguments names
const std::string MODULE_ARG_FORMAT     = "format";

//! Allowed 'format' argument values
const std::string F_VRML_NAME           = "vrml";
const std::string F_STL_NAME            = "stl";


//==============================================================================
/*
 * Implementation of the class CSaveTriMesh.
 */
CSaveTriMesh::CSaveTriMesh(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CSaveTriMesh::~CSaveTriMesh()
{
}


bool CSaveTriMesh::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 1 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // Is 'format' parameter given?
    if( !m_Arguments.exists(MODULE_ARG_FORMAT) )
    {
        MDS_CERR('<' << m_sFilename << "> Format was not specified: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // Type of resulting data
    std::string sFormatName;
    m_Arguments.value(MODULE_ARG_FORMAT, sFormatName);
    if( sFormatName == F_VRML_NAME )
    {
        m_eDataFormat = F_VRML;
    }
    else if( sFormatName == F_STL_NAME )
    {
        m_eDataFormat = F_STL;
    }
    else
    {
        MDS_CERR('<' << m_sFilename << "> Bad '" << MODULE_ARG_FORMAT << "' parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // O.K.
    return true;
}


bool CSaveTriMesh::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Create a new triangular mesh
    mds::seg::CTriMeshPtr spMesh;

    // Wait for data
    if( pIChannel->wait(1000) )
    {
        // Read mesh from the input channel
        if( readInput(pIChannel, spMesh.get()) )
        {
            // Write the mesh to the output channel
            switch( m_eDataFormat )
            {
                case F_VRML:
                    if( !spMesh->saveVRML(*pOChannel) )
                    {
                        MDS_CERR('<' << m_sFilename << "> Failed to write the triangular mesh" << std::endl);
                    }
                    break;

                case F_STL:
                    if( !spMesh->saveSTL(*pOChannel) )
                    {
                        MDS_CERR('<' << m_sFilename << "> Failed to write the triangular mesh" << std::endl);
                    }
                    break;

                default:
                    if( !writeOutput(pOChannel, spMesh.get()) )
                    {
                        MDS_CERR('<' << m_sFilename << "> Failed to write the triangular mesh" << std::endl);
                    }
                    break;
            }
        }
        else
        {
            MDS_CERR('<' << m_sFilename << "> Failed to read input mesh" << std::endl);
            return false;
        }
    }
    else
    {
        MDS_LOG_NOTE("Wait timeout");
    }

    // Returning 'true' means to continue processing the input channel
    return true;
}


void CSaveTriMesh::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CSaveTriMesh::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR(std::endl);
    MDS_CERR("Extended usage: [-format strValue]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("  -format  Specifies format of the output data. This module" << std::endl);
    MDS_CERR("           is able to produce 2D triangular mesh in VRML, STL and" << std::endl);
    MDS_CERR("           MDSTk native format. You can choose one of those types" << std::endl);
    MDS_CERR("           by giving:" << std::endl);
    MDS_CERR("             " << F_VRML_NAME << " or" << std::endl);
    MDS_CERR("             " << F_STL_NAME << "." << std::endl);
    MDS_CERR("Input: triangular mesh." << std::endl);
    MDS_CERR("Output: VRML or STL." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CSaveTriMeshPtr spModule(new CSaveTriMesh(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

