//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2006 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsTetraMeshAdapt.cpp               \n
 * Section: mTetraMeshAdapt                     \n
 * Date:    2006/06/17                          \n
 *
 * $Id: mdsTetraMeshAdapt.cpp 1494 2009-05-21 14:36:00Z spanel $
 *
 * Description:
 * - Adaptation of tetrahedral mesh to an underlying volume data.
 * - An empty module at present!
 */

#include "mdsTetraMeshAdapt.h"

// STL
#include <climits>
#include <iostream>
#include <string>


//==============================================================================
/*
 * Global constants.
 */

//! Module description.
const std::string MODULE_DESCRIPTION    = "Adapts tetrahedral mesh to underlying volume data.";

//! Additional command line arguments.
const std::string MODULE_ARGUMENTS      = "";
//const std::string MODULE_ARGUMENTS      = "minlength:threshold:rho";

//! Additional arguments names.
const std::string MODULE_ARG_MIN_LENGTH = "minlength";
const std::string MODULE_ARG_THRESHOLD  = "threshold";
const std::string MODULE_ARG_RHO        = "rho";

//! Default arguments values.
//const double DEFAULT_MIN_LENGTH         = mds::seg::CTetraMesh::DEFAULT_MIN_EDGE_LENGTH;
//const double DEFAULT_THRESHOLD          = mds::seg::CTetraMesh::DEFAULT_HOMOGENEITY_THRESHOLD;
//const double DEFAULT_RHO                = mds::seg::CTetraMesh::DEFAULT_RHO;

//! Allowed arguments values
//const double MIN_LENGTH                 = 1.0;
//const double MAX_LENGTH                 = 50.0;

//const double MIN_THRESHOLD              = 0.0;
//const double MAX_THRESHOLD              = 1.0;

//const double MIN_RHO                    = 0.0;
//const double MAX_RHO                    = 20.0;


//==============================================================================
/*
 * Implementation of the class CTetraMeshAdapt.
 */
CTetraMeshAdapt::CTetraMeshAdapt(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CTetraMeshAdapt::~CTetraMeshAdapt()
{
}


bool CTetraMeshAdapt::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 3 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // Parse command line arguments
/*    m_dMinLength = DEFAULT_MIN_LENGTH;
    m_Arguments.value(MODULE_ARG_MIN_LENGTH, m_dMinLength);
    if( m_dMinLength < MIN_LENGTH || m_dMinLength > MAX_LENGTH )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_MIN_LENGTH << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_dHomogeneityThreshold = DEFAULT_THRESHOLD;
    m_Arguments.value(MODULE_ARG_THRESHOLD, m_dHomogeneityThreshold);
    if( m_dHomogeneityThreshold < MIN_THRESHOLD || m_dHomogeneityThreshold > MAX_THRESHOLD )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_THRESHOLD << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_dRho = DEFAULT_RHO;
    m_Arguments.value(MODULE_ARG_RHO, m_dRho);
    if( m_dRho < MIN_RHO || m_dRho > MAX_RHO )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_RHO << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }*/

    // O.K.
    return true;
}


bool CTetraMeshAdapt::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read underlying volume from the input channel
    if( !readInput(pIChannel, m_spVolume.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input volume" << std::endl);
        return false;
    }


    // I/O channels
    pIChannel = getInput(1);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read underlying edge volume from the input channel
    if( !readInput(pIChannel, m_spEdgeVolume.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input edge volume" << std::endl);
        return false;
    }


    // I/O channels
    pIChannel = getInput(2);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read mesh from the input channel
    if( !readInput(pIChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input tetrahedral mesh" << std::endl);
        return false;
    }

    // Create the volume margin
    m_spVolume->mirrorMargin();

    // Set the underlying mesh volume
    m_spMesh->setVolume(m_spVolume);
    m_spMesh->setEdgeVolume(m_spEdgeVolume.get());

    // Evaluate tetrahedra features
//    m_spMesh->evaluateFeatures();

    // Adapt the triangular mesh
//    m_spMesh->tetrahedronSplitting(m_dMinLength, m_dHomogeneityThreshold, m_dRho);

    // Write the result to the output channel
    if( !writeOutput(pOChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write the tetrahedral mesh" << std::endl);
    }

    // Returning 'true' means to continue processing the input channel
    return false;
}


void CTetraMeshAdapt::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTetraMeshAdapt::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR(std::endl);
    MDS_CERR("Extended usage:" << std::endl);
//    MDS_CERR("Extended usage: [-minlength dValue] [-threshold dValue] [-rho dValue]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("Input: volume data, edge volume, tetrahedral mesh." << std::endl);
    MDS_CERR("Output: tetrahedral mesh." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CTetraMeshAdaptPtr spModule(new CTetraMeshAdapt(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

