//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)     \n
 * Copyright (c) 2003-2008 by Michal Spanel      \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz   \n
 * File:    mdsTetraMeshApproachEdges.cpp        \n
 * Section: mTetraMeshApproachEdges              \n
 * Date:    2008/02/12                           \n
 *
 * $Id: mdsTetraMeshApproachEdges.cpp 1494 2009-05-21 14:36:00Z spanel $
 *
 * Description:
 * - Introduces all image edge points not present in the mesh.
 */

#include "mdsTetraMeshApproachEdges.h"

// STL
#include <climits>
#include <iostream>
#include <string>


//==============================================================================
/*
 * Global constants.
 */

//! Module description
const std::string MODULE_DESCRIPTION    = "Surface recovery by stitching.";

//! Additional command line arguments
const std::string MODULE_ARGUMENTS      = "minlength:checkiso";

//! Additional arguments names
const std::string MODULE_ARG_MIN_LENGTH = "minlength";
const std::string MODULE_ARG_CHECK_ISO  = "checkiso";

//! Default arguments values
const double DEFAULT_MIN_LENGTH         = mds::seg::CTetraMesh::DEFAULT_MIN_EDGE_LENGTH;

//! Allowed arguments values
const double MIN_LENGTH                 = 0.1;
const double MAX_LENGTH                 = 50.0;


//==============================================================================
/*
 * Implementation of the class CTetraMeshApproachEdges.
 */
CTetraMeshApproachEdges::CTetraMeshApproachEdges(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CTetraMeshApproachEdges::~CTetraMeshApproachEdges()
{
}


bool CTetraMeshApproachEdges::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 3 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // Command line arguments
    m_dMinLength = DEFAULT_MIN_LENGTH;
    m_Arguments.value(MODULE_ARG_MIN_LENGTH, m_dMinLength);
    if( m_dMinLength < MIN_LENGTH || m_dMinLength > MAX_LENGTH )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_MIN_LENGTH << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_bCheckIsotropicMinEdgeLength = m_Arguments.exists(MODULE_ARG_CHECK_ISO);

    // O.K.
    return true;
}


bool CTetraMeshApproachEdges::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read underlying edge volume from the input channel
    if( !readInput(pIChannel, m_spEdgeVolume.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read edge data" << std::endl);
        return false;
    }


    // I/O channels
    pIChannel = getInput(1);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read volume from the input channel
    if( !readInput(pIChannel, m_spControlSpace.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input control space" << std::endl);
        return false;
    }


    // I/O channels
    pIChannel = getInput(2);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read underlying edge volume from the input channel
    if( !readInput(pIChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input mesh" << std::endl);
        return false;
    }


    // Set the underlying volume
    m_spMesh->setEdgeVolume(m_spEdgeVolume.get());
    m_spMesh->setControlSpace(m_spControlSpace);
    m_spMesh->setMinEdgeLength(m_dMinLength);
//    m_spMesh->setConstraintFunc(mds::seg::CTetraMesh::checkBoundaryFacetFunc);

    // Optimize the mesh
    m_spMesh->approachEdges(m_bCheckIsotropicMinEdgeLength);

    // Write the result to the output channel
    if( !writeOutput(pOChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write output tetrahedral mesh" << std::endl);
    }

    // Returning 'true' means to continue processing the input channel
    return false;
}


void CTetraMeshApproachEdges::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTetraMeshApproachEdges::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR(std::endl);
    MDS_CERR("Extended usage: [-minlength dValue] [-checkiso]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("  -minlength  Minimal allowed length of a shortest tetrahedron edge" << std::endl);
    MDS_CERR("              in the mesh. Default value is " << DEFAULT_MIN_LENGTH << "." << std::endl);
    MDS_CERR(std::endl);
    MDS_CERR("  -checkiso" << std::endl);
    MDS_CERR(std::endl);
    MDS_CERR("Input: edge volume, control space, tetrahedral mesh." << std::endl);
    MDS_CERR("Output: tetrahedral mesh." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CTetraMeshApproachEdgesPtr spModule(new CTetraMeshApproachEdges(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

