//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2007 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsTetraMeshChangeRegion.h          \n
 * Section: mTetraMeshChangeRegion              \n
 * Date:    2007/03/22                          \n
 *
 * $Id: mdsTetraMeshChangeRegion.cpp 1494 2009-05-21 14:36:00Z spanel $
 *
 * Description:
 * - Overrides classification of a specified region.
 */

#include "mdsTetraMeshChangeRegion.h"

// MDSTk
#include <MDSTk/Math/mdsVector.h>
#include <MDSTk/Math/mdsVectorFunctions.h>

// STL
#include <cstring>
#include <iostream>
#include <string>
#include <sstream>


//==============================================================================
/*
 * Global constants.
 */

//! Module description
const std::string MODULE_DESCRIPTION    = "Overrides classification of specified regions.";

//! Additional command line arguments
const std::string MODULE_ARGUMENTS      = "min:max:region";

//! Additional arguments names
const std::string MODULE_ARG_MIN        = "min";
const std::string MODULE_ARG_MAX        = "max";
const std::string MODULE_ARG_REGION     = "region";

//! Default argument values
const int DEFAULT_MIN                   = -1;
const int DEFAULT_MAX                   = -1; 
const int DEFAULT_REGION                = -1; 


//==============================================================================
/*
 * Implementation of the class CTetraMeshChangeRegion.
 */
CTetraMeshChangeRegion::CTetraMeshChangeRegion(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CTetraMeshChangeRegion::~CTetraMeshChangeRegion()
{
}


bool CTetraMeshChangeRegion::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 1 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // Parse command line arguments
    m_iMin = DEFAULT_MIN;
    m_Arguments.value(MODULE_ARG_MIN, m_iMin);
    if( m_iMin < -1 )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_MIN << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_iMax = DEFAULT_MAX;
    m_Arguments.value(MODULE_ARG_MAX, m_iMax);
    if( m_iMax < -1 )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_MAX << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_iRegion = DEFAULT_REGION;
    m_Arguments.value(MODULE_ARG_REGION, m_iRegion);
    if( m_iRegion < -1 )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_REGION << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // O.K.
    return true;
}


bool CTetraMeshChangeRegion::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read mesh from the input channel
    if( !readInput(pIChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input tetrahedral mesh" << std::endl);
        return false;
    }

    // Find individual isolated regions in the mesh
    m_spMesh->changeRegion(m_iMin, m_iMax, m_iRegion);
    
    // Write the result to the output channel
    if( !writeOutput(pOChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write the output tetrahedral mesh" << std::endl);
    }

    // Returning 'true' means to continue processing the input channel
    return true;
}


void CTetraMeshChangeRegion::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTetraMeshChangeRegion::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR(std::endl);
    MDS_CERR("Extended usage:" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("Input: tetrahedral mesh." << std::endl);
    MDS_CERR("Output: tetrahedral mesh." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CTetraMeshChangeRegionPtr spModule(new CTetraMeshChangeRegion(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

