//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2006 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsTetraMeshClassifyEM.h            \n
 * Section: mTetraMeshClassifyEM                \n
 * Date:    2006/06/08                          \n
 *
 * $Id: mdsTetraMeshClassifyEM.cpp 1517 2009-10-02 13:58:11Z spanel $
 *
 * Description:
 * - Classification of tetrahedra in a mesh.
 */

#include "mdsTetraMeshClassifyEM.h"

// MDSTk Extensions
#include <MDSTk/VectorSegmentation/mdsTetraMembershipEM.h>

// STL
#include <climits>
#include <iostream>
#include <string>


//==============================================================================
/*
 * Global constants.
 */

//! Module description.
const std::string MODULE_DESCRIPTION    = "Module that classifies tetrahedra into segments.";

//! Additional command line arguments.
const std::string MODULE_ARGUMENTS      = "add:clusters";

//! Additional arguments names.
const std::string MODULE_ARGUMENT_ADD   = "add";
const std::string MODULE_ARGUMENT_C     = "clusters";

//! Default arguments values.
const double DEFAULT_ADD                = 0.01;
const mds::tSize DEFAULT_C              = 0;

//! Allowed arguments values.
const double MIN_ADD                    = 0.0;
const double MAX_ADD                    = 1.0;


//==============================================================================
/*
 * Implementation of the class CTetraMeshClassifyEM.
 */
CTetraMeshClassifyEM::CTetraMeshClassifyEM(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CTetraMeshClassifyEM::~CTetraMeshClassifyEM()
{
}


bool CTetraMeshClassifyEM::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 2 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // Parse command line arguments
    m_dMinAddChange = DEFAULT_ADD;
    m_Arguments.value(MODULE_ARGUMENT_ADD, m_dMinAddChange);
    if( m_dMinAddChange < MIN_ADD || m_dMinAddChange > MAX_ADD )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARGUMENT_ADD << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_NumOfClusters = DEFAULT_C;
    m_Arguments.value(MODULE_ARGUMENT_C, m_NumOfClusters);
    if( m_NumOfClusters < 0 )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARGUMENT_C << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // O.K.
    return true;
}


bool CTetraMeshClassifyEM::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read slice from the input channel
    if( !readInput(pIChannel, m_spVolume.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input volume" << std::endl);
        return false;
    }

    // I/O channels
    pIChannel = getInput(1);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read mesh from the input channel
    if( !readInput(pIChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input tetrahedral mesh" << std::endl);
        return false;
    }

    // Create the volume margin
    m_spVolume->mirrorMargin();

    // Set the underlying volume
    m_spMesh->setVolume(m_spVolume);

    // Evaluate tetrahedra features
    m_spMesh->evaluateFeatures();

    // Region membership function
    mds::seg::CTetraMembershipEM Membership(m_dMinAddChange);

    // Classify the triangles
    if( m_NumOfClusters > 0 )
    {
        Membership(m_spMesh->getFirstTetrahedron(), m_NumOfClusters);
    }
    else
    {
        Membership(m_spMesh->getFirstTetrahedron());
    }

    m_spMesh->setNumOfSegments(Membership.getNumOfSegments());

    // Write the result to the output channel
    if( !writeOutput(pOChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write the output tetrahedral mesh" << std::endl);
    }

    // Returning 'true' means to continue processing the input channel
    return false;
}


void CTetraMeshClassifyEM::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTetraMeshClassifyEM::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR("Extended usage: [-add dValue] [-clusters iValue]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("  -add  Parameter of the EM algorithm. Minimal required change of the log-likelihood function" << std::endl);
    MDS_CERR("        which results in a new gaussian component addition. See its documentation." << std::endl);
    MDS_CERR("        Value in the range " << MIN_ADD << ".." << MAX_ADD << " can be given. Default value is " << DEFAULT_ADD << "." << std::endl);
    MDS_CERR("Input: volume data, tetrahedral mesh." << std::endl);
    MDS_CERR("Output: tetrahedral mesh." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CTetraMeshClassifyEMPtr spModule(new CTetraMeshClassifyEM(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

