//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2006 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsTetraMeshEdgeSplitting.cpp       \n
 * Section: mTetraMeshEdgeSplitting             \n
 * Date:    2006/06/07                          \n
 *
 * $Id: mdsTetraMeshEdgeSplitting.cpp 1535 2009-12-11 15:07:36Z spanel $
 *
 * Description:
 * - Adaptation of tetrahedral mesh to an underlying volume data.
 * - Edge splitting algorithm.
 */

#include "mdsTetraMeshEdgeSplitting.h"

// MDSTk
#include <MDSTk/VectorSegmentation/mdsTetraEdgeIterator.h>
#include <MDSTk/VectorSegmentation/mdsTetraTriIterator.h>

// STL
#include <climits>
#include <iostream>
#include <string>


//==============================================================================
/*
 * Global constants.
 */

//! Module description
const std::string MODULE_DESCRIPTION    = "Module provides edge splitting algorithm for adaptation of a tetrahedral mesh to an underlying volume data.";

//! Additional command line arguments
const std::string MODULE_ARGUMENTS      = "minlength:threshold";

//! Additional arguments names
const std::string MODULE_ARG_MIN_LENGTH = "minlength";
const std::string MODULE_ARG_THRESHOLD  = "threshold";

//! Default arguments values
const double DEFAULT_MIN_LENGTH         = mds::seg::CTetraMesh::DEFAULT_MIN_EDGE_LENGTH;
const double DEFAULT_THRESHOLD          = mds::seg::CTetraMesh::DEFAULT_SPLITTING_THRESHOLD;

//! Allowed arguments values
const double MIN_LENGTH                 = 0.1;
const double MAX_LENGTH                 = 50.0;

const double MIN_THRESHOLD              = 0.0;
const double MAX_THRESHOLD              = 1.0;


//==============================================================================
/*
 * Implementation of the class CTetraMeshEdgeSplitting.
 */
CTetraMeshEdgeSplitting::CTetraMeshEdgeSplitting(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CTetraMeshEdgeSplitting::~CTetraMeshEdgeSplitting()
{
}


bool CTetraMeshEdgeSplitting::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 3 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // Parse command line arguments
    m_dMinLength = DEFAULT_MIN_LENGTH;
    m_Arguments.value(MODULE_ARG_MIN_LENGTH, m_dMinLength);
    if( m_dMinLength < MIN_LENGTH || m_dMinLength > MAX_LENGTH )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_MIN_LENGTH << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_dOrientThreshold = DEFAULT_THRESHOLD;
    m_Arguments.value(MODULE_ARG_THRESHOLD, m_dOrientThreshold);
    if( m_dOrientThreshold < MIN_THRESHOLD || m_dOrientThreshold > MAX_THRESHOLD )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_THRESHOLD << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // O.K.
    return true;
}


bool CTetraMeshEdgeSplitting::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read underlying volume from the input channel
    if( !readInput(pIChannel, m_spVolume.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input volume" << std::endl);
        return false;
    }


    // I/O channels
    pIChannel = getInput(1);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read underlying edge volume from the input channel
    if( !readInput(pIChannel, m_spEdgeVolume.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input edge volume" << std::endl);
        return false;
    }


    // I/O channels
    pIChannel = getInput(2);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read mesh from the input channel
    if( !readInput(pIChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input tetrahedral mesh" << std::endl);
        return false;
    }

    // Create the volume margin
    m_spVolume->mirrorMargin();

    // Set the underlying mesh image
    m_spMesh->setVolume(m_spVolume);
    m_spMesh->setEdgeVolume(m_spEdgeVolume.get());
    m_spMesh->setMinEdgeLength(m_dMinLength);

    // Adapt the triangular mesh
//    m_spMesh->edgeSplitting(m_dOrientThreshold);

    // 2009/12/11
    // Tetrahedron rasterization
    // - Only for testing purposes!
    m_spMesh->markBoundaryVertices();
    mds::img::CDensityVolumePtr spNewVolume(new mds::img::CDensityVolume(*m_spVolume));
    spNewVolume->fillEntire(mds::img::CPixelTraits<mds::img::tDensityPixel>::getPixelMin());
    vctl::MCTetra *pTetrahedron = m_spMesh->getFirstTetrahedron();
    while( pTetrahedron )
    {
        for( int i = 0; i < 4; ++i )
        {
            vctl::MCVertex *pVertices[3];
            pTetrahedron->GetTri(i, pVertices);

            if( m_spMesh->isBoundaryTriangle(pVertices[0], pVertices[1], pVertices[2]) )
            {
                mds::seg::CTetraTriIterator It(spNewVolume, pVertices[0], pVertices[1], pVertices[2], m_spMesh);
                while( !It.isEnd() )
                {
                    *It = mds::img::CPixelTraits<mds::img::tDensityPixel>::getPixelMax();
                    ++It;
                }
            }
        }
        pTetrahedron = (vctl::MCTetra *)pTetrahedron->GetNext();
    }

    // Tetrahedron rasterization
    // - Only for testing purposes!
/*    mds::img::CDensityVolumePtr spNewVolume(new mds::img::CDensityVolume(*m_spVolume));
    spNewVolume->fillEntire(mds::img::CPixelTraits<mds::img::tDensityPixel>::getPixelMin());
    vctl::MCTetra *pTetrahedron = m_spMesh->getFirstTetrahedron();
    while( pTetrahedron )
    {
        for( int i = 0; i < 4; ++i )
        {
            vctl::MCVertex *pVertices[3];
            pTetrahedron->GetTri(i, pVertices);

            if( m_spMesh->checkBoundaryTriangle(pVertices[0], pVertices[1], pVertices[2], 0.1) )
            {
                mds::seg::CTetraTriIterator It(spNewVolume, pVertices[0], pVertices[1], pVertices[2], m_spMesh);
                while( !It.isEnd() )
                {
                    *It = mds::img::CPixelTraits<mds::img::tDensityPixel>::getPixelMax();
                    ++It;
                }
            }
        }
        pTetrahedron = (vctl::MCTetra *)pTetrahedron->GetNext();
    }*/

    // Write the result to the output channel
//    if( !writeOutput(pOChannel, m_spMesh.get()) )
    if( !writeOutput(pOChannel, spNewVolume.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write the tetrahedral mesh" << std::endl);
    }

    // Returning 'true' means to continue processing the input channel
    return false;
}


void CTetraMeshEdgeSplitting::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTetraMeshEdgeSplitting::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR(std::endl);
    MDS_CERR("Extended usage: [-minlength dValue] [-threshold dValue]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("Input: volume data, edge volume, tetrahedral mesh." << std::endl);
    MDS_CERR("Output: tetrahedral mesh." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CTetraMeshEdgeSplittingPtr spModule(new CTetraMeshEdgeSplitting(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

