//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2009 by Michal Spanel          \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2009/10/01                          \n
 *
 * $Id:$
 *
 * Description:
 * - Extracts image features of all tetrahedra.
 */

#include "mdsTetraMeshExtractFeatures.h"

// MDSTk Extensions
#include <MDSTk/VectorSegmentation/mdsTetraInfo.h>

// STL
#include <climits>
#include <iostream>
#include <string>
#include <sstream>


//==============================================================================
/*
 * Global constants.
 */

//! Module description.
const std::string MODULE_DESCRIPTION    = "Module extracts tetrahedra features.";

//! Additional command line arguments.
const std::string MODULE_ARGUMENTS      = "chedges";

//! Additional arguments names.
const std::string MODULE_ARGUMENT_EDGES = "chedges";


//==============================================================================
/*
 * Implementation of the class CTetraMeshExtractFeatures.
 */
CTetraMeshExtractFeatures::CTetraMeshExtractFeatures(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
    , m_bCheckEdges(false)
{
    allowArguments(MODULE_ARGUMENTS);
}


CTetraMeshExtractFeatures::~CTetraMeshExtractFeatures()
{
}


bool CTetraMeshExtractFeatures::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 3 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // Parse command line arguments
    m_bCheckEdges = m_Arguments.exists(MODULE_ARGUMENT_EDGES);

    // O.K.
    return true;
}


bool CTetraMeshExtractFeatures::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return false;
    }

    // Read volume from the input channel
    if( !readInput(pIChannel, m_spVolume.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input volume" << std::endl);
        return false;
    }

    // I/O channels
    pIChannel = getInput(1);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return false;
    }

    // Read edge volume from the input channel
    if( !readInput(pIChannel, m_spEdgeVolume.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input edge volume" << std::endl);
        return false;
    }

    // I/O channels
    pIChannel = getInput(2);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return false;
    }

    // Read mesh from the input channel
    if( !readInput(pIChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input tetrahedral mesh" << std::endl);
        return false;
    }

    // Create the volume margin
    m_spVolume->mirrorMargin();
    m_spEdgeVolume->mirrorMargin();

    // Set the underlying volume
    m_spMesh->setVolume(m_spVolume);
    m_spMesh->setEdgeVolume(m_spEdgeVolume);

    // Evaluate features
    m_spMesh->evaluateFeatures(false);

    // Assign unique indexes to all tetrahedra
    m_spMesh->forEachTetrahedron(mds::seg::CAssignTetraIndex(0));

    // Write tetrahedron features to a stream
    std::stringstream Stream;
    int Flags = (m_bCheckEdges) ? mds::seg::CWriteTetraInfo::CHECK_EDGE_FACES : 0;
    mds::seg::CWriteTetraInfo Writer(Stream, Flags | mds::seg::CWriteTetraInfo::DEFAULT_FLAGS);
    Writer.setMesh(m_spMesh.get());
    Writer.writeHeading();
    m_spMesh->forEachTetrahedron(Writer);

    // Write the result to the output channel
    if( !pOChannel->write(Stream.str().c_str(), (int)Stream.str().length()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write the output info" << std::endl);
    }

    // Returning 'true' means to continue processing the input channel
    return false;
}


void CTetraMeshExtractFeatures::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTetraMeshExtractFeatures::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR("Extended usage: [-chedges]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("  -chedges  Checks if tetrahedron faces correspond to an image edge." << std::endl);
    MDS_CERR("Input: volume data, edge volume data, tetrahedral mesh." << std::endl);
    MDS_CERR("Output: text info." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CTetraMeshExtractFeaturesPtr spModule(new CTetraMeshExtractFeatures(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}
