//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2007 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsTetraMeshFindIsolatedRegions.h   \n
 * Section: mTetraMeshFindIsolatedRegions       \n
 * Date:    2007/03/22                          \n
 *
 * $Id: mdsTetraMeshFindIsolatedRegions.cpp 1494 2009-05-21 14:36:00Z spanel $
 *
 * Description:
 * - Identifies individual isolated regions of tetrahedra in the classified mesh.
 */

#include "mdsTetraMeshFindIsolatedRegions.h"

// MDSTk
#include <MDSTk/Math/mdsVector.h>
#include <MDSTk/Math/mdsVectorFunctions.h>

// STL
#include <cstring>
#include <iostream>
#include <string>
#include <sstream>


//==============================================================================
/*
 * Global constants.
 */

//! Module description
const std::string MODULE_DESCRIPTION    = "Identifies individual isolated regions of tetrahedra in the classified mesh.";

//! Additional command line arguments
const std::string MODULE_ARGUMENTS      = "";

//! Additional arguments names

//! Default argument values


//==============================================================================
/*
 * Implementation of the class CTetraMeshFindIsolatedRegions.
 */
CTetraMeshFindIsolatedRegions::CTetraMeshFindIsolatedRegions(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CTetraMeshFindIsolatedRegions::~CTetraMeshFindIsolatedRegions()
{
}


bool CTetraMeshFindIsolatedRegions::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 1 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // Command line arguments

    // O.K.
    return true;
}


bool CTetraMeshFindIsolatedRegions::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read mesh from the input channel
    if( !readInput(pIChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input tetrahedral mesh" << std::endl);
        return false;
    }

    // Find individual isolated regions in the mesh
    m_spMesh->findIsolatedRegions();
    
    // Write the result to the output channel
    if( !writeOutput(pOChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write the output tetrahedral mesh" << std::endl);
    }

    // Returning 'true' means to continue processing the input channel
    return true;
}


void CTetraMeshFindIsolatedRegions::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTetraMeshFindIsolatedRegions::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR(std::endl);
    MDS_CERR("Extended usage:" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("Input: tetrahedral mesh." << std::endl);
    MDS_CERR("Output: tetrahedral mesh." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CTetraMeshFindIsolatedRegionsPtr spModule(new CTetraMeshFindIsolatedRegions(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

