//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2009 by Michal Spanel          \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * Date:    2009/10/06                          \n
 *
 * $Id:$
 *
 * Description:
 * - Imports classification results to the mesh.
 */

#include "mdsTetraMeshImportClassification.h"

// MDSTk Extensions
#include <MDSTk/VectorSegmentation/mdsTetraFeatures.h>

// STL
#include <climits>
#include <iostream>
#include <string>
#include <sstream>


//==============================================================================
/*
 * Global constants.
 */

//! Module description.
const std::string MODULE_DESCRIPTION    = "Module imports classification results to the mesh.";

//! Additional command line arguments.
const std::string MODULE_ARGUMENTS      = "";

//! Additional arguments names.
//const std::string MODULE_ARGUMENT_XYZ = "xyz";


//==============================================================================
/*
 * Implementation of the class CTetraMeshImportClassification.
 */
CTetraMeshImportClassification::CTetraMeshImportClassification(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CTetraMeshImportClassification::~CTetraMeshImportClassification()
{
}


bool CTetraMeshImportClassification::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 2 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // Parse command line arguments
//    m_bXYZ = m_Arguments.exists(MODULE_ARGUMENT_XYZ);

    // O.K.
    return true;
}


bool CTetraMeshImportClassification::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read mesh from the input channel
    if( !readInput(pIChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input tetrahedral mesh" << std::endl);
        return false;
    }

    // I/O channels
    pIChannel = getInput(1);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read results of the classification...
    std::string Data;
    int iCount = 0;
    char pcData[1025];
    while( (iCount = pIChannel->read(pcData, 1024)) > 0 )
    {
        pcData[iCount] = '\0';
        Data += std::string(pcData);
    }
    if( Data.empty() )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input classification" << std::endl);
        return false;
    }

    // Modify the classification...
    std::istringstream Stream(Data);
    vctl::MCTetra *pTetrahedron = m_spMesh->getFirstTetrahedron();
    while( pTetrahedron )
    {
        // Get region index
        double Index;
        Stream >> Index;
        if( Stream.fail() )
        {
            MDS_CERR('<' << m_sFilename << "> Failed to decode input classification" << std::endl);
            return false;
        }

        // Modify tetrahedron classification
        mds::seg::setTetrahedronRegion(pTetrahedron, int(Index));

        // Get next one
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Write the result to the output channel
    if( !writeOutput(pOChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write the output mesh" << std::endl);
    }

    // Returning 'true' means to continue processing the input channel
    return false;
}


void CTetraMeshImportClassification::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTetraMeshImportClassification::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR("Extended usage: [-xyz]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("  -xyz  ..." << std::endl);
    MDS_CERR("Input: tetrahedral mesh, text." << std::endl);
    MDS_CERR("Output: tetrahedral mesh." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CTetraMeshImportClassificationPtr spModule(new CTetraMeshImportClassification(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}
