//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2006 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsTetraMeshInfo.h                  \n
 * Section: mTetraMeshInfo                      \n
 * Date:    2006/06/02                          \n
 *
 * $Id: mdsTetraMeshInfo.cpp 1532 2009-12-10 08:32:58Z spanel $
 *
 * Description:
 * - Shows information on tetrahedral mesh.
 */

#include "mdsTetraMeshInfo.h"

// MDSTk
#include <MDSTk/Math/mdsVector.h>
#include <MDSTk/Math/mdsVectorFunctions.h>

// STL
#include <cstring>
#include <iostream>
#include <string>
#include <sstream>


//==============================================================================
/*
 * Global constants.
 */

//! Module description
const std::string MODULE_DESCRIPTION    = "Shows information on input tetrahedral mesh.";

//! Additional command line arguments
const std::string MODULE_ARGUMENTS      = "hist:volume:angles";

//! Additional arguments names
const std::string MODULE_ARG_HIST       = "hist";
const std::string MODULE_ARG_VOLUME     = "volume";
const std::string MODULE_ARG_ANGLES     = "angles";

//! Default argument values

//! Maximal length of visualized histogram bin
const int MAX_BIN_LENGTH                = 60;


//==============================================================================
/*
 * Implementation of the class CTetraMeshInfo.
 */
CTetraMeshInfo::CTetraMeshInfo(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CTetraMeshInfo::~CTetraMeshInfo()
{
}


bool CTetraMeshInfo::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 1 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // Command line arguments
    m_bShowHistogram = m_Arguments.exists(MODULE_ARG_HIST);
    m_bShowVolume = m_Arguments.exists(MODULE_ARG_VOLUME);
    m_bDihedralAngles = m_Arguments.exists(MODULE_ARG_ANGLES);

    // O.K.
    return true;
}


bool CTetraMeshInfo::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read mesh from the input channel
    if( !readInput(pIChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input tetrahedral mesh" << std::endl);
        return false;
    }

    // Get minimal and maximal coordinates
    double dMinX = 0.0, dMinY = 0.0, dMinZ = 0.0;
    double dMaxX = 0.0, dMaxY = 0.0, dMaxZ = 0.0;
    m_spMesh->getMinCoordinates(dMinX, dMinY, dMinZ);
    m_spMesh->getMaxCoordinates(dMaxX, dMaxY, dMaxZ);

    // Create the output string
    std::stringstream Stream;
    Stream << "-- Information on tetrahedral mesh --" << std::endl
        << " Number of vertices:     " << m_spMesh->getNumOfVertices() << std::endl
        << " Number of tetrahedra: " << m_spMesh->getNumOfTetrahedra() << std::endl
        << " Minimal coordinates:    " << dMinX << ", " << dMinY << " ," << dMinZ << std::endl
        << " Maximal coordinates:    " << dMaxX << ", " << dMaxY << " ," << dMaxZ << std::endl
        << " Number of segments:     " << m_spMesh->getNumOfSegments()
        << std::endl;

    // Compute histogram of regions in the mesh
    if( m_bShowHistogram && m_spMesh->getNumOfSegments() > 0 )
    {
        // Histogram
        mds::math::CVector<double> Histogram(m_spMesh->getNumOfSegments() + 1);
        Histogram.fill(0);

        // Evaluate histogram
        vctl::MCTetra *pTetra = m_spMesh->getFirstTetrahedron();
        while( pTetra )
        {
            // Get tetrahedron region
            int iRegion = mds::seg::getTetrahedronRegion(pTetra);
            if( iRegion >= -1 && iRegion < m_spMesh->getNumOfSegments() )
            {
                if( m_bShowVolume )
                {
                    Histogram(iRegion + 1) += pTetra->GetVolume();
                }
                else
                {
                    Histogram(iRegion + 1) += 1;
                }
            }

            // Get the next tetrahedron
            pTetra = pTetra->GetNext();
        }

        // Print the histogram
        if( m_bShowVolume )
        {
            Stream << "-- Volume of regions --" << std::endl;
        }
        else
        {
            Stream << "-- Number of tetrahedra assigned to regions --" << std::endl;
        }
        Stream.fill('0');
        Stream.setf(std::ios::right | std::ios::fixed);

        // Prepare bin visualization string
        char *pcBin = new char[MAX_BIN_LENGTH + 1];
        memset(pcBin, '-', MAX_BIN_LENGTH);
        pcBin[MAX_BIN_LENGTH] = '\0';

        // Maximal histogram value
        double dMax = mds::math::getMax<double>(Histogram);
        double dInvMax = (dMax > 0.0) ? 1.0 / dMax : 0.0;

        // Print the histogram
        for( mds::tSize i = 0; i < Histogram.getSize(); ++i )
        {
            double dValue = Histogram(i);
            Stream << ' ' << std::setw(5) << i - 1 << '>';
            int iLength = int(dValue * MAX_BIN_LENGTH * dInvMax);
            pcBin[iLength] = '\0';
            Stream << pcBin;
            pcBin[iLength] = '-';
            Stream << "| " << int(dValue) << std::endl;
        }

        // Delete the helper string
        delete[] pcBin;
    }

    // Compute all dihedral angles
    if( m_bDihedralAngles )
    {
        Stream << "-- Dihedral angles (region, a1, ... , a6, min) --" << std::endl;

        // Evaluate angles
        vctl::MCTetra *pTetra = m_spMesh->getFirstTetrahedron();
        while( pTetra )
        {
            // Get tetrahedron region
            int iRegion = mds::seg::getTetrahedronRegion(pTetra);

            // Compute all dihedral angles and the minimum one
            Stream << iRegion << " ";
            double dMinAngle = 180.0;
            for( int i = 0; i < 4; ++i )
            {
                vctl::MCVector3D ni;
                pTetra->GetNormal(i, ni);
                ni.Normalization();
                for( int j = i + 1; j < 4; ++j )
                {
                    vctl::MCVector3D nj;
                    pTetra->GetNormal(j, nj);
                    nj.Normalization();

                    // Compute the angle -1..1 ~ 0..180
                    double dAngle = (1.0 + vctl::MCVector3D::DotProduct(ni, nj)) * 90.0;
                    Stream << dAngle << " ";

                    // Find minimum
                    dMinAngle = mds::math::getMin(dAngle, dMinAngle);
                }
            }
            Stream << dMinAngle << std::endl;

            // Get the next tetrahedron
            pTetra = pTetra->GetNext();
        }
    }

    // Write it to the output channel
    if( !pOChannel->write(Stream.str().c_str(), (int)Stream.str().length()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write the output" << std::endl);
        return false;
    }

    // Returning 'true' means to continue processing the input channel
    return true;
}


void CTetraMeshInfo::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTetraMeshInfo::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR(std::endl);
    MDS_CERR("Extended usage: [-hist] [-volume] [-angles]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("Input: tetrahedral mesh." << std::endl);
    MDS_CERR("Output: text." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CTetraMeshInfoPtr spModule(new CTetraMeshInfo(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

