//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)        \n
 * Copyright (c) 2003-2007 by Michal Spanel         \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz      \n
 * File:    mdsTetraMeshIsotropicEdgeSplitting.cpp  \n
 * Section: mTetraMeshIsotropicEdgeSplitting        \n
 * Date:    2007/03/03                              \n
 *
 * $Id: mdsTetraMeshIsotropicEdgeSplitting.cpp 1494 2009-05-21 14:36:00Z spanel $
 *
 * Description:
 * - Adaptation of tetrahedral mesh to an underlying volume data.
 * - Isotropic edge splitting algorithm.
 */

#include "mdsTetraMeshIsotropicEdgeSplitting.h"

// MDSTk
#include <MDSTk/VectorSegmentation/mdsTetraEdgeIterator.h>
#include <MDSTk/VectorSegmentation/mdsTetraTriIterator.h>

// STL
#include <climits>
#include <iostream>
#include <string>


//==============================================================================
/*
 * Global constants.
 */

//! Module description
const std::string MODULE_DESCRIPTION    = "Module provides edge splitting algorithm for adaptation of a tetrahedral mesh to an underlying volume data.";

//! Additional command line arguments
const std::string MODULE_ARGUMENTS      = "minlength:threshold:boundary:cdt";

//! Additional arguments names
const std::string MODULE_ARG_MIN_LENGTH = "minlength";
const std::string MODULE_ARG_THRESHOLD  = "threshold";
const std::string MODULE_ARG_BOUNDARY   = "boundary";
const std::string MODULE_ARG_CDT        = "cdt";

//! Default arguments values
const double DEFAULT_MIN_LENGTH         = mds::seg::CTetraMesh::DEFAULT_MIN_EDGE_LENGTH;
const double DEFAULT_THRESHOLD          = 0.5;

//! Allowed arguments values
const double MIN_LENGTH                 = 0.1;
const double MAX_LENGTH                 = 50.0;

const double MIN_THRESHOLD              = 0.0;
const double MAX_THRESHOLD              = 1.0;


//==============================================================================
/*
 * Implementation of the class CTetraMeshIsotropicEdgeSplitting.
 */
CTetraMeshIsotropicEdgeSplitting::CTetraMeshIsotropicEdgeSplitting(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CTetraMeshIsotropicEdgeSplitting::~CTetraMeshIsotropicEdgeSplitting()
{
}


bool CTetraMeshIsotropicEdgeSplitting::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 4 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // Parse command line arguments
    m_dMinLength = DEFAULT_MIN_LENGTH;
    m_Arguments.value(MODULE_ARG_MIN_LENGTH, m_dMinLength);
    if( m_dMinLength < MIN_LENGTH || m_dMinLength > MAX_LENGTH )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_MIN_LENGTH << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_dThreshold = DEFAULT_THRESHOLD;
    m_Arguments.value(MODULE_ARG_THRESHOLD, m_dThreshold);
    if( m_dThreshold < MIN_THRESHOLD || m_dThreshold > MAX_THRESHOLD )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_THRESHOLD << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_bBoundary = m_Arguments.exists(MODULE_ARG_BOUNDARY);
    m_bCDT = m_Arguments.exists(MODULE_ARG_CDT);

    // O.K.
    return true;
}


bool CTetraMeshIsotropicEdgeSplitting::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read underlying volume from the input channel
    if( !readInput(pIChannel, m_spVolume.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input volume" << std::endl);
        return false;
    }


    // I/O channels
    pIChannel = getInput(1);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read underlying edge volume from the input channel
    if( !readInput(pIChannel, m_spEdgeVolume.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input edge volume" << std::endl);
        return false;
    }


    // I/O channels
    pIChannel = getInput(2);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read underlying edge volume from the input channel
    if( !readInput(pIChannel, m_spControlSpace.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input control space" << std::endl);
        return false;
    }


    // I/O channels
    pIChannel = getInput(3);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read mesh from the input channel
    if( !readInput(pIChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input tetrahedral mesh" << std::endl);
        return false;
    }

    // Create the volume margin
    m_spVolume->mirrorMargin();
    m_spEdgeVolume->mirrorMargin();

    // Set the underlying mesh image
    m_spMesh->setVolume(m_spVolume);
    m_spMesh->setEdgeVolume(m_spEdgeVolume.get());
    m_spMesh->setControlSpace(m_spControlSpace);
    m_spMesh->setMinEdgeLength(m_dMinLength);
    m_spMesh->setCornerMinEdgeLength(0.3 * m_dMinLength);
//    m_spMesh->setCornerMinEdgeLength(m_dMinLength);
    m_spMesh->setVertexNeighbourhood(5);

    if( m_bCDT )
    {
        m_spMesh->setConstraintFunc(mds::seg::CTetraMesh::checkBoundaryFacetFunc);
    }

    // Adapt the triangular mesh
    m_spMesh->isotropicEdgeSplitting(m_dThreshold, m_bBoundary);

    // Write the result to the output channel
    if( !writeOutput(pOChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write the tetrahedral mesh" << std::endl);
    }

    // Returning 'true' means to continue processing the input channel
    return false;
}


void CTetraMeshIsotropicEdgeSplitting::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTetraMeshIsotropicEdgeSplitting::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR(std::endl);
    MDS_CERR("Extended usage: [-minlength dValue] [-threshold dValue] [-boundary] [-cdt]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("Input: volume data, edge volume, control space, tetrahedral mesh." << std::endl);
    MDS_CERR("Output: tetrahedral mesh." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CTetraMeshIsotropicEdgeSplittingPtr spModule(new CTetraMeshIsotropicEdgeSplitting(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

