//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)     \n
 * Copyright (c) 2003-2007 by Michal Spanel      \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz   \n
 * File:    mdsTetraMeshIsotropicVertexMoving.h  \n
 * Section: mTetraMeshIsotropicVertexMoving      \n
 * Date:    2007/02/23                           \n
 *
 * $Id: mdsTetraMeshIsotropicVertexMoving.cpp 1494 2009-05-21 14:36:00Z spanel $
 *
 * Description:
 * - Tetrahedral mesh optimization.
 */

#include "mdsTetraMeshIsotropicVertexMoving.h"

// STL
#include <climits>
#include <iostream>
#include <string>


//==============================================================================
/*
 * Global constants.
 */

//! Module description
const std::string MODULE_DESCRIPTION    = "Tetrahedral mesh adaptation using vertex moving.";

//! Additional command line arguments
const std::string MODULE_ARGUMENTS      = "minlength:neighbourhood";

//! Additional arguments names
const std::string MODULE_ARG_MIN_LENGTH    = "minlength";
const std::string MODULE_ARG_NEIGHBOURHOOD = "neighbourhood";

//! Default argument values
const double DEFAULT_MIN_LENGTH         = mds::seg::CTetraMesh::DEFAULT_MIN_EDGE_LENGTH;
const int DEFAULT_NEIGHBOURHOOD         = mds::seg::CTetraMesh::DEFAULT_VERTEX_NEIGHBOURHOOD;

//! Allowed arguments values
const double MIN_LENGTH                 = 0.1;
const double MAX_LENGTH                 = 50.0;
const int MIN_NEIGHBOURHOOD             = 1;
const int MAX_NEIGHBOURHOOD             = 50;


//==============================================================================
/*
 * Implementation of the class CTetraMeshIsotropicVertexMoving.
 */
CTetraMeshIsotropicVertexMoving::CTetraMeshIsotropicVertexMoving(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CTetraMeshIsotropicVertexMoving::~CTetraMeshIsotropicVertexMoving()
{
}


bool CTetraMeshIsotropicVertexMoving::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 3 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // Command line parameters
    m_dMinEdgeLength = DEFAULT_MIN_LENGTH;
    m_Arguments.value(MODULE_ARG_MIN_LENGTH, m_dMinEdgeLength);
    if( m_dMinEdgeLength < MIN_LENGTH || m_dMinEdgeLength > MAX_LENGTH )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_MIN_LENGTH << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_iNeighbourhood = DEFAULT_NEIGHBOURHOOD;
    m_Arguments.value(MODULE_ARG_NEIGHBOURHOOD, m_iNeighbourhood);
    if( m_iNeighbourhood < MIN_NEIGHBOURHOOD || m_iNeighbourhood > MAX_NEIGHBOURHOOD )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_NEIGHBOURHOOD << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // O.K.
    return true;
}


bool CTetraMeshIsotropicVertexMoving::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read volume from the input channel
    if( !readInput(pIChannel, m_spEdgeVolume.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input edge volume" << std::endl);
        return false;
    }


    // I/O channels
    pIChannel = getInput(1);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read volume from the input channel
    if( !readInput(pIChannel, m_spControlSpace.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read control space" << std::endl);
        return false;
    }


    // I/O channels
    pIChannel = getInput(2);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read mesh from the input channel
    if( !readInput(pIChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input tetrahedral mesh" << std::endl);
        return false;
    }

    // Set the underlying volume
    m_spMesh->setEdgeVolume(m_spEdgeVolume.get());
    m_spMesh->setControlSpace(m_spControlSpace);
    m_spMesh->setMinEdgeLength(m_dMinEdgeLength);
//    m_spMesh->setCornerMinEdgeLength(0.3 * m_dMinEdgeLength);
    m_spMesh->setCornerMinEdgeLength(0.5 * m_dMinEdgeLength);
    m_spMesh->setVertexNeighbourhood(m_iNeighbourhood);
//    m_spMesh->setConstraintFunc(mds::seg::CTetraMesh::checkBoundaryFacetFunc);

    // Optimize the mesh
    m_spMesh->isotropicVertexMoving();

    // Write the result to the output channel
    if( !writeOutput(pOChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write the tetrahedral mesh" << std::endl);
    }

    // Returning 'true' means to continue processing the input channel
    return false;
}


void CTetraMeshIsotropicVertexMoving::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTetraMeshIsotropicVertexMoving::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR(std::endl);
    MDS_CERR("Extended usage: [-minlength dValue] [-neighbourhood dValue]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("  -minlength  If the longest tetrahedron edge is shorter then a given value" << std::endl);
    MDS_CERR("              the tetrahedron is not optimized. Default value of this" << std::endl);
    MDS_CERR("              parameter is " << DEFAULT_MIN_LENGTH << "." << std::endl);
    MDS_CERR("Input: edge volume, control space, tetrahedral mesh." << std::endl);
    MDS_CERR("Output: tetrahedral mesh." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CTetraMeshIsotropicVertexMovingPtr spModule(new CTetraMeshIsotropicVertexMoving(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

