//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2006 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsTetraMeshRasterization.h         \n
 * Section: mTetraMeshRasterization             \n
 * Date:    2006/06/12                          \n
 *
 * $Id: mdsTetraMeshRasterization.cpp 1510 2009-09-21 08:35:53Z spanel $
 *
 * Description:
 * - Rasterization of tetrahedra in a mesh.
 */

#include "mdsTetraMeshRasterization.h"

// MDSTk Extensions
#include <MDSTk/VectorSegmentation/mdsTetraIterator.h>
#include <MDSTk/VectorSegmentation/mdsTetraEdgeIterator.h>
#include <MDSTk/VectorSegmentation/mdsTetraTriIterator.h>
#include <MDSTk/VectorSegmentation/mdsTetraFeatures.h>

// STL
#include <climits>
#include <iostream>
#include <string>
#include <vector>
#include <algorithm>


//==============================================================================
/*
 * Global constants.
 */

//! Module description.
const std::string MODULE_DESCRIPTION    = "Module rasterizes tetrahedra in an input mesh.";

//! Additional command line arguments.
const std::string MODULE_ARGUMENTS      = "region:random:alpha:edges:triangles:interpolate";

//! Additional arguments names.
const std::string MODULE_ARG_REGION     = "region";
const std::string MODULE_ARG_RANDOM     = "random";
const std::string MODULE_ARG_ALPHA      = "alpha";
const std::string MODULE_ARG_EDGES      = "edges";
const std::string MODULE_ARG_TRIANGLES  = "triangles";
const std::string MODULE_ARG_INTERPOLATE = "interpolate";

//! Default arguments values.
const int DEFAULT_REGION                = -1;
const double DEFAULT_ALPHA              = 0.5;

//! Maximal allowed voxel value.
mds::img::tDensityPixel VOXEL_MAX       = mds::img::CPixelTraits<mds::img::tDensityPixel>::getPixelMax();
const double ALPHA_MIN                  = 0.0;
const double ALPHA_MAX                  = 1.0;


//==============================================================================
/*
 * Implementation of the class CTetraMeshRasterization.
 */
CTetraMeshRasterization::CTetraMeshRasterization(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CTetraMeshRasterization::~CTetraMeshRasterization()
{
}


bool CTetraMeshRasterization::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channels
    if( getNumOfInputs() != 2 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // Parse command line arguments
    m_iRegion = DEFAULT_REGION;
    m_Arguments.value(MODULE_ARG_REGION, m_iRegion);
    if( m_iRegion < -1 )
    {
        MDS_CERR('<' << m_sFilename << "> Bad '" << MODULE_ARG_REGION << "' parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_bRandom = m_Arguments.exists(MODULE_ARG_RANDOM);
    m_bEdges = m_Arguments.exists(MODULE_ARG_EDGES);
    m_bTriangles = m_Arguments.exists(MODULE_ARG_TRIANGLES);
    m_bInterpolate = m_Arguments.exists(MODULE_ARG_INTERPOLATE);

    m_dAlpha = DEFAULT_ALPHA;
    m_Arguments.value(MODULE_ARG_ALPHA, m_dAlpha);
    if( m_dAlpha > ALPHA_MAX || m_dAlpha < ALPHA_MIN )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_ALPHA << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // O.K.
    return true;
}


bool CTetraMeshRasterization::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read volume from the input channel
    if( !readInput(pIChannel, m_spVolume.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input volume" << std::endl);
        return false;
    }

    // I/O channels
    pIChannel = getInput(1);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read mesh from the input channel
    if( !readInput(pIChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input tetrahedral mesh" << std::endl);
        return false;
    }

    // Set mesh properties
    m_spMesh->setVoxelSize(m_spVolume->getDX(), m_spVolume->getDY(), m_spVolume->getDZ());

    // Random number generator
    mds::math::CUniformPRNG RandomNumber;

    // Blending coefficient
    double dInvAlpha = 1.0 - m_dAlpha;

    std::vector<mds::img::tDensityPixel> List;

    // Tetrahedra rasterization
    vctl::MCTetra *pTetrahedron = m_spMesh->getFirstTetrahedron();
    while( pTetrahedron )
    {
        // Get tetrahedron segment
        int iTetraRegion = mds::seg::getTetrahedronRegion(pTetrahedron);

        // Interpolating density data
        if( m_bInterpolate )
        {
/*            vctl::MCVertex *pVertices[4];
            pTetrahedron->GetVerticeS(pVertices);

            vctl::MCPoint3D Point[4];
            Point[0] = *pVertices[0];
            Point[1] = *pVertices[1];
            Point[2] = *pVertices[2];
            Point[3] = *pVertices[3];
            m_spMesh->conv2Volume(&Point[0]);
            m_spMesh->conv2Volume(&Point[1]);
            m_spMesh->conv2Volume(&Point[2]);
            m_spMesh->conv2Volume(&Point[3]);

            mds::img::tDensityPixel Value[4];
            Value[0] = m_spVolume->interpolate(Point[0]);
            Value[1] = m_spVolume->interpolate(Point[1]);
            Value[2] = m_spVolume->interpolate(Point[2]);
            Value[3] = m_spVolume->interpolate(Point[3]);*/

/*            vctl::MCPoint3D Point;
            pTetrahedron->GetCenter(Point);
            m_spMesh->conv2Volume(&Point);
            mds::img::tDensityPixel Value = m_spVolume->interpolate(Point);*/

/*            double dMean = 0.0;
            int iCount = 0;
            mds::seg::CTetraIterator It(m_spVolume, pTetrahedron, m_spMesh);
            while( !It.isEnd() )
            {
                dMean += *It;
                ++iCount;
                ++It;
            }*/

            List.clear();
            mds::seg::CTetraIterator It(m_spVolume, pTetrahedron, m_spMesh);
            while( !It.isEnd() )
            {
                List.push_back(*It);
                ++It;
            }

//            if( iCount > 0 )
            if( List.size() > 0 )
            {
//                mds::img::tDensityPixel Value = mds::img::tDensityPixel(dMean / iCount);

                std::sort(List.begin(), List.end());
                mds::img::tDensityPixel Value = List[List.size() / 2];

                mds::seg::CTetraIterator It(m_spVolume, pTetrahedron, m_spMesh);
                while( !It.isEnd() )
                {
                    vctl::MCPoint3D p(It.getX(), It.getY(), It.getZ());

    /*                double w0 = p.SquareDistance(Point[0]);
                    double w1 = p.SquareDistance(Point[1]);
                    double w2 = p.SquareDistance(Point[2]);
                    double w3 = p.SquareDistance(Point[3]);
                    double norm = 1.0 / (w0 + w1 + w2 + w3);
                    w0 *= norm;
                    w1 *= norm;
                    w2 *= norm;
                    w3 *= norm;
                    double ww0 = w1 + w2 + w3;
                    double ww1 = w2 + w3 + w0;
                    double ww2 = w3 + w0 + w1;
                    double ww3 = w0 + w1 + w2;

                    *It = mds::img::tDensityPixel(ww0 * Value[0] + ww1 * Value[1] + ww2 * Value[2] + ww3 * Value[3]);*/
                    *It = Value;
                    ++It;
                }
            }
        }

        // Check the segment
        else if( m_iRegion == -1 || m_iRegion == iTetraRegion )
        {
            mds::img::tDensityPixel Value;
            if( m_bRandom )
            {
                Value = mds::img::tDensityPixel(RandomNumber.random(0, VOXEL_MAX));
            }
            else
            {
                Value = mds::img::tDensityPixel(iTetraRegion + 1);
                if( Value > VOXEL_MAX )
                {
                    Value = VOXEL_MAX;
                }
            }

            mds::seg::CTetraIterator It(m_spVolume, pTetrahedron, m_spMesh);
            while( !It.isEnd() )
            {
                mds::img::tDensityPixel OrigValue = *It;
                *It = mds::img::tDensityPixel(m_dAlpha * Value + dInvAlpha * OrigValue);
                ++It;
            }
        }

        // Get pointer to the next tetrahedon
        pTetrahedron = pTetrahedron->GetNext();
    }

    // Draw the edges
    if( m_bEdges )
    {
        m_spMesh->visualize(m_spVolume);
    }

    // Draw the faces
    if( m_bTriangles )
    {
        // Use the white color
        mds::img::tDensityPixel White = m_spVolume->color2Voxel(mds::img::C_WHITE);

        pTetrahedron = m_spMesh->getFirstTetrahedron();
        while( pTetrahedron )
        {
            // Rasterize all tetrahedron faces
            for( int i = 0; i < 4; ++i )
            {
                // Get triangle vertices
                vctl::MCVertex *pVertices[3];
                pTetrahedron->GetTri(i, pVertices);

                // Create temporary triangle
                vctl::MCTri Triangle(pVertices[0], pVertices[1], pVertices[2]);

                // Traverse all triangle voxels
                mds::seg::CTetraTriIterator It(m_spVolume, &Triangle, m_spMesh);
                while( !It.isEnd() )
                {
                    *It = White;
                    ++It;
                }
            }

            // Get the next tetrahedon
            pTetrahedron = pTetrahedron->GetNext();
        }
    }

    // Write the result to the output channel
    if( !writeOutput(pOChannel, m_spVolume.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write the output tetrahedral mesh" << std::endl);
    }

    // Returning 'true' means to continue processing the input channel
    return false;
}


void CTetraMeshRasterization::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTetraMeshRasterization::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR("Extended usage: [-region iValue] [-random] [-alpha dValue]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("  -region  Index of the region/segment in the mesh whose tetrahedra" << std::endl);
    MDS_CERR("           will be rasterized. Default values is " << DEFAULT_REGION << "." << std::endl);
    MDS_CERR(std::endl);
    MDS_CERR("  -random  Assigns region to every tetrahedron randomly." << std::endl);
    MDS_CERR(std::endl);
    MDS_CERR("  -alpha  Blending ratio. Values in the range <" << ALPHA_MIN << ", " << ALPHA_MAX << "> are allowed." << std::endl);
    MDS_CERR("          Default value is " << DEFAULT_ALPHA << '.' << std::endl);
    MDS_CERR(std::endl);
    MDS_CERR("  -edges" << std::endl);
    MDS_CERR(std::endl);
    MDS_CERR("  -triangles" << std::endl);
    MDS_CERR("Input: volume data, tetrahedral mesh." << std::endl);
    MDS_CERR("Output: volume data." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CTetraMeshRasterizationPtr spModule(new CTetraMeshRasterization(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

