//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2006 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsTetraMeshRegionMerging.cpp       \n
 * Section: mTetraMeshRegionMerging             \n
 * Date:    2006/12/28                          \n
 *
 * $Id: mdsTetraMeshRegionMerging.cpp 1537 2009-12-14 15:11:52Z spanel $
 *
 * Description:
 * - Region merging algorithm.
 */

#include "mdsTetraMeshRegionMerging.h"

// STL
#include <climits>
#include <iostream>
#include <string>


//==============================================================================
/*
 * Global constants.
 */

//! Module description
const std::string MODULE_DESCRIPTION        = "Module merges regions until a given similarity predicate is true.";

//! Additional command line arguments
const std::string MODULE_ARGUMENTS          = "bthreshold:simthreshold:rho:regions:mergeall:delta";

//! Additional arguments names
const std::string MODULE_ARG_BTHRESHOLD     = "bthreshold";
const std::string MODULE_ARG_SIMTHRESHOLD   = "simthreshold";
const std::string MODULE_ARG_RHO            = "rho";
const std::string MODULE_ARG_REGIONS        = "regions";
const std::string MODULE_ARG_MERGEALL       = "mergeall";
const std::string MODULE_ARG_DELTA          = "delta";

//! Default arguments values
const double DEFAULT_BTHRESHOLD             = mds::seg::CTetraMesh::DEFAULT_DISTANCE_THRESHOLD;
const double DEFAULT_SIMTHRESHOLD           = mds::seg::CTetraMeanValueSimilarity::DEFAULT_THRESHOLD;
const double DEFAULT_RHO                    = mds::seg::CTetraMeanValueSimilarity::DEFAULT_RHO;
const double DEFAULT_DELTA                  = mds::seg::CTetraMesh::DEFAULT_DELTA_THRESHOLD;
const int DEFAULT_REGIONS                   = -1;

//! Allowed arguments values
const double MIN_THRESHOLD                  = 0.0;
const double MAX_THRESHOLD                  = 10.0;


//==============================================================================
/*
 * Implementation of the class CTetraMeshRegionMerging.
 */
CTetraMeshRegionMerging::CTetraMeshRegionMerging(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CTetraMeshRegionMerging::~CTetraMeshRegionMerging()
{
}


bool CTetraMeshRegionMerging::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 3 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // Parse command line arguments
    m_dBoundaryThreshold = DEFAULT_BTHRESHOLD;
    m_Arguments.value(MODULE_ARG_BTHRESHOLD, m_dBoundaryThreshold);
    if( m_dBoundaryThreshold < MIN_THRESHOLD || m_dBoundaryThreshold > MAX_THRESHOLD )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_BTHRESHOLD << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_dSimilarityThreshold = DEFAULT_SIMTHRESHOLD;
    m_Arguments.value(MODULE_ARG_SIMTHRESHOLD, m_dSimilarityThreshold);
    if( m_dSimilarityThreshold < MIN_THRESHOLD || m_dSimilarityThreshold > MAX_THRESHOLD )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_SIMTHRESHOLD << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_dRho = DEFAULT_RHO;
    m_Arguments.value(MODULE_ARG_RHO, m_dRho);
    if( m_dRho < 0.0 )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_RHO << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_iNumOfRegions = DEFAULT_REGIONS;
    m_Arguments.value(MODULE_ARG_REGIONS, m_iNumOfRegions);
    if( m_iNumOfRegions < -1 )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_REGIONS << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_dDelta = DEFAULT_DELTA;
    m_Arguments.value(MODULE_ARG_DELTA, m_dDelta);
    if( m_dDelta < 0.0 )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_DELTA << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_bMergeAdjacent = !m_Arguments.exists(MODULE_ARG_MERGEALL);

    // O.K.
    return true;
}


bool CTetraMeshRegionMerging::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read underlying volume from the input channel
    if( !readInput(pIChannel, m_spVolume.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input volume" << std::endl);
        return false;
    }


    // I/O channels
    pIChannel = getInput(1);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read underlying edge volume from the input channel
    if( !readInput(pIChannel, m_spEdgeVolume.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input edge volume" << std::endl);
        return false;
    }


    // I/O channels
    pIChannel = getInput(2);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read mesh from the input channel
    if( !readInput(pIChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input tetrahedral mesh" << std::endl);
        return false;
    }

    // Create the volume margin
    m_spVolume->mirrorMargin();

    // Set the underlying mesh image
    m_spMesh->setVolume(m_spVolume.get());
    m_spMesh->setEdgeVolume(m_spEdgeVolume.get());

    // Prepare a similarity criterion
    mds::seg::CTetraMeanValueSimilarity Criterion1(m_dSimilarityThreshold, m_dRho);
    mds::seg::CTetraVarianceSimilarity Criterion2(m_dSimilarityThreshold);
    mds::seg::CTetraMixtureSimilarity Criterion(Criterion1, Criterion2, 1.0, 0.0);

    // Agglomerative merging of regions
    m_spMesh->regionMerging(Criterion, m_dBoundaryThreshold, m_dDelta, m_iNumOfRegions, m_bMergeAdjacent);

    // Classify very small triangles
//    m_spMesh->classifyAdjacent(m_dBoundaryThreshold);

    // Write the result to the output channel
    if( !writeOutput(pOChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write the tetrahedral mesh" << std::endl);
    }

    // Returning 'true' means to continue processing the input channel
    return false;
}


void CTetraMeshRegionMerging::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTetraMeshRegionMerging::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR(std::endl);
    MDS_CERR("Extended usage: [-bthreshold dValue] [-simthreshold dValue] [-rho dValue]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("Input: volume data, edge volume, tetrahedral mesh." << std::endl);
    MDS_CERR("Output: tetrahedral mesh." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CTetraMeshRegionMergingPtr spModule(new CTetraMeshRegionMerging(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

