//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2005 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsTetraMeshView.cpp                 n
 * Section: mTetraMeshView                      \n
 * Date:    2004/07/30                          \n
 *
 * $Id: mdsTetraMeshView.cpp 800 2008-02-29 08:32:00Z spanel $
 *
 * File description:
 * - Reads a tetra mesh and underlying vlume from the input channel.
 * - Visualizes it using GLUT and OpenGL.
 */

#include "mdsTetraMeshView.h"

#include <MDSTk/Math/mdsRandom.h>
#include <MDSTk/Image/mdsVolumeFunctions.h>

// OpenGL
#include <GL/gl.h>
#include <GL/glut.h>

// STL
#include <iostream>
#include <string>


//==============================================================================
/*
 * Macro definitions.
 */

//! Macro enables the perspective projection if defined.
//#define ENABLE_PERSPECTIVE


//==============================================================================
/*
 * Global definitions.
 */

//! GLUT window title.
const char GLUT_WINDOW_TITLE[]              = "MDSTk TetraMesh View Window";

//! Module description.
const std::string MODULE_DESCRIPTION        = "Module visualizes an input tetra mesh and underlying volume using GLUT and OpenGL.";

//! Additional command line arguments.
const std::string MODULE_ARGUMENTS          = "alpha:coloring";

//! Additional arguments.
const std::string MODULE_ARGUMENT_ALPHA     = "alpha";
const std::string MODULE_ARGUMENT_COLORING  = "coloring";

//! Default GLUT window position.
const int GLUT_WINDOW_XPOS                  = 100;
const int GLUT_WINDOW_YPOS                  = 100;

//! Default GLUT window size.
int GLUT_WINDOW_XSIZE                       = 512;
int GLUT_WINDOW_YSIZE                       = 512;

//! State of the mouse buttons.
enum EMouseState
{
    MOUSE_NORMAL = 0,
    MOUSE_ROTATION,
    MOUSE_ZOOM
};

//! Type of the slice coloring method
enum ESliceColoring
{
    COL_SEGMENTED,
    COL_SEGMENTED2,
};

//! Allowed 'coloring' argument values
const std::string COL_SEGMENTED_NAME        = "segmented";
const std::string COL_SEGMENTED2_NAME       = "segmented2";

//! Default color argument
const std::string DEFAULT_COLORING_NAME     = COL_SEGMENTED_NAME;

//! Maximum voxel value.
const unsigned int PIXEL_MAX                = mds::img::CPixelTraits<mds::img::tPixel8>::getPixelMax();

//! Maximal number of input channels
const int MAX_CHANNELS                      = 2;

//! Default alpha argument value
const unsigned int DEFAULT_ALPHA            = 32;


//==============================================================================
/*
 * Global variables.
 */

//! Current GLUT window size.
int iGlutWindowXSize                        = GLUT_WINDOW_XSIZE;
int iGlutWindowYSize                        = GLUT_WINDOW_YSIZE;

#ifdef ENABLE_PERSPECTIVE

//! Field of view.
GLfloat fGLFoV                              = 70.0f;

#endif // ENABLE_PERSPECTIVE

//! Near and far view frustum planes.
GLfloat fGLNearPlane                        = 1.0f;
GLfloat fGLFarPlane                         = 2048.0f;

//! State of the mouse buttons.
EMouseState eMouseState                     = MOUSE_NORMAL;

//! Texture/polygon color alpha component
GLfloat fGLMeshAlpha;

//! Coloring method
ESliceColoring eColoring;

//! Alpha
unsigned int uAlpha;

//! Plane positions.
double pPlanes[3];

//! Texture numbers.
GLuint puGLTextures[3];

//! Size of the used texture part.
GLfloat pfGLTextureXSize[3];
GLfloat pfGLTextureYSize[3];

//! Mouse cursor position.
int iPosX, iPosY, iPosZ;
int iOldX = 0, iOldY = 0, iOldZ = 0;
int iNewX = 0, iNewY = 0, iNewZ = 0;

//! The number of input channels
int iChannels;

#ifndef ENABLE_PERSPECTIVE

//! Default model position in 3D space.
const GLfloat fGLPosition                   = fGLFarPlane * -0.5;

//! Zooming step.
const GLfloat fGLZoomStep                   = 0.01f;

//! Minimal zoom factor.
const GLfloat fGLMinZoom                    = 0.2f;

//! Maximal zoom factor.
const GLfloat fGLMaxZoom                    = 100.f;

//! Zoom factor.
GLfloat fGLZoom                             = 1.0f;

#endif // ENABLE_PERSPECTIVE

//! Step length
const double STEP_X                         = 1.0;
const double STEP_Y                         = 1.0;
const double STEP_Z                         = 1.0;


//==============================================================================
/*
 * Global functions.
 */

//! Texture creation.
void createTexture(mds::img::EPlane Plane,
                   mds::img::CDImage *pImage
                   );

//! Display a given plane.
void displayPlane(mds::img::EPlane Plane);


//==============================================================================
/*
 * GLUT functions.
 */

//! Callback functions
void onInit();
void onResize(int width, int height);
void onDisplay(void);
void onKeyboard(unsigned char key, int x, int y);
void onSpecialKeyboard(int key, int x, int y);
void onMouseButton(int iButton, int iState, int x, int y);
void onMouseMotion(int x, int y);


//==============================================================================
/*
 * Implementation of the class CTetraMeshView.
 */
CTetraMeshView *CTetraMeshView::m_pView = NULL;


CTetraMeshView::CTetraMeshView(const std::string& sDescription)
    : mds::mod::CView(sDescription)   
{
    MDS_ASSERT(m_pView == NULL);
    m_pView = this;

    allowArguments(MODULE_ARGUMENTS);
}


CTetraMeshView::~CTetraMeshView()
{
}


bool CTetraMeshView::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    iChannels = getNumOfInputs();

    // Test of existence of input channel
    if( iChannels < 1 || iChannels > MAX_CHANNELS )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input channels" << std::endl);
        return false;
    }
    

    // Alpha channel
    uAlpha = DEFAULT_ALPHA;
    m_Arguments.value(MODULE_ARGUMENT_ALPHA, uAlpha);
    if( uAlpha > 0xff )
    {
        MDS_CERR('<' << m_sFilename << "> Bad 'alpha' parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // Type of slice coloring
    std::string sColoringName = DEFAULT_COLORING_NAME;
    m_Arguments.value(MODULE_ARGUMENT_COLORING, sColoringName);
    if( sColoringName == COL_SEGMENTED_NAME )
    {
        eColoring = COL_SEGMENTED;
    }
    else if( sColoringName == COL_SEGMENTED2_NAME )
    {
        eColoring = COL_SEGMENTED2;
    }
    else
    {
        MDS_CERR('<' << m_sFilename << "> Bad 'coloring' parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // O.K.
    return true;
}


bool CTetraMeshView::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // Input channel
    mds::mod::CChannel *pIChannel = getInput(0);

    // Wait while channel is not closed
    bool bMeshRead = false;
    while( pIChannel->isConnected() )
    {
        // Wait for data
        if( pIChannel->wait(1000) )
        {
            // Read tetra mesh from the input channel
            if( readInput(pIChannel, &m_TetraMesh) )
            {
                bMeshRead = true;
                break;                
            }
            else
            {
                MDS_CERR('<' << m_sFilename << "> Failed to read input tetra mesh" << std::endl);
                return false; 
            }
        }
        else
        {
            MDS_LOG_NOTE("Wait timeout");
            return false;
        }
    }

    if( iChannels == 2 )
    {
        // Input channel
        pIChannel = getInput(1);

        // Wait while channel is not closed
        while( pIChannel->isConnected() )
        {
            // Wait for data
            if( pIChannel->wait(1000) )
            {
                // Read volume from the input channel
                if( readInput(pIChannel, &m_Volume) )
                {
                   break;
                }
                else
                {
                    MDS_CERR('<' << m_sFilename << "> Failed to read input volume" << std::endl);
                    return false;
                }
                
            }
            else
            {
                MDS_LOG_NOTE("Wait timeout");
                return false;
            }
        }
    }

    if( bMeshRead )
    {
        // initialization
        initTetraMeshView(m_TetraMesh.getNumOfSegments());

        // Initialize GLUT library
        glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
        glutCreateWindow(GLUT_WINDOW_TITLE);
        glutDisplayFunc(onDisplay);
        glutReshapeFunc(onResize);
        glutKeyboardFunc(onKeyboard);
        glutSpecialFunc(onSpecialKeyboard);
        glutMouseFunc(onMouseButton);
        glutMotionFunc(onMouseMotion);

        // Initialize module
        onInit();

        // Run GLUT event processing loop
        glutMainLoop();
    }

    return false;
}


void CTetraMeshView::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTetraMeshView::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR("Extended usage: [-alpha Alpha]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("  -alpha  Adjusts transparency of the visualized mesh." << std::endl);
    MDS_CERR("          Allowed values are within the range 0..255. If no" << std::endl);
    MDS_CERR("          value is specified the default number " << DEFAULT_ALPHA << " is used." << std::endl);
    MDS_CERR("Input: tetra mesh, volume data." << std::endl);
    MDS_CERR("Output: none." << std::endl);
    
}


// Pseudo coloring
/*void pseudoColor(float fSegment, float& fR, float& fG, float &fB)
{
    if( fSegment < 0.0f )
    {
        fR = 0.0f;
        fG = 0.0f;
        fB = 0.0f;
    }
    else if( fSegment <= 1.0f )
    {
        fR = 0.0f;
        fG = fSegment;
        fB = 0.0f;
    }
    else if( fSegment <= 2.0f )
    {
        fR = 0.0f;
        fG = 1.0f;
        fB = fSegment - 1.0f;
    }
    else if( fSegment <= 3.0f )
    {
        fR = 0.0f;
        fG = 1.0f - (fSegment - 2.0f);
        fB = 1.0f;
    }
    else if( fSegment <= 4.0f )
    {
        fR = fSegment - 3.0f;
        fG = 0.0f;
        fB = 1.0f;
    }
    else if( fSegment <= 5.0f )
    {
        fR = 1.0f;
        fG = 0.0f;
        fB = 1.0f - (fSegment - 4.0f);
    }
    else if( fSegment <= 6.0f )
    {
        fR = 1.0f;
        fG = fSegment - 5.0f;
        fB = fSegment - 5.0f;
    }
    else
    {
        fR = 1.0f;
        fG = 1.0f;
        fB = 1.0f;
    }
}*/


void CTetraMeshView::initTetraMeshView(mds::tSize NumOfRegions)
{
    if( iChannels == 2 )
    {
        m_maxX = m_Volume.getXSize() - 1;
        m_maxY = m_Volume.getYSize() - 1;
        m_maxZ = m_Volume.getZSize() - 1;
        m_dx = m_Volume.getDX();
        m_dy = m_Volume.getDY();
        m_dz = m_Volume.getDZ();
    }
    else
    {
        vctl::MCVertex min, max;        
        getTetraS()->GetVerticeS()->GetRange(min, max);

        m_maxX = max.GetX();
        m_maxY = max.GetY();
        m_maxZ = max.GetZ();
        m_dx = 1.0;
        m_dy = 1.0;
        m_dz = 1.0;
    }

    fGLMeshAlpha = (GLfloat)uAlpha / 255.0f;

    // set colors of segments
    // default color
//    m_colorTable.SetDefaultColor(mds::cs::colorGL4f(1.0, 1.0, 1.0, 1.0));
    m_colorTable.SetDefaultColor(mds::cs::colorGL4f(1.0, 1.0, 1.0, fGLMeshAlpha));

    // segment 0
/*    m_colorTable.push_back(mds::cs::colorGL4f(1.0, 0.0, 0.0, fGLMeshAlpha));
    // segment ...
    m_colorTable.push_back(mds::cs::colorGL4f(0.0, 1.0, 0.0, fGLMeshAlpha));
    m_colorTable.push_back(mds::cs::colorGL4f(0.0, 0.0, 1.0, fGLMeshAlpha));
    m_colorTable.push_back(mds::cs::colorGL4f(1.0, 0.0, 1.0, fGLMeshAlpha));
    m_colorTable.push_back(mds::cs::colorGL4f(1.0, 1.0, 0.0, fGLMeshAlpha));
    m_colorTable.push_back(mds::cs::colorGL4f(0.0, 1.0, 1.0, fGLMeshAlpha));*/
    
    // Random coloring
    mds::math::CUniformPRNG Random;
    for( mds::tSize i = 0; i < NumOfRegions; ++i )
    {
        float fR, fG, fB;
        if( eColoring == COL_SEGMENTED )
        {
            fB = float(Random.random(0.0, 1.0));
            fR = float(Random.random(0.0, 1.0));
            fG = float(Random.random(0.0, 1.0));
        }
        else
        {
            float fValue = float(i + 1) / float(NumOfRegions + 1);
            fB = fValue;
            fR = fValue;
            fG = fValue;
        }
        m_colorTable.push_back(mds::cs::colorGL4f(fR, fG, fB, fGLMeshAlpha));
    }
    
    // assign color table instance to triangle containers
    m_triXY.SetColorTable(& m_colorTable);
    m_triXZ.SetColorTable(& m_colorTable);
    m_triYZ.SetColorTable(& m_colorTable);

    // initialize cross sections
    m_csXY.SetTetraS(getTetraS());
    m_csXY.SetPlane(vctl::MCPlane(vctl::MCPlane::XY_PLANE));
    m_csXY.SetTriContainer(&m_triXY);

    m_csXZ.SetTetraS(getTetraS());
    m_csXZ.SetPlane(vctl::MCPlane(vctl::MCPlane::XZ_PLANE));
    m_csXZ.SetTriContainer(&m_triXZ);

    m_csYZ.SetTetraS(getTetraS());
    m_csYZ.SetPlane(vctl::MCPlane(vctl::MCPlane::YZ_PLANE));
    m_csYZ.SetTriContainer(&m_triYZ); 
}


//==============================================================================
/*
 * Function main() which creates and executes view application.
 */
int main(int argc, char *argv[])
{
    // Creation of a view module using smart pointer
    CTetraMeshViewPtr spView(new CTetraMeshView(MODULE_DESCRIPTION));

    // Initialize GLUT library
    glutInit(&argc, argv);

    // Initialize and run the view
    if( spView->init(argc, argv) )
    {
        spView->run();
    }

    // Console application finished
    return 0;
}


//==============================================================================
/*
 * Implementation of the global functions.
 */


// Creation of an OpenGL RGBA texture
void createTexture(mds::img::EPlane Plane,
                   mds::img::CDImage *pImage
                   )
{
    // Set texture image parameters
    glBindTexture(GL_TEXTURE_2D, puGLTextures[Plane]);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

    // Image size
    mds::tSize XSize = pImage->getXSize();
    mds::tSize YSize = pImage->getYSize();

    // Texture dimensions
    mds::tSize TexXSize = 2, TexYSize = 2;
    while( TexXSize < XSize || TexYSize < YSize )
    {
        TexXSize *= 2;
        TexYSize *= 2;
    }

    // Allowed maximal texture dimension
    int iMaxTexDim;
    glGetIntegerv(GL_MAX_TEXTURE_SIZE, &iMaxTexDim);

    // Modify the image dimension
    if( (int)TexXSize > iMaxTexDim )
    {
        XSize = TexXSize = iMaxTexDim;
        YSize = TexYSize = iMaxTexDim;
    }

    // Used texture size
    pfGLTextureXSize[Plane] = (GLfloat)XSize / (GLfloat)TexXSize;
    pfGLTextureYSize[Plane] = (GLfloat)YSize / (GLfloat)TexYSize;

    // Allocate texture RGB data
    unsigned char *pucRGBTexture = new unsigned char[TexYSize * TexXSize * 3];
    memset(pucRGBTexture, 0, TexYSize * TexXSize * 3);

    // Volume data
    mds::img::CDensityVolumePtr spVolume(CTetraMeshView::getViewPtr()->getVolume());

    // Create required texture
    for( mds::tSize j = 0; j < YSize; ++j )
    {
        unsigned char *p = &pucRGBTexture[j * TexXSize * 3];
        for( mds::tSize i = 0; i < XSize; ++i )
        {
            unsigned char ucTemp = (unsigned char)pImage->get(i, j);
            *(p++) = ucTemp;
            *(p++) = ucTemp;
            *(p++) = ucTemp;
        }
    }

    // Define 2D texture image
    glTexImage2D(GL_TEXTURE_2D, 0, 3,
                 TexXSize, TexYSize,
                 0, GL_RGB, GL_UNSIGNED_BYTE,
                 pucRGBTexture
                 );

    // Free the memory
    delete[] pucRGBTexture;
}


// Display a plane
void displayPlane(mds::img::EPlane Plane)
{
    // smart pointer to tetra view
    CTetraMeshViewPtr spTetraMeshView(CTetraMeshView::getViewPtr());
    
    switch( Plane )
    {
        case mds::img::PLANE_XY:
            // display texture
            if(iChannels == 2)
            {
                glEnable(GL_TEXTURE_2D);
                glBegin(GL_QUADS);
                glColor3f(1.0f, 1.0f, 1.0f);
                glTexCoord2f(0.0f, 0.0f);
                glVertex3f(0.0f, 0.0f, pPlanes[Plane] * spTetraMeshView->getDZ());
        
                glTexCoord2f(pfGLTextureXSize[Plane], 0.0f);
                glVertex3f(spTetraMeshView->getMaxX() * spTetraMeshView->getDX(), 0.0f, pPlanes[Plane] * spTetraMeshView->getDZ());
        
                glTexCoord2f(pfGLTextureXSize[Plane], pfGLTextureYSize[Plane]);
                glVertex3f(spTetraMeshView->getMaxX() * spTetraMeshView->getDX(), spTetraMeshView->getMaxY() * spTetraMeshView->getDY(), pPlanes[Plane] * spTetraMeshView->getDZ());
        
                glTexCoord2f(0.0f, pfGLTextureYSize[Plane]);
                glVertex3f(0.0f, spTetraMeshView->getMaxY() * spTetraMeshView->getDY(), pPlanes[Plane] * spTetraMeshView->getDZ());
                glEnd();
                glDisable(GL_TEXTURE_2D);
            }
        
            // display border
            glBegin(GL_LINE_LOOP);
            glColor3f(1.0f, 0.0f, 0.0f);
            glVertex3f(0.0f, 0.0f, pPlanes[Plane] * spTetraMeshView->getDZ());
            glVertex3f(spTetraMeshView->getMaxX() * spTetraMeshView->getDX(), 0.0f, pPlanes[Plane] * spTetraMeshView->getDZ());
            glVertex3f(spTetraMeshView->getMaxX() * spTetraMeshView->getDX(), spTetraMeshView->getMaxY() * spTetraMeshView->getDY(), pPlanes[Plane] * spTetraMeshView->getDZ());
            glVertex3f(0.0f, spTetraMeshView->getMaxY() * spTetraMeshView->getDY(), pPlanes[Plane] * spTetraMeshView->getDZ());
            glEnd();
            break;

        case mds::img::PLANE_XZ:
            // display texture
            if(iChannels == 2)
            {
                glEnable(GL_TEXTURE_2D);
                glBegin(GL_QUADS);
                glColor3f(1.0f, 1.0f, 1.0f);
                glTexCoord2f(0.0f, 0.0f);
                glVertex3f(0.0f, pPlanes[Plane] * spTetraMeshView->getDY(), 0.0f);
        
                glTexCoord2f(pfGLTextureXSize[Plane], 0.0f);
                glVertex3f(spTetraMeshView->getMaxX() * spTetraMeshView->getDX(), pPlanes[Plane] * spTetraMeshView->getDY(), 0.0f);
        
                glTexCoord2f(pfGLTextureXSize[Plane], pfGLTextureYSize[Plane]);
                glVertex3f(spTetraMeshView->getMaxX() * spTetraMeshView->getDX(), pPlanes[Plane] * spTetraMeshView->getDY(), spTetraMeshView->getMaxZ() * spTetraMeshView->getDZ());
        
                glTexCoord2f(0.0f, pfGLTextureYSize[Plane]);
                glVertex3f(0.0f, pPlanes[Plane] * spTetraMeshView->getDY(), spTetraMeshView->getMaxZ() * spTetraMeshView->getDZ());
                glEnd();
                glDisable(GL_TEXTURE_2D);
            }
    
            // display border
            glBegin(GL_LINE_LOOP);
            glColor3f(0.0f, 1.0f, 0.0f);
            glVertex3f(0.0f, pPlanes[Plane] * spTetraMeshView->getDY(), 0.0f);
            glVertex3f(spTetraMeshView->getMaxX() * spTetraMeshView->getDX(), pPlanes[Plane] * spTetraMeshView->getDY(), 0.0f);
            glVertex3f(spTetraMeshView->getMaxX() * spTetraMeshView->getDX(), pPlanes[Plane] * spTetraMeshView->getDY(), spTetraMeshView->getMaxZ() * spTetraMeshView->getDZ());
            glVertex3f(0.0f, pPlanes[Plane] * spTetraMeshView->getDY(), spTetraMeshView->getMaxZ() * spTetraMeshView->getDZ());
            glEnd();
            break;
    
        case mds::img::PLANE_YZ:
            // display texture
            if(iChannels == 2)
            {
                glEnable(GL_TEXTURE_2D);
                glBegin(GL_QUADS);
                glColor3f(1.0f, 1.0f, 1.0f);
                glTexCoord2f(0.0f, 0.0f);
                glVertex3f(pPlanes[Plane] * spTetraMeshView->getDX(), 0.0f, 0.0f);
        
                glTexCoord2f(pfGLTextureXSize[Plane], 0.0f);
                glVertex3f(pPlanes[Plane] * spTetraMeshView->getDX(), spTetraMeshView->getMaxY() * spTetraMeshView->getDY(), 0.0f);
        
                glTexCoord2f(pfGLTextureXSize[Plane], pfGLTextureYSize[Plane]);
                glVertex3f(pPlanes[Plane] * spTetraMeshView->getDX(), spTetraMeshView->getMaxY() * spTetraMeshView->getDY(), spTetraMeshView->getMaxZ() * spTetraMeshView->getDZ());
        
                glTexCoord2f(0.0f, pfGLTextureYSize[Plane]);
                glVertex3f(pPlanes[Plane] * spTetraMeshView->getDX(), 0.0f, spTetraMeshView->getMaxZ() * spTetraMeshView->getDZ());
                glEnd();
                glDisable(GL_TEXTURE_2D);
            }
    
            // display border
            glBegin(GL_LINE_LOOP);
            glColor3f(0.0f, 0.0f, 1.0f);
            glVertex3f(pPlanes[Plane] * spTetraMeshView->getDX(), 0.0f, 0.0f);
            glVertex3f(pPlanes[Plane] * spTetraMeshView->getDX(), spTetraMeshView->getMaxY() * spTetraMeshView->getDY(), 0.0f);
            glVertex3f(pPlanes[Plane] * spTetraMeshView->getDX(), spTetraMeshView->getMaxY() * spTetraMeshView->getDY(), spTetraMeshView->getMaxZ() * spTetraMeshView->getDZ());
            glVertex3f(pPlanes[Plane] * spTetraMeshView->getDX(), 0.0f, spTetraMeshView->getMaxZ() * spTetraMeshView->getDZ());
            glEnd();
            break;
    
        default:
            break;
    }


    // Tetra data
    // coordinates of triangle vertices stored in an array
    GLdouble * triangles = spTetraMeshView->getTriCont(Plane)->getVertices();
    // colors of corresponding vertices stored in array
    GLfloat * color = spTetraMeshView->getTriCont(Plane)->getColors();
    // number of triangles
    int numTriangles = spTetraMeshView->getTriCont(Plane)->getNumTriangles();

    // display blended with volume data
    if(iChannels == 2)
    {
        // disable depth buffer
        glDepthMask(GL_FALSE);
    
        // enable polygon offset
        glEnable(GL_POLYGON_OFFSET_FILL);
        glPolygonOffset(-1.0f, -1.0f);

        // enable blending
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);        
    }
    else
    {
        // draw wireframe
        glPolygonMode(GL_FRONT, GL_LINE);
        glPolygonMode(GL_BACK, GL_LINE);
    }

    // draw triangular mesh
    glBegin(GL_TRIANGLES);  
    for(int i = 0; i < numTriangles; ++i)
    {
        glColor4f(color[4*i], color[4*i+1], color[4*i+2], color[4*i+3]);
        glVertex3d(triangles[i*3*3   ], triangles[i*3*3 +1], triangles[i*3*3 +2]);
        glVertex3d(triangles[i*3*3 +3], triangles[i*3*3 +4], triangles[i*3*3 +5]);
        glVertex3d(triangles[i*3*3 +6], triangles[i*3*3 +7], triangles[i*3*3 +8]);            
    }
    glEnd();
    
    if(iChannels == 2)
    {
        // disable polygon offset
        glDisable(GL_POLYGON_OFFSET_FILL);
    
        // disable blending
        glDisable(GL_BLEND);
    
        // enable depth buffer
        glDepthMask(GL_TRUE);
    }
}


// GLUT and texture initialization
void onInit()
{
    // Set clear color
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

    // Enable z-buffer
    glClearDepth(1.0f);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);

    // Enable antialiasing
    glEnable(GL_POINT_SMOOTH);
    glEnable(GL_LINE_SMOOTH);
    glEnable(GL_POLYGON_SMOOTH);

    // Enable the polygon filling
    glPolygonMode(GL_FRONT, GL_FILL);
    glPolygonMode(GL_BACK, GL_FILL);

    // Turn the culling off
    glDisable(GL_CULL_FACE);

    // Set global texture image parameters
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

    // Make performace hints
#ifdef ENABLE_PERSPECTIVE
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
#endif // ENABLE_PERSPECTIVE

    // Set the line width
//    glLineWidth(1.1f);
    glLineWidth(1.0f);

    // Generate texture names
    glGenTextures(3, puGLTextures);


    CTetraMeshViewPtr spTetraMeshView(CTetraMeshView::getViewPtr());

    // Initialize the plane positions
    pPlanes[mds::img::PLANE_XY] = spTetraMeshView->getMaxZ() * 0.5 + 0.2;
    pPlanes[mds::img::PLANE_XZ] = spTetraMeshView->getMaxY() * 0.5 + 0.2;
    pPlanes[mds::img::PLANE_YZ] = spTetraMeshView->getMaxX() * 0.5 + 0.2;

    if(iChannels == 2)
    {
        // Volume data
        mds::img::CDensityVolumePtr spVolume(spTetraMeshView->getVolume());

        // Create the textures
        mds::img::CDImagePtr spPlane;
        spPlane->create(spVolume->getXSize(), spVolume->getYSize());
        spVolume->getPlaneXY((mds::tSize)pPlanes[mds::img::PLANE_XY], *spPlane);
        createTexture(mds::img::PLANE_XY, spPlane);

        spPlane->create(spVolume->getXSize(), spVolume->getZSize());
        spVolume->getPlaneXZ((mds::tSize)pPlanes[mds::img::PLANE_XZ], *spPlane);
        createTexture(mds::img::PLANE_XZ, spPlane);

        spPlane->create(spVolume->getYSize(), spVolume->getZSize());
        spVolume->getPlaneYZ((mds::tSize)pPlanes[mds::img::PLANE_YZ], *spPlane);
        createTexture(mds::img::PLANE_YZ, spPlane);
    }


    // compute the cross section
    vctl::MCVertex v(pPlanes[mds::img::PLANE_YZ]*spTetraMeshView->getDX(),pPlanes[mds::img::PLANE_XZ]*spTetraMeshView->getDY(),pPlanes[mds::img::PLANE_XY]*spTetraMeshView->getDZ());
    spTetraMeshView->getCS(mds::img::PLANE_XY)->GetPlane()->MoveTo(&v);
    spTetraMeshView->getCS(mds::img::PLANE_XY)->make();
    spTetraMeshView->getCS(mds::img::PLANE_XZ)->GetPlane()->MoveTo(&v);
    spTetraMeshView->getCS(mds::img::PLANE_XZ)->make();
    spTetraMeshView->getCS(mds::img::PLANE_YZ)->GetPlane()->MoveTo(&v);
    spTetraMeshView->getCS(mds::img::PLANE_YZ)->make();


    // Initialize glut window
    glutPositionWindow(GLUT_WINDOW_XPOS, GLUT_WINDOW_YPOS);
    glutReshapeWindow(GLUT_WINDOW_XSIZE, GLUT_WINDOW_YSIZE);

    // Initial zoom
#ifdef ENABLE_PERSPECTIVE
    iOldZ = iNewZ = (int)(spTetraMeshView->getMaxY() * spTetraMeshView->getDY() * -1.2);
#else
    fGLZoom = GLUT_WINDOW_YSIZE / (spTetraMeshView->getMaxY() * spTetraMeshView->getDY()) * 0.6;
#endif // ENABLE_PERSPECTIVE
}


// Sets coordinate system depending on the window size
void onResize(int iWidth, int iHeight)
{
    // Current window size
    iGlutWindowXSize = iWidth;
    iGlutWindowYSize = iHeight;

    // Set the viewport
    glViewport(0, 0, iWidth, iHeight);

    // Projection matrix modification
    glMatrixMode(GL_PROJECTION);

    // Clear the projection matrix
    glLoadIdentity();

#ifdef ENABLE_PERSPECTIVE
    // Perspective projection
    gluPerspective(fGLFoV, (GLfloat)iWidth / (GLfloat)iHeight, fGLNearPlane, fGLFarPlane);
#else
    // Map abstract coordinates to the window coordinates
    glOrtho(0, iWidth, 0, iHeight, fGLNearPlane, fGLFarPlane);

    // Translate the origin
    glTranslatef(iWidth / 2, iHeight / 2, 0);
#endif // ENABLE_PERSPECTIVE
}


// Displays the window
void onDisplay(void)
{
    // Clear the color buffer and z-buffer
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Enable textures
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

    // Model view matrix modification
    glMatrixMode(GL_MODELVIEW);

    // Clear the model view matrix
    glLoadIdentity();

    // Mouse zooming
#ifdef ENABLE_PERSPECTIVE
    glTranslatef(0.0f, 0.0f, iNewZ);
#else
    glTranslatef(0.0f, 0.0f, fGLPosition);
#endif // ENABLE_PERSPECTIVE

    // Mouse rotation
    glRotatef(iNewX, 0.0f, 1.0f, 0.0f);
    glRotatef(iNewY, 1.0f, 0.0f, 0.0f);

    // Invert y-axis
#ifdef ENABLE_PERSPECTIVE
    glScalef(1, -1, 1);
#else
    glScalef(fGLZoom, -fGLZoom, fGLZoom);
#endif // ENABLE_PERSPECTIVE

    CTetraMeshViewPtr spTetraMeshView(CTetraMeshView::getViewPtr());
    
    // Translate to the center
    glTranslatef((GLfloat)(spTetraMeshView->getMaxX() * spTetraMeshView->getDX() * -0.5),
                 (GLfloat)(spTetraMeshView->getMaxY() * spTetraMeshView->getDY() * -0.5),
                 (GLfloat)(spTetraMeshView->getMaxZ() * spTetraMeshView->getDZ() * -0.5)
                );

    // Draw the pPlanes
    glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
    glBindTexture(GL_TEXTURE_2D, puGLTextures[mds::img::PLANE_XY]);
    displayPlane(mds::img::PLANE_XY);

    glBindTexture(GL_TEXTURE_2D, puGLTextures[mds::img::PLANE_XZ]);
    displayPlane(mds::img::PLANE_XZ);

    glBindTexture(GL_TEXTURE_2D, puGLTextures[mds::img::PLANE_YZ]);
    displayPlane(mds::img::PLANE_YZ);

    // Draw changes
    glFlush();
    glutSwapBuffers();
}


// Called on ASCII key pressed
void onKeyboard(unsigned char ucKey, int x, int y)
{
    CTetraMeshViewPtr spTetraMeshView(CTetraMeshView::getViewPtr());

    // Volume data
    mds::img::CDensityVolumePtr spVolume(spTetraMeshView->getVolume());

    // Translate to lower case
    if( ucKey > 'A' && ucKey <= 'Z' )
    {
        ucKey += 'a' - 'A';
    }

    switch( ucKey )
    {
        // ESC, ^C and 'c' exits the application
        case 'c':
        case 27:
        case 3:
            exit(0);
            break;

        // Window mode
        case 'w':
            glutReshapeWindow(GLUT_WINDOW_XSIZE, GLUT_WINDOW_YSIZE);
            glutPositionWindow(GLUT_WINDOW_XPOS, GLUT_WINDOW_YPOS);
            break;

        // Fullscreen mode
        case 'f':
            GLUT_WINDOW_XSIZE = iGlutWindowXSize;
            GLUT_WINDOW_YSIZE = iGlutWindowYSize;
            glutFullScreen();
            break;

        // Initialize XY view
        case '1':
            iNewX = iOldX = 0;
            iNewY = iOldY = 0;
            glutPostRedisplay();
            break;

        // Initialize XZ view
        case '2':
            iNewX = iOldX = 0;
            iNewY = iOldY = 90;
            glutPostRedisplay();
            break;

        // Initialize YZ view
        case '3':
            iNewX = iOldX = 90;
            iNewY = iOldY = 0;
            glutPostRedisplay();
            break;

        // XY plane position
        case 'q':
            if( pPlanes[mds::img::PLANE_XY] <= (spTetraMeshView->getMaxZ() - STEP_Z) )
            {
                pPlanes[mds::img::PLANE_XY] += STEP_Z;
    
                if(iChannels == 2)
                {
                    mds::img::CDImagePtr spPlane(new mds::img::CDImage(spVolume->getXSize(), spVolume->getYSize()));
                    spVolume->getPlaneXY((mds::tSize)pPlanes[mds::img::PLANE_XY], *spPlane);
        
                    createTexture(mds::img::PLANE_XY, spPlane);
                }

                
                vctl::MCVertex v(pPlanes[mds::img::PLANE_YZ]*spTetraMeshView->getDX(),
                                 pPlanes[mds::img::PLANE_XZ]*spTetraMeshView->getDY(),
                                 pPlanes[mds::img::PLANE_XY]*spTetraMeshView->getDZ()
                                 );

                spTetraMeshView->getCS(mds::img::PLANE_XY)->GetPlane()->MoveTo(&v);
                spTetraMeshView->getCS(mds::img::PLANE_XY)->make();   
            }
            glutPostRedisplay();
            break;
    
        case 'a':
            if( pPlanes[mds::img::PLANE_XY] >= STEP_Z )
            {
                pPlanes[mds::img::PLANE_XY] -= STEP_Z;
    
                if(iChannels == 2)
                {
                    mds::img::CDImagePtr spPlane(new mds::img::CDImage(spVolume->getXSize(), spVolume->getYSize()));
                    spVolume->getPlaneXY((mds::tSize)pPlanes[mds::img::PLANE_XY], *spPlane);
    
                    createTexture(mds::img::PLANE_XY, spPlane);
                }
                
                vctl::MCVertex v(pPlanes[mds::img::PLANE_YZ]*spTetraMeshView->getDX(),
                                 pPlanes[mds::img::PLANE_XZ]*spTetraMeshView->getDY(),
                                 pPlanes[mds::img::PLANE_XY]*spTetraMeshView->getDZ()
                                 );
                
                spTetraMeshView->getCS(mds::img::PLANE_XY)->GetPlane()->MoveTo(&v);
                spTetraMeshView->getCS(mds::img::PLANE_XY)->make();                
            }
            glutPostRedisplay();
            break;

        // Help
        case 'h':
            MDS_CERR("Keyboard Shortcuts:" << std::endl);
            MDS_CERR(" h        Shows this help." << std::endl);
            MDS_CERR(" q and a  Changes the XY plane position." << std::endl);
            MDS_CERR(" Arrows   Changes position of the YZ and XZ planes." << std::endl);
            MDS_CERR(" 1-3      Initializes the eye position." << std::endl);
            MDS_CERR(" f        Switches to the fullscreen mode." << std::endl);
            MDS_CERR(" w        Switches back to the window." << std::endl);
            MDS_CERR(" c        Exits the program." << std::endl);
            break;
    }
}


// Called on direction ucKey pressed
void onSpecialKeyboard(int ucKey, int x, int y)
{
    CTetraMeshViewPtr spTetraMeshView(CTetraMeshView::getViewPtr());

    // Volume data
    mds::img::CDensityVolumePtr spVolume(spTetraMeshView->getVolume());

    switch( ucKey )
    {
        // XZ plane position
        case GLUT_KEY_DOWN:
            if( pPlanes[mds::img::PLANE_XZ] <= (spTetraMeshView->getMaxY() - STEP_Y) )
            {
                pPlanes[mds::img::PLANE_XZ] += STEP_Y;
    
                if(iChannels == 2)
                {
                    mds::img::CDImagePtr spPlane(new mds::img::CDImage(spVolume->getXSize(), spVolume->getZSize()));
                    spVolume->getPlaneXZ((mds::tSize)pPlanes[mds::img::PLANE_XZ], *spPlane);
    
                    createTexture(mds::img::PLANE_XZ, spPlane);
                }
                vctl::MCVertex v(pPlanes[mds::img::PLANE_YZ]*spTetraMeshView->getDX(),
                                 pPlanes[mds::img::PLANE_XZ]*spTetraMeshView->getDY(),
                                 pPlanes[mds::img::PLANE_XY]*spTetraMeshView->getDZ()
                                 );
            
                spTetraMeshView->getCS(mds::img::PLANE_XZ)->GetPlane()->MoveTo(&v);
                spTetraMeshView->getCS(mds::img::PLANE_XZ)->make();
            }
            glutPostRedisplay();
            break;

        case GLUT_KEY_UP:
            if( pPlanes[mds::img::PLANE_XZ] >= STEP_Y)
            {
                pPlanes[mds::img::PLANE_XZ] -= STEP_Y;
    
                if(iChannels == 2)
                {
                    mds::img::CDImagePtr spPlane(new mds::img::CDImage(spVolume->getXSize(), spVolume->getZSize()));
                    spVolume->getPlaneXZ((mds::tSize)pPlanes[mds::img::PLANE_XZ], *spPlane);
    
                    createTexture(mds::img::PLANE_XZ, spPlane);
                }

                vctl::MCVertex v (pPlanes[mds::img::PLANE_YZ]*spTetraMeshView->getDX(),
                                    pPlanes[mds::img::PLANE_XZ]*spTetraMeshView->getDY(),
                                    pPlanes[mds::img::PLANE_XY]*spTetraMeshView->getDZ());

                spTetraMeshView->getCS(mds::img::PLANE_XZ)->GetPlane()->MoveTo(&v);
                spTetraMeshView->getCS(mds::img::PLANE_XZ)->make();
            }
            glutPostRedisplay();
            break;

        // YZ plane position
        case GLUT_KEY_RIGHT:
            if( pPlanes[mds::img::PLANE_YZ] <= (spTetraMeshView->getMaxX() - STEP_X) )
            {
                pPlanes[mds::img::PLANE_YZ] += STEP_X;
    
                if(iChannels == 2)
                {
                    mds::img::CDImagePtr spPlane(new mds::img::CDImage(spVolume->getYSize(), spVolume->getZSize()));
                    spVolume->getPlaneYZ((mds::tSize)pPlanes[mds::img::PLANE_YZ], *spPlane);
    
                    createTexture(mds::img::PLANE_YZ, spPlane);
                }

                vctl::MCVertex v(pPlanes[mds::img::PLANE_YZ]*spTetraMeshView->getDX(),
                                 pPlanes[mds::img::PLANE_XZ]*spTetraMeshView->getDY(),
                                 pPlanes[mds::img::PLANE_XY]*spTetraMeshView->getDZ()
                                 );
            
                spTetraMeshView->getCS(mds::img::PLANE_YZ)->GetPlane()->MoveTo(&v);
                spTetraMeshView->getCS(mds::img::PLANE_YZ)->make();
            
            }
            glutPostRedisplay();
            break;
    
        case GLUT_KEY_LEFT:
            if( pPlanes[mds::img::PLANE_YZ] >= STEP_X )
            {
                pPlanes[mds::img::PLANE_YZ] -= STEP_X;
    
                if(iChannels == 2)
                {
                    mds::img::CDImagePtr spPlane(new mds::img::CDImage(spVolume->getYSize(), spVolume->getZSize()));
                    spVolume->getPlaneYZ((mds::tSize)pPlanes[mds::img::PLANE_YZ], *spPlane);
    
                    createTexture(mds::img::PLANE_YZ, spPlane);
                }

                vctl::MCVertex v(pPlanes[mds::img::PLANE_YZ]*spTetraMeshView->getDX(),
                                 pPlanes[mds::img::PLANE_XZ]*spTetraMeshView->getDY(),
                                 pPlanes[mds::img::PLANE_XY]*spTetraMeshView->getDZ()
                                 );
            
                spTetraMeshView->getCS(mds::img::PLANE_YZ)->GetPlane()->MoveTo(&v);
                spTetraMeshView->getCS(mds::img::PLANE_YZ)->make();
            
            }
            glutPostRedisplay();
            break;
    }
}


// Called on mouse button action
void onMouseButton(int iButton, int iState, int x, int y)
{
    // Roatation
    if( iButton == GLUT_LEFT_BUTTON )
    {
        if( iState == GLUT_DOWN )
        {
            eMouseState = MOUSE_ROTATION;
            iPosX = x;
            iPosY = y;
        }
        else
        {
            eMouseState = MOUSE_NORMAL;
            iOldX = iNewX;
            iOldY = iNewY;
        }
    }

    // Zoom
    else if( iButton == GLUT_RIGHT_BUTTON )
    {
        if( iState == GLUT_DOWN )
        {
            eMouseState = MOUSE_ZOOM;
            iPosZ = y;
        }
        else
        {
            eMouseState = MOUSE_NORMAL;
            iOldZ = iNewZ;
        }
    }

    // Redraw the scene
    glutPostRedisplay();
}


// Called on mouse motion
void onMouseMotion(int x, int y)
{
    if( eMouseState == MOUSE_ROTATION )
    {
        // Compute the rotation
        iNewX = iOldX + x - iPosX;
        iNewY = iOldY + y - iPosY;
    }

    else if( eMouseState == MOUSE_ZOOM)
    {
#ifdef ENABLE_PERSPECTIVE
        iNewZ = iOldZ + y - iPosZ;
#else
        fGLZoom += (y - iPosZ) * fGLZoomStep;
        if( fGLZoom < fGLMinZoom )
        {
            fGLZoom = fGLMinZoom;
        }
        else if( fGLZoom > fGLMaxZoom )
        {
            fGLZoom = fGLMaxZoom;
        }
        iPosZ = y;
#endif // ENABLE_PERSPECTIVE
    }

    // Redisplay the scene
    glutPostRedisplay();
}

