//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2005 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsTetraMeshView.h                  \n
 * Section: mTetraMeshView                      \n
 * Date:    2004/07/30                          \n
 *
 * $Id: mdsTetraMeshView.h 1550 2009-12-21 15:40:37Z spanel $
 *
 * File description:
 * - Reads a tetra mesh and underlying volume from the input channel.
 * - Visualizes it using GLUT and OpenGL.
 */

#ifndef MDS_TETRAMESHVIEW_H
#define MDS_TETRAMESHVIEW_H

#include <MDSTk/Base/mdsSetup.h>
#include <MDSTk/Module/mdsView.h>
#include <MDSTk/Image/mdsSlice.h>
#include <MDSTk/Image/mdsDensityVolume.h>
#include <MDSTk/VectorSegmentation/mdsTetraMesh.h>

#include <MDSTk/CrossSection/mdsCrossSection.h>
#include <MDSTk/CrossSection/mdsTriContainerGL.h>


//==============================================================================
/*!
 * Module visualizes an input tetra mesh (and volume data) using GLUT and OpenGL.
 */
class CTetraMeshView : public mds::mod::CView
{
public:
    //! Smart pointer type.
    //! - Declares type tSmartPtr.
    MDS_SHAREDPTR(CTetraMeshView);

public:
    //! Default constructor.
    CTetraMeshView(const std::string& sDescription);

    //! Virtual destructor.
    virtual ~CTetraMeshView();

    //! Returns pointer to the current view.
    static CTetraMeshView *getViewPtr()
    {
        MDS_ASSERT(m_pView);
        return m_pView;
    }

    //! Returns pointer to the input volume.
    mds::img::CDensityVolume *getVolume()
    {
        return &m_Volume;
    }

    //! Returns pointer to the input tetra mesh.
    vctl::MCTetraS *getTetraS()
    {
        return m_TetraMesh.getTetrahedra();
    }

	//! Returns pointer to the cross section in coordinate plane
    mds::cs::CCrossSection * getCS(mds::img::EPlane Plane)
    {
        switch(Plane)
        {
            case mds::img::PLANE_XY: return &m_csXY;
            case mds::img::PLANE_XZ: return &m_csXZ;
            case mds::img::PLANE_YZ: return &m_csYZ;
            default: /*never could get here*/ return NULL;
        }
    }

	//! Returns pointer to triangles of cross section in coordinate plane
    mds::cs::CTriContainerGL * getTriCont(mds::img::EPlane Plane)
    {
        switch(Plane)
        {
            case mds::img::PLANE_XY: return &m_triXY;
            case mds::img::PLANE_XZ: return &m_triXZ;
            case mds::img::PLANE_YZ: return &m_triYZ;
            default: /*never could get here*/ return NULL;
        }
    }

	//! Returns maximal coordinate of the model
    double getMaxX() { return m_maxX; }
    double getMaxY() { return m_maxY; }
    double getMaxZ() { return m_maxZ; }

	//! Returns size of a voxel
    double getDX() { return m_dx; }
    double getDY() { return m_dy; }
    double getDZ() { return m_dz; }

protected:
    //! Virtual method called on startup.
    virtual bool startup();

    //! Virtual method called by the processing thread.
    virtual bool main();

    //! Called on console shutdown.
    virtual void shutdown();

    //! Called on writing a usage statement.
    virtual void writeExtendedUsage(std::ostream& Stream);

protected:
    //! Static pointer to the created view.
    static CTetraMeshView *m_pView;

    //! Visualized volume.
    mds::img::CDensityVolume m_Volume;

    //! Tetra mesh.
    mds::seg::CTetraMesh m_TetraMesh;

    //! Containers of triangles
    mds::cs::CTriContainerGL m_triXY;
    mds::cs::CTriContainerGL m_triXZ;
    mds::cs::CTriContainerGL m_triYZ;

    //! Triangulated crosssections
    mds::cs::CCrossSection m_csXY;
    mds::cs::CCrossSection m_csXZ;
    mds::cs::CCrossSection m_csYZ;

    //! Color table
    mds::cs::CColorTable<mds::cs::colorGL4f> m_colorTable;

	//! Maximal coordinates of the model
    double m_maxX, m_maxY, m_maxZ;

    //! Size of a voxel
    double m_dx, m_dy, m_dz;

    //! Initialization
    void initTetraMeshView(mds::tSize NumOfRegions);
};


//==============================================================================
/*!
 * Smart pointer to view application.
 */
typedef CTetraMeshView::tSmartPtr  CTetraMeshViewPtr;


#endif // MDS_VOLUMEVIEW_H

