//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2005 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsTriMeshAdapt.h                   \n
 * Section: mTriMeshAdapt                       \n
 * Date:    2005/09/28                          \n
 *
 * $Id: mdsTriMeshAdapt.cpp 1494 2009-05-21 14:36:00Z spanel $
 *
 * Description:
 * - Adaptation of a triangular mesh to an underlying image.
 * - An empty module at present!
 */

#include "mdsTriMeshAdapt.h"

// STL
#include <climits>
#include <iostream>
#include <string>


//==============================================================================
/*
 * Global constants.
 */

//! Module description
const std::string MODULE_DESCRIPTION    = "Adapts triangular mesh to underlying image data.";

//! Additional command line arguments
const std::string MODULE_ARGUMENTS      = "";

//! Additional arguments names
const std::string MODULE_ARG_           = "";


//==============================================================================
/*
 * Implementation of the class CTriMeshAdapt.
 */
CTriMeshAdapt::CTriMeshAdapt(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CTriMeshAdapt::~CTriMeshAdapt()
{
}


bool CTriMeshAdapt::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 2 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // O.K.
    return true;
}


bool CTriMeshAdapt::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read underlying slice from the input channel
    if( !readInput(pIChannel, m_spSlice.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input slice" << std::endl);
        return false;
    }

    // I/O channels
    pIChannel = getInput(1);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read mesh from the input channel
    if( !readInput(pIChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input triangular mesh" << std::endl);
        return false;
    }

    // Set the underlying mesh image
    m_spMesh->setImage((mds::img::CDImage *)m_spSlice);

    // Adapt the triangular mesh
    // ...

    // Write the result to the output channel
    if( !writeOutput(pOChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write the triangular mesh" << std::endl);
    }

    // Returning 'true' means to continue processing the input channel
    return false;
}


void CTriMeshAdapt::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTriMeshAdapt::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR(std::endl);
    MDS_CERR("Extended usage:" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("Input: image data, triangular mesh." << std::endl);
    MDS_CERR("Output: triangular mesh." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CTriMeshAdaptPtr spModule(new CTriMeshAdapt(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

