//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2005 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsTriMeshEdgeSplitting.cpp         \n
 * Section: mTriMeshEdgeSplitting               \n
 * Date:    2005/12/15                          \n
 *
 * $Id: mdsTriMeshEdgeSplitting.cpp 1494 2009-05-21 14:36:00Z spanel $
 *
 * Description:
 * - Adaptation of a triangular mesh to an underlying image.
 * - Edge splitting algorithm.
 */

#include "mdsTriMeshEdgeSplitting.h"

// MDSTk
#include <MDSTk/VectorSegmentation/mdsTriEdgeIterator.h>

// STL
#include <climits>
#include <iostream>
#include <string>


//==============================================================================
/*
 * Global constants.
 */

//! Module description
const std::string MODULE_DESCRIPTION    = "Module provides edge splitting algorithm for adaptation of a triangular mesh to an underlying image.";

//! Additional command line arguments
const std::string MODULE_ARGUMENTS      = "minlength:threshold";

//! Additional arguments names
const std::string MODULE_ARG_MIN_LENGTH = "minlength";
const std::string MODULE_ARG_THRESHOLD  = "threshold";

//! Default arguments values
const double DEFAULT_MIN_LENGTH         = mds::seg::CTriMesh::DEFAULT_MIN_EDGE_LENGTH;
const double DEFAULT_THRESHOLD          = mds::seg::CTriMesh::DEFAULT_SPLITTING_THRESHOLD;

//! Allowed arguments values
const double MIN_LENGTH                 = 1.0;
const double MAX_LENGTH                 = 50.0;

const double MIN_THRESHOLD              = 0.0;
const double MAX_THRESHOLD              = 1.0;


//==============================================================================
/*
 * Implementation of the class CTriMeshEdgeSplitting.
 */
CTriMeshEdgeSplitting::CTriMeshEdgeSplitting(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CTriMeshEdgeSplitting::~CTriMeshEdgeSplitting()
{
}


bool CTriMeshEdgeSplitting::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 3 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // Parse command line arguments
    m_dMinLength = DEFAULT_MIN_LENGTH;
    m_Arguments.value(MODULE_ARG_MIN_LENGTH, m_dMinLength);
    if( m_dMinLength < MIN_LENGTH || m_dMinLength > MAX_LENGTH )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_MIN_LENGTH << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_dOrientThreshold = DEFAULT_THRESHOLD;
    m_Arguments.value(MODULE_ARG_THRESHOLD, m_dOrientThreshold);
    if( m_dOrientThreshold < MIN_THRESHOLD || m_dOrientThreshold > MAX_THRESHOLD )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_THRESHOLD << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // O.K.
    return true;
}


bool CTriMeshEdgeSplitting::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read underlying slice from the input channel
    if( !readInput(pIChannel, m_spSlice.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input slice" << std::endl);
        return false;
    }


    // I/O channels
    pIChannel = getInput(1);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read underlying edge slice from the input channel
    if( !readInput(pIChannel, m_spEdgeSlice.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input edge slice" << std::endl);
        return false;
    }


    // I/O channels
    pIChannel = getInput(2);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read mesh from the input channel
    if( !readInput(pIChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input triangular mesh" << std::endl);
        return false;
    }


    // Set the underlying mesh image
    m_spMesh->setImage((mds::img::CDImage *)m_spSlice);
    m_spMesh->setEdgeImage((mds::img::CDImage *)m_spEdgeSlice);

    // Triangle rasterization
    // - Only for testing purposes!
/*    mds::img::CSlicePtr spNewSlice(*m_spSlice);
    spNewSlice->fillEntire(mds::img::CPixelTraits<mds::img::tDensityPixel>::getPixelMin());
    vctl::MCTri *pTriangle = m_spMesh->getFirstTriangle();
    while( pTriangle )
    {
        vctl::MCVertex *pVertices[3];
        pTriangle->GetVerticeS(pVertices);
 
        for( int i = 0; i < 3; ++i )
        {
            int j = (i + 1) % 3;
            
            vctl::MCEdge Edge(pVertices[i], pVertices[j]);
            if( m_spMesh->checkBoundaryEdge(&Edge, 0.5) )
            {
                mds::seg::CTriEdgeIterator It(spNewSlice, &Edge);
                while( !It.isEnd() )
                {
                    *It = mds::img::CPixelTraits<mds::img::tDensityPixel>::getPixelMax();
                    ++It;
                }
            }
        }
 
        pTriangle = (vctl::MCTri *)pTriangle->GetNext();
    }*/

    // Adapt the triangular mesh
    m_spMesh->edgeSplitting(m_dMinLength, m_dOrientThreshold);

    // Write the result to the output channel
    if( !writeOutput(pOChannel, m_spMesh.get()) )
//    if( !writeOutput(pOChannel, spNewSlice.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write the triangular mesh" << std::endl);
    }

    // Returning 'true' means to continue processing the input channel
    return false;
}


void CTriMeshEdgeSplitting::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTriMeshEdgeSplitting::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR(std::endl);
    MDS_CERR("Extended usage: [-minlength dValue] [-threshold dValue]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("Input: image data, edge image, triangular mesh." << std::endl);
    MDS_CERR("Output: triangular mesh." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CTriMeshEdgeSplittingPtr spModule(new CTriMeshEdgeSplitting(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

