//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2005 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsTriMeshInfo.h                    \n
 * Section: mTriMeshInfo                        \n
 * Date:    2005/10/13                          \n
 *
 * $Id: mdsTriMeshInfo.cpp 1494 2009-05-21 14:36:00Z spanel $
 *
 * Description:
 * - Shows information on triangular mesh.
 */

#include "mdsTriMeshInfo.h"

#include <climits>
#include <cstring>
#include <cstdio>
#include <cmath>

// STL
#include <iostream>
#include <string>


//==============================================================================
/*
 * Global constants.
 */

//! Module description
const std::string MODULE_DESCRIPTION    = "Shows information on input triangular mesh.";

//! Additional command line arguments
const std::string MODULE_ARGUMENTS      = "";

//! Additional arguments names

//! Default argument values


//==============================================================================
/*
 * Implementation of the class CTriMeshInfo.
 */
CTriMeshInfo::CTriMeshInfo(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CTriMeshInfo::~CTriMeshInfo()
{
}


bool CTriMeshInfo::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 1 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // O.K.
    return true;
}


bool CTriMeshInfo::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read mesh from the input channel
    if( !readInput(pIChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input triangular mesh" << std::endl);
        return false;
    }

    // Get minimal and maximal coordinates
    double dMinX = 0.0, dMinY = 0.0, dMaxX = 0.0, dMaxY = 0.0;
    m_spMesh->getMinCoordinates(dMinX, dMinY);
    m_spMesh->getMaxCoordinates(dMaxX, dMaxY);

    // Create the output string
    char pcOut[1024];
    sprintf(pcOut,
            "-- Information on triangular mesh --\n"
            " Number of vertices:  %d\n"
            " Number of triangles: %d\n"
            " Minimal coordinates: %.2f, %.2f\n"
            " Maximal coordinates: %.2f, %.2f\n"
            " Number of regions:   %d\n",
            m_spMesh->getNumOfVertices(),
            m_spMesh->getNumOfTriangles(),
            dMinX, dMinY,
            dMaxX, dMaxY,
            m_spMesh->getNumOfSegments()
           );

    // Write it to the output channel
    if( !pOChannel->write(pcOut, (int)strlen(pcOut)) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write the output" << std::endl);
        return false;
    }

    // Returning 'true' means to continue processing the input channel
    return true;
}


void CTriMeshInfo::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTriMeshInfo::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR(std::endl);
    MDS_CERR("Extended usage:]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("Input: triangular mesh." << std::endl);
    MDS_CERR("Output: text." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CTriMeshInfoPtr spModule(new CTriMeshInfo(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

