//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2005 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsTriMeshOptimize.h                \n
 * Section: mTriMeshOptimize                    \n
 * Date:    2005/10/11                          \n
 *
 * $Id: mdsTriMeshOptimize.cpp 1494 2009-05-21 14:36:00Z spanel $
 *
 * Description:
 * - Triangular mesh optimization.
 */

#include "mdsTriMeshOptimize.h"

// STL
#include <climits>
#include <iostream>
#include <string>


//==============================================================================
/*
 * Global constants.
 */

//! Module description
const std::string MODULE_DESCRIPTION    = "Triangular mesh optimization.";

//! Additional command line arguments
const std::string MODULE_ARGUMENTS      = "quality:edge";

//! Additional arguments names
const std::string MODULE_ARG_Q          = "quality";
const std::string MODULE_ARG_E          = "edge";

//! Default argument values
const double DEFAULT_ARG_Q              = 5.0;
const double DEFAULT_ARG_E              = 10.0;


//==============================================================================
/*
 * Implementation of the class CTriMeshOptimize.
 */
CTriMeshOptimize::CTriMeshOptimize(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CTriMeshOptimize::~CTriMeshOptimize()
{
}


bool CTriMeshOptimize::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
//    if( getNumOfInputs() != 2 || getNumOfOutputs() != 1 )
    if( getNumOfInputs() != 1 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // Command line parameters
    m_dQuality = DEFAULT_ARG_Q;
    m_Arguments.value(MODULE_ARG_Q, m_dQuality);
    if( m_dQuality < 1.0 )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_Q << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_dEdge = DEFAULT_ARG_E;
    m_Arguments.value(MODULE_ARG_E, m_dEdge);
    if( m_dEdge < 1.0 )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_E << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // O.K.
    return true;
}


bool CTriMeshOptimize::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read slice from the input channel
/*    if( !readInput(pIChannel, m_spSlice.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input slice" << std::endl);
        return false;
    }
 
    // I/O channels
    pIChannel = getInput(1);
 
    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }
 
    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }*/

    // Read mesh from the input channel
    if( !readInput(pIChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input triangular mesh" << std::endl);
        return false;
    }

    // Set the underlying mesh image
//    m_spMesh->setImage((mds::img::CDImage *)m_spSlice);

    // Optimize the triangular mesh
    m_spMesh->normalize(m_dQuality, m_dEdge);

    // Write the result to the output channel
    if( !writeOutput(pOChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write the triangular mesh" << std::endl);
    }

    // Returning 'true' means to continue processing the input channel
    return false;
}


void CTriMeshOptimize::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTriMeshOptimize::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR(std::endl);
    MDS_CERR("Extended usage: [-quality dValue] [-edge dValue]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("  -quality  Required minimal quality of all triangles in the mesh." << std::endl);
    MDS_CERR("            The quality means is ratio of the longest traingle" << std::endl);
    MDS_CERR("            edge and the inscribed circle radius. It is normalized" << std::endl);
    MDS_CERR("            so that quality of equilateral triangle is equal to" << std::endl);
    MDS_CERR("            one. Default value is " << DEFAULT_ARG_Q << "." << std::endl);
    MDS_CERR(std::endl);
    MDS_CERR("  -edge  If the longest triangle edge is shorter then a given value" << std::endl);
    MDS_CERR("         the triangle is not optimized. Default value of this" << std::endl);
    MDS_CERR("         parameter is " << DEFAULT_ARG_E << "." << std::endl);
    MDS_CERR("Input: triangular mesh." << std::endl);
    MDS_CERR("Output: triangular mesh." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CTriMeshOptimizePtr spModule(new CTriMeshOptimize(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

