//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2006 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsTriMeshRegionClassifyFCM.h       \n
 * Section: mTriMeshRegionClassifyFCM           \n
 * Date:    2006/12/01                          \n
 *
 * $Id: mdsTriMeshRegionClassifyFCM.cpp 1494 2009-05-21 14:36:00Z spanel $
 *
 * Description:
 * - Classification of sub-regions in a mesh.
 */

#include "mdsTriMeshRegionClassifyFCM.h"

// MDSTk Extensions
#include <MDSTk/VectorSegmentation/mdsTriMembershipFCM2.h>

// STL
#include <climits>
#include <iostream>
#include <string>


//==============================================================================
/*
 * Global constants.
 */

//! Module description
const std::string MODULE_DESCRIPTION    = "Module that classifies sub-regions into segments.";

//! Additional command line arguments
const std::string MODULE_ARGUMENTS      = "weight:clusters";

//! Additional arguments names
const std::string MODULE_ARGUMENT_W     = "weight";
const std::string MODULE_ARGUMENT_C     = "clusters";

//! Default arguments values
const double DEFAULT_W                  = 1.5;
const mds::tSize DEFAULT_C              = 0;

//! Allowed arguments values
const double MIN_W                      = 1.0;
const double MAX_W                      = 10.0;


//==============================================================================
/*
 * Implementation of the class CTriMeshRegionClassifyFCM.
 */
CTriMeshRegionClassifyFCM::CTriMeshRegionClassifyFCM(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CTriMeshRegionClassifyFCM::~CTriMeshRegionClassifyFCM()
{
}


bool CTriMeshRegionClassifyFCM::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 2 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    // Parse command line arguments
    m_dWeight = DEFAULT_W;
    m_Arguments.value(MODULE_ARGUMENT_W, m_dWeight);
    if( m_dWeight < MIN_W || m_dWeight > MAX_W )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARGUMENT_W << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_NumOfClusters = DEFAULT_C;
    m_Arguments.value(MODULE_ARGUMENT_C, m_NumOfClusters);
    if( m_NumOfClusters < 0 )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARGUMENT_C << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // O.K.
    return true;
}


bool CTriMeshRegionClassifyFCM::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read slice from the input channel
    if( !readInput(pIChannel, m_spSlice.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input slice" << std::endl);
        return false;
    }

    // I/O channels
    pIChannel = getInput(1);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read mesh from the input channel
    if( !readInput(pIChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input triangular mesh" << std::endl);
        return false;
    }

    // Set the underlying mesh image
    m_spMesh->setImage((mds::img::CDImage *)m_spSlice);

    // Evaluate sub-region features
    m_spMesh->evaluateSegmentFeatures();

    // Region membership function
    mds::seg::CTriMembershipFCM2 Membership(m_spMesh->getSegments(), m_dWeight);

    // Classify the triangles
    if( m_NumOfClusters > 0 )
    {
        Membership(m_spMesh->getFirstTriangle(), m_NumOfClusters);
    }
    else
    {
        Membership(m_spMesh->getFirstTriangle());
    }

    m_spMesh->setNumOfSegments(Membership.getNumOfSegments());

    // Write the result to the output channel
    if( !writeOutput(pOChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write the triangular mesh" << std::endl);
    }

    // Returning 'true' means to continue processing the input channel
    return false;
}


void CTriMeshRegionClassifyFCM::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTriMeshRegionClassifyFCM::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR("Extended usage: [-weight dValue] [-clusters iValue]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("  -weight  Parameter of the fuzzy C-means clustering algorithm." << std::endl);
    MDS_CERR("           See its documentation. Value in the range " << MIN_W << ".." << MAX_W << " can be" << std::endl);
    MDS_CERR("           given. Default value is " << DEFAULT_W << "." << std::endl);
    MDS_CERR("Input: image data, pre-classified triangular mesh." << std::endl);
    MDS_CERR("Output: triangular mesh." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CTriMeshRegionClassifyFCMPtr spModule(new CTriMeshRegionClassifyFCM(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

