//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2006 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsTriMeshRegionMerging.cpp         \n
 * Section: mTriMeshRegionMerging               \n
 * Date:    2006/11/20                          \n
 *
 * $Id: mdsTriMeshRegionMerging.cpp 1494 2009-05-21 14:36:00Z spanel $
 *
 * Description:
 * - Region merging algorithm.
 */

#include "mdsTriMeshRegionMerging.h"

// STL
#include <climits>
#include <iostream>
#include <string>


//==============================================================================
/*
 * Global constants.
 */

//! Module description
const std::string MODULE_DESCRIPTION        = "Module merges regions until a given similarity predicate is true.";

//! Additional command line arguments
const std::string MODULE_ARGUMENTS          = "bthreshold:simthreshold:rho:regions";

//! Additional arguments names
const std::string MODULE_ARG_BTHRESHOLD     = "bthreshold";
const std::string MODULE_ARG_SIMTHRESHOLD   = "simthreshold";
const std::string MODULE_ARG_RHO            = "rho";
const std::string MODULE_ARG_REGIONS        = "regions";

//! Default arguments values
const double DEFAULT_BTHRESHOLD             = mds::seg::CTriMesh::DEFAULT_DISTANCE_THRESHOLD;
const double DEFAULT_SIMTHRESHOLD           = mds::seg::CTriMeanValueSimilarity::DEFAULT_THRESHOLD;
const double DEFAULT_RHO                    = mds::seg::CTriMeanValueSimilarity::DEFAULT_RHO;
const int DEFAULT_REGIONS                   = -1;

//! Allowed arguments values
const double MIN_THRESHOLD                  = 0.0;
const double MAX_THRESHOLD                  = 10.0;


//==============================================================================
/*
 * Implementation of the class CTriMeshRegionMerging.
 */
CTriMeshRegionMerging::CTriMeshRegionMerging(const std::string& sDescription)
    : mds::mod::CModule(sDescription)
{
    allowArguments(MODULE_ARGUMENTS);
}


CTriMeshRegionMerging::~CTriMeshRegionMerging()
{
}


bool CTriMeshRegionMerging::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // Test of existence of input and output channel
    if( getNumOfInputs() != 3 || getNumOfOutputs() != 1 )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input and output channels" << std::endl);
        return false;
    }

    m_dBoundaryThreshold = DEFAULT_BTHRESHOLD;
    m_Arguments.value(MODULE_ARG_BTHRESHOLD, m_dBoundaryThreshold);
    if( m_dBoundaryThreshold < MIN_THRESHOLD || m_dBoundaryThreshold > MAX_THRESHOLD )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_BTHRESHOLD << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_dSimilarityThreshold = DEFAULT_SIMTHRESHOLD;
    m_Arguments.value(MODULE_ARG_SIMTHRESHOLD, m_dSimilarityThreshold);
    if( m_dSimilarityThreshold < MIN_THRESHOLD || m_dSimilarityThreshold > MAX_THRESHOLD )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_SIMTHRESHOLD << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_dRho = DEFAULT_RHO;
    m_Arguments.value(MODULE_ARG_RHO, m_dRho);
    if( m_dRho < 0.0 )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_RHO << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    m_iNumOfRegions = DEFAULT_REGIONS;
    m_Arguments.value(MODULE_ARG_REGIONS, m_iNumOfRegions);
    if( m_iNumOfRegions < -1 )
    {
        MDS_CERR('<' << m_sFilename << "> Bad " << MODULE_ARG_REGIONS << " parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // O.K.
    return true;
}


bool CTriMeshRegionMerging::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // I/O channels
    mds::mod::CChannel *pIChannel = getInput(0);
    mds::mod::CChannel *pOChannel = getOutput(0);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read underlying slice from the input channel
    if( !readInput(pIChannel, m_spSlice.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input slice" << std::endl);
        return false;
    }


    // I/O channels
    pIChannel = getInput(1);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read underlying edge slice from the input channel
    if( !readInput(pIChannel, m_spEdgeSlice.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input edge slice" << std::endl);
        return false;
    }


    // I/O channels
    pIChannel = getInput(2);

    // Is any input?
    if( !pIChannel->isConnected() )
    {
        return false;
    }

    // Wait for data
    if( !pIChannel->wait(1000) )
    {
        MDS_LOG_NOTE("Wait timeout");
        return true;
    }

    // Read mesh from the input channel
    if( !readInput(pIChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to read input triangular mesh" << std::endl);
        return false;
    }


    // Set the underlying mesh image
    m_spMesh->setImage((mds::img::CDImage *)m_spSlice);
    m_spMesh->setEdgeImage((mds::img::CDImage *)m_spEdgeSlice);

    // Prepare a similarity criterion
    mds::seg::CTriMeanValueSimilarity Criterion(m_dSimilarityThreshold, m_dRho);

    // Aglomerative merging of regions
    m_spMesh->regionMerging(Criterion, m_dBoundaryThreshold, m_iNumOfRegions);
    
    // Classify very small triangles
    m_spMesh->classifyAdjacent(m_dBoundaryThreshold);
    
    // Write the result to the output channel
    if( !writeOutput(pOChannel, m_spMesh.get()) )
    {
        MDS_CERR('<' << m_sFilename << "> Failed to write the triangular mesh" << std::endl);
    }

    // Returning 'true' means to continue processing the input channel
    return false;
}


void CTriMeshRegionMerging::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTriMeshRegionMerging::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR(std::endl);
    MDS_CERR("Extended usage: [-bthreshold dValue] [-simthreshold dValue] [-rho dValue] [-regions iValue]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("Input: image data, edge image, triangular mesh." << std::endl);
    MDS_CERR("Output: triangular mesh." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes console application.
 */
int main(int argc, char *argv[])
{
    // Creation of a module using smart pointer
    CTriMeshRegionMergingPtr spModule(new CTriMeshRegionMerging(MODULE_DESCRIPTION));

    // Initialize and execute the module
    if( spModule->init(argc, argv) )
    {
        spModule->run();
    }

    // Console application finished
    return 0;
}

