//==============================================================================
/*! \file
 * Medical Data Segmentation Toolkit (MDSTk)    \n
 * Copyright (c) 2003-2005 by Michal Spanel     \n
 *
 * Author:  Michal Spanel, spanel@fit.vutbr.cz  \n
 * File:    mdsTriMeshView.cpp                  \n
 * Section: mTriMeshView                        \n
 * Date:    2005/08/30                          \n
 *
 * $Id: mdsTriMeshView.cpp 1494 2009-05-21 14:36:00Z spanel $
 *
 * Description:
 * - Reads a triangular mesh and underlying slice from the input channel.
 * - Draws both using GLUT and OpenGL.
 */

#include "mdsTriMeshView.h"

// MDSTk Extensions
#include <MDSTk/VectorSegmentation/mdsTriFeatures.h>

// OpenGL
#include <GL/gl.h>
#include <GL/glut.h>

// STL
#include <iostream>
#include <string>


//==============================================================================
/*
 * Global constants.
 */

//! GLUT window title
const char GLUT_WINDOW_TITLE[]              = "MDSTk Triangular Mesh View Window";

//! Module description
const std::string MODULE_DESCRIPTION        = "Module visualizes an input triangular mesh and underlying image using GLUT and OpenGL.";

//! Additional command line arguments
const std::string MODULE_ARGUMENTS          = "coloring:alpha";

//! Additional arguments
const std::string MODULE_ARGUMENT_COLORING  = "coloring";
const std::string MODULE_ARGUMENT_ALPHA     = "alpha";

//! Default GLUT window position
const int GLUT_WINDOW_XPOS                  = 100;
const int GLUT_WINDOW_YPOS                  = 100;

//! Default GLUT window size
int GLUT_WINDOW_XSIZE                       = 512;
int GLUT_WINDOW_YSIZE                       = 512;

//! State of the mouse buttons
enum EMouseState
{
    MOUSE_NORMAL = 0,
    MOUSE_MOVE,
    MOUSE_ZOOM
};

//! Type of the slice coloring method
enum ESliceColoring
{
    COL_GRAYSCALE = 1,
    COL_PSEUDO,
    COL_SEGMENTED,
    COL_SEGMENTED2
};

//! Allowed 'coloring' argument values
const std::string COL_GRAYSCALE_NAME        = "grayscale";
const std::string COL_PSEUDO_NAME           = "pseudo";
const std::string COL_SEGMENTED_NAME        = "segmented";
const std::string COL_SEGMENTED2_NAME       = "segmented2";

//! Default color argument
const std::string DEFAULT_COLORING_NAME     = COL_GRAYSCALE_NAME;

//! Default alpha argument value
const unsigned int DEFAULT_ALPHA            = 128;

//! Maximum pixel value
const unsigned int PIXEL_MAX                = mds::img::CPixelTraits<mds::img::tPixel8>::getMax();

//! Maximal number of input channels
const int MAX_CHANNELS                      = 2;


//==============================================================================
/*
 * Global variables.
 */

//! Current GLUT window size
int iGlutWindowXSize                        = GLUT_WINDOW_XSIZE;
int iGlutWindowYSize                        = GLUT_WINDOW_YSIZE;

//! Near and far view frustum planes
GLfloat fGLNearPlane                        = 1.0f;
GLfloat fGLFarPlane                         = 2048.0f;

//! State of the mouse buttons
EMouseState eMouseState                     = MOUSE_NORMAL;

//! Texture number
GLuint uGLTexture;

//! Display list number
GLuint uGLLineList, uGLTriList;

//! Size of the used texture part
GLfloat fGLTextureXSize;
GLfloat fGLTextureYSize;

//! Center point of the triangular mesh
GLfloat fGLMeshXSize;
GLfloat fGLMeshYSize;

//! Texture/polygon color alpha component
GLfloat fGLTextureAlpha;

//! Texture/polygon color alpha component
GLfloat fGLMeshAlpha;

//! Coloring method
ESliceColoring eColoring;

//! Alpha
unsigned int uAlpha;

//! The number of input channels
int iChannels;

//! Default camera position in 3D space
const GLfloat fGLPosition                   = fGLFarPlane * -0.5;

//! Zooming step
const GLfloat fGLZoomStep                   = 0.01f;

//! Minimal zoom factor
const GLfloat fGLMinZoom                    = 0.2f;

//! Maximal zoom factor
const GLfloat fGLMaxZoom                    = 100.f;

//! Zoom factor
GLfloat fGLZoom                             = 1.0f;

//! Mouse cursor position
int iPosX, iPosY, iPosZ;
int iOldX = 0, iOldY = 0, iOldZ = 0;
int iNewX = 0, iNewY = 0, iNewZ = 0;


//==============================================================================
/*
 * Global functions.
 */

//! Texture creation
void createTexture(ESliceColoring Coloring, mds::img::CSlice *pSlice);

//! Triangular mesh creation
void compileMesh(ESliceColoring Coloring, mds::seg::CTriMesh *pMesh);


//==============================================================================
/*
 * GLUT functions.
 */

//! Callback functions
void onInit();
void onResize(int iWidth, int iHeight);
void onDisplay(void);
void onKeyboard(unsigned char ucKey, int iX, int iY);
void onMouseButton(int iButton, int iState, int x, int y);
void onMouseMotion(int x, int y);


//==============================================================================
/*
 * Implementation of the class CTriMeshView.
 */
CTriMeshView *CTriMeshView::m_pView = NULL;


CTriMeshView::CTriMeshView(const std::string& sDescription)
    : mds::mod::CView(sDescription)
{
    MDS_ASSERT(m_pView == NULL);
    m_pView = this;

    allowArguments(MODULE_ARGUMENTS);
}


CTriMeshView::~CTriMeshView()
{
}


bool CTriMeshView::startup()
{
    // Note
    MDS_LOG_NOTE("Module startup");

    // The number of input channels
    iChannels = getNumOfInputs();

    // Test of existence of input channel
    if( iChannels < 1 || iChannels > MAX_CHANNELS )
    {
        MDS_CERR('<' << m_sFilename << "> Wrong number of input channels" << std::endl);
        return false;
    }

    // Type of slice coloring
    std::string sColoringName = DEFAULT_COLORING_NAME;
    m_Arguments.value(MODULE_ARGUMENT_COLORING, sColoringName);
    if( sColoringName == COL_GRAYSCALE_NAME )
    {
        eColoring = COL_GRAYSCALE;
    }
    else if( sColoringName == COL_PSEUDO_NAME )
    {
        eColoring = COL_PSEUDO;
    }
    else if( sColoringName == COL_SEGMENTED_NAME )
    {
        eColoring = COL_SEGMENTED;
    }
    else if( sColoringName == COL_SEGMENTED2_NAME )
    {
        eColoring = COL_SEGMENTED2;
    }
    else
    {
        MDS_CERR('<' << m_sFilename << "> Bad 'coloring' parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // Alpha channel
    uAlpha = DEFAULT_ALPHA;
    m_Arguments.value(MODULE_ARGUMENT_ALPHA, uAlpha);
    if( uAlpha > 0xff )
    {
        MDS_CERR('<' << m_sFilename << "> Bad 'alpha' parameter value: type -h for help" << std::endl);
        printUsage();
        return false;
    }

    // O.K.
    return true;
}


bool CTriMeshView::main()
{
    // Note
    MDS_LOG_NOTE("Module main function");

    // Input channel
    mds::mod::CChannel *pIChannel = getInput(0);

    // Wait while channel is not closed
    bool bMeshRead = false;
    while( pIChannel->isConnected() )
    {
        // Wait for data
        if( pIChannel->wait(1000) )
        {
            // Read the triangular mesh from input channel
            if( readInput(pIChannel, m_spMesh.get()) )
            {
                bMeshRead = true;
                break;
            }
            else
            {
                MDS_CERR('<' << m_sFilename << "> Failed to read input triangular mesh" << std::endl);
                return false;
            }
        }
        else
        {
            MDS_LOG_NOTE("Wait timeout");
        }
    }

    if( iChannels == 2 )
    {
        // Second input channel
        pIChannel = getInput(1);

        while( pIChannel->isConnected() )
        {
            // Wait for data
            if( pIChannel->wait(1000) )
            {
                // Read the slice from input channel
                if( readInput(pIChannel, m_spSlice.get()) )
                {
                    break;
                }
                else
                {
                    MDS_CERR('<' << m_sFilename << "> Failed to read input slice" << std::endl);
                    return false;
                }
            }
            else
            {
                MDS_LOG_NOTE("Wait timeout");
            }
        }
    }

    // Draw the mesh
    if( bMeshRead )
    {
        // Initialize GLUT library
        glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
        glutCreateWindow(GLUT_WINDOW_TITLE);
        glutDisplayFunc(onDisplay);
        glutReshapeFunc(onResize);
        glutKeyboardFunc(onKeyboard);
        glutMouseFunc(onMouseButton);
        glutMotionFunc(onMouseMotion);

        // Initialize module
        onInit();

        // Run GLUT event processing loop
        glutMainLoop();
    }

    return false;
}


void CTriMeshView::shutdown()
{
    // Note
    MDS_LOG_NOTE("Module shutdown");
}


void CTriMeshView::writeExtendedUsage(std::ostream& Stream)
{
    MDS_CERR("Extended usage: [-coloring Type] [-alpha Alpha]" << std::endl);
    MDS_CERR("Options:" << std::endl);
    MDS_CERR("  -coloring  Selects the slice coloring method. Allowed types are" << std::endl);
    MDS_CERR("               " << COL_GRAYSCALE_NAME << "," << std::endl);
    MDS_CERR("               " << COL_PSEUDO_NAME << " and" << std::endl);
    MDS_CERR("               " << COL_SEGMENTED_NAME << "." << std::endl);
    MDS_CERR("             Pseudo-coloring shows an underlying slice colored instead of" << std::endl);
    MDS_CERR("             grayscale and segmented shows the correspondence" << std::endl);
    MDS_CERR("             of triangles to regions." << std::endl);
    MDS_CERR(std::endl);
    MDS_CERR("  -alpha  Adjusts transparency of the visualized mesh." << std::endl);
    MDS_CERR("          Allowed values are within the range 0..255. If no" << std::endl);
    MDS_CERR("          value is specified the default number " << DEFAULT_ALPHA << " is used." << std::endl);
    MDS_CERR("Input: triangular mesh, image data." << std::endl);
    MDS_CERR("Output: none." << std::endl);
}


//==============================================================================
/*
 * Function main() which creates and executes view application.
 */
int main(int argc, char *argv[])
{
    // Creation of a view module using smart pointer
    CTriMeshViewPtr spView(new CTriMeshView(MODULE_DESCRIPTION));

    // Initialize GLUT library
    glutInit(&argc, argv);

    // Initialize and run the view
    if( spView->init(argc, argv) )
    {
        spView->run();
    }

    // Console application finished
    return 0;
}


//==============================================================================
/*
 * GLUT functions implementation.
 */

// Pseudo coloring
void pseudoColor(double dSegment, unsigned char *pucRGB)
{
    if( dSegment < 0.0 )
    {
        *(pucRGB) = (unsigned char)(0);
        *(pucRGB + 1) = (unsigned char)(0);
        *(pucRGB + 2) = (unsigned char)(0);
    }
    else if( dSegment <= 1.0 )
    {
        *(pucRGB) = (unsigned char)(0);
        *(pucRGB + 1) = (unsigned char)(255.0 * dSegment);
        *(pucRGB + 2) = (unsigned char)(0);
    }
    else if( dSegment <= 2.0 )
    {
        *(pucRGB) = (unsigned char)(0);
        *(pucRGB + 1) = (unsigned char)(255);
        *(pucRGB + 2) = (unsigned char)(255.0 * (dSegment - 1.0));
    }
    else if( dSegment <= 3.0 )
    {
        *(pucRGB) = (unsigned char)(0);
        *(pucRGB + 1) = (unsigned char)(255.0 - 255.0 * (dSegment - 2.0));
        *(pucRGB + 2) = (unsigned char)(255);
    }
    else if( dSegment <= 4.0 )
    {
        *(pucRGB) = (unsigned char)(255.0 * (dSegment - 3.0));
        *(pucRGB + 1) = (unsigned char)(0);
        *(pucRGB + 2) = (unsigned char)(255);
    }
    else if( dSegment <= 5.0 )
    {
        *(pucRGB) = (unsigned char)(255);
        *(pucRGB + 1) = (unsigned char)(0);
        *(pucRGB + 2) = (unsigned char)(255.0 - 255.0 * (dSegment - 4.0));
    }
    else if( dSegment <= 6.0 )
    {
        *(pucRGB) = (unsigned char)(255);
        *(pucRGB + 1) = (unsigned char)(255.0 * (dSegment - 5.0));
        *(pucRGB + 2) = (unsigned char)(255.0 * (dSegment - 5.0));
    }
    else
    {
        *(pucRGB) = (unsigned char)(255);
        *(pucRGB + 1) = (unsigned char)(255);
        *(pucRGB + 2) = (unsigned char)(255);
    }
}


void pseudoColor2(double dSegment, unsigned char *pucRGB)
{
    static const double dConst = 255.0 / 6.0;

    unsigned char ucValue = 255;
    if( dSegment <= 6.0 )
    {
        ucValue = (unsigned char)(dConst * dSegment);
    }

    *(pucRGB) = ucValue;
    *(pucRGB + 1) = ucValue;
    *(pucRGB + 2) = ucValue;
}


// Creation of a GL RGB texture
void createTexture(ESliceColoring Coloring, mds::img::CSlice *pSlice)
{
    // Set texture image parameters
    glBindTexture(GL_TEXTURE_2D, uGLTexture);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

    // Image size
    mds::tSize XSize = pSlice->getXSize();
    mds::tSize YSize = pSlice->getYSize();

    // Texture dimensions
    mds::tSize TexXSize = 2, TexYSize = 2;
    while( TexXSize < XSize || TexYSize < YSize )
    {
        TexXSize *= 2;
        TexYSize *= 2;
    }

    // Allowed maximal texture dimension
    int iMaxTexDim;
    glGetIntegerv(GL_MAX_TEXTURE_SIZE, &iMaxTexDim);

    // Modify the image dimension
    if( TexXSize > iMaxTexDim )
    {
        XSize = TexXSize = iMaxTexDim;
        YSize = TexYSize = iMaxTexDim;
    }

    // Used texture size
    fGLTextureXSize = (GLfloat)pSlice->getXSize() / (GLfloat)TexXSize;
    fGLTextureYSize = (GLfloat)pSlice->getYSize() / (GLfloat)TexYSize;

    // Allocate texture RGBA data
    unsigned char *pucRGBTexture = new unsigned char[TexXSize * TexYSize * 3];
    memset(pucRGBTexture, 0, TexXSize * TexYSize * 3);

    // Create required texture
    switch( Coloring )
    {
        // Pseudo color RGB texture
        case COL_PSEUDO: {
            // Pseudo coloring of grayscale image
            double dConst = 6.0 / (double)PIXEL_MAX;
            for( mds::tSize j = 0; j < YSize; ++j )
            {
                unsigned char *p = &pucRGBTexture[j * TexXSize * 3];
                for( mds::tSize i = 0; i < XSize; ++i )
                {
                    double dSegment = dConst * (double)pSlice->get(i, j);
                    pseudoColor(dSegment, p);
                    p += 3;
                }
            }
        } break;

        // Grayscale texture
        default: {
            for( mds::tSize j = 0; j < YSize; ++j )
            {
                unsigned char *p = &pucRGBTexture[j * TexXSize * 3];
                for( mds::tSize i = 0; i < XSize; ++i )
                {
                    unsigned char ucTemp = (unsigned char)pSlice->get(i, j);
                    *(p++) = ucTemp;
                    *(p++) = ucTemp;
                    *(p++) = ucTemp;
                }
            }
        } break;
    }

    // Define 2D texture image
    glTexImage2D(GL_TEXTURE_2D, 0, 3,
                 TexXSize,
                 TexYSize,
                 0, GL_RGB, GL_UNSIGNED_BYTE,
                 pucRGBTexture
                 );

    // Free the memory
    delete[] pucRGBTexture;
}


// Display list creation
void compileMesh(ESliceColoring eColoring, mds::seg::CTriMesh *pMesh)
{
    static const float fByteConv = 1.0f / 255.0f;

    // Triangles pseudo coloring
    mds::tSize NumOfRegions = pMesh->getNumOfSegments();
    double dConst = 6.0 / NumOfRegions;

    // Create a new display list
    uGLLineList = 1;
    glNewList(uGLLineList, GL_COMPILE);

    // Polygon filling
    glPolygonMode(GL_FRONT, GL_LINE);
    glPolygonMode(GL_BACK, GL_LINE);

    // Enable polygon offset
    glPolygonOffset(-1, -1);
    glEnable(GL_POLYGON_OFFSET_LINE);

    // Enable the blending
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_BLEND);

    // Set the mesh color
    glColor4f(1.0f, 1.0f, 1.0f, fGLMeshAlpha);

    // Insert all triangles to the list
    glBegin(GL_TRIANGLES);

    vctl::MCTri *pTriangle = pMesh->getFirstTriangle();
    while( pTriangle )
    {
        // Get triangle vertices
        vctl::MCVertex *pVertices[3];
        pTriangle->GetVerticeS(pVertices);

        // OpenGL Triangle
        glVertex3d(pVertices[0]->GetX(), pVertices[0]->GetY(), 0.0);
        glVertex3d(pVertices[1]->GetX(), pVertices[1]->GetY(), 0.0);
        glVertex3d(pVertices[2]->GetX(), pVertices[2]->GetY(), 0.0);

        // Get next one
        pTriangle = (vctl::MCTri *)pTriangle->GetNext();
    }

    glEnd();

    // Disable blending
    glDisable(GL_BLEND);

    // Disable polygon offset
    glDisable(GL_POLYGON_OFFSET_LINE);

    glEndList();

    // Enable the polygon filling
    if( eColoring == COL_SEGMENTED || eColoring == COL_SEGMENTED2 )
    {
        // Create a new display list
        uGLTriList = 2;
        glNewList(uGLTriList, GL_COMPILE);

        // Polygon filling
        glPolygonMode(GL_FRONT, GL_FILL);
        glPolygonMode(GL_BACK, GL_FILL);

        // Enable polygon offset
        glPolygonOffset(-1, -1);
        glEnable(GL_POLYGON_OFFSET_FILL);

        // Enable the blending
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glEnable(GL_BLEND);

        // Insert all triangles to the list
        glBegin(GL_TRIANGLES);

        vctl::MCTri *pTriangle = pMesh->getFirstTriangle();
        while( pTriangle )
        {
            unsigned char pucRGB[3];
            double dSegment = dConst * mds::seg::getTriangleRegion(pTriangle);
            if( eColoring == COL_SEGMENTED )
            {
                pseudoColor(dSegment, pucRGB);
            }
            else
            {
                pseudoColor2(dSegment, pucRGB);
            }

            glColor4f(pucRGB[0] * fByteConv,
                      pucRGB[1] * fByteConv,
                      pucRGB[2] * fByteConv,
                      fGLMeshAlpha
                     );

            // Get triangle vertices
            vctl::MCVertex *pVertices[3];
            pTriangle->GetVerticeS(pVertices);

            // OpenGL Triangle
            glVertex3d(pVertices[0]->GetX(), pVertices[0]->GetY(), 0.0);
            glVertex3d(pVertices[1]->GetX(), pVertices[1]->GetY(), 0.0);
            glVertex3d(pVertices[2]->GetX(), pVertices[2]->GetY(), 0.0);

            // Get next one
            pTriangle = (vctl::MCTri *)pTriangle->GetNext();
        }

        glEnd();

        // Disable blending
        glDisable(GL_BLEND);

        // Disable polygon offset
        glDisable(GL_POLYGON_OFFSET_FILL);

        glEndList();
    }
}


// GLUT and texture initialization
void onInit()
{
    // Set clear color
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

    // Enable z-buffer
    glClearDepth(1.0f);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);

    // Enable antialiasing
    glEnable(GL_POINT_SMOOTH);
    glEnable(GL_LINE_SMOOTH);
    glEnable(GL_POLYGON_SMOOTH);

    // Turn the culling off
    glDisable(GL_CULL_FACE);

    // Set global texture image parameters
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

    // Set the line width
    glLineWidth(1.0f);

    // Generate texture names
    glGenTextures(1, &uGLTexture);

    // Minimal and maximal vertex coordinates
    double dMinX, dMinY, dMaxX, dMaxY;
    CTriMeshView::getViewPtr()->getMesh()->getMinCoordinates(dMinX, dMinY);
    CTriMeshView::getViewPtr()->getMesh()->getMaxCoordinates(dMaxX, dMaxY);

    // Size of the triangular mesh
    fGLMeshXSize = (GLfloat)(dMaxX - dMinX);
    fGLMeshYSize = (GLfloat)(dMaxY - dMinY);

    // Modify glut window size to keep image dimensions ratio
    if( fGLMeshXSize > fGLMeshYSize )
    {
        GLUT_WINDOW_YSIZE = int(GLUT_WINDOW_XSIZE * fGLMeshYSize / fGLMeshXSize);
    }
    else if (fGLMeshYSize > fGLMeshXSize )
    {
        GLUT_WINDOW_XSIZE = int(GLUT_WINDOW_YSIZE * fGLMeshXSize / fGLMeshYSize);
    }

    // Create the mesh
    fGLMeshAlpha = (GLfloat)uAlpha / 255.0f;
    compileMesh(eColoring, CTriMeshView::getViewPtr()->getMesh());

    // Create the background polygon texture
    if( iChannels == 2 )
    {
        fGLTextureAlpha = 1.0f;
        createTexture(eColoring, CTriMeshView::getViewPtr()->getSlice());
    }

    // Initialize glut window
    glutPositionWindow(GLUT_WINDOW_XPOS, GLUT_WINDOW_YPOS);
    glutReshapeWindow(GLUT_WINDOW_XSIZE, GLUT_WINDOW_YSIZE);

    // Initial zoom
    fGLZoom = GLUT_WINDOW_YSIZE / (dMaxY - dMinY + 1.0) * 0.9;
}


// Sets coordinate system depending on the window size
void onResize(int iWidth, int iHeight)
{
    iGlutWindowXSize = iWidth;
    iGlutWindowYSize = iHeight;

    // Set the viewport
    glViewport(0, 0, iWidth, iHeight);

    // Projection matrix modification
    glMatrixMode(GL_PROJECTION);

    // Clear the projection matrix
    glLoadIdentity();

    // Map abstract coordinates to the window coordinates
    glOrtho(0, iWidth, 0, iHeight, fGLNearPlane, fGLFarPlane);

    // Translate the origin
    glTranslatef(iWidth / 2, iHeight / 2, 0);
}


// Displays the window
void onDisplay(void)
{
    // Clear the color buffer and z-buffer
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Model view matrix modification
    glMatrixMode(GL_MODELVIEW);

    // Clear the model view matrix
    glLoadIdentity();

    // Mouse movement
    glTranslatef(iNewX, -iNewY, fGLPosition);

    // Invert y-axis
    glScalef(fGLZoom, -fGLZoom, fGLZoom);

    // Translate to the center
    glTranslatef(fGLMeshXSize * -0.5f, fGLMeshYSize * -0.5f, 0.0f);

    // Draw the underlying slice
    if( iChannels == 2 )
    {
        // Enable the polygon filling
        glPolygonMode(GL_FRONT, GL_FILL);
        glPolygonMode(GL_BACK, GL_FILL);

        // Enable the blending
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glEnable(GL_BLEND);

        // Color
        glColor4f(1.0f, 1.0f, 1.0f, fGLTextureAlpha);

        // Enable textures
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
        glEnable(GL_TEXTURE_2D);

        // Draw polygon representing the underlying slice
        glBindTexture(GL_TEXTURE_2D, uGLTexture);
        glBegin(GL_QUADS);

        glTexCoord2f(0.0f, 0.0f);
        glVertex3i(0, 0, 0);

        glTexCoord2f(fGLTextureXSize, 0.0f);
        glVertex3i(CTriMeshView::getViewPtr()->getSlice()->getXSize() - 1, 0, 0);

        glTexCoord2f(fGLTextureXSize, fGLTextureYSize);
        glVertex3i(CTriMeshView::getViewPtr()->getSlice()->getXSize() - 1,
                   CTriMeshView::getViewPtr()->getSlice()->getYSize() - 1,
                   0
                  );

        glTexCoord2f(0.0f, fGLTextureYSize);
        glVertex3i(0, CTriMeshView::getViewPtr()->getSlice()->getYSize() - 1, 0);

        glEnd();

        // Disable textures
        glDisable(GL_TEXTURE_2D);

        // Disable blending
        glDisable(GL_BLEND);
    }

    // Draw the mesh
    glCallList(uGLLineList);
    if( eColoring == COL_SEGMENTED || eColoring == COL_SEGMENTED2)
    {
        glCallList(uGLTriList);
    }

    // Draw changes
    glFlush();
    glutSwapBuffers();
}


// Called on ASCII key pressed
void onKeyboard(unsigned char ucKey, int iX, int iY)
{
    // Translate to lower case
    ucKey = (ucKey > 'A' && ucKey <= 'Z') ? ucKey + 'a' - 'A' : ucKey;

    switch( ucKey )
    {
        // ESC, ^C and 'c' exits the application
        case 'c':
        case 27:
        case 3:
            exit(0);
            break;

        // Window mode
        case 'w':
            glutReshapeWindow(GLUT_WINDOW_XSIZE, GLUT_WINDOW_YSIZE);
            glutPositionWindow(GLUT_WINDOW_XPOS, GLUT_WINDOW_YPOS);
            break;

        // Fullscreen mode
        case 'f':
            GLUT_WINDOW_XSIZE = iGlutWindowXSize;
            GLUT_WINDOW_YSIZE = iGlutWindowYSize;
            glutFullScreen();
            break;

        // Center the view
        case 'v': {
            // Minimal and maximal vertex coordinates
            double dMinX, dMinY, dMaxX, dMaxY;
            CTriMeshView::getViewPtr()->getMesh()->getMinCoordinates(dMinX, dMinY);
            CTriMeshView::getViewPtr()->getMesh()->getMaxCoordinates(dMaxX, dMaxY);

            // Center the view
            fGLZoom = GLUT_WINDOW_YSIZE / (dMaxY - dMinY + 1.0) * 0.9;
            iOldX = iNewX = 0;
            iOldY = iNewY = 0;
    
            // Redraw the scene
            glutPostRedisplay();
        } break;

        // Help
        case 'h':
            MDS_CERR("Keyboard Shortcuts:" << std::endl);
            MDS_CERR(" h  Shows this help." << std::endl);
            MDS_CERR(" f  Switches to the fullscreen mode." << std::endl);
            MDS_CERR(" w  Switches back to the window." << std::endl);
            MDS_CERR(" c  Exits the program." << std::endl);
            break;
    }
}


// Called on mouse button action
void onMouseButton(int iButton, int iState, int x, int y)
{
    // Movement
    if( iButton == GLUT_LEFT_BUTTON )
    {
        if( iState == GLUT_DOWN )
        {
            eMouseState = MOUSE_MOVE;
            iPosX = x;
            iPosY = y;
        }
        else
        {
            eMouseState = MOUSE_NORMAL;
            iOldX = iNewX;
            iOldY = iNewY;
        }
    }

    // Zoom
    else if( iButton == GLUT_RIGHT_BUTTON )
    {
        if( iState == GLUT_DOWN )
        {
            eMouseState = MOUSE_ZOOM;
            iPosZ = y;
        }
        else
        {
            eMouseState = MOUSE_NORMAL;
            iOldZ = iNewZ;
        }
    }

    // Redraw the scene
    glutPostRedisplay();
}


// Called on mouse motion
void onMouseMotion(int x, int y)
{
    // Movement
    if( eMouseState == MOUSE_MOVE )
    {
        iNewX = iOldX + x - iPosX;
        iNewY = iOldY + y - iPosY;
    }

    // Zoom
    else if( eMouseState == MOUSE_ZOOM )
    {
        fGLZoom += (y - iPosZ) * fGLZoomStep;
        if( fGLZoom < fGLMinZoom )
        {
            fGLZoom = fGLMinZoom;
        }
        else if( fGLZoom > fGLMaxZoom )
        {
            fGLZoom = fGLMaxZoom;
        }
        iPosZ = y;
    }

    // Redisplay the scene
    glutPostRedisplay();
}

