////////////////////////////////////////////////////////////
// $Id: mcentity.h 1622 2010-02-11 09:38:33Z spanel $
////////////////////////////////////////////////////////////

#ifndef MCENTITY_H
#define MCENTITY_H

////////////////////////////////////////////////////////////
// include

// MDSTk
#include <MDSTk/Base/mdsSmallObject.h>
#include <MDSTk/Module/mdsSerializer.h>
#include <MDSTk/Math/mdsBase.h>

#include "mclistnode.h"
#include "mcpoint3d.h"
#include "mcvector3d.h"

#include <vector>
#include <stdexcept>
#include <cmath>
#include <cassert>
#include <sstream>


////////////////////////////////////////////////////////////
// macro definitions

//! Maximal required precision. The value affects comparison of vertex coordinates, etc.
#define         DOUBLE_MAX                 1000000000
//#define         DOUBLE_MAX                 1000

//! Minimal required precision. The value affects comparison of vertex coordinates, etc.
#define         DOUBLE_MIN                 0.000000001
//#define         DOUBLE_MIN                 0.001

//! A predefined constant.
#define         SQRT_3                    1.73205080756887729353


////////////////////////////////////////////////////////////
// namespace vctl

namespace vctl
{

//! Basic flags used to mark an entity.
enum MEEntityFlags
{
    NO_FLAG             = 0,
    ENTITY_MARK_FLAG    = 1,
    VERTEX_CORNER_FLAG  = 2,
    VERTEX_EDGE_FLAG    = 4,
};


////////////////////////////////////////////////////////////
// forward declarations

class MCEdge;
class MCTri;
class MCTetra;

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
/*!
 * Base class for all vector entities like edges, etc.
 * - Derived from the MCListNode class.
 */
template <typename ENTITY_TYPE>
class MCEntity : public MCListNode<ENTITY_TYPE>
{
protected:
    //! Internal entity flags.
    int flag;

    //! Entity number.
    int object;

    //! A user specific value attribute of the entity.
    double value;

    //! A user specific pointer attribute of the entity.
    //! - You should manage the underlying memory on your own.
    //! - Please, remember to clear the pointer before destructor is called.
    void * value_ptr;

public:
    //! Default constructor.
    MCEntity() : flag(NO_FLAG), value(0), value_ptr(NULL), object(-1) {}

    //! Virtual destructor.
    virtual ~MCEntity() { assert(value_ptr == NULL); }


    //! Serialization of entity attributes.
    template <class S>
    void SerializeEntity(mds::mod::CChannelSerializer<S>& Writer)
    {
        // write entity flag
        Writer.write(flag);

        // write entity object index
        Writer.write(object);
    }

    //! Deserializes the entity attributes.
    template <class S>
    void DeserializeEntity(mds::mod::CChannelSerializer<S>& Reader)
    {
        // read entity object flag
        Reader.read(flag);

        // read entity object index
        Reader.read(object);
    }


    //! Returns the object number.
    int GetObject() const { return object; }

    //! Sets the object number.
    void SetObject(int _value) { object = _value; }

    //! Sets the value attribute.
    void SetValue( const double& _value ) { value = _value; }

    //! Returns the current value attribute.
    double& GetValue() { return value; }
    const double& GetValue() const { return value; }

    //! Sets the pointer attribute.
    void SetValuePtr( void *  _ptr ) { value_ptr = _ptr; }

    //! Returns the current value of the pointer attribute.
    void * GetValuePtr() const { return value_ptr; }

    //! Adds a given flag using the binary or operation.
    void SetFlag( int set_flag ) { flag |= set_flag; }

    //! Returns the current value of internal flags.
    int GetFlag() { return flag; }

    //! Changes the internal flags completely.
    void ReplaceFlag( int new_flag ) { flag = new_flag; }

    //! Clears all bits specified in the mask.
    void MaskFlag( int mask_flag ) { flag &= (0xffff - mask_flag); }

    //! Checks if a given flag is present.
    bool TestFlag( int test_flag ) const { return ((flag & test_flag) != 0); }

    //! Copies all entity attributes.
    void SetEntityAttributes(MCEntity<ENTITY_TYPE> *entity)
    {
        assert(entity);
        flag = entity->flag;
        object = entity->object;
        value = entity->value;
    }


    //! Compares two entities and returns true if they are equal.
    virtual bool TestIdentity( ENTITY_TYPE * test_entity ) const = 0;

    //! Calculates hash code identifieng the entity.
    virtual unsigned int GetHashCode() const = 0;

    //! Returns textual representation of all entity attributes.
    virtual std::string ToString() const = 0;

    //! Returns a simple hash code of a given block of bytes.
    static unsigned int MakeStringHashCode(const char * hash_string, int hash_string_size)
    {
        static const unsigned int       hash_shift = 6;                           // hodnota posunu pro vypocet hash kodu
        static const unsigned int       hash_mask = ~0U << (32U - hash_shift);    // maska pro vypocet hash kodu
        unsigned int                    hash_code = 0;                            // vysledny hash kod

        assert(hash_string != NULL);

        // cyklus byte daneho stringu pro vypocet hash kodu
        for (int i = 0; i < hash_string_size; ++i)
            hash_code = (hash_code & hash_mask) ^ (hash_code << hash_shift) ^ hash_string[i];

        return hash_code;
    }
};

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
/*!
 * A vertex representation  used by polygonal meshes.
 * - Derived from MCPoint3D and MCEntity classes.
 */
class MDS_EXPORT MCVertex : public MCEntity<MCVertex>, public MCPoint3D
{
protected:
    //! Base type.
    typedef MCEntity<MCVertex> tBase;

    //! Node index (used during saving).
    unsigned int index;

    //! Vertex coordinates hash code.
    unsigned int m_hash_code;

    //! Pointer to a first edge sharing the vertex.
    MCEdge * edge_list;

    //! Pointer to a first triangle sharing the vertex.
    MCTri * tri_list;

    //! Pointer to a first tetrahedron sharing the vertex.
    MCTetra * tetra_list;

public:
    //! Default constructor.
    MCVertex() : index(0), m_hash_code(0), edge_list(NULL), tri_list(NULL), tetra_list(NULL) {}

    //! Copy constructor.
    MCVertex(const MCPoint3D & _point) : MCPoint3D(_point), index(0), m_hash_code(0), edge_list(NULL), tri_list(NULL), tetra_list(NULL) {}

    //! Just another constructor.
    MCVertex(const double& _x, const double& _y, const double& _z) : MCPoint3D(_x, _y, _z), index(0), edge_list(NULL), tri_list(NULL), tetra_list(NULL) {}

    //! Destructor.
    ~MCVertex() {}


    //! Writes the vertex data.
    template <class S>
    void SerializeEntity(mds::mod::CChannelSerializer<S>& Writer)
    {
        // saving vertex coordinates
        Writer.write(this->x());
        Writer.write(this->y());
        Writer.write(this->z());

        // serialize MCEntity data
        tBase::SerializeEntity(Writer);
    }

    //! Reads the Vertex data.
    template <class S>
    void DeserializeEntity(mds::mod::CChannelSerializer<S>& Reader)
    {
        // loading vertex coordinates
        Reader.read(this->x());
        Reader.read(this->y());
        Reader.read(this->z());

        // deserialize MCEntity data
        tBase::DeserializeEntity(Reader);
    }


    //! Sets the vertex index.
    void SetIndex( unsigned int _index ) { index = _index; }

    //! Returns the vertex index;
    unsigned int GetIndex() const { return index; }


    //! Sets pointer to an already registered edge linked to the vertex.
    void SetRegisteredEdge(MCEdge * _edge) { edge_list = _edge; }

    //! Returns pointer to a first registered edge.
    MCEdge * GetRegisteredEdge() { return edge_list; }
    const MCEdge * GetRegisteredEdge() const { return edge_list; }

    //! Sets pointer to an already registered triangle linked to the vertex.
    void SetRegisteredTri(MCTri * _tri) { tri_list = _tri; }

    //! Returns pointer to a registered triangle.
    MCTri * GetRegisteredTri() { return tri_list; }
    const MCTri * GetRegisteredTri() const { return tri_list; }

    //! Sets pointer to a registered tetrahedron already linked to the vertex.
    void SetRegisteredTetra(MCTetra * _tetra) { tetra_list = _tetra; }

    //! Returns pointer to a registered tetrahedron.
    MCTetra * GetRegisteredTetra() { return tetra_list; }
    const MCTetra * GetRegisteredTetra() const { return tetra_list; }

    //! Returns vector of all linked edges.
    void GetRegisteredEdgeList(std::vector<MCEdge *> & _edge_list);

    //! Returns vector of all linked triangles.
    void GetRegisteredTriList(std::vector<MCTri *> & _tri_list);

    //! Returns vector of all linked tetrahedra.
    void GetRegisteredTetraList(std::vector<MCTetra *> & _tetra_list);

    //! Returns vector of triangles registered for this vertex,
    //! excluding those including a given exclude vertex.
    void GetRegisteredTriListExclude(std::vector<MCTri *> & _tri_list, MCVertex * exclude_vertex);


    //! Checks identity of two vertices.
    bool TestIdentity( MCVertex * _test ) const
    {
        assert(_test);
        #ifndef NO_VERTEX_IDENTITY
            return ((mds::math::getAbs(x() - _test->GetX()) < DOUBLE_MIN)
                    && (mds::math::getAbs(y() - _test->GetY()) < DOUBLE_MIN)
                    && (mds::math::getAbs(z() - _test->GetZ()) < DOUBLE_MIN));
        #else
            return false;
        #endif
    }

    //! Returns hash code of vertex coordinates.
    unsigned int GetHashCode() const { return m_hash_code; }

    //! Sets the hash code equal to a given vertex.
    void SetHashCode(MCVertex & _vertex) { m_hash_code = _vertex.m_hash_code; }

    //! Calculates hash code of vertex coordinates.
    void MakeHashCode()
    {
        // zaokrouhleni souradnic vrcholu a ulozeni vysledku do pracovniho pole pro vypocet hash kodu
        double double_array[3];
        double_array[0] = floor(x() * DOUBLE_MAX + 0.5);
        double_array[1] = floor(y() * DOUBLE_MAX + 0.5);
        double_array[2] = floor(z() * DOUBLE_MAX + 0.5);

        // vypocet hash kodu a vraceni vysledku
        m_hash_code = MakeStringHashCode(reinterpret_cast<char *>(double_array), sizeof(double)*3);
    }

    //! Returns text description of the vertex.
    std::string ToString() const
    {
        std::ostringstream text_buff;
        text_buff << "Vertex " << this << " : " << x() << " / " << y() << " / " << z();
        return text_buff.str();
    }
};

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
/*!
 * Edge of a polygonal mesh.
 * - Derived from the MCEntity class.
 */
class MDS_EXPORT MCEdge : public MCEntity<MCEdge>, public mds::base::CSmallObject<>
{
protected:
    //! Base class.
    typedef MCEntity<MCEdge> tBase;

    //! End-points of the edge.
    MCVertex * u[2];

    //! Pointers to edges sharing the same vertex.
    MCEdge * s[2];

public:
    //! Default constructor.
    MCEdge() { u[0] = u[1] = NULL; s[0] = s[1] = NULL; }

    //! Copy constructor.
    //! - Initializes pointers to vertices only!
    MCEdge(const MCEdge & _hrana) { _hrana.GetVerticeS(u); s[0] = s[1] = NULL; }

    //! Constructor.
    //! - Initializes pointers to vertices only!
    MCEdge(MCEdge * _hrana) { assert(_hrana); _hrana->GetVerticeS(u); s[0] = s[1] = NULL; }

    //! Constructor.
    //! - Initializes pointers to vertices only!
    MCEdge(MCVertex * _u0, MCVertex * _u1)
    {
        assert(_u0 && _u1);
        u[0] = _u0; u[1] = _u1;
        s[0] = NULL; s[1] = NULL;
    }

    //! Constructor.
    //! - Initializes pointers to vertices only!
    MCEdge(MCVertex * _u[2])
    {
        assert(_u[0] && _u[1]);
        u[0] = _u[0]; u[1] = _u[1];
        s[0] = s[1] = NULL;
    }

    //! Empty destructor.
    ~MCEdge() {}


    //! Writes the edge data.
    template <class S>
    void SerializeEntity(mds::mod::CChannelSerializer<S>& Writer)
    {
        assert(u[0] && u[1]);

        // write vertex indexes
        Writer.write(u[0]->GetIndex());
        Writer.write(u[1]->GetIndex());

        // serialize MCEntity data
        tBase::SerializeEntity(Writer);
    }

    //! Reads the edge data.
    template <class S>
    void DeserializeEntity(mds::mod::CChannelSerializer<S>& Reader, std::vector<MCVertex *> & index_array)
    {
        // read vertex indexes
        int index0, index1;
        Reader.read(index0);
        Reader.read(index1);

        assert(index0 < int(index_array.size()) && index0 >= 0);
        assert(index1 < int(index_array.size()) && index1 >= 0);

        // translation of indexes to pointers
        u[0] = index_array[index0];
        u[1] = index_array[index1];

        // deserialize MCEntity data
        tBase::DeserializeEntity(Reader);
    }


    //! Sets pointer to one of the edge vertices.
    //! @param index - index of the vertex (0 or 1).
    void SetVertex( int index, MCVertex * new_uzel )
    {
        assert((index >= 0) && (index < 2) && new_uzel);
        u[index] = new_uzel;
    }

    //! Sets the both edge end-points.
    void SetVerticeS( MCVertex * _u0, MCVertex * _u1 )      { assert(_u0 && _u1); u[0] = _u0; u[1] = _u1; }

    //! Sets both edge end-points.
    void SetVerticeS( MCVertex * _u[2] )                    { assert(_u[0] && _u[1]); u[0] = _u[0]; u[1] = _u[1]; }

    //! Returns pointer to a subscripted vertex.
    MCVertex * GetVertex( int index )                       { assert((index >= 0) && (index < 2)); return u[index]; }

    //! Returns pointers to edge vertices.
    void GetVerticeS( MCVertex ** _u0, MCVertex ** _u1 ) const
    {
        assert(_u0 && _u1);
        *_u0 = u[0]; *_u1 = u[1];
    }

    //! Returns pointers to edge vertices.
    void GetVerticeS( MCVertex * _u[2] ) const              { _u[0] = u[0]; _u[1] = u[1]; }

    //! Exchanges pointer to edge vertex.
    //!  @param old_uzel - pointer to the original vertex.
    //!  @param new_uzel - pointer to a new vertex.
    void ChangeVertex(MCVertex * old_uzel, MCVertex * new_uzel)
    {
        int old_index = IsVertex(old_uzel);
        assert(old_index != -1);
        assert(new_uzel);
        u[old_index] = new_uzel;
    }


    //! Checks if a given vertex belongs to the edge.
    //! @return index of the vertex within the edge, or -1 otherwise.
    int IsVertex( MCVertex * test_uzel ) const             { assert(test_uzel); return ( (test_uzel == u[0]) ? 0 : ((test_uzel == u[1]) ? 1 : -1) ); }

    //! Checks if a given vertex belongs to the edge.
    bool IsVertexBool( MCVertex * test_uzel ) const        { assert(test_uzel); return ((test_uzel == u[0]) || (test_uzel == u[1])); }

    //! Returns pointer to the opossite vertex of the edge.
    //! @param _uzel - pointer to a one of the edge vertices.
    //! @return NULL on failure.
    MCVertex * GetRestVertex( MCVertex * _uzel )           { assert(_uzel); return ((u[0] == _uzel) ? u[1] : ((u[1] == _uzel) ? u[0] : NULL)); }

    //! Returns geometrical center of the edge.
    void GetCenter( MCPoint3D & _stred )                   { _stred.SetXYZ( (u[0]->GetX()+u[1]->GetX())*0.5, (u[0]->GetY()+u[1]->GetY())*0.5, (u[0]->GetZ()+u[1]->GetZ())*0.5 ); }

    //! Returns length of the edge.
    double GetLength()                                     { return u[0]->Distance(*u[1]); }

    //! Divides the edge with respect to a specified ratio <0,1>.
    //! - Returns coordinates of calculated point.
    void GetEdgePoint(MCPoint3D & bod_h, double pomer)
    {
        bod_h.SetX(u[0]->GetX() + pomer*(u[1]->GetX() - u[0]->GetX()));
        bod_h.SetY(u[0]->GetY() + pomer*(u[1]->GetY() - u[0]->GetY()));
        bod_h.SetZ(u[0]->GetZ() + pomer*(u[1]->GetZ() - u[0]->GetZ()));
    }


    //! Registers the edge into its vertices.
    void Registration()
    {
        s[0] = u[0]->GetRegisteredEdge();
        s[1] = u[1]->GetRegisteredEdge();
        u[0]->SetRegisteredEdge(this);
        u[1]->SetRegisteredEdge(this);
    }

    //! Deregistration of the edge in its vertices.
    void DeRegistration();


    //! Checks identity of two edges.
    bool TestIdentity( MCEdge * test_edge ) const
    {
        assert(test_edge);
        return (IsVertexBool(test_edge->GetVertex(0)) && IsVertexBool(test_edge->GetVertex(1)));
    }

    //! Returns hash code of the edge.
    //! - Not yet implemented!
    //! @todo Doplnit implementaci generovani hash kodu.
    unsigned int GetHashCode() const                       { return 0; }

    //! Returns text description of the edge.
    virtual std::string ToString() const
    {
        std::ostringstream text_buff;
        text_buff << "Edge " << this << " : " << u[0] << " / " << u[1];
        return text_buff.str();
    }

    //! Returns pointer to a first edge registered in a specified vertex.
    //! - NULL on failure.
    MCEdge * GetVertexEdge( MCVertex * _uzel )             { int index = IsVertex(_uzel); return((index != -1) ? s[index] : NULL); }

    //! Sets an edge registered in a specified vertex.
    void SetVertexEdge(MCVertex * _uzel, MCEdge * _hrana)  { int index = IsVertex(_uzel); assert(index != -1); s[index] = _hrana; }
};

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
/*!
 * Triangle in 3D space, component of a polygonal mesh.
 * - Derived from MCEntity class.
 */
class MDS_EXPORT MCTri : public MCEntity<MCTri>, public mds::base::CSmallObject<>
{
protected:
    //! Base class.
    typedef MCEntity<MCTri> tBase;

    //! Pointers to all triangle vertices.
    MCVertex * u[3];

    //! Pointers to triangles sharing the corresponding vertex.
    MCTri * s[3];

public:
    //! Default constructor.
    //! - Initializes all pointers to NULL!
    MCTri() { u[0] = u[1] = u[2] = NULL; s[0] = s[1] = s[2] = NULL; }

    //! Copy constructor.
    //! - Initializes only pointers to triangle vertices.
    MCTri(const MCTri& _tri)
    {
        u[0] = _tri.u[0];
        u[1] = _tri.u[1];
        u[2] = _tri.u[2];
        s[0] = s[1] = s[2] = NULL;
    }

    //! Constructor.
    //! - Initializes only pointers to triangle vertices.
    MCTri(MCTri * _tri)
    {
        assert(_tri);
        u[0] = _tri->u[0];
        u[1] = _tri->u[1];
        u[2] = _tri->u[2];
        s[0] = s[1] = s[2] = NULL;
    }

    //! Constructor.
    //! - Initializes only pointers to triangle vertices.
    MCTri(MCVertex * _u0, MCVertex * _u1, MCVertex * _u2)
    {
        assert(_u0 && _u1 && _u2);
        u[0] = _u0;
        u[1] = _u1;
        u[2] = _u2;
        s[0] = s[1] = s[2] = NULL;
    }

    //! Constructor.
    //! - Initializes only pointers to triangle vertices.
    MCTri(MCVertex * _u[3])
    {
        assert(_u[0] && _u[1] && _u[2]);
        u[0] = _u[0];
        u[1] = _u[1];
        u[2] = _u[2];
        s[0] = s[1] = s[2] = NULL;
    }

    //! Empty destructor.
    virtual ~MCTri() {}


    //! Loads triangle (indexes of vertices) from a given stream.
    //! - An array of vertices must be loaded previously.
    void Load(std::ifstream & proud, std::vector<MCVertex *> & _pole);

    //! Saves triangle (indexes of vertices) into a stream.
    //! - An index must be assigned to each vertex.
    void Save(std::ofstream & proud);


    //! Serializes the triangle.
    //! - An index must be assigned to each vertex.
    template <class S>
    void SerializeEntity(mds::mod::CChannelSerializer<S>& Writer)
    {
        assert(u[0] && u[1] && u[2]);

        // write vertex indexes into
        Writer.write(u[0]->GetIndex());
        Writer.write(u[1]->GetIndex());
        Writer.write(u[2]->GetIndex());

        // serialize MCEntity data
        MCEntity<MCTri>::SerializeEntity(Writer);
    }

    //! Deserializes the triangle.
    //! - An array of vertices must be loaded previously.
    template <class S>
    void DeserializeEntity(mds::mod::CChannelSerializer<S>& Reader, std::vector<MCVertex *> & index_array)
    {
        // read vertices index from given channel
        int index0, index1, index2;
        Reader.read(index0);
        Reader.read(index1);
        Reader.read(index2);

        assert((index0 < (int)index_array.size()) && (index0 >= 0));
        assert((index1 < (int)index_array.size()) && (index1 >= 0));
        assert((index2 < (int)index_array.size()) && (index2 >= 0));

        // translation from vertices index on their pointers
        u[0] = index_array[index0];
        u[1] = index_array[index1];
        u[2] = index_array[index2];

        // deserialize MCEntity data
        tBase::DeserializeEntity(Reader);
    }


    //! Sets pointer to one of the vertices.
    void SetVertex( int index, MCVertex * new_uzel )       { assert((index >= 0) && (index < 3) && new_uzel); u[index] = new_uzel; }

    //! Sets pointers to all triangle vertices.
    void SetVerticeS( MCVertex * _u0, MCVertex * _u1, MCVertex * _u2 ) { assert(_u0 && _u1 && _u2); u[0] = _u0; u[1] = _u1; u[2] = _u2; }

    //! Sets pointers to all triangle vertices.
    void SetVerticeS( MCVertex * _u[3] )                    { assert(_u[0] && _u[1] && _u[2]); u[0] = _u[0]; u[1] = _u[1]; u[2] = _u[2]; }

    //! Return pointer to a specified vertex.
    MCVertex * GetVertex( int index )                       { assert((index >= 0) && (index < 3)); return u[index]; }

    //! Returns pointer to all triangle vertices.
    void GetVerticeS( MCVertex ** _u0, MCVertex ** _u1, MCVertex ** _u2 ) const
    {
        assert(_u0 && _u1 && _u2);
        *_u0 = u[0]; *_u1 = u[1]; *_u2 = u[2];
    }

    //! Returns pointer to all triangle vertices.
    void GetVerticeS( MCVertex * _u[3] ) const              { _u[0] = u[0]; _u[1] = u[1]; _u[2] = u[2]; }


    //! Exchanges one of the triangle vertices.
    //! - This method doesn't changes registration of tringles in vertices.
    //!   You must do the registration on your own.
    void ChangeVertex(MCVertex * old_uzel, MCVertex * new_uzel)
    {
        int old_index = IsVertex(old_uzel);
        assert(old_index != -1);
        assert(new_uzel);
        u[old_index] = new_uzel;
    }

    //! Checks if a given vertex belongs to the triangle.
    //! @return index of the vertex within the triangle, or -1 otherwise.
    int IsVertex( MCVertex * test_uzel ) const              { assert(test_uzel); return ((test_uzel == u[0]) ? 0 : ((test_uzel == u[1]) ? 1 : ((test_uzel == u[2]) ? 2 : -1)) ); }

    //! Checks if a given vertex belongs to the triangle.
    bool IsVertexBool( MCVertex * test_uzel ) const        { assert(test_uzel); return ((test_uzel == u[0]) || (test_uzel == u[1]) || (test_uzel == u[2])); }


    //! Returns a remaining vertex, or NULL.
    MCVertex * GetRestVertex(MCVertex * _u0, MCVertex * _u1);

    //! Returns two remaining vertices.
    void GetRestEdge(MCVertex * _uzel, MCVertex ** _u0, MCVertex ** _u1);


    //! Returns length and vertices of the longest edge.
    double GetMaxEdge( MCVertex ** _u0, MCVertex ** _u1 );

    //! Returns pointer to a longest edge, or NULL on failure.
    double GetMaxEdge( MCEdge ** max_hrana );

    //! Returns index of the longest edge.
    int GetMaxEdge();

    //! Returns length of the longest edge.
    double GetMaxEdgeLength();

    //! Returns length of the shortest edge.
    double GetMinEdgeLength();


    //! Returns all triangle edges, or NULL if no edges exist.
    void GetEdges( MCEdge * hrany[3] );

    //! Returns a specified triangle edge, or NULL if no such edge exists.
    MCEdge * GetEdge( int hrana );

    //! Returns pointers to all triangle vertices.
    void GetEdge(int index, MCVertex ** _u0, MCVertex ** _u1);

    //! Returns perimeter of the triangle.
    double GetCircuit();

    //! Returns area of the triangle.
    double GetArea();

    //! Returns coordinates of the triangle's center.
    void GetCenter(MCPoint3D & _center);


    //! Returns normal of the triangle normalized so that its length is equal to one.
    MCVector3D GetNormal();

    //! Returns normal of the triangle normalized so that its length is equal to one.
    void GetNormal(MCVector3D & normala);

    //! Returns parameters of the triangle's plane.
    void GetPlane( double & A, double & B, double & C, double & D );

    //! Returns quality of the triangle.
    double GetQuality()
    {
        double plocha = GetArea();
        if( plocha == 0 ) return 1000000;
        return ((SQRT_3 * GetMaxEdgeLength() * GetCircuit()) / (12.0 * plocha));
    }


    //! Returns pointers to neighbouring triangles.
    void GetNeighbours( MCTri * _pole[3] );

    //! Returns pointer to a specific triangle.
    MCTri * GetNeighbour( int _index );


    //! Changes orientation of the triangle.
    void InverseOrientation();

    //! Compares orientation of two triangles.
    //! - 1 - same orientation of triangles.
    //! - 0 - perpendicular triangles.
    //! - -1 - reverse orientation.
    int CompareTriOrientation(MCTri * test_tri);


    //! Registers the triangle into its vertices.
    void Registration()
    {
        s[0] = u[0]->GetRegisteredTri();
        s[1] = u[1]->GetRegisteredTri();
        s[2] = u[2]->GetRegisteredTri();
        u[0]->SetRegisteredTri(this);
        u[1]->SetRegisteredTri(this);
        u[2]->SetRegisteredTri(this);
    }

    //! Deregistration of triangle from vertices.
    void DeRegistration();


    //! Checks of two triangles are equal.
    bool TestIdentity( MCTri * test_tri ) const
    {
        assert(test_tri);
        return (test_tri->IsVertexBool(u[0]) && test_tri->IsVertexBool(u[1]) && test_tri->IsVertexBool(u[2]));
    }

    //! Returns triangle hash code.
    //! - Not yet implemented!
    //! @todo Doplnit implementaci generovani hash kodu.
    unsigned int GetHashCode() const                        { return 0; }

    //! Returns text description of the triangle.
    virtual std::string ToString() const
    {
        std::ostringstream text_buff;
        text_buff << "Tri " << this << " : " << u[0] << " / " << u[1] << " / " << u[2];
        return text_buff.str();
    }

    //! Returns a registered triangle in a given vertex.
    //! - NULL on failure.
    MCTri * GetVertexTri( MCVertex * _uzel )
    {
        int index = IsVertex(_uzel);
        return ((index != -1) ? s[index] : NULL);
    }

    //! Sets pointer to a triangle registered in a given vertex.
    void SetVertexTri(MCVertex * _uzel, MCTri * _tri)
    {
        int index = IsVertex(_uzel);
        assert(index != -1);
        s[index] = _tri;
    }
};

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
/*!
 * Class representing a tetrahedron.
 * - Derived from MCEntity class.
 */
class MDS_EXPORT MCTetra: public MCEntity<MCTetra>, public mds::base::CSmallObject<>
{
protected:
    //! Base class.
    typedef MCEntity<MCTetra> tBase;

    //! Pointers to all tetrahedron vertices.
    MCVertex * u[4];

    //! Pointers to tetrahedra sharing the same vertices.
    MCTetra * s[4];

    //! Direct pointers to neighbouring tetrahedra.
    MCTetra * n[4];

    //! Value flag for signing the tetrahedron.
    unsigned int value_flag;

public:
    //! Default constructor.
    //! - Initializes all pointers to NULL.
    MCTetra() : value_flag(0)
    {
        u[0] = u[1] = u[2] = u[3] = NULL;
        s[0] = s[1] = s[2] = s[3] = NULL;
        n[0] = n[1] = n[2] = n[3] = NULL;
    }

    //! Copy constructor.
    //! - Initializes pointers to tetrahedron vertices only.
    MCTetra(const MCTetra& _tet) : value_flag(0)
    {
        _tet.GetVerticeS(u);
        s[0] = s[1] = s[2] = s[3] = NULL;
        n[0] = n[1] = n[2] = n[3] = NULL;
    }

    //! Constructor.
    //! - Initializes pointers to tetrahedron vertices only.
    MCTetra(MCTetra * _tet) : value_flag(0)
    {
        assert(_tet);
        _tet->GetVerticeS(u);
        s[0] = s[1] = s[2] = s[3] = NULL;
        n[0] = n[1] = n[2] = n[3] = NULL;
    }

    //! Constructor.
    //! - Initializes pointers to tetrahedron vertices only.
    MCTetra(MCVertex * _u0, MCVertex * _u1, MCVertex * _u2, MCVertex * _u3) : value_flag(0)
    {
        SetVerticeS(_u0, _u1, _u2, _u3);
        s[0] = s[1] = s[2] = s[3] = NULL;
        n[0] = n[1] = n[2] = n[3] = NULL;
    }

    //! Constructor.
    //! - Initializes pointers to tetrahedron vertices only.
    MCTetra(MCVertex * _u[4]) : value_flag(0)
    {
        assert(_u[0] && _u[1] && _u[2] && _u[3]);
        u[0] = _u[0];
        u[1] = _u[1];
        u[2] = _u[2];
        u[3] = _u[3];
        s[0] = s[1] = s[2] = s[3] = NULL;
        n[0] = n[1] = n[2] = n[3] = NULL;
    }

    //! Virtual destructor.
    virtual ~MCTetra() {}


    //! Returns the value flag.
    unsigned int GetValueFlag()                                           { return value_flag; }

    //! Sets the value flag.
    void SetValueFlag(unsigned int _value)                                { value_flag = _value; }


    //! Writes the tetrahedron data.
    //! - An unique index must be assigned to each vertex first.
    template <class S>
    void SerializeEntity(mds::mod::CChannelSerializer<S>& Writer)
    {
        assert(u[0] && u[1] && u[2] && u[3]);

        // write indexes of vertices
        Writer.write(u[0]->GetIndex());
        Writer.write(u[1]->GetIndex());
        Writer.write(u[2]->GetIndex());
        Writer.write(u[3]->GetIndex());

        // serialize MCEntity data
        tBase::SerializeEntity(Writer);
    }

    //! Reads the tetra data.
    //! - An array of vertices must be loaded previously.
    template <class S>
    void DeserializeEntity(mds::mod::CChannelSerializer<S>& Reader, std::vector<MCVertex *> & index_array)
    {
        // read vertices index from given channel
        int index0, index1, index2, index3;
        Reader.read(index0);
        Reader.read(index1);
        Reader.read(index2);
        Reader.read(index3);

        assert(index0 < (int)index_array.size() && index0 >= 0);
        assert(index1 < (int)index_array.size() && index1 >= 0);
        assert(index2 < (int)index_array.size() && index2 >= 0);
        assert(index3 < (int)index_array.size() && index3 >= 0);

        // translation of index to pointer
        u[0] = index_array[index0];
        u[1] = index_array[index1];
        u[2] = index_array[index2];
        u[3] = index_array[index3];

        // deserialize MCEntity data
        tBase::DeserializeEntity(Reader);
    }


    //! Sets pointer to a specified vertex.
    void SetVertex( int index, MCVertex * new_uzel )       { assert((index >= 0) && (index < 4) && new_uzel); u[index] = new_uzel; }

    //! Sets pointers to all tetrahedron vertices.
    void SetVerticeS( MCVertex * _u0, MCVertex * _u1, MCVertex * _u2, MCVertex * _u3 )
    {
        assert(_u0 && _u1 && _u2 && _u3);
        u[0] = _u0; u[1] = _u1; u[2] = _u2; u[3] = _u3;
    }

    //! Sets pointers to all tetrahedron vertices.
    void SetVerticeS( MCVertex * _u[4] )
    {
        assert(_u[0] && _u[1] && _u[2] && _u[3]);
        u[0] = _u[0]; u[1] = _u[1]; u[2] = _u[2]; u[3] = _u[3];
    }

    //! Returns pointer to a specified vertex.
    MCVertex * GetVertex( int index )                      { assert((index >= 0) && (index < 4)); return u[index]; }

    //! Returns pointers to all tetrahedron vertices.
    void GetVerticeS( MCVertex ** _u0, MCVertex ** _u1, MCVertex ** _u2, MCVertex ** _u3 ) const
    {
        assert(_u0 && _u1 && _u2 && _u3);
        *_u0 = u[0]; *_u1 = u[1]; *_u2 = u[2]; *_u3 = u[3];
    }

    //! Returns pointers to all tetrahedron vertices.
    void GetVerticeS( MCVertex * _u[4] ) const             { _u[0] = u[0]; _u[1] = u[1]; _u[2] = u[2]; _u[3] = u[3]; }


    //! Returns pointer to the remaining vertex.
    MCVertex * GetRestVertex(MCVertex * _u0, MCVertex * _u1, MCVertex * _u2);

    //! Returns pointer to the vertex opposing a specified face.
    MCVertex * GetRestVertex( int index );


    //! Exchanges a specified vertex with a new one.
    //! - The method doesn't provide any registration/deregistration of tetrahedron in vertices.
    //!   You must do this on your own!
    void ChangeVertex(MCVertex * old_uzel, MCVertex * new_uzel)
    {
        int old_index = IsVertex(old_uzel);
        assert(old_index != -1);
        assert(new_uzel);
        u[old_index] = new_uzel;
    }


    //! Returns index of a specified vertex comparing pointers, or -1 if no such vertex was found.
    int IsVertex( MCVertex * test_uzel ) const
    {
        assert(test_uzel);
        return ((test_uzel == u[0]) ? 0 : ((test_uzel == u[1]) ? 1 : ((test_uzel == u[2]) ? 2: ((test_uzel == u[3]) ? 3 : -1))));
    }

    //! Checks if a given vertex belongs to the tetrahedra.
    bool IsVertexBool( MCVertex * test_uzel ) const
    {
        assert(test_uzel);
        return ((test_uzel == u[0]) || (test_uzel == u[1]) || (test_uzel == u[2]) || (test_uzel == u[3]));
    }

    //! Checks if a given triangle belongs to the tetrahedra.
    bool IsTriBool( MCVertex * _u0, MCVertex * _u1, MCVertex * _u2 )
    {
        return (IsVertexBool(_u0) && IsVertexBool(_u1) && IsVertexBool(_u2));
    }

    //! Checks if a given triangle belongs to the tetrahedra.
    bool IsTriBool( MCVertex * _u[3] )                     { return IsTriBool(_u[0], _u[1], _u[2]); }


    //! Returns langth of the longest edge.
    double GetMaxEdgeLength();

    //! Returns langth of the shortest edge.
    double GetMinEdgeLength();

    //! Returns the longest edge.
    void GetMaxEdge( MCVertex ** _u0, MCVertex ** _u1 );

    //! Returns the shortest edge.
    void GetMinEdge( MCVertex ** _u0, MCVertex ** _u1 );


    //! Returns pointers to all vertices of a specified tetrahedron face.
    void GetTri( int index, MCVertex ** _u0, MCVertex ** _u1, MCVertex ** _u2 );

    //! Returns pointers to all vertices of a specified tetrahedron face.
    void GetTri( int index, MCVertex * _u[3] ) { GetTri(index, _u, _u + 1, _u + 2); }

    //! Returns indexes of vertices of a tetrahedron face.
    void GetTri( int index, int * _u0, int * _u1, int * _u2 );

    //! Returns indexes of vertices of a tetrahedron face.
    void GetTri( int index, int _u[3] ) { GetTri(index, _u, _u + 1, _u + 2); }


    //! Returns pointers to vertices opposing a specified vertex.
    void GetRestTri( int index, MCVertex ** _u0, MCVertex ** _u1, MCVertex ** _u2 );

    //! Returns pointers to vertices opposing a specified vertex.
    void GetRestTri( int index, MCVertex * _u[3] ) { GetRestTri(index, _u, _u + 1, _u + 2); }


    //! Returns tetrahedron surface area.
    double GetArea();

    //! Returns tetrahedron volume.
    double GetVolume();

    //! Returns coordinates of tetrahedron center.
    void GetCenter(MCPoint3D & _center);

    //! Returns normal of a specified tetrahedron face.
    //! - The returned vector is normalized.
    MCVector3D GetNormal( int index );

    //! Returns normal of a specified tetrahedron face.
    //! - The returned vector is not normalized.
    MCVector3D GetNormal2( int index );

    //! Returns normal of a specified tetrahedron face.
    //! - The returned vector is normalized.
    void GetNormal( int index, MCVector3D & normal );

    //! Returns normal of a specified tetrahedron face.
    //! - The returned vector is not normalized.
    void GetNormal2( int index, MCVector3D & normal );


    //! Returns neighbouring tetrahedra.
    //! - Traversing all tetrahedra registered in the tetrahedron vertices.
    void GetNeighboursByVertices( MCTetra * _pole[4] );

    //! Returns a specified neighbouring tetrahedron.
    //! - Traversing all tetrahedra registered in the tetrahedron vertices.
    MCTetra * GetNeighbourByVertices( int _index );

    //! Returns neighbouring tetrahedra.
    //! - Directly via internal pointers.
    void GetNeighbours( MCTetra * _n[4] )    { _n[0] = n[0]; _n[1] = n[1]; _n[2] = n[2]; _n[3] = n[3]; }

    //! Returns pointers to neighbouring tetrahedra.
    //! - Directly via internal pointers.
    const MCTetra ** GetNeighbours() const   { return const_cast<const MCTetra **>(n); }

    //! Returns a specified neighbouring tetrahedron.
    //! - Directly via internal pointers.
    MCTetra * GetNeighbour( int _index )     { assert((_index >= 0) && (_index < 4)); return n[_index]; }


    //! Inverses orientation of the tetrahedron.
    void InverseOrientation();

    //! Registeres the tetrahedron in all its vertices.
    void Registration()
    {
        for (int i = 0; i < 4; i++)
        {
            s[i] = u[i]->GetRegisteredTetra();
            u[i]->SetRegisteredTetra(this);
        }
        NeighboursRegistration();
    }

    //! Deregisteres the tetrahedron from its vertices.
    void DeRegistration();


    //! Checks identity of two tetrahedra.
    bool TestIdentity( MCTetra * test_tetra ) const
    {
        assert(test_tetra);
        return ((test_tetra->IsVertexBool(u[0])) && (test_tetra->IsVertexBool(u[1])) && (test_tetra->IsVertexBool(u[2])) && (test_tetra->IsVertexBool(u[3])));
    }

    //! Returns the tetrahedron hash code.
    //! - Not yet implemented!
    //! @todo Doplnit implementaci generovani hash kodu.
    unsigned int GetHashCode() const                       { return 0; };

    //! Returns text description of the tetrahedra.
    virtual std::string ToString() const
    {
        std::ostringstream text_buff;
        text_buff << "Tetra " << this << " : " << u[0] << " / " << u[1] << " / " << u[2] << " / " << u[3];
        return text_buff.str();
    }

    //! Returns a tetrahedron registered in a specified tetrahedron vertex (sharing the same vertex).
    //! - NULL on failure.
    MCTetra * GetVertexTetra( MCVertex * _uzel )           { int index = IsVertex(_uzel); return((index != -1) ? s[index] : NULL); }

    //! Sets pointer to a tetrahedron registered in a specified tetrahedron vertex (sharing the same vertex).
    //! - NULL on failure.
    void SetVertexTetra(MCVertex * _uzel, MCTetra * _tet)  { int index = IsVertex(_uzel); assert(index != -1); s[index] = _tet; }

protected:
    //! Tetrahedron neighbours registration.
    //! - Actualizes all neighbours and also neighbours of all newly find neighbours.
    void NeighboursRegistration();

    //! Tetrahedron neighbours deregistration.
    //! - Actualizes all neighbours of its actual neighbours.
    //! - The function expects, that the tetra is deregistered from his vertices
    //!   by using the function DeRegistration() first.
    void NeighboursDeRegistration();

    //! Tetrahedra neighbours actualization.
    //! - Only for this tetra.
    void NeighboursActualization() { GetNeighboursByVertices(n); }
};

} // namespace vctl

#endif

////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////

